#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "config.h"
#include "files.h"

#define SECTION_PATHS 1
#define SECTION_NODES 2
#define SECTION_AREAS 3
#define SECTION_HATCH 4

XFIX_REC *xfirst, *xrec;
XNODE_REC *firstnode, *noderec;
HATCH_REC *firsthatch, *hatchrec;
char inbound[256], outbound[256], safein[256], ticpath[256], badtics[256];

int str2addr(char *addrstrp, unsigned short *zone, unsigned short *net, unsigned short *node, unsigned short *point)
{
    char *pstr;
    char addrstr[50],pos;

    strcpy(addrstr,addrstrp); pos = 0;
    for (pstr = addrstr; ; pstr++)
    {
        if (*pstr >= '0' && *pstr <= '9') continue;
        if (*pstr == ':' && pos == 0) { *pstr = ' '; pos = 1; continue; }
        if (*pstr == '/' && pos == 1) { *pstr = ' '; pos = 2; continue; }
        if (*pstr == '.' && pos == 2) { *pstr = ' '; pos = 3; continue; }
        if (*pstr == '@') { *pstr = '\0'; break; }
        if (*pstr == '\0') break;
        printf("Error : str2addr(\"%s\") position %d ('%c')\n",addrstrp,(int) (pstr-addrstr)+1,*pstr);
        return 0;
    }

    *zone = 0; *net = 0; *node = 0; *point = 0;
    return sscanf(addrstr,"%hu %hu %hu %hu",zone,net,node,point) >= 3;
}

/* Convert string to list of addresses */
ADDR_REC *get_list(char *str, int *stat)
{
    ADDR_REC *first, *rec, *tmp;
    char *addr;

    first = NULL; rec = NULL; *stat = 1;
    while (str != NULL && *str != '\0')
    {
        addr = str;
        str = strchr(str, ' ');
        if (str != NULL)
        {
            *str++ = '\0';
            while (*str == ' ') str++;
        }

        tmp = new ADDR_REC;
        if (tmp == NULL)
        {
            *stat = 0;
            return NULL;
        }
        if (first == NULL) first = tmp;
        if (rec != NULL) rec->next = tmp;
        rec = tmp; rec->next = NULL;
        if (!str2addr(addr, &rec->zone, &rec->net, &rec->node, &rec->point))
        {
            *stat = 0;
            return NULL;
        }
    }

    return first;
}

/* Read configuration */
int read_config(char *fname)
{
    FILE *Fcfg;
    XFIX_REC *xtmp;
    XNODE_REC *tmpnode;
    HATCH_REC *tmphatch;
    char line[512], export[256], tag[61], path[256], *strp, *data, addr1[30], addr2[30], password[9];
    unsigned lines;
    int section, stat;
    ADDR_REC ownaddr, expaddr;

    xfirst = NULL; xrec = NULL; firstnode = NULL; noderec = NULL;
    firsthatch = NULL; hatchrec = NULL; inbound[0] = '\0'; outbound[0] = '\0';
    safein[0] = '\0'; ticpath[0] = '\0'; badtics[0] = '\0'; lines = 0;

    Fcfg = fopen(fname, "rb"); section = 0;
    if (Fcfg == NULL)
    {
        printf("Can't open configuration file %s\n", fname);
        return 0;
    }

    while (_fgets(line, sizeof(line), Fcfg) != NULL)
    {
        lines++;
        if (line[0] == '\0' || line[0] == ';') continue;

        if (line[0] == '[')
        {
            if (stricmp(line, "[paths]") == 0)
            {
                section = SECTION_PATHS;
                continue;
            }
            if (stricmp(line, "[nodes]") == 0)
            {
                section = SECTION_NODES;
                continue;
            }
            if (stricmp(line, "[areas]") == 0)
            {
                section = SECTION_AREAS;
                continue;
            }
            if (stricmp(line, "[hatch]") == 0)
            {
                section = SECTION_HATCH;
                continue;
            }
        __err:
            printf("Error in file area configuration file '%s' line %u\n", fname, lines);
            return 0;
        }

        switch (section)
        {
            case SECTION_PATHS:
                strp = strchr(line, '=');
                if (strp == NULL) goto __err;
                data = strp+1;
                while (*data == ' ') data++;
                while (*(strp-1) == ' ') strp--; *strp = '\0';

                if (stricmp(line, "Inbound") == 0)
                {
                    strcpy(inbound, data);
                    break;
                }

                if (stricmp(line, "Outbound") == 0)
                {
                    strcpy(outbound, data);
                    break;
                }

                if (stricmp(line, "SafeIn") == 0)
                {
                    strcpy(safein, data);
                    break;
                }

                if (stricmp(line, "TicPath") == 0)
                {
                    strcpy(ticpath, data);
                    break;
                }

                if (stricmp(line, "BadTics") == 0)
                {
                    strcpy(badtics, data);
                    break;
                }

                break;
            case SECTION_NODES:
                if (sscanf(line, "%s %s %s", addr1, addr2, password) != 3) goto __err;
                if (!str2addr(addr1, &expaddr.zone, &expaddr.net, &expaddr.node, &expaddr.point)) goto __err;
                if (!str2addr(addr2, &ownaddr.zone, &ownaddr.net, &ownaddr.node, &ownaddr.point)) goto __err;

                tmpnode = new XNODE_REC;
                if (tmpnode == NULL)
                {
                    printf("Not enough memory to allocate %d bytes!\n", sizeof(XNODE_REC));
                    return 0;
                }
                if (firstnode == NULL) firstnode = tmpnode;
                if (noderec != NULL) noderec->next = tmpnode;
                tmpnode->next = NULL;

                memcpy(&tmpnode->expaddr, &expaddr, sizeof(expaddr));
                memcpy(&tmpnode->ownaddr, &ownaddr, sizeof(ownaddr));
                strcpy(tmpnode->password, password);
                noderec = tmpnode;
                break;
            case SECTION_AREAS:
                export[0] = '\0';
                if (sscanf(line, "%s %s %[^\n]", tag, path, export) < 3) goto __err;

                /* Add to linked list */
                xtmp = new XFIX_REC;
                if (xtmp == NULL)
                {
                    printf("Not enough memory to allocate %d bytes!\n", sizeof(XFIX_REC));
                    return 0;
                }
                if (xfirst == NULL) xfirst = xtmp;
                if (xrec != NULL) xrec->next = xtmp;
                xtmp->next = NULL;

                xtmp->path = new char[strlen(path)+1];
                if (xtmp->path == NULL)
                {
                    printf("Not enough memory to allocate %d bytes!\n", sizeof(XFIX_REC));
                    return 0;
                }
                strcpy(xtmp->path, path);
                strcpy(xtmp->tag, tag);

                strp = strchr(export, '|');
                if (strp != NULL) *strp++ = '\0';
                xtmp->import = get_list(export, &stat);
                if (!stat) goto __err;

                if (strp == NULL)
                    xtmp->export = NULL;
                else
                {
                    while (*strp == ' ') strp++;
                    if (*strp != '\0')
                    {
                        xtmp->export = get_list(strp, &stat);
                        if (!stat) goto __err;
                    }
                }

                xrec = xtmp;
                break;
            case SECTION_HATCH:
                if (sscanf(line, "%s %s", tag, path) != 2) goto __err;

                /* Add to linked list */
                tmphatch = (HATCH_REC *) malloc(sizeof(HATCH_REC)+strlen(path)+1);
                if (tmphatch == NULL)
                {
                    printf("Not enough memory to allocate %d bytes!\n", sizeof(HATCH_REC)+strlen(path)+1);
                    return 0;
                }
                if (firsthatch == NULL) firsthatch = tmphatch;
                if (hatchrec != NULL) hatchrec->next = tmphatch;
                tmphatch->next = NULL; hatchrec = tmphatch;

                strcpy(hatchrec->area, tag);
                strcpy(hatchrec->fname, path);
                break;
            default:
                goto __err;
        }
    }

    if (inbound[0] == '\0')
    {
        printf("No inbound directory specified in configuration file!\n");
        return 0;
    }

    if (ticpath[0] == '\0')
    {
        printf("No tic directory specified in configuration file!\n");
        return 0;
    }

    if (badtics[0] == '\0')
    {
        printf("No bad tics directory specified in configuration file!\n");
        return 0;
    }

    return 1;
}
