#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include "files.h"

#ifndef __linux__
#  include <dos.h>  /* sleep() */
#else
#  define stricmp strcasecmp
#  define strnicmp strncasecmp
#endif

int copyfile(char *source, char *dest)
{
    FILE *Fsrc,*Fdst;
    char *buf;
    size_t readed;

    if ((Fsrc = fopen(source, "rb")) == NULL) return 0;
    if ((Fdst = fopen(dest, "w+b")) == NULL) { fclose(Fsrc); return 0; }
#ifdef __linux__
    chmod(dest, 0660);
#endif
    buf = new char[4096];

    do
    {
        readed = fread(buf,1,4096,Fsrc);
        fwrite(buf,1,readed,Fdst);
    }
    while (readed == 4096);

    delete[] buf;
    fclose(Fsrc); fclose(Fdst);

    return 1;
}

int lockfile(int handle)
{
    int tries;

    tries = 0;
    while (flock(handle, 0, 1) != 0)
    {
        tries++;
        if (tries == 20)
        {
            printf("\nPlease wait, trying to lock file...\n");
        }
        if (tries == 60*5)
        {
            printf("Okay! Now I've had it! Couldn't lock it in 5 minutes..\n");
        }
        sleep(1);
    }

    return 1;
}

/* Get string from file, remove all CRs and LFs */
char *_fgets(char *buf, size_t n, FILE *fp)
{
    char *strp;

    if (fgets(buf,n,fp) == NULL) return NULL;
    strp = buf;
    while (*strp != '\0')
    {
        if (*strp == '\r' || *strp == '\n')
        {
            *strp = '\0';
            break;
        }
        strp++;
    }
    return buf;
}

#ifdef __linux__
char *search_ign_file(char *fname)
{
    DIR *dirp;
    struct dirent *direntp;
    char *strp;

    /* Open directory */
    strp = strrchr(fname,SLASH);
    if (strp != NULL)
    {
        *strp = '\0';
        dirp = opendir(fname);
        *strp++ = SLASH;
    }
    else
    {
        strp = fname;
        dirp = opendir(".");
    }

    if (dirp == NULL) return NULL;

    /* Read directory */
    for (;;)
    {
        direntp = readdir(dirp);
        if (direntp == NULL) break;

        if (stricmp(direntp->d_name, strp) == 0)
        {
            /* Found file! */
            strcpy(strp, direntp->d_name);
            closedir(dirp);
            return fname;
        }
    }
    closedir(dirp);

    return NULL;
}
#endif

char *__fgets(char *buf, size_t n, int handle)
{
    char *strp;
    size_t readed,left;
    unsigned long fpos;

    fpos = lseek(handle,0,SEEK_CUR);
    readed = read(handle,buf,n);
    if (readed == 0) return NULL;

    strp = buf; left = readed;
    while (left != 0)
    {
        left--;
        if (*strp == '\r' || *strp == '\n')
        {
            if (*strp == '\r' && left > 0 && *(strp+1) == '\n') left--;
            *strp = '\0';
            break;
        }
        strp++;
    }

    if (left != 0) lseek(handle,fpos+readed-left,SEEK_SET);

    return buf;
}

int delete_space(int handle, long pos, long size, long fsize)
{
    char tmpbuf[128];
    char *buffer;

    int bufsize;
    long movesz, lpos;

    bufsize = 16384;
    movesz = fsize - pos - size;
    if (bufsize > movesz) bufsize = (int) movesz;
    if (bufsize == 0) return 0;

    if ((bufsize <= (int) sizeof(tmpbuf)) || ((buffer = new char[bufsize]) == NULL))
    {
        bufsize = sizeof(tmpbuf);
        buffer = tmpbuf;
    }

    for (lpos=1; lpos<=movesz/bufsize; lpos++)
    {
        lseek(handle, pos+size+(lpos-1)*bufsize, SEEK_SET);
        read(handle, buffer, bufsize);
        lseek(handle, pos+(lpos-1)*bufsize, SEEK_SET);
        if (write(handle, buffer, bufsize) != bufsize)
        {
            if (buffer != tmpbuf) delete[] buffer;
            return 0;
        }
    }

    if (movesz % bufsize > 0)
    {
        lseek(handle, pos+size+(movesz/bufsize)*bufsize,SEEK_SET);
        read(handle, buffer, (int) (movesz % bufsize));
        lseek(handle, pos+(movesz/bufsize)*bufsize, SEEK_SET);
        if (write(handle, buffer, (int) (movesz % bufsize)) != (int) (movesz % bufsize))
        {
            if (buffer != tmpbuf) delete[] buffer;
            return 0;
        }
    }

    if (buffer != tmpbuf) delete[] buffer;
    return 1;
}
