#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "config.h"
#include "files.h"
#include "ticks.h"

/* Compare if file matches search string */
int compare_filenames(char *fname, char *wildcard)
{
    char *strp,*strp2,*wptr;
    int quit,wild;

    if (wildcard[0] == '\0') return 1;

    while (wildcard != NULL)
    {
        wptr = strchr(wildcard,' ');
        if (wptr != NULL) *wptr = '\0';

        strp = fname;
        quit = 0;
        wild = 0;
        while (!quit)
        {
            switch(*wildcard)
            {
                case '\0':
                    if (wptr != NULL) *wptr = ' ';
                    if ((*strp != '\0') && (!wild)) return 0;
                    return 1;
                case '*':
                    strp2 = strrchr(fname,'.');
                    if ((strp2 == NULL) || (strp2 < strp))
                    {
                        strp = &fname[strlen(fname)];
                    }
                    else
                    {
                        strp = strp2;
                    }
                    wild = 1;
                    break;
                case '.':
                    if ((*strp != '.') && (*strp != '\0'))
                    {
                        quit = 1;
                        break;
                    }
                case '?':
                    if (*strp != '\0') strp++;
                    wild = 0;
                    break;
                default:
                    if (toupper(*strp) != toupper(*wildcard))
                    {
                        quit = 1;
                        break;
                    }
                    strp++;
                    wild = 0;
                    break;
            }
            wildcard++;
        }
        if (wptr != NULL)
        {
            *wptr = ' ';
            wildcard = wptr+1;
        }
        else
        {
            wildcard = NULL;
        }
    }

    return 0;
}

/* Create xfix.dat record */
int create_datrec(char *fname)
{
    DIR *dirp;
    struct dirent *direntp;
    struct stat statbuf;
    char tmp[256], *strp;
    XDAT_REC datrec;
    int handle;

    strp = strrchr(fname, SLASH);
    if (strp == NULL)
    {
        printf("create_datrec(%s) : NO SLASH ?? This should NEVER happen!!\n", fname);
        return 0;
    }
    *strp++ = '\0';

    sprintf(tmp, "%s"SSLASH"xfix.dat", hatchrec->fname);
    handle = creat(tmp, READWRITE);
    if (handle == -1)
    {
        printf("Can't create file %s\n", tmp);
        *(strp-1) = SLASH;
        return 0;
    }

    dirp = opendir(hatchrec->fname);
    if (dirp == NULL)
    {
        printf("Can't open directory %s\n", hatchrec->fname);
        *(strp-1) = SLASH;
        return 0;
    }

    for (;;)
    {
        direntp = readdir(dirp);
        if (direntp == NULL) break;

        sprintf(tmp, "%s"SSLASH"%s", hatchrec->fname, direntp->d_name);
        if (stat(tmp, &statbuf) == 0)
        {
            if (!ISDIR(statbuf))
            {
                memset(&datrec, 0, sizeof(datrec));
                strcpy(datrec.fname, direntp->d_name);
                datrec.time = statbuf.st_mtime;
                write(handle, &datrec, sizeof(datrec));
            }
        }
    }
    close(handle);
    closedir(dirp);
    *(strp-1) = SLASH;

    return 1;
}

/* Write record to xfix.dat */
int update_datrec(char *path, char *fname)
{
    char tmp[256];
    struct stat statbuf;
    XDAT_REC datrec;
    int Fdat;

    sprintf(tmp, "%s"SSLASH"%s", path, fname);
    if (stat(tmp, &statbuf) != 0) return 0;

    sprintf(tmp, "%s"SSLASH"xfix.dat", path);
    Fdat = sopen(tmp, O_RDWR | O_BINARY, SH_DENYRW);
    if (Fdat == -1)
    {
        Fdat = creat(tmp, READWRITE);
        if (Fdat == -1) return 0;
    }

    if (!lockfile(Fdat))
    {
        close(Fdat);
        printf("Can't lock file %s\n", tmp);
        return 0;
    }

    lseek(Fdat, 0, SEEK_END);

    memset(&datrec, 0, sizeof(datrec));
    strcpy(datrec.fname, fname);
    datrec.time = statbuf.st_mtime;
    write(Fdat, &datrec, sizeof(datrec));

    unlockfile(Fdat);
    close(Fdat);

    return 1;
}

/* Hatch manager */
int hatch_mgr(void)
{
    DIR *dirp;
    struct dirent *direntp;
    struct stat statbuf;
    int Fdat;
    char tmp[256], *strp, ontime;
    XDAT_REC datrec;
    unsigned long fsize, pos;

    hatchrec = firsthatch;
    while (hatchrec != NULL)
    {
        strp = strrchr(hatchrec->fname, SLASH);
        if (strp == NULL)
            printf("Invalid file name in hatch manager: %s\n", hatchrec->fname);
        else
        {
            *strp++ = '\0';
            sprintf(tmp, "%s"SSLASH"xfix.dat", hatchrec->fname);

            Fdat = sopen(tmp, O_RDWR | O_BINARY, SH_DENYRW);
            if (Fdat == -1)
            {
                /* No previous xfix.dat file, create it. */
                *(strp-1) = SLASH;
                if (!create_datrec(hatchrec->fname)) return 0;

                hatchrec = hatchrec->next;
                continue;
            }
            if (!lockfile(Fdat))
            {
                close(Fdat);
                printf("Can't lock file %s\n",tmp);
                return 0;
            }

            dirp = opendir(hatchrec->fname);
            if (dirp == NULL)
            {
                unlockfile(Fdat); close(Fdat);
                printf("Can't open directory %s\n", strp);
                return 0;
            }
            for (;;)
            {
                direntp = readdir(dirp);
                if (direntp == NULL) break;

                fsize = lseek(Fdat, 0, SEEK_END);
                if (compare_filenames(direntp->d_name, strp))
                {
                    lseek(Fdat, 0, SEEK_SET); pos = 0; ontime = 0;
                    while (read(Fdat, &datrec, sizeof(datrec)) == sizeof(datrec))
                    {
                        sprintf(tmp, "%s"SSLASH"%s", hatchrec->fname, datrec.fname);
                        if (stat(tmp, &statbuf) != 0)
                        {
                            /* Old information, remove this record. */
                            delete_space(Fdat, pos*sizeof(datrec), sizeof(datrec), fsize);
                            lseek(Fdat, pos*sizeof(datrec), SEEK_SET);
                            fsize -= sizeof(datrec);
                        }
                        else
                        {
#ifdef __linux__
                            /* Case sensitive compare under Linux */
                            if (strcmp(datrec.fname, direntp->d_name) == 0)
#else
                            if (stricmp(datrec.fname, direntp->d_name) == 0)
#endif
                            {
                                ontime = (statbuf.st_mtime == (time_t) datrec.time);
                                break;
                            }
                            pos++;
                        }
                        if (fsize <= pos*sizeof(datrec)) break;
                    }
                    ftrunc(Fdat, fsize);

                    if (!ontime)
                    {
                        /* File not sent yet, send it.. */
                        strcpy(datrec.fname, direntp->d_name);
                        datrec.time = statbuf.st_mtime;
                        sprintf(tmp, "%s"SSLASH"%s", hatchrec->fname, direntp->d_name);
                        if (create_tick(tmp, hatchrec->area))
                        {
                            /* Update xfix.dat record */
                            lseek(Fdat, pos*sizeof(datrec), SEEK_SET);
                            write(Fdat, &datrec, sizeof(datrec));
                        }
                    }
                }
            }
            closedir(dirp);
            *(strp-1) = SLASH;

            unlockfile(Fdat); close(Fdat);
        }
        hatchrec = hatchrec->next;
    }

    return 1;
}
