#include <stdio.h>
#include <string.h>
#include <time.h>

#include "config.h"
#include "files.h"
#include "outbound.h"
#include "filebase.h"
#include "hatch.h"
#include "crc32.h"

void delete_list(ADDR_REC **rec)
{
    ADDR_REC *tmp;

    while (*rec != NULL)
    {
        tmp = (*rec)->next;
        delete *rec;
        *rec = tmp;
    }
}

/* Find file area record */
int get_file_area(char *area)
{
    xrec = xfirst;
    while (xrec != NULL)
    {
        if (stricmp(area, xrec->tag) == 0) return 1;
        xrec = xrec->next;
    }

    return 0;
}

/* Find export address from node records */
int get_node_rec(ADDR_REC *addr)
{
    noderec = firstnode;
    while (noderec != NULL)
    {
        if (noderec->expaddr.zone == addr->zone &&
            noderec->expaddr.net == addr->net &&
            noderec->expaddr.node == addr->node &&
            noderec->expaddr.point == addr->point)
        {
            return 1;
        }
        noderec = noderec->next;
    }

    return 0;
}

/* Modify tick file for specific node */
int update_tick(FILE *Ftic, XNODE_REC *node, ADDR_REC *seenby, ADDR_REC *ticseen, int paths, char *fname)
{
    FILE *Fout;
    char tmp[257], tickname[256], addr[30], date[50];
    ADDR_REC *srec;
    int firstseen, ownaddr;
    struct stat statbuf;
    time_t tim;

    /* Check if our address is already in tic's seenby lines */
    ownaddr = 0;
    while (ticseen != NULL)
    {
        if (ticseen->zone == node->ownaddr.zone &&
            ticseen->net == node->ownaddr.net &&
            ticseen->node == node->ownaddr.node &&
            ticseen->point == node->ownaddr.point)
        {
            ownaddr = 1;
            break;
        }
        ticseen = ticseen->next;
    }

    tim = time(NULL);
    do
    {
        sprintf(tickname, "%s"SSLASH"xf%06lx.tic", ticpath, tim++ & 0x00ffffff);
    }
    while (stat(tickname, &statbuf) == 0);

    Fout = fopen(tickname, "w+b");
    if (Fout == NULL)
    {
        printf("Can't create file %s\n", tickname);
        return 0;
    }
#ifdef __linux__
    chmod(tickname, 0660);
#endif

    if (node->ownaddr.point == 0)
        sprintf(addr, "%u:%u/%u", node->ownaddr.zone, node->ownaddr.net, node->ownaddr.node);
    else
        sprintf(addr, "%u:%u/%u.%u", node->ownaddr.zone, node->ownaddr.net, node->ownaddr.node, node->ownaddr.point);

    fseek(Ftic, 0, SEEK_SET); firstseen = 1;
    while (_fgets(tmp, sizeof(tmp), Ftic) != NULL)
    {
        if (strnicmp(tmp, "Path ", 5) == 0)
        {
            paths--;
            if (paths == 0)
            {
                tim = time(NULL);
                fprintf(Fout, "%s\r\n", tmp);
                strcpy(date, ctime(&tim)); date[strlen(date)-1] = '\0';
                sprintf(tmp, "Path %s %lu %s", addr, tim+3600*24*365+3600*24*2, date);
            }
        }
        else if (strnicmp(tmp, "From ", 5) == 0)
        {
            sprintf(tmp, "From %s", addr);
        }
        else if (strnicmp(tmp, "Seenby ", 7) == 0)
        {
            if (firstseen)
            {
                /* Add our own Seenbys */
                if (!ownaddr)
                {
                    if (node->ownaddr.point == 0)
                        fprintf(Fout, "Seenby %u:%u/%u\r\n", node->ownaddr.zone, node->ownaddr.net, node->ownaddr.node);
                    else
                        fprintf(Fout, "Seenby %u:%u/%u.%u\r\n", node->ownaddr.zone, node->ownaddr.net, node->ownaddr.node, node->ownaddr.point);
                }

                srec = seenby; firstseen = 0;
                while (srec != NULL)
                {
                    if (srec->point == 0)
                        fprintf(Fout, "Seenby %u:%u/%u\r\n", srec->zone, srec->net, srec->node);
                    else
                        fprintf(Fout, "Seenby %u:%u/%u.%u\r\n", srec->zone, srec->net, srec->node, srec->point);
                    srec = srec->next;
                }
            }
        }
        else if (strnicmp(tmp, "Pw ", 3) == 0)
        {
            if (node->password[0] == '\0') continue;
            sprintf(tmp, "Pw %s", node->password);
        }
        fprintf(Fout, "%s\r\n", tmp);
    }
    fclose(Fout);

    update_outbound(tickname, fname, &node->expaddr);

    printf("Forwarded file %s to node %u:%u/%u.%u\n",
           fname, node->expaddr.zone, node->expaddr.net, node->expaddr.node, node->expaddr.point);

    return 1;
}

/* Forward tick file to export addresses */
int send_tick(char *tickname, char *fname)
{
    FILE *Ftic;
    char line[257], area[61], *strp;
    ADDR_REC *firstseen, *seenby, *tmpseen, *firstexport, *exprec;
    int paths;

    Ftic = fopen(tickname, "rb");
    if (Ftic == NULL)
    {
        printf("Can't open tick file %s\n", tickname);
        return 0;
    }

    firstseen = NULL; seenby = NULL; paths = 0;
    while (_fgets(line, sizeof(line), Ftic) != NULL)
    {
        strp = strchr(line, ' ');
        if (strp == NULL) continue;

        *strp++ = '\0';
        if (stricmp(line, "Area") == 0)
        {
            strcpy(area, strp);
            continue;
        }

        if (stricmp(line, "Path") == 0)
        {
            paths++;
            continue;
        }

        if (stricmp(line, "Seenby") == 0)
        {
            tmpseen = new ADDR_REC;
            if (firstseen == NULL) firstseen = tmpseen;
            if (seenby != NULL) seenby->next = tmpseen;
            tmpseen->next = NULL; seenby = tmpseen;

            str2addr(strp, &seenby->zone, &seenby->net, &seenby->node, &seenby->point);
            continue;
        }
    }

    if (!get_file_area(area))
    {
        printf("Unknown area %s\n", area);
        delete_list(&firstseen);
        return 0;
    }

    /* Build export list */
    seenby = xrec->export; firstexport = NULL; exprec = NULL;
    while (seenby != NULL)
    {
        tmpseen = firstseen;
        while (tmpseen != NULL)
        {
            if (tmpseen->zone == seenby->zone && tmpseen->net == seenby->net &&
                tmpseen->node == seenby->node && tmpseen->point == seenby->point)
            {
                break;
            }
            tmpseen = tmpseen->next;
        }
        if (tmpseen == NULL)
        {
            tmpseen = new ADDR_REC;
            if (tmpseen == NULL)
            {
                printf("Not enough memory to allocate %d bytes!\n", sizeof(ADDR_REC));
                delete_list(&firstexport);
                delete_list(&firstseen);
                return 0;
            }
            if (firstexport == NULL) firstexport = tmpseen;
            if (exprec != NULL) exprec->next = tmpseen;
            exprec = tmpseen; exprec->next = NULL;

            exprec->zone = seenby->zone; exprec->net = seenby->net;
            exprec->node = seenby->node; exprec->point = seenby->point;
        }
        seenby = seenby->next;
    }

    tmpseen = firstexport;
    while (tmpseen != NULL)
    {
        if (get_node_rec(tmpseen))
            update_tick(Ftic, noderec, firstexport, firstseen, paths, fname);
        else
            printf("Can't find export addres %u:%u/%u.%u from nodes!\n",
                   firstexport->zone, firstexport->net, firstexport->node, firstexport->point);

        tmpseen = tmpseen->next;
    }

    fclose(Ftic);
    delete_list(&firstseen);
    delete_list(&firstexport);

    return 1;
}

/* Toss .tic file into file area */
int process_tic(char *dir, char *ticname)
{
    FILE *Ftic;
    char *line, area[61], origin[40], fname[256], password[9], *desc, tmp[256];
    ADDR_REC from, *addr;
    struct stat statbuf;

    desc = new char[1024]; desc[0] = '\0';
    line = new char[1024];
    if (desc == NULL || line == NULL)
    {
        printf("Not enough memory to allocate 1024 bytes!\n");
        return 0;
    }

    sprintf(line, "%s"SSLASH"%s", dir, ticname);
    Ftic = fopen(line, "rt");
    if (Ftic == NULL)
    {
        printf("Can't open file %s\n", line);
        delete[] desc;
        delete[] line;
        return 0;
    }

    memset(&from, 0, sizeof(from));
    while (_fgets(line, 1024, Ftic) != NULL)
    {
        if (strnicmp(line, "Area ",5) == 0)
        {
            strcpy(area, line+5);
            continue;
        }

        if (strnicmp(line, "From ",5) == 0)
        {
            str2addr(line+5, &from.zone, &from.net, &from.node, &from.point);
            continue;
        }

        if (strnicmp(line, "Origin", 7) == 0)
        {
            strcpy(origin, line+7);
            continue;
        }

        if (strnicmp(line, "File ", 5) == 0)
        {
            strcpy(fname, line+5);
#ifdef __linux__
            char *strp;

            sprintf(tmp, "%s"SSLASH"%s", dir, fname);
            search_ign_file(tmp);
            strp = strrchr(tmp, SLASH);
            if (strp == NULL) strp = tmp; else strp++;
            strcpy(fname, strp);
#endif
            continue;
        }

        if (strnicmp(line, "Desc ", 5) == 0)
        {
            if (desc[0] == '\0') strcpy(desc, line+5);
            continue;
        }

        if (strnicmp(line, "Ldesc ", 6) == 0)
        {
            strcpy(desc, line+6);
            continue;
        }

        if (strnicmp(line, "Pw ", 3) == 0)
        {
            strcpy(password, line+3);
            continue;
        }
    }
    fclose(Ftic);

    sprintf(tmp, "%s"SSLASH"%s", dir, fname);
    if (stat(tmp, &statbuf) != 0)
    {
        printf("Can't find file %s\n", tmp);
        goto __err;
    }

    if (!get_file_area(area))
    {
        printf("Unknown area '%s'\n", area);
    __badtic:
        /* Copy .tic to bad */
        sprintf(line, "%s"SSLASH"%s", dir, ticname);
        sprintf(tmp, "%s"SSLASH"%s", badtics, ticname);
        if (rename(line, tmp) != 0)
            if (copyfile(line, tmp)) remove(line);

        /* Copy file to bad */
        sprintf(line, "%s"SSLASH"%s", dir, fname);
        sprintf(tmp, "%s"SSLASH"%s", badtics, fname);
        if (rename(line, tmp) != 0)
            if (copyfile(line, tmp)) remove(line);
    }
    else
    {
        printf("Tossing file %s in area %s\n", fname, area);

        if (!get_node_rec(&from))
        {
            printf("Unknown address %u:%u/%u.%u\n",
                   from.zone, from.net, from.node, from.point);
            goto __badtic;
        }

        /* Check if password is correct */
        if (stricmp(noderec->password, password) != 0)
        {
            printf("Invalid password %s <-> %s\n", password, noderec->password);
            goto __badtic;
        }

        /* Check if file importing is allowed from this node.. */
        addr = xrec->import;
        while (addr != NULL)
        {
            if (addr->zone == from.zone && addr->net == from.net &&
                addr->node == from.node && addr->point == from.point)
            {
                break;
            }
        }
        if (addr == NULL)
        {
            printf("Address %u:%u/%u.%u not found from import list\n",
                   from.zone, from.net, from.node, from.point);
            goto __badtic;
        }

        /* Copy file to right path */
        sprintf(line, "%s"SSLASH"%s", dir, fname);
        sprintf(tmp, "%s"SSLASH"%s", xrec->path, fname);
        if (rename(line, tmp) != 0)
            if (copyfile(line, tmp)) remove(line); else
            {
                printf("Can't copy file %s to %s\n", line, tmp);
            __err:
                delete[] desc;
                delete[] line;
                return 0;
            }

        update_datrec(xrec->path, fname);

        add_description(fname, desc, area);

        /* Send file to export nodes */
        sprintf(line, "%s"SSLASH"%s", dir, ticname);
        send_tick(line, tmp);

        /* Remove tick file */
        remove(line);
    }

    delete[] desc;
    delete[] line;

    return 1;
}

/* Create tick file */
int create_tick(char *fname, char *area)
{
    FILE *Ftic;
    char tmp[256], tickname[256], date[30], *strp;
    ADDR_REC *addr, *tmpaddr;
    struct stat statbuf;
    time_t _tim;

    if (!get_file_area(area))
    {
        printf("Unknown area %s\n", area);
        return 0;
    }

    addr = xrec->export;
    while (addr != NULL)
    {
        _tim = time(NULL);
        do
        {
            sprintf(tickname, "%s"SSLASH"xf%06lx.tic", ticpath, _tim++ & 0x00ffffff);
        }
        while (stat(tickname, &statbuf) == 0);

        Ftic = fopen(tickname, "w+b");
        if (Ftic == NULL)
        {
            printf("Can't create tick file %s\n", tmp);
            return 0;
        }
#ifdef __linux__
        chmod(tickname, 0660);
#endif

        if (!get_node_rec(addr))
        {
            printf("Can't find export addres %u:%u/%u.%u from nodes!\n",
                   addr->zone, addr->net, addr->node, addr->point);
            addr = addr->next;
            continue;
        }

        strp = strrchr(fname, SLASH);
        if (strp == NULL) strp = fname; else strp++;
        fprintf(Ftic, "Area %s\r\n", area);
        if (noderec->ownaddr.point == 0)
            sprintf(tmp, "%u:%u/%u", noderec->ownaddr.zone, noderec->ownaddr.net, noderec->ownaddr.node);
        else
            sprintf(tmp, "%u:%u/%u.%u", noderec->ownaddr.zone, noderec->ownaddr.net, noderec->ownaddr.node, noderec->ownaddr.point);
        fprintf(Ftic, "Origin %s\r\n", tmp);
        fprintf(Ftic, "From %s\r\n", tmp);
        fprintf(Ftic, "File %s\r\n", strp);
        fprintf(Ftic, "Desc %s\r\n", getdesc(fname));
        fprintf(Ftic, "Crc %08lX\r\n", filecrc(fname));
        fprintf(Ftic, "Created by "CREATED"\r\n");
        strcpy(date, ctime(&_tim)); date[strlen(date)-1] = '\0';
        fprintf(Ftic, "Path %s %lu %s\r\n", tmp, _tim+3600*24*365+3600*24*2, date);
        fprintf(Ftic, "Seenby %s\r\n", tmp);
        tmpaddr = xrec->export;
        while (tmpaddr != NULL)
        {
            if (tmpaddr->point == 0)
                fprintf(Ftic, "Seenby %u:%u/%u\r\n", tmpaddr->zone, tmpaddr->net, tmpaddr->node);
            else
                fprintf(Ftic, "Seenby %u:%u/%u.%u\r\n", tmpaddr->zone, tmpaddr->net, tmpaddr->node, tmpaddr->point);
            tmpaddr = tmpaddr->next;
        }
        fprintf(Ftic, "Pw %s\r\n", noderec->password);

        update_outbound(tickname, fname, addr);

        printf("Hatched file %s in area %s to node %u:%u/%u.%u\n",
               fname, area, addr->zone, addr->net, addr->node, addr->point);

        addr = addr->next;
    }

    return 1;
}
