#include <stdio.h>
#include <string.h>

#include "files.h"
#include "logfile.h"
#include "mareas.h"
#include "config.h"

#ifdef __linux__
#  define stricmp strcasecmp
#  define strnicmp strncasecmp
#endif

typedef struct AFIX_REC
{
    struct AFIX_REC *pNext;

    char szArea[61];
}
AFIX_REC;

extern unsigned short wOrigZone, wOrigNet, wOrigNode, wOrigPoint;
extern unsigned short wDestZone, wDestNet, wDestNode, wDestPoint;

static FILE *Ffix;

static AFIX_REC *afixrec, *first_afix;

int FindAddr(ADDR_REC *pList)
{
    while (pList != NULL)
    {
        if (pList->wZone == wOrigZone && pList->wNet == wOrigNet &&
            pList->wNode == wOrigNode && pList->wPoint == wOrigPoint)
        {
            /* Got it! */
            return 1;
        }
        pList = pList->pNext;
    }

    return 0;
}

int group_access(char *group)
{
    char *strp,gp[4];
    int pos;

    /* Check has node access to this message area */
    strp = pNodeRecord->pszAccessGroups; pos = 0;
    for (;;)
    {
        if (*strp == ',' || *strp == '\0')
        {
            gp[pos] = '\0'; pos = 0;
            if (stricmp(gp, group) == 0)
            {
                /* Found group! */
                return 1;
            }
            if (*strp == '\0') break;
        }
        gp[pos++] = *strp++;
        if (pos == 4) { gp[0] = '\0'; pos = 1; }
    }

    return 0;
}

void ProcessCommand(char *cmd)
{
    char tmp[256],*strp;

    if (strnicmp(cmd, "LIST", 4) == 0)
    {
        sprintf(tmp, "List of all areas available for node %u:%u/%u.%u", wOrigZone, wOrigNet, wOrigNode, wOrigPoint);
        fprintf(Ffix, "%s\r",tmp);
        strp = tmp; while (*strp != '\0') *strp++ = '-';
        fprintf(Ffix, "%s\r",tmp);
        pMsgAreaRecord = pFirstMsgArea;
        while (pMsgAreaRecord != NULL)
        {
            if (group_access(pMsgAreaRecord->szGroup))
                fprintf(Ffix, "%c%s\r", FindAddr(pMsgAreaRecord->pExport) ? '+' : ' ', pMsgAreaRecord->szAreaTag);
            pMsgAreaRecord = pMsgAreaRecord->pNext;
        }
        fprintf(Ffix, "\r");
    }

    if (strnicmp(cmd, "QUERY", 5) == 0)
    {
        sprintf(tmp, "List of areas connected for node %u:%u/%u.%u", wOrigZone, wOrigNet, wOrigNode, wOrigPoint);
        fprintf(Ffix, "%s\r",tmp);
        strp = tmp; while (*strp != '\0') *strp++ = '-';
        fprintf(Ffix, "%s\r",tmp);
        pMsgAreaRecord = pFirstMsgArea;
        while (pMsgAreaRecord != NULL)
        {
            if (FindAddr(pMsgAreaRecord->pExport))
                fprintf(Ffix, "%s\r", pMsgAreaRecord->szAreaTag);
            pMsgAreaRecord = pMsgAreaRecord->pNext;
        }
        fprintf(Ffix, "\r");
    }

    if (strnicmp(cmd, "UNLINKED", 8) == 0)
    {
        sprintf(tmp, "List of areas not connected for node %u:%u/%u.%u", wOrigZone, wOrigNet, wOrigNode, wOrigPoint);
        fprintf(Ffix, "%s\r",tmp);
        strp = tmp; while (*strp != '\0') *strp++ = '-';
        fprintf(Ffix, "%s\r",tmp);
        pMsgAreaRecord = pFirstMsgArea;
        while (pMsgAreaRecord != NULL)
        {
            if (group_access(pMsgAreaRecord->szGroup) && !FindAddr(pMsgAreaRecord->pExport))
                fprintf(Ffix, "%s\r", pMsgAreaRecord->szAreaTag);
            pMsgAreaRecord = pMsgAreaRecord->pNext;
        }
        fprintf(Ffix, "\r");
    }
}

void AreaCommand(char *name)
{
    AFIX_REC *tmpafix;

    if (strlen(name) > sizeof(tmpafix->szArea)-1)
    {
        printf("AreaFix : Area name too long: '%s', ignored\n",name);
        WriteLog('*',"AreaFix : Area name too long: '%s', ignored",name);
        return;
    }

    /* Add area change command to linked list */
    tmpafix = new AFIX_REC;
    if (tmpafix == NULL)
    {
        printf("AreaFix : AreaComand() : malloc ( %d ) : Not enough memory!\n", sizeof(AFIX_REC));
        return;
    }
    tmpafix->pNext = NULL;
    strcpy(tmpafix->szArea, name);
    if (afixrec != NULL) afixrec->pNext = tmpafix;
    if (first_afix == NULL) first_afix = tmpafix;
    afixrec = tmpafix;
}

int add_node(char *line, char *area)
{
    unsigned short zone, net, node, point;
    char addr[30],pos,group[4];

    /* Get group name */
    while (*line == ' ') line++;
    pos = 0;
    while (*line != ' ')
    {
        group[pos++] = *line++;
        if (pos == 4)
        {
            /* Too long group name! */
            return 0;
        }
    }
    group[pos] = '\0';
    while (*line == ' ') line++;

    if (!group_access(group))
    {
        /* No access */
        fprintf(Ffix, "No access to area: %s\r", area);
        return 0;
    }

    pos = 0;
    for (;;)
    {
        if (*line == ' ' || *line == '\0')
        {
            if (pos == 0)
            {
                if (*line == '\0') break;
                line++;
                continue;
            }

            /* Check if address is already connected.. */
            addr[pos] = '\0';
            if (str2addr(addr,&zone,&net,&node,&point) &&
                zone == wOrigZone && net == wOrigNet &&
                node == wOrigNode && point == wOrigPoint)
            {
                /* It is! */
                fprintf(Ffix, "Area already connected: %s\r", area);
                return 0;
            }

            if (*line == '\0') break;
            pos = 0; line++;
            continue;
        }
        addr[pos++] = *line++;
        if (pos == sizeof(addr)) pos = 0;
    }

    if (wOrigPoint == 0)
        sprintf(line," %u:%u/%u", wOrigZone, wOrigNet, wOrigNode);
    else
        sprintf(line," %u:%u/%u.%u", wOrigZone, wOrigNet, wOrigNode, wOrigPoint);

    fprintf(Ffix, "Area connected: %s\r", area);

    return 1;
}

int drop_node(char *line, char *area)
{
    unsigned short zone, net, node, point;
    char addr[30],pos;

    /* Bypass group name */
    while (*line == ' ') line++;
    while (*line != ' ') line++;
    while (*line == ' ') line++;

    pos = 0;
    for (;;)
    {
        if (*line == ' ' || *line == '\0')
        {
            /* Check if address found.. */
            addr[pos] = '\0';
            if (str2addr(addr,&zone,&net,&node,&point) &&
                zone == wOrigZone && net == wOrigNet &&
                node == wOrigNode && point == wOrigPoint)
            {
                /* Found! */
                memmove(line-pos-1,line,strlen(line)+1);
                fprintf(Ffix, "Area dropped: %s\r", area);
                return 1;
            }

            if (*line == '\0') break;
            pos = 0; line++;
            continue;
        }
        addr[pos++] = *line++;
        if (pos == sizeof(addr)) pos = 0;
    }

    fprintf(Ffix, "Area already dropped: %s\r", area);
    return 0;
}

void ProcessAreaCommands(void)
{
    FILE *Fin,*Fout;
    AFIX_REC *afix,*prev,*temp;
    char *line,*strp;

    fprintf(Ffix, "Area changes\r------------\r");

    /* Open configuration file */
    Fin = fopen("xtoss.cfg", "rt");
    if (Fin == NULL)
    {
        fprintf(Ffix, "\rProcessAreaCommands() : Couldn't open 'xtoss.cfg', can't configure areas!\r");
        printf("ProcessAreaCommands() : Couldn't open 'xtoss.cfg', can't configure areas!\n");
        WriteLog('!',"ProcessAreaCommands() : Couldn't open 'xtoss.cfg', can't configure areas!");
        return;
    }

    /* Create output configuration file */
    Fout = fopen("xtoss.cf$", "w+t");
    if (Fout == NULL)
    {
        fclose(Fin);
        fprintf(Ffix, "\rProcessAreaCommands() : Couldn't create 'xtoss.cf$', can't configure areas!\r");
        printf("ProcessAreaCommands() : Couldn't create 'xtoss.cf$', can't configure areas!\n");
        WriteLog('!',"ProcessAreaCommands() : Couldn't create 'xtoss.cf$', can't configure areas!");
        return;
    }

    line = new char[8192];
    if (line == NULL)
    {
        fclose(Fin);
        fclose(Fout);
        fprintf(Ffix, "\rProcessAreaCommands() : Not enough memory, can't configure areas!\r");
        printf("ProcessAreaCommands() : Not enough memory, can't configure areas!\n");
        WriteLog('!',"ProcessAreaCommands() : Not enough memory, can't configure areas!");
        return;
    }

    /* Seek to [export] section */
    for (;;)
    {
        if (_fgets(line,8192,Fin) == NULL)
        {
            /* Error! */
            fclose(Fin); fclose(Fout); delete[] line;
            fprintf(Ffix, "\rProcessAreaCommands() : Error in configuration file ?? Can't configure areas!\r");
            printf("ProcessAreaCommands() : [export] section not found ?? Error in configuration file ?? Can't configure areas!\n");
            WriteLog('!',"ProcessAreaCommands() : [export] section not found ?? Error in configuration file ?? Can't configure areas!");
            return;
        }

        fprintf(Fout, "%s\n", line);
        if (stricmp(line,"[export]") == 0) break;
    }

    while (_fgets(line,8192,Fin) != NULL)
    {
        if (line[0] != '\0' && line[0] != ';' && (strp = strchr(line,' ')) != NULL)
        {
            *strp = '\0';

            afix = first_afix; prev = NULL;
            while (afix != NULL)
            {
                if (stricmp(line,afix->szArea+1) == 0)
                {
                    if (afix->szArea[0] == '+')
                    {
                        /* Add node to export list */
                        add_node(strp+1, line);
                        if (prev == NULL)
                            first_afix = afix->pNext;
                        else
                            prev->pNext = afix->pNext;
                        temp = afix;
                        afix = afix->pNext;
                        delete temp;
                        continue;
                    }
                    else if (afix->szArea[0] == '-')
                    {
                        /* Drop node from export list */
                        drop_node(strp+1, line);
                        if (prev == NULL)
                            first_afix = afix->pNext;
                        else
                            prev->pNext = afix->pNext;
                        temp = afix;
                        afix = afix->pNext;
                        delete temp;
                        continue;
                    }
                    else
                    {
                        printf("Unknown command (This should NEVER happen!!): %s\n",afix->szArea);
                        WriteLog('!',"Unknown command (This should NEVER happen!!): %s",afix->szArea);
                    }
                }
                prev = afix; afix = afix->pNext;
            }
            *strp = ' ';
        }

        fprintf(Fout, "%s\n", line);
    }
    delete[] line;

    fclose(Fin);
    fclose(Fout);

    /* xtoss.cf$ -> xtoss.cfg -> xtoss.cf~ */
    remove("xtoss.cf~");
    rename("xtoss.cfg","xtoss.cf~");
    rename("xtoss.cf$","xtoss.cfg");

    /* Print areas not found */
    afix = first_afix;
    while (afix != NULL)
    {
        fprintf(Ffix, "Area not found: %s\r",afix->szArea+1);
        afix = afix->pNext;
    }

    fprintf(Ffix, "\r");
}

int AreaFixMessage(char *pszSubject)
{
    char c,*strp,buf[256];
    int iBufPos;
    MSG_TEXT *pMsgText;

    afixrec = NULL;
    first_afix = NULL;

    pAddrRecord = pFirstAddr;
    while (pAddrRecord != NULL)
    {
        if (pAddrRecord->wZone == wDestZone && pAddrRecord->wNet == wDestNet &&
            pAddrRecord->wNode == wDestNode && pAddrRecord->wPoint == wDestPoint)
        {
            /* Was ours, netmail in destination.. */
            break;
        }
        pAddrRecord = pAddrRecord->pNext;
    }

    if (pAddrRecord == NULL) return 0;

    /* Find node. */
    pNodeRecord = pFirstNode;
    while (pNodeRecord != NULL)
    {
        if (pNodeRecord->wZone == wOrigZone && pNodeRecord->wNet == wOrigNet &&
            pNodeRecord->wNode == wOrigNode && pNodeRecord->wPoint == wOrigPoint)
        {
            /* Was ours, netmail in destination.. */
            break;
        }
        pNodeRecord = pNodeRecord->pNext;
    }

    if (pNodeRecord == NULL)
    {
        printf("AreaFix message from unknown node: %u:%u/%u.%u\n", wOrigZone, wOrigNet, wOrigNode, wOrigPoint);
        WriteLog('*',"AreaFix message from unknown node: %u:%u/%u.%u", wOrigZone, wOrigNet, wOrigNode, wOrigPoint);
    }

    /* Check if password is correct. */
    c = pszSubject[strlen(pNodeRecord->szAreaFixPassword)];
    if (strnicmp(pNodeRecord->szAreaFixPassword, pszSubject, strlen(pNodeRecord->szAreaFixPassword)) != 0 || (c != ' ' && c != '\0'))
    {
        printf("Wrong password in AreaFix message from node: %u:%u/%u.%u: %s <-> %s\n", wOrigZone, wOrigNet, wOrigNode, wOrigPoint, pNodeRecord->szAreaFixPassword, pszSubject);
        WriteLog('*',"Wrong password in AreaFix message from node: %u:%u/%u.%u: %s <-> %s", wOrigZone, wOrigNet, wOrigNode, wOrigPoint, pNodeRecord->szAreaFixPassword, pszSubject);
    }

    Ffix = fopen("areafix.msg","w+b");
    if (Ffix == NULL)
    {
        printf("Can't create 'areafix.msg'\n");
        WriteLog('!',"Can't create `areafix.msg`");
    }

    printf("Processing AreaFix message from %u:%u/%u.%u\n", wOrigZone, wOrigNet, wOrigNode, wOrigPoint);
    WriteLog('+',"Processing AreaFix message from %u:%u/%u.%u", wOrigZone, wOrigNet, wOrigNode, wOrigPoint);

    pMsgText = pFirstMsgText; iBufPos = 0;
    while (pMsgText != NULL)
    {
        strp = pMsgText->sText;
        while (strp < pMsgText->sText+sizeof(pMsgText->sText))
        {
            if (*strp == '\0') break; /* EOT */
            if (*strp == 13) /* EOL */
            {
                buf[iBufPos] = '\0';
                strp++;
                if (iBufPos == 0 || buf[0] == 1)
                {
                    iBufPos = 0;
                    continue;
                }
                if (buf[0] == '%')
                    ProcessCommand(buf+1);
                else if (buf[0] == '-' || buf[0] == '+')
                    AreaCommand(buf);
                else
                {
                    /* Default: Add area */
                    memmove(buf+1, buf, strlen(buf)+1); buf[0] = '+';
                    AreaCommand(buf);
                }
                iBufPos = 0;
                continue;
            }
            buf[iBufPos++] = *strp++;
        }

        pMsgText = pMsgText->pNext;
    }

    if (afixrec != NULL)
    {
        /* Process area changing commands */
        ProcessAreaCommands();
    }

    if (ftell(Ffix) == 0)
    {
        fclose(Ffix);
        remove("areafix.msg");
        return 0;
    }

    fclose(Ffix);
    return 1;
}
