#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "os.h"
#include "files.h"
#include "config.h"
#include "version.h"

#ifdef VBOARD
#include "vboard.h"
#endif

#define SECTION_ADDRESSES     1
#define SECTION_PATHS         2
#define SECTION_PACKERS       3
#define SECTION_NODES         4
#define SECTION_EXPORT        5
#define SECTION_OTHER         6

int max_dupes;

int iAddresses;
ADDR_REC *pFirstAddr;
ADDR_REC *pAddrRecord;

int iNodes;
NODE_REC *pFirstNode;
NODE_REC *pNodeRecord;

int iPackers;
PACKER_REC *pFirstPacker;
PACKER_REC *pPackerRecord;

int iMsgAreas;
MSGAREA_REC *pFirstMsgArea;
MSGAREA_REC *pMsgAreaRecord;
MSGAREA_REC *pNetmailArea, *pBadMsgArea, *pDupeMsgArea;

char *pszDupeFile;
char *pszBaseFile;
char *pszInbound;
char *pszSafeIn;
char *pszMsgPath;
char *pszOutbound;
char *pszTemp;
char *pszLogFile;
char *pszSeqFile;
char *pszSpool;
char *pszExportList;
#ifdef VBOARD
char *pszVBLogFile;
#endif

char szMsgAreaPath[256];

char szTearLine[76];
char szOrigin[70];
char szOrganization[80];
char szDomain[80];
int low_flo, bink_outb, always_slash, domain_outbounds, no_outboundzones;
int fSavePacked;

#ifdef YAK
char YakData[256];
#else
int registered;
int checkkeyfile(char *regname, unsigned char *data);
char regname[100];
#endif

int str2addr(char *addrstrp, unsigned short *zone, unsigned short *net, unsigned short *node, unsigned short *point)
{
    char *pstr;
    char addrstr[50],pos;

    strcpy(addrstr,addrstrp); pos = 0;
    for (pstr = addrstr; ; pstr++)
    {
        if (*pstr >= '0' && *pstr <= '9') continue;
        if (*pstr == ':' && pos == 0) { *pstr = ' '; pos = 1; continue; }
        if (*pstr == '/' && pos == 1) { *pstr = ' '; pos = 2; continue; }
        if (*pstr == '.' && pos == 2) { *pstr = ' '; pos = 3; continue; }
        if (*pstr == '@') { *pstr = '\0'; break; }
        if (*pstr == '\0') break;
        printf("Error : str2addr(\"%s\") position %d ('%c')\n",addrstrp,(int) (pstr-addrstr)+1,*pstr);
        return 0;
    }

    *zone = 0; *net = 0; *node = 0; *point = 0;
    return sscanf(addrstr,"%hu %hu %hu %hu",zone,net,node,point) >= 3;
}

#ifdef VBOARD
int ReadVBoardConfig(char *pszFileName)
{
    PACKER Packer;
    PACKER_REC *pTempPacker;
    CONF Conf;
    MSGAREA_REC *pTempMsgArea;
    NCONFIG NodeConfig;
    char szString[256];
    int hConfig,i,iLen,iPosition;

    /* Open VBoard's configuration file */
    hConfig = sopen(pszFileName, O_RDONLY | O_BINARY, SH_DENYNO);
    if (hConfig == -1)
    {
        printf("File not found: '%s'\n", pszFileName);
        return 0;
    }

    /* Check this really is VBoard's file.. */
    read(hConfig,szString,10);
    if (strncmp(szString,"VBoard\n\r",8) != 0)
    {
        close(hConfig);
        printf("'%s': This really isn't VBoard configuration file?!?!\n",pszFileName);
        return 0;
    }

    /* Read node configuration */
    if (read(hConfig,&NodeConfig,sizeof(NCONFIG)) != sizeof(NCONFIG))
    {
        close(hConfig);
        printf("'%s': Error reading nconfig\n",pszFileName);
        return 0;
    }

    lseek(hConfig,10 + sizeof(NCONFIG) + sizeof(RANGES) +
          MAXEVENTS*sizeof(EVENT) + MAXPROTOCOLS*sizeof(PROTOCOL),SEEK_SET);

    /* Read packers */
    for (i = 0; i < MAXPACKERS; i++)
    {
        if (read(hConfig,&Packer,sizeof(PACKER)) != sizeof(PACKER)) break;
        if (Packer.Name[0] != '\0')
        {
            iPackers++;

            /* Allocate memory for packer record */
            pTempPacker = new PACKER_REC;
            if (pTempPacker == NULL)
            {
                /* Not enough memory */
                printf("ReadVBoardConfig() : packer : malloc( %d ) : Not enough memory!\n", sizeof(PACKER_REC));
                return 0;
            }

            /* Update linked list */
            if (pPackerRecord != NULL) pPackerRecord->pNext = pTempPacker;
            if (pFirstPacker == NULL) pFirstPacker = pTempPacker;
            pPackerRecord = pTempPacker;

            /* Update packer record */
            pPackerRecord->pNext = NULL;
            strcpy(pPackerRecord->szIdentify,(char *) Packer.IdBytes);
            /* Get packer string */
            iLen = strlen((char *) Packer.PackExec)+1;
            pPackerRecord->pszPackString = new char[iLen];
            if (pPackerRecord->pszPackString == NULL)
            {
                /* Not enough memory */
                printf("ReadVBoardConfig() : packer string : malloc( %d ) : Not enough memory!\n",iLen);
                return 0;
            }
            memcpy(pPackerRecord->pszPackString,Packer.PackExec,iLen);
            /* Get unpacker string */
            iLen = strlen((char *) Packer.UnPackExec);
            pPackerRecord->pszUnPackString = new char[iLen];
            if (pPackerRecord->pszUnPackString == NULL)
            {
                /* Not enough memory */
                printf("ReadVBoardConfig() : unpacker string : malloc( %d ) : Not enough memory!\n",iLen);
                return 0;
            }
            memcpy(pPackerRecord->pszUnPackString,Packer.UnPackExec,iLen);
        }
    }

    close(hConfig);

    /* No MainDir set? */
    if (NodeConfig.MainDir[0] == '\0')
    {
        printf("'%s': Invalid MainDir\n",pszFileName);
        return 0;
    }

    /* Get message area configuration file name */
#ifndef __linux__
    if (NodeConfig.MainDir[strlen((char *) NodeConfig.MainDir)-1] != SLASH)
        sprintf(szMsgAreaPath,"%s"SSLASH, NodeConfig.MainDir);
    else
        strcpy(szMsgAreaPath, (char *) NodeConfig.MainDir);
#else
    strcpy(szMsgAreaPath,"/vboard/main/");
#endif
    sprintf(szString,"%sconfinfo.dat",szMsgAreaPath);

    /* Open confinfo.dat */
    hConfig = sopen(szString,O_RDONLY | O_BINARY, SH_DENYNO);
    if (hConfig == -1)
    {
        printf("Could not open message area configuration file '%s'\n",szString);
        return 0;
    }

    /* Read message areas */
    iPosition = 0;
    while (read(hConfig,&Conf,sizeof(CONF)) == sizeof(CONF))
    {
        iPosition++;

        /* Skip local areas */
        if ((Conf.Flags2 & CONFFLAG2_NETAREA) == 0) continue;

        iMsgAreas++;

        /* Allocate memory for msg area record */
        pTempMsgArea = new MSGAREA_REC;
        if (pTempMsgArea == NULL)
        {
            /* Not enough memory */
            printf("ReadVBoardConfig() : msg area : malloc( %d ) : Not enough memory!\n", sizeof(MSGAREA_REC));
            return 0;
        }

        /* Update linked list */
        if (pMsgAreaRecord != NULL) pMsgAreaRecord->pNext = pTempMsgArea;
        if (pFirstMsgArea == NULL) pFirstMsgArea = pTempMsgArea;
        pMsgAreaRecord = pTempMsgArea;

        /* Update address record */
        pMsgAreaRecord->pNext = NULL;
        strcpy(pMsgAreaRecord->szAreaTag, (char *) Conf.AreaTag);
        strcpy(pMsgAreaRecord->szOrigin, (char *) Conf.OriginLine);
        pMsgAreaRecord->szGroup[0] = '\0';
        pMsgAreaRecord->wAreaNum = (unsigned short) (iPosition-1);
        pMsgAreaRecord->OwnAddr.wZone = Conf.Own.Zone;
        pMsgAreaRecord->OwnAddr.wNet = Conf.Own.Net;
        pMsgAreaRecord->OwnAddr.wNode = Conf.Own.Node;
        pMsgAreaRecord->OwnAddr.wPoint = Conf.Own.Point;
        pMsgAreaRecord->pExport = NULL;

        if (pNetmailArea == NULL && Conf.Flags1 & CONFFLAG1_POSTCONF)
        {
            /* Netmail area found */
            pNetmailArea = pMsgAreaRecord;
        }

        if (pDupeMsgArea == NULL && stricmp((char *) Conf.AreaTag,"DUPEECHO") == 0)
        {
            /* Dupe area found */
            pDupeMsgArea = pMsgAreaRecord;
        }

        if (pBadMsgArea == NULL && stricmp((char *) Conf.AreaTag,"BADECHO") == 0)
        {
            /* Bad area found */
            pBadMsgArea = pMsgAreaRecord;
        }
    }

    close(hConfig);

    return 1;
}
#endif

#ifdef YAK

char net_number[50][24];

int ReadYAKConfig(char *pszFileName, char *pszLine)
{
    FILE *F;
    MSGAREA_REC *pTempMsgArea;
    char *strp, *path, *tag, *ownaddr, ok, msgpath[256];
    unsigned short zone, net, node, point;
    int addr, sect, nets;

    msgpath[0] = '\0';
    F = fopen(pszFileName, "rb");
    if (F == NULL)
    {
        printf("YAK configuration file not found: '%s'\n", pszFileName);
        return 0;
    }

    sect = 0; nets = 0; YakData[0] = '\0';
    while (_fgets(pszLine, 8192, F) != NULL)
    {
        if (pszLine[0] == ';' || pszLine[0] == '\0') continue;
        if (pszLine[0] == '[')
        {
            if (stricmp(pszLine, "[paths]") == 0) sect = 1;
            else if (stricmp(pszLine, "[addresses]") == 0) sect = 2;
            else sect = 3;
            continue;
        }

        if (sect == 1 && strnicmp(pszLine, "MsgAreas", 8) == 0)
        {
            strp = pszLine+8;
            while (*strp == ' ') strp++;
            if (*strp == '=')
            {
                strp++;
                while (*strp == ' ') strp++;
                strcpy(msgpath, strp);
            }
        }
        else if (sect == 1 && strnicmp(pszLine, "DataPath", 8) == 0)
        {
            strp = pszLine+8;
            while (*strp == ' ') strp++;
            if (*strp == '=')
            {
                strp++;
                while (*strp == ' ') strp++;
                strcpy(YakData, strp);
            }
        }
        else if (sect == 2) strcpy(net_number[nets++], pszLine);
    }

    fclose(F);

    /* Open YAK's message area configuration file */
    F = fopen(msgpath, "rb");
    if (F == NULL)
    {
        printf("YAK message area configuration file not found: '%s'\n", msgpath);
        return 0;
    }

    while (_fgets(pszLine, 8192, F) != NULL)
    {
        if (*pszLine == '\0' || *pszLine == ';') continue;

        /* Get area path */
        strp = pszLine; while (*strp == ' ') strp++;
        path = strp;
        strp = strchr(strp, ' '); if (strp == NULL) continue;
        *strp++ = '\0';
        while (*strp == ' ') strp++;

        /* Get area tag */
        tag = strp;
        strp = strchr(strp, ' '); if (strp == NULL) continue;
        *strp++ = '\0';
        while (*strp == ' ') strp++;

        /* Skip BW tag */
        while (*strp != ' ') strp++;
        while (*strp == ' ') strp++;

        /* Get own address */
        ownaddr = strp;
        strp = strchr(strp, ' '); if (strp == NULL) continue;
        *strp++ = '\0';
        while (*strp == ' ') strp++;

        /* Skip read access group */
        while (*strp != ' ') strp++;
        while (*strp == ' ') strp++;

        /* Skip write access group */
        while (*strp != ' ') strp++;
        while (*strp == ' ') strp++;

        /* Get flags */
        ok = 0;
        while (*strp != '\0' && *strp != ' ')
        {
            if (toupper(*strp) == 'N') { ok = 2; break; }
            if (toupper(*strp) == 'U') { ok = 1; break; }
            if (toupper(*strp) == 'E') { ok = 1; break; }
            strp++;
        }
        if (!ok) continue; /* Local area */

        iMsgAreas++;

        /* Allocate memory for msg area record */
        pTempMsgArea = new MSGAREA_REC;
        if (pTempMsgArea == NULL)
        {
            /* Not enough memory */
            printf("ReadYAKConfig() : msg area : malloc( %d ) : Not enough memory!\n", sizeof(MSGAREA_REC));
            return 0;
        }

        /* Update linked list */
        if (pMsgAreaRecord != NULL) pMsgAreaRecord->pNext = pTempMsgArea;
        if (pFirstMsgArea == NULL) pFirstMsgArea = pTempMsgArea;
        pMsgAreaRecord = pTempMsgArea;

        if (ok == 1)
        {
            addr = atoi(ownaddr);
            if (addr == 0) continue;
            str2addr(net_number[addr-1], &zone, &net, &node, &point);
        }

        /* Update address record */
        pMsgAreaRecord->pNext = NULL;
        strcpy(pMsgAreaRecord->szAreaTag, tag);
        pMsgAreaRecord->szGroup[0] = '\0';
        pMsgAreaRecord->pszPath = new char[strlen(path)+1];
        if (pMsgAreaRecord->pszPath == NULL)
        {
            /* Not enough memory */
            printf("ReadYAKConfig() : msg area path : malloc( %d ) : Not enough memory!\n", strlen(path)+1);
            return 0;
        }
        strcpy(pMsgAreaRecord->pszPath, path);
        pMsgAreaRecord->OwnAddr.wZone = zone;
        pMsgAreaRecord->OwnAddr.wNet = net;
        pMsgAreaRecord->OwnAddr.wNode = node;
        pMsgAreaRecord->OwnAddr.wPoint = point;
        pMsgAreaRecord->pExport = NULL;

        if (pNetmailArea == NULL && ok == 2)
        {
            /* Netmail area found */
            pNetmailArea = pMsgAreaRecord;
        }

        if (pDupeMsgArea == NULL && stricmp(tag,"DUPEECHO") == 0)
        {
            /* Dupe area found */
            pDupeMsgArea = pMsgAreaRecord;
        }

        if (pBadMsgArea == NULL && stricmp(tag,"BADECHO") == 0)
        {
            /* Bad area found */
            pBadMsgArea = pMsgAreaRecord;
        }
    }

    fclose(F);

    return 1;
}
#endif

int ReadConfig(char *pszFileName)
{
    FILE *hConfig;
    char *szLine,*pStr,*pszData,*pszAccessGroups,*pStr2;
    char szPassword[9],szAreaFixPassword[21];
    unsigned long dwLine;
    int iSubConfig,iDomainLen,iDataLen,iPacker;
    unsigned short wZone,wNet,wNode,wPoint;
    unsigned short wYourZone,wYourNet,wYourNode,wYourPoint;
    struct stat statbuf;

    PACKER_REC *pTempPacker;
    ADDR_REC *pTempAddr;
    NODE_REC *pTempNode;

    /* Open configuration file */
    hConfig = fopen(pszFileName, "rt");
    if (hConfig == NULL)
    {
        printf("Configuration file %s not found.\n", pszFileName);
        return 0;
    }

    /* Initialize variables */
    iSubConfig = 0; dwLine = 0; iMsgAreas = 0; szTearLine[0] = '\0';
    low_flo = 0; bink_outb = 1; always_slash = 0; domain_outbounds = 0;
    no_outboundzones = 0; fSavePacked = 1;

    pFirstAddr = NULL; pAddrRecord = NULL;
    pFirstNode = NULL; pNodeRecord = NULL;
    pFirstPacker = NULL; pPackerRecord = NULL;
    pFirstMsgArea = NULL; pMsgAreaRecord = NULL;
    pNetmailArea = NULL; pBadMsgArea = NULL; pDupeMsgArea = NULL;

    iAddresses = 0; iNodes = 0; iPackers = 0;

    pszInbound = NULL; pszSafeIn = NULL; pszOutbound = NULL; pszTemp = NULL;
    pszLogFile = NULL; pszBaseFile = NULL; pszDupeFile = NULL; pszSpool = NULL;
    pszSeqFile = NULL; pszExportList = NULL; pszMsgPath = NULL;
#ifdef VBOARD
    pszVBLogFile = NULL; registered = 0;
#endif

    /* Read configuration */
    szLine = new char[8192];
    if (szLine == NULL)
    {
        printf("ReadConfig() : malloc( 8192 ) : Not enough memory!\n");
        return 0;
    }

    while (_fgets(szLine,8192,hConfig) != NULL)
    {
        dwLine++;

        /* Remove trailing spaces */
        pStr = szLine+strlen(szLine)-1;
        while (pStr > szLine && *pStr == ' ') pStr--;
        *(pStr+1) = '\0';

        /* Skip empty lines and comments */
        if (szLine[0] == '\0' || szLine[0] == ';') continue;

        if (szLine[0] == '[')
        {
            /* New section */
            if (stricmp(szLine,"[addresses]") == 0)
            {
                iSubConfig = SECTION_ADDRESSES;
                continue;
            }

            if (stricmp(szLine,"[paths]") == 0)
            {
                iSubConfig = SECTION_PATHS;
                continue;
            }

            if (stricmp(szLine,"[packers]") == 0)
            {
                /* Release memory used by packers */
                while (pFirstPacker != NULL)
                {
                    pPackerRecord = pFirstPacker->pNext;
                    delete[] pFirstPacker->pszPackString;
                    delete[] pFirstPacker->pszUnPackString;
                    delete pFirstPacker;
                    pFirstPacker = pPackerRecord;
                }
                iPackers = 0;
                iSubConfig = SECTION_PACKERS;
                continue;
            }

            if (stricmp(szLine,"[nodes]") == 0)
            {
                iSubConfig = SECTION_NODES;
                continue;
            }

            if (stricmp(szLine,"[export]") == 0)
            {
                iSubConfig = SECTION_EXPORT;
                continue;
            }

            if (stricmp(szLine,"[other]") == 0)
            {
                iSubConfig = SECTION_OTHER;
                continue;
            }

            goto __err;
        }

        switch (iSubConfig)
        {
#ifdef VBOARD
            case 0:
                pStr = strchr(szLine,'=');
                if (pStr == NULL) goto __err;
                pszData = pStr+1;
                pStr--;

                /* Remove spaces before '=' */
                while (pStr >= szLine && *pStr == ' ') pStr--;
                *(pStr+1) = '\0';

                /* Remove spaces after '=' */
                while (*pszData == ' ') pszData++;
                *(pszData-1) = '\0';

                if (*pszData == '\0' || *szLine == '\0') goto __err;

                if (stricmp(szLine,"RegisterKey") == 0)
                {
                    registered = checkkeyfile(regname, (unsigned char *) pszData);
                    break;
                }
                goto __err;
#endif
            case SECTION_ADDRESSES:
                /* New address */
#ifdef VBOARD
                if (!registered && iAddresses > 0)
                {
                    printf("Only one node address is allowed in unregistered version (register or suffer ;)\n");
                    return 0;
                }
#endif
                iAddresses++;

                /* Get domain */
                pStr = strchr(szLine,'@');
                if (pStr == NULL)
                    iDomainLen = 0;
                else
                {
                    *pStr++ = '\0';
                    iDomainLen = strlen(pStr);
                }

                /* Allocate memory for address */
                pTempAddr = (ADDR_REC *) malloc(sizeof(ADDR_REC)+iDomainLen+1);
                if (pTempAddr == NULL)
                {
                    /* Not enough memory */
                    printf("SECTION_ADDRESSES : malloc( %d ) : Not enough memory!\n",sizeof(ADDR_REC)+iDomainLen+1);
                    return 0;
                }

                /* Update linked list */
                if (pAddrRecord != NULL) pAddrRecord->pNext = pTempAddr;
                if (pFirstAddr == NULL) pFirstAddr = pTempAddr;
                pAddrRecord = pTempAddr;

                /* Update address record */
                pAddrRecord->pNext = NULL;
                if (!str2addr(szLine,&pAddrRecord->wZone,&pAddrRecord->wNet,&pAddrRecord->wNode,&pAddrRecord->wPoint)) goto __err;
                if (pStr == NULL)
                    pAddrRecord->szDomain[0] = '\0';
                else
                    strcpy(pAddrRecord->szDomain,pStr);
                break;
            case SECTION_PATHS:
                /* Path configuration */
                pStr = strchr(szLine,'=');
                if (pStr == NULL) goto __err;
                pszData = pStr+1;
                pStr--;

                /* Remove spaces before '=' */
                while (pStr >= szLine && *pStr == ' ') pStr--;
                *(pStr+1) = '\0';

                /* Remove spaces after '=' */
                while (*pszData == ' ') pszData++;
                *(pszData-1) = '\0';

                if (*pszData == '\0' || *szLine == '\0') goto __err;

#ifdef VBOARD
                if (stricmp(szLine, "VBoardConfig") == 0)
                {
                    /* Read VBoard's config files */
                    if (!ReadVBoardConfig(pszData))
                    {
                        printf("Error reading VBoard's configuration files\n");
                        return 0;
                    }
                    break;
                }
#endif

#ifdef YAK
                if (stricmp(szLine, "YAKConfig") == 0)
                {
                    /* Read VBoard's config files */
                    if (!ReadYAKConfig(pszData, szLine))
                    {
                        printf("Error reading YAK's configuration files\n");
                        return 0;
                    }
                    break;
                }
#endif

                if (stricmp(szLine,"Inbound") == 0)
                {
                    /* Inbound path */
                    iDataLen = strlen(pszData)+1;
                    if (iDataLen > 1 && pszData[iDataLen-2] != SLASH)
                    {
                        /* Add slash */
                        pszData[iDataLen-1] = SLASH;
                        iDataLen++;
                    }
                    pszInbound = new char[iDataLen];
                    if (pszInbound == NULL)
                    {
                        /* Not enough memory */
                        printf("SECTION_PATHS : Inbound : malloc( %d ) : Not enough memory!\n",iDataLen);
                        return 0;
                    }
                    memcpy(pszInbound, pszData, iDataLen);
                    break;
                }

                if (stricmp(szLine,"UUCPSeqFile") == 0)
                {
                    /* UUCP seq. file */
                    iDataLen = strlen(pszData)+1;
                    pszSeqFile = new char[iDataLen];
                    if (pszSeqFile == NULL)
                    {
                        /* Not enough memory */
                        printf("SECTION_PATHS : SeqFile : malloc( %d ) : Not enough memory!\n",iDataLen);
                        return 0;
                    }
                    memcpy(pszSeqFile, pszData, iDataLen);
                    break;
                }

                if (stricmp(szLine,"UUCPSpoolPath") == 0)
                {
                    /* UUCP spool path */
                    iDataLen = strlen(pszData)+1;
                    if (iDataLen > 1 && pszData[iDataLen-2] != SLASH)
                    {
                        /* Add slash */
                        pszData[iDataLen-1] = SLASH;
                        iDataLen++;
                    }
                    pszSpool = new char[iDataLen];
                    if (pszSpool == NULL)
                    {
                        /* Not enough memory */
                        printf("SECTION_PATHS : Spool : malloc( %d ) : Not enough memory!\n",iDataLen);
                        return 0;
                    }
                    memcpy(pszSpool, pszData, iDataLen);
                    break;
                }

                if (stricmp(szLine,"EMailExportList") == 0)
                {
                    /* E-Mail export list */
                    iDataLen = strlen(pszData)+1;
                    pszExportList = new char[iDataLen];
                    if (pszExportList == NULL)
                    {
                        /* Not enough memory */
                        printf("SECTION_PATHS : E-Mail export list : malloc( %d ) : Not enough memory!\n",iDataLen);
                        return 0;
                    }
                    memcpy(pszExportList, pszData, iDataLen);
                    break;
                }

                if (stricmp(szLine,"BaseFile") == 0)
                {
                    /* Base File name */
                    iDataLen = strlen(pszData)+1;
                    pszBaseFile = new char[iDataLen];
                    if (pszBaseFile == NULL)
                    {
                        /* Not enough memory */
                        printf("SECTION_PATHS : BaseFile : malloc( %d ) : Not enough memory!\n",iDataLen);
                        return 0;
                    }
                    memcpy(pszBaseFile, pszData, iDataLen);
                    break;
                }

                if (stricmp(szLine,"DupeFile") == 0)
                {
                    /* Inbound path */
                    iDataLen = strlen(pszData)+1;
                    pszDupeFile = new char[iDataLen];
                    if (pszDupeFile == NULL)
                    {
                        /* Not enough memory */
                        printf("SECTION_PATHS : DupeFile : malloc( %d ) : Not enough memory!\n",iDataLen);
                        return 0;
                    }
                    memcpy(pszDupeFile, pszData, iDataLen);
                    break;
                }

                if (stricmp(szLine,"SafeIn") == 0)
                {
                    /* Inbound path */
                    iDataLen = strlen(pszData)+1;
                    if (iDataLen > 1 && pszData[iDataLen-2] != SLASH)
                    {
                        /* Add slash */
                        pszData[iDataLen-1] = SLASH;
                        iDataLen++;
                    }
                    pszSafeIn = new char[iDataLen];
                    if (pszSafeIn == NULL)
                    {
                        /* Not enough memory */
                        printf("SECTION_PATHS : SafeIn : malloc( %d ) : Not enough memory!\n",iDataLen);
                        return 0;
                    }
                    memcpy(pszSafeIn, pszData, iDataLen);
                    break;
                }

                if (stricmp(szLine,"MsgPath") == 0)
                {
                    /* .msg netmail path */
                    iDataLen = strlen(pszData)+1;
                    if (iDataLen > 1 && pszData[iDataLen-2] != SLASH)
                    {
                        /* Add slash */
                        pszData[iDataLen-1] = SLASH;
                        iDataLen++;
                    }
                    pszMsgPath = new char[iDataLen];
                    if (pszMsgPath == NULL)
                    {
                        /* Not enough memory */
                        printf("SECTION_PATHS : MsgPath : malloc( %d ) : Not enough memory!\n",iDataLen);
                        return 0;
                    }
                    memcpy(pszMsgPath, pszData, iDataLen);
                    break;
                }

                if (stricmp(szLine,"Outbound") == 0)
                {
                    /* Outbound path */
                    iDataLen = strlen(pszData)+1;
                    if (iDataLen > 1 && pszData[iDataLen-2] == SLASH)
                    {
                        /* Remove slash */
                        iDataLen--;
                        pszData[iDataLen-1] = '\0';
                    }
                    pszOutbound = new char[iDataLen];
                    if (pszOutbound == NULL)
                    {
                        /* Not enough memory */
                        printf("SECTION_PATHS : Outbound : malloc( %d ) : Not enough memory!\n",iDataLen);
                        return 0;
                    }
                    memcpy(pszOutbound, pszData, iDataLen);
                    break;
                }

                if (stricmp(szLine,"Temp") == 0)
                {
                    /* Temp path */
                    iDataLen = strlen(pszData)+1;
                    if (iDataLen > 1 && pszData[iDataLen-2] != SLASH)
                    {
                        /* Add slash */
                        pszData[iDataLen-1] = SLASH;
                        iDataLen++;
                    }
                    pszTemp = new char[iDataLen];
                    if (pszTemp == NULL)
                    {
                        /* Not enough memory */
                        printf("SECTION_PATHS : Temp : malloc( %d ) : Not enough memory!\n",iDataLen);
                        return 0;
                    }
                    memcpy(pszTemp, pszData, iDataLen);
                    pszData[iDataLen-1] = '\0';
                    if (stat(pszData, &statbuf) != 0) MakeDir(pszData);
                    break;
                }

                if (stricmp(szLine,"LogFile") == 0)
                {
                    /* Log file */
                    iDataLen = strlen(pszData)+1;
                    pszLogFile = new char[iDataLen];
                    if (pszLogFile == NULL)
                    {
                        /* Not enough memory */
                        printf("SECTION_PATHS : LogFile : malloc( %d ) : Not enough memory!\n", iDataLen);
                        return 0;
                    }
                    memcpy(pszLogFile, pszData, iDataLen);
                    break;
                }

#ifdef VBOARD
                if (stricmp(szLine,"VBLogFile") == 0)
                {
                    /* Log file */
                    iDataLen = strlen(pszData)+1;
                    pszVBLogFile = new char[iDataLen];
                    if (pszVBLogFile == NULL)
                    {
                        /* Not enough memory */
                        printf("SECTION_PATHS : VBLogFile : malloc( %d ) : Not enough memory!\n", iDataLen);
                        return 0;
                    }
                    memcpy(pszVBLogFile, pszData, iDataLen);
                    break;
                }
#endif
                goto __err;
            case SECTION_PACKERS:
                /* New packer */
                iPackers++;

                /* Allocate memory for packer */
                pTempPacker = new PACKER_REC;
                if (pTempPacker == NULL)
                {
                    /* Not enough memory */
                    printf("SECTION_PACKERS : malloc( %d ) : Not enough memory!\n",sizeof(PACKER_REC));
                    return 0;
                }

                /* Update linked list */
                if (pPackerRecord != NULL) pPackerRecord->pNext = pTempPacker;
                if (pFirstPacker == NULL) pFirstPacker = pTempPacker;
                pPackerRecord = pTempPacker;
                pPackerRecord->pNext = NULL;

                /* Get identify string */
                pStr = strchr(szLine,' ');
                if (pStr == NULL) goto __err;
                *pStr = '\0'; while (*(pStr+1) == ' ') pStr++;
                if (strlen(szLine) > sizeof(pPackerRecord->szIdentify)-1)
                {
                    printf("Error in configuration file '%s' line %lu: Packer identifer too long\n",pszFileName,dwLine);
                    return 0;
                }
                strcpy(pPackerRecord->szIdentify, szLine);
                pszData = pStr+1;

                /* Get packer command line */
                pStr = strchr(pszData,'|');
                if (pStr == NULL) goto __err;
                pStr2 = pStr; while (*(pStr+1) == ' ') pStr++;
                while (*(pStr2-1) == ' ') pStr2--; *pStr2 = '\0';
                pPackerRecord->pszPackString = new char[strlen(pszData)+1];
                if (pPackerRecord->pszPackString == NULL)
                {
                    printf("SECTION_PACKERS : pack string : malloc( %d ) : Not enough memory!\n", strlen(pszData)+1);
                    return 0;
                }
                strcpy(pPackerRecord->pszPackString, pszData);
                pszData = pStr+1;
                while (*(pStr-1) == ' ') pStr--; *pStr = '\0';

                /* Get unpacker command line */
                pStr = pszData+strlen(pszData);
                while (*(pStr-1) == ' ') pStr--; *pStr = '\0';
                pPackerRecord->pszUnPackString = new char[strlen(pszData)+1];
                if (pPackerRecord->pszUnPackString == NULL)
                {
                    printf("SECTION_PACKERS : unpack string : malloc( %d ) : Not enough memory!\n", strlen(pszData)+1);
                    return 0;
                }
                strcpy(pPackerRecord->pszUnPackString, pszData);
                break;
            case SECTION_NODES:
                /* New node */
                iNodes++;

                /* Get address */
                pStr = strchr(szLine,' ');
                if (pStr == NULL) goto __err;
                *pStr = '\0'; while (*(pStr+1) == ' ') pStr++;
                pszData = pStr+1;
                if (!str2addr(szLine,&wZone,&wNet,&wNode,&wPoint)) goto __err;

                /* Get your address */
                pStr = strchr(pszData,' ');
                if (pStr == NULL) goto __err;
                *pStr = '\0'; while (*(pStr+1) == ' ') pStr++;
                if (!str2addr(pszData,&wYourZone,&wYourNet,&wYourNode,&wYourPoint)) goto __err;
                pszData = pStr+1;

                /* Get packer number */
                pStr = strchr(pszData,' ');
                if (pStr == NULL) goto __err;
                *pStr = '\0'; while (*(pStr+1) == ' ') pStr++;
                if (sscanf(pszData,"%d",&iPacker) != 1) goto __err;
                if (iPacker < 1 || iPacker > iPackers)
                {
                    printf("Error in configuration file '%s' line %lu: Packer number out of range (%d of %d)\n",pszFileName,dwLine,iPacker,iPackers);
                    return 0;
                }
                pszData = pStr+1;

                /* Get password */
                pStr = strchr(pszData,' ');
                if (pStr == NULL) goto __err;
                *pStr = '\0';
                while (*(pStr+1) == ' ') pStr++;
                if (strlen(pszData) > 8)
                {
                    printf("Error in configuration file '%s' line %lu: Password too long\n",pszFileName,dwLine);
                    return 0;
                }
                strcpy(szPassword, pszData);
                if (szPassword[0] == '-' && szPassword[1] == '\0') szPassword[0] = '\0';
                pszData = pStr+1;

                /* Get areafix password */
                pStr = strchr(pszData,' ');
                if (pStr == NULL) goto __err;
                *pStr = '\0';
                while (*(pStr+1) == ' ') pStr++;
                if (strlen(pszData) > 20)
                {
                    printf("Error in configuration file '%s' line %lu: Areafix password too long\n",pszFileName,dwLine);
                    return 0;
                }
                strcpy(szAreaFixPassword, pszData);
                if (szAreaFixPassword[0] == '-' && szAreaFixPassword[1] == '\0') szAreaFixPassword[0] = '\0';
                pszData = pStr+1;

                /* Get access groups */
                pStr = strchr(pszData,' ');
                if (pStr != NULL)
                {
                    *pStr = '\0';
                    while (*(pStr+1) == ' ') pStr++;
                }
                pszAccessGroups = new char[strlen(pszData)+1];
                if (pszAccessGroups == NULL)
                {
                    /* Not enough memory */
                    printf("SECTION_NODES : malloc( %d ) : Not enough memory!\n", strlen(pszData)+1);
                    return 0;
                }
                strcpy(pszAccessGroups, pszData);

                /* Get routing */
                if (pStr != NULL)
                {
                    pszData = pStr+1;
                    iDataLen = strlen(pszData)+1;
                }
                else
                {
                    /* No routings here */
                    iDataLen = 1;
                    pszData = "";
                }

                /* Allocate memory for node */
                pTempNode = (NODE_REC *) malloc(sizeof(NODE_REC)+iDataLen+1);
                if (pTempNode == NULL)
                {
                    /* Not enough memory */
                    printf("SECTION_NODES : malloc( %d ) : Not enough memory!\n",sizeof(NODE_REC)+iDataLen+1);
                    return 0;
                }

                /* Update linked list */
                if (pNodeRecord != NULL) pNodeRecord->pNext = pTempNode;
                if (pFirstNode == NULL) pFirstNode = pTempNode;
                pNodeRecord = pTempNode;

                /* Update address record */
                pNodeRecord->pNext = NULL;
                pNodeRecord->wZone = wZone; pNodeRecord->wNet = wNet;
                pNodeRecord->wNode = wNode; pNodeRecord->wPoint = wPoint;
                pNodeRecord->wYourZone = wYourZone; pNodeRecord->wYourNet = wYourNet;
                pNodeRecord->wYourNode = wYourNode; pNodeRecord->wYourPoint = wYourPoint;
                pNodeRecord->iPacker = iPacker;
                pNodeRecord->pszAccessGroups = pszAccessGroups;
                strcpy(pNodeRecord->szPassword,szPassword);
                strcpy(pNodeRecord->szAreaFixPassword,szAreaFixPassword);
                memcpy(pNodeRecord->szRoute,pszData,iDataLen);
                break;
            case SECTION_EXPORT:
                /* New export */
                pStr = strchr(szLine,' ');
                if (pStr == NULL) goto __err;
                *pStr++ = '\0'; while (*pStr == ' ') pStr++;

                /* Search area */
                pszData = szLine;
                if (*pszData == '*') pszData++;
                pMsgAreaRecord = pFirstMsgArea;
                while (pMsgAreaRecord != NULL)
                {
                    if (stricmp(pMsgAreaRecord->szAreaTag,pszData) == 0)
                    {
                        break;
                    }
                    pMsgAreaRecord = pMsgAreaRecord->pNext;
                }
                if (pMsgAreaRecord == NULL)
                {
                    /* Couldn't find it */
                    printf("File '%s' line %lu: Area not found: '%s'\n",pszFileName,dwLine,szLine);
                    return 0;
                }

                if (*(pszData-1) == '*')
                {
                    /* Change area tag name.. */
                    pszData = pStr; pStr = strchr(pszData,' ');
                    if (pStr == NULL) goto __err;
                    *pStr++ = '\0'; while (*pStr == ' ') pStr++;
                    if (strlen(pszData) > sizeof(pMsgAreaRecord->szAreaTag)-1)
                    {
                        printf("Error in configuration file '%s' line %lu: New area tag name too long (max. %d chars)\n",pszFileName,dwLine,sizeof(pMsgAreaRecord->szAreaTag)-1);
                        return 0;
                    }
                    strcpy(pMsgAreaRecord->szAreaTag, pszData);
                }

                /* Get group */
                pszData = pStr; pStr = strchr(pszData,' ');
                if (pStr != NULL)
                {
                    *pStr++ = '\0';
                    while (*pStr == ' ') pStr++;
                }
                if (strlen(pszData) > 3)
                {
                    printf("Error in configuration file '%s' line %lu: Group name too long (max. 3 chars)\n",pszFileName,dwLine);
                    return 0;
                }
                strcpy(pMsgAreaRecord->szGroup, pszData);

                /* Get export list */
                pszData = pStr; pAddrRecord = NULL;
                while (pszData != NULL)
                {
                    pStr = strchr(pszData,' ');
                    if (pStr != NULL)
                    {
                        *pStr++ = '\0';
                        while (*pStr == ' ') pStr++;
                    }

                    if (!str2addr(pszData,&wZone,&wNet,&wNode,&wPoint)) goto __err;

                    pNodeRecord = pFirstNode;
                    while (pNodeRecord != NULL)
                    {
                        if (pNodeRecord->wZone == wZone &&
                            pNodeRecord->wNet == wNet &&
                            pNodeRecord->wNode == wNode &&
                            pNodeRecord->wPoint == wPoint)
                        {
                            /* Got it! */
                            break;
                        }
                        pNodeRecord = pNodeRecord->pNext;
                    }

                    if (pNodeRecord == NULL)
                    {
                        /* Node not found */
                        printf("File '%s' line %lu: Export node not found (add it to [nodes] section): %u:%u/%u.%u\n",
                               pszFileName, dwLine, wZone, wNet, wNode, wPoint);
                        return 0;
                    }

                    /* Allocate memory for node */
                    pTempAddr = new ADDR_REC;
                    if (pTempAddr == NULL)
                    {
                        /* Not enough memory */
                        printf("SECTION_EXPORT : malloc( %d ) : Not enough memory!\n",sizeof(NODE_REC));
                        return 0;
                    }

                    /* Update linked list */
                    if (pAddrRecord != NULL) pAddrRecord->pNext = pTempAddr;
                    if (pMsgAreaRecord->pExport == NULL) pMsgAreaRecord->pExport = pTempAddr;
                    pAddrRecord = pTempAddr;

                    /* Update address record */
                    pAddrRecord->pNext = NULL;
                    pAddrRecord->wZone = wZone;
                    pAddrRecord->wNet = wNet;
                    pAddrRecord->wNode = wNode;
                    pAddrRecord->wPoint = wPoint;

                    pszData = pStr;
                }
                break;
            case SECTION_OTHER:
                /* Other settings */
                pStr = strchr(szLine,'=');
                if (pStr == NULL) goto __err;
                pszData = pStr+1;
                pStr--;

                /* Remove spaces before '=' */
                while (pStr >= szLine && *pStr == ' ') pStr--;
                *(pStr+1) = '\0';

                /* Remove spaces after '=' */
                while (*pszData == ' ') pszData++;
                *(pszData-1) = '\0';

                if (*szLine == '\0') goto __err;

                if (stricmp(szLine,"TearLine") == 0)
                {
                    /* Tear line made by tosser */
                    if (strlen(pszData) > sizeof(szTearLine)-1)
                    {
                        /* Too long origin line */
                        printf("Error in configuration file '%s' line %lu: Tear line too long!\n",pszFileName,dwLine);
                        return 0;
                    }
#ifdef VBOARD
                    if (!registered)
                    {
                        printf("Tear line can be changed only in registered version!\n");
                        return 0;
                    }
#endif
                    strcpy(szTearLine, pszData);
                    break;
                }
                if (stricmp(szLine,"Origin") == 0)
                {
                    /* Tear line made by tosser */
                    if (strlen(pszData) > sizeof(szOrigin)-1)
                    {
                        /* Too long origin line */
                        printf("Error in configuration file '%s' line %lu: Origin too long!\n",pszFileName,dwLine);
                        return 0;
                    }
                    strcpy(szOrigin, pszData);
                    break;
                }
                if (stricmp(szLine,"Organization") == 0)
                {
                    /* Tear line made by tosser */
                    if (strlen(pszData) > sizeof(szOrganization)-1)
                    {
                        /* Too long origin line */
                        printf("Error in configuration file '%s' line %lu: Organization too long!\n",pszFileName,dwLine);
                        return 0;
                    }
                    strcpy(szOrganization, pszData);
                    break;
                }
                if (stricmp(szLine,"Domain") == 0)
                {
                    /* Tear line made by tosser */
                    if (strlen(pszData) > sizeof(szDomain)-1)
                    {
                        /* Too long origin line */
                        printf("Error in configuration file '%s' line %lu: Domain too long!\n",pszFileName,dwLine);
                        return 0;
                    }
                    strcpy(szDomain, pszData);
                    break;
                }
                if (stricmp(szLine,"Dupes") == 0)
                {
                    if (sscanf(pszData, "%d", &max_dupes) != 1) max_dupes = 2048;
                    break;
                }
                if (stricmp(szLine,"LowFLO") == 0)
                {
                    low_flo = pszData[0] == 'Y';
                    break;
                }
                if (stricmp(szLine,"SavePacked") == 0)
                {
                    fSavePacked = pszData[0] == 'Y';
                    break;
                }
                if (stricmp(szLine,"BinkleyOutbounds") == 0)
                {
                    bink_outb = pszData[0] == 'Y';
                    break;
                }
                if (stricmp(szLine,"DomainOutbounds") == 0)
                {
                    domain_outbounds = pszData[0] == 'Y';
                    break;
                }
                if (stricmp(szLine,"NoOutboundZones") == 0)
                {
                    no_outboundzones = pszData[0] == 'Y';
                    break;
                }
                if (stricmp(szLine,"UseAlwaysSlash") == 0)
                {
                    always_slash = pszData[0] == 'Y';
                    break;
                }
                goto __err;
            default:
            __err:
                printf("Error in configuration file '%s' line %lu\n",pszFileName,dwLine);
                return 0;
        }
    }
    delete[] szLine;

    fclose(hConfig);

    if (pNetmailArea == NULL)
    {
        /* Netmail area not found */
        printf("No netmail area!\n");
        return 0;
    }

    if (pBadMsgArea == NULL)
    {
        /* Badecho not found */
        printf("No badecho area!\n");
        return 0;
    }

#ifdef VBOARD
    if (!registered)
    {
#ifdef __OS2__
        sprintf(szTearLine, "VBoard+XToss/2 %s", VERSION);
#elif defined (__DOS__)
        sprintf(szTearLine, "VBoard+XToss/D %s", VERSION);
#elif defined (__NT__)
        sprintf(szTearLine, "VBoard+XToss/W32 %s", VERSION);
#endif
    }
#endif

    return 1;
}

void CloseConfig(void)
{
    /* Release memory used by addresses */
    while (pFirstAddr != NULL)
    {
        pAddrRecord = pFirstAddr->pNext;
        free(pFirstAddr);
        pFirstAddr = pAddrRecord;
    }

    /* Release memory used by nodes */
    while (pFirstNode != NULL)
    {
        pNodeRecord = pFirstNode->pNext;
        delete[] pFirstNode->pszAccessGroups;
        free(pFirstNode);
        pFirstNode = pNodeRecord;
    }

    /* Release memory used by packers */
    while (pFirstPacker != NULL)
    {
        pPackerRecord = pFirstPacker->pNext;
        delete[] pFirstPacker->pszPackString;
        delete[] pFirstPacker->pszUnPackString;
        delete pFirstPacker;
        pFirstPacker = pPackerRecord;
    }

    /* Release memory used by addresses */
    while (pFirstMsgArea != NULL)
    {
        pMsgAreaRecord = pFirstMsgArea->pNext;

        /* Release memory used by export list */
        pFirstAddr = pFirstMsgArea->pExport;
        while (pFirstAddr != NULL)
        {
            pAddrRecord = pFirstAddr->pNext;
            delete pFirstAddr;
            pFirstAddr = pAddrRecord;
        }

#ifndef VBOARD
        delete[] pFirstMsgArea->pszPath;
#endif
        delete pFirstMsgArea;
        pFirstMsgArea = pMsgAreaRecord;
    }

    delete[] pszBaseFile;
    delete[] pszInbound;
    delete[] pszSafeIn;
    delete[] pszMsgPath;
    delete[] pszOutbound;
    delete[] pszTemp;
    delete[] pszLogFile;
    delete[] pszSeqFile;
    delete[] pszSpool;
#ifdef VBOARD
    delete[] pszVBLogFile;
#endif
}
