#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <malloc.h>
#include <ctype.h>

#include "os.h"
#include "mareas.h"
#include "logfile.h"
#include "version.h"
#include "files.h"
#include "config.h"

#ifdef __linux__
#  define stricmp strcasecmp
#  define strnicmp strncasecmp
#endif

typedef struct LIST_REC
{
    struct LIST_REC *next;

    char name[100];
    char alias[100];
    unsigned short zone, net, node, point;
}
LIST_REC;

static LIST_REC *first = NULL;

static char subject[100];
static char from[100];

extern char *szMonth[12];

static char *weekday[7] =
{
    "Sun","Mon","Tue","Wed","Thu","Fri","Sat"
};


void write_kludges(FILE *ftxt, char *from, char *subj, char *recv)
{
    time_t _tim;
    struct tm *tim;
    char *strp, *strp2, buf[256];
    LIST_REC *list;

    list = first; strp = NULL;
    while (list != NULL)
    {
        if (stricmp(list->name, from) == 0)
        {
            strp = list->alias;
            break;
        }
        list = list->next;
    }

    if (strp == NULL)
    {
        /* Not found from export.lst, create field from user.name */
        //WriteLog('*', "Warning: user `%s` not found from export.lst", from);
        strcpy(buf, from);
        strp = buf;
        while (*strp != '\0')
        {
            *strp = (char) tolower(*strp);
            if (*strp == ' ') *strp = '.';
            strp++;
        }
        strp = buf;
    }

    _tim = time(NULL); tim = localtime(&_tim);
    if (tim->tm_year < 80) tim->tm_year += 2000; else tim->tm_year += 1900;

    strp2 = strchr(szDomain, '.'); if (strp2 != NULL) *strp2 = '\0';
    fprintf(ftxt, "From %s!%s %s %s %02d %02d:%02d:%02d %d\n", szDomain, strp,
            weekday[tim->tm_wday], szMonth[tim->tm_mon], tim->tm_mday,
            tim->tm_hour, tim->tm_min, tim->tm_sec, tim->tm_year);
    if (strp2 != NULL) *strp2 = '.';
    fprintf(ftxt, "To: %s\n", recv);
    fprintf(ftxt, "Subject: %s\n", subj);
    fprintf(ftxt, "From: %s@%s (%s)\n", strp, szDomain, from);
    fprintf(ftxt, "Message-ID: <xt%08lx@%s>\n", time(NULL), szDomain);
    //fprintf(ftxt, "References: ....\n");
    fprintf(ftxt, "Date: %02d %s %04d %02d:%02d:%02d %s\n", tim->tm_mday,
            szMonth[tim->tm_mon], tim->tm_year, tim->tm_hour, tim->tm_min, tim->tm_sec, tzname[0]);
    fprintf(ftxt, "Organization: %s\n", szOrganization);
    fprintf(ftxt, "X-Mailer: XToss "VERSION"\n");
    fprintf(ftxt, "\n");
}

int update_seq(void)
{
    FILE *ftxt;
    int seq;

    /* Read seq file */
    ftxt = fopen(pszSeqFile, "r+b");
    if (ftxt == NULL)
    {
        ftxt = fopen(pszSeqFile, "w+b");
        if (ftxt == NULL)
        {
            WriteLog('!',"Can't create seq file `%s`", pszSeqFile);
            return 0;
        }
        fwrite("0", 1, 1, ftxt);
        seq = 0;
    }
    else
    {
        if (fscanf(ftxt, "%d", &seq) != 1) seq = 0; else seq++;
        if (seq > 9999) seq = 0;
        fseek(ftxt, 0, SEEK_SET); ftrunc(fileno(ftxt), 0);
        fprintf(ftxt, "%d", seq);
    }
    fclose(ftxt);

    return seq;
}

LIST_REC *get_rec(char *alias)
{
    LIST_REC *rec, *sec;
    char *strp, tmp[100];

    rec = first; sec = NULL;
    while (rec != NULL)
    {
        if (stricmp(rec->alias, alias) == 0) break;
        if (sec == NULL)
        {
            strcpy(tmp, rec->name);
            strp = tmp;
            while (*strp != '\0')
            {
                if (*strp == ' ') *strp = '.';
                strp++;
            }
            if (stricmp(tmp, alias) == 0) sec = rec;
        }
        rec = rec->next;
    }
    if (rec == NULL) rec = sec;

    return rec;
}

extern unsigned long CurrentMessage;

int SendEmail(char *from, char *to, char *subj, int vb)
{
    char recv[100], *strp, *strp2, buf[256], tmp[256];
    int start, seq, iLastSpace;
    FILE *ftxt;
    int iBufPos, back;
    MSG_TEXT *pMsgText;
    LIST_REC *rec;
    unsigned readed;

    if (pszSeqFile == NULL || pszSpool == NULL) return 0;

    WriteLog('+',"Processing E-Mail from %s to %s", from, to);

    strcpy(recv, to); start = 1; ftxt = NULL;
    pMsgText = pFirstMsgText; iBufPos = 0; readed = 0; seq = 0;
    iLastSpace = -1; back = 0;
    while (pMsgText != NULL)
    {
        strp = pMsgText->sText;
        while (strp < pMsgText->sText+sizeof(pMsgText->sText))
        {
            if (vb)
            {
                readed++;
                if (readed > wMsgTextLength) break;
            }
            else if (*strp == '\0') break; /* EOT */

            if (*strp == 13 || *strp == '\0') /* EOL */
            {
                buf[iBufPos] = '\0';
                strp++;
                if (buf[0] != 1)
                {
                    /* Not a kludge */
                    if (strnicmp(buf, "TO:", 3) == 0 && start)
                    {
                        /* To-field */
                        if (buf[3] == ' ')
                            strcpy(recv, buf+4);
                        else
                            strcpy(recv, buf+3);
                    }
                    else
                    {
                        /* Normal line */
                        if (start)
                        {
                        __start:
                            start = 0;

                            /* Check if receiver is in same domain */
                            strp2 = strchr(recv, '@');
                            if (strp2 != NULL)
                            {
                                if (stricmp(strp2+1, szDomain) == 0)
                                {
                                    /* Same domain, send this in netmail to destination */
                                    *strp2 = '\0';
                                    rec = get_rec(recv);
                                    *strp2 = '@';
                                    if (rec == NULL)
                                    {
                                        /* Not found from alias.lst */
                                        strcpy(szMsgTo, recv);
                                        strp2 = szMsgTo;
                                        while (*strp2 != '\0')
                                        {
                                            if (*strp2 == '.') *strp2 = ' ';
                                            strp2++;
                                        }
                                    }
                                    else
                                    {
                                        strcpy(szMsgTo, rec->name);
                                        if (rec->zone != 0)
                                        {
                                            /* Not for a local user, so use netmail.. */
                                            strcpy(tmp, szMsgSubj);
                                            sprintf(szMsgSubj, "%d:%d/%d.%d %s", rec->zone, rec->net, rec->node, rec->point, tmp);
                                            return 0;
                                        }
                                    }
                                    /* Local e-mail */
#ifndef YAK
                                    WriteMessageHeader(CurrentMessage);
#endif
                                    return 1;
                                }
                            }

                            seq = update_seq();

                            /* Create .DAT file */
                            sprintf(tmp, "%s%04d.DAT", pszSpool, seq);
                            ftxt = fopen(tmp, "w+b");
                            if (ftxt == NULL)
                            {
                                WriteLog('!',"Can't create file `%s`", tmp);
                                return 0;
                            }

                            /* Write kludges */
                            write_kludges(ftxt, from, subj, recv);
                        }
                        fprintf(ftxt, "%s\n", buf);
                        if (back) goto __back;
                    }
                }
                iBufPos = 0; iLastSpace = -1;
                continue;
            }
            if (*strp != 10)
            {
                if (*strp == ' ') iLastSpace = iBufPos;
                if (iBufPos == 79)
                {
                    /* Gotta split line */

                    if (iLastSpace == -1)
                    {
                        /* No spaces in line, raw split */
                        buf[iBufPos] = '\0';
                        if (start) goto __start; else fprintf(ftxt, "%s\n", buf);
                    }
                    else
                    {
                        /* Split from last space */
                        buf[iLastSpace] = '\0';
                        if (start)
                        {
                            back = 1;
                            goto __start;
                        }
                        else
                            fprintf(ftxt, "%s\n", buf);
                    __back:
                        back = 0;
                        if (*strp != ' ')
                        {
                            memmove(buf, buf+iLastSpace+1, iBufPos-iLastSpace-1);
                            iBufPos -= iLastSpace+1;
                            buf[iBufPos] = '\0';
                            strp2 = strrchr(buf, ' ');
                            iLastSpace = strp2 == NULL ? -1 : (int) (strp2-buf);
                        }
                        else
                        {
                            iBufPos = -1; iLastSpace = -1;
                        }
                    }
                }
                if (iBufPos != -1) buf[iBufPos++] = *strp; else iBufPos++;
                strp++;
            }
        }

        pMsgText = pMsgText->pNext;
    }
    if (iBufPos != 0)
    {
        buf[iBufPos] = '\0';
        fprintf(ftxt, "%s\n", buf);
    }
    if (ftxt != NULL) fclose(ftxt);

    strp = strchr(szDomain, '.'); if (strp != NULL) *strp = '\0';

    /* Create .XQT file */
    sprintf(buf, "%s%04d.XQT", pszSpool, seq);
    ftxt = fopen(buf, "w+b");
    if (ftxt == NULL)
    {
        WriteLog('!',"Can't create file `%s`", buf);
        return 0;
    }
    fprintf(ftxt, "U uucp %s\n", szDomain);
    fprintf(ftxt, "F D.%s%04d\n", szDomain, seq);
    fprintf(ftxt, "I D.%s%04d\n", szDomain, seq);
    fprintf(ftxt, "C rmail %s\n", recv);
    fclose(ftxt);

    /* Create .CMD file */
    sprintf(buf, "%s%04d.CMD", pszSpool, seq);
    ftxt = fopen(buf, "w+b");
    if (ftxt == NULL)
    {
        WriteLog('!', "Can't create file `%s`", buf);
        return 0;
    }
    fprintf(ftxt, "S %04d.DAT D.%s%04d uucp - %04d.DAT 0666\n", seq, szDomain, seq, seq);
    fprintf(ftxt, "S %04d.XQT X.%s%04d uucp - %04d.XQT 0666\n", seq, szDomain, seq, seq);
    fclose(ftxt);

    if (strp != NULL) *strp = '.';

    return 1;
}

static void get_name(char *input, char *output)
{
    char *strp, *strp2;

    strp = strchr(input, '<');
    if (strp != NULL)
    {
        /* name <addr@inet> */
        strp2 = strchr(strp, '>');
        if (strp2 != NULL) *strp2 = '\0';
        strcpy(output, strp+1);
        if (strp2 != NULL) *strp2 = '>';
    }
    else
    {
        /* addr@inet (name) */
        strp = strchr(input, '(');
        if (strp != NULL) *strp = '\0';
        strcpy(output, input);
        if (strp != NULL) *strp = '(';
    }

    /* Strip trailing spaces */
    strp = output;
    while (*strp != '\0') strp++;
    while (*(strp-1) == ' ') strp--;
    *strp = '\0';
}

int OutUUCP(FILE *F, char *receiver, unsigned long start, unsigned long end)
{
    unsigned long oldpos = ftell(F);
    int readed, len, ret;
    char *strp, recvr[100], sender[100];
    LIST_REC *rec;
    MSG_TEXT *mtext, *mprev;

    get_name(receiver, recvr);
    get_name(from, sender);

    /* Strip domain */
    strp = strchr(recvr, '@');
    if (strp != NULL) *strp = '\0';

    rec = get_rec(recvr);
    if (rec != NULL)
    {
        printf("Sending message to %s via %u:%u/%u.%u\n",
               receiver, rec->zone, rec->net, rec->node, rec->point);
    }

    fseek(F, start, SEEK_SET);

    mprev = NULL;

    wMsgTextLength = 0; pFirstMsgText = NULL;
    while (end > start)
    {
        mtext = (MSG_TEXT *) malloc(sizeof(MSG_TEXT));
        if (mtext == NULL)
        {
            WriteLog('!', "OutUUCP() : malloc( %d ) : Not enough memory!", sizeof(MSG_TEXT));
            printf("OutUUCP() : malloc( %d ) : Not enough memory!\n", sizeof(MSG_TEXT));
            exit(1);
        }
        mtext->pNext = NULL;
        if (pFirstMsgText == NULL) pFirstMsgText = mtext; else mprev->pNext = mtext;
        mprev = mtext;

        readed = (int) ((end-start > 4096) ? 4096 : end-start);
        readed = fread(mtext->sText, 1, readed, F);
        if (readed < 4096) mtext->sText[readed] = '\0';
        if (readed == 0) break;

        wMsgTextLength += readed;
        end -= readed;
    }
    fseek(F, oldpos, SEEK_SET);

    if (rec != NULL)
        strcpy(recvr, rec->name);
    else
    {
        /* Not found from export.lst, user user.name style */
        strp = recvr;
        while (*strp != '\0')
        {
            if (*strp == '.') *strp = ' ';
            strp++;
        }
    }

    tMsgDate = time(NULL);
    if (rec == NULL || rec->zone == 0)
        len = 0;
    else
        len = sprintf(szMsgSubj, "%u:%u/%u.%u ", rec->zone, rec->net, rec->node, rec->point);
    strncpy(szMsgSubj+len, subject, sizeof(szMsgSubj)-1-len); szMsgSubj[sizeof(szMsgSubj)-1] = '\0';
    strncpy(szMsgFrom, sender, 35); szMsgFrom[35] = '\0';
    strncpy(szMsgTo, recvr, 35); szMsgTo[35] = '\0';
    fMsgSent = rec == NULL || rec->zone == 0; fMsgPrivate = 1;

    ret = WriteMessage(0, 0);

    while (pFirstMsgText != NULL)
    {
        mtext = pFirstMsgText->pNext;
        free(pFirstMsgText);
        pFirstMsgText = mtext;
    }

    return ret;
}

int ReadUUCP(char *path, char *fname)
{
    FILE *F;
    char str[256], *strp, *strp2, receiver[100];
    int last, ret;
    unsigned long start, prev;

    sprintf(str, "%s%s", path, fname);
    F = fopen(str, "rb");
    if (F == NULL)
    {
        WriteLog('!', "Can't open file `%s`", str);
        printf("Can't open file `%s`!\n", str);
        return 0;
    }
    printf("Processing file %s\n", str);

    ret = 1;
    from[sizeof(from)-1] = '\0'; subject[sizeof(subject)-1] = '\0';
    prev = 0; start = 0; last = 0; receiver[0] = '\0';
    while (_fgets(str, sizeof(str), F) != NULL)
    {
        if (strnicmp(str, "Received: ", 10) == 0 ||
            (last && strncmp(str, "          ", 10) == 0 && str[10] != ' '))
        {
            last = 1;
            strp = strstr(str, " for ");
            if (strp != NULL)
            {
                strp += 5;
                strp2 = strchr(strp, ';');
                if (strp2 != NULL)
                {
                    /* Receiver name */
                    *strp2 = '\0';
                    strcpy(receiver, strp);
                    *strp2 = ';';
                }
            }
        }
        else
        {
            last = 0;
            if (strnicmp(str, "From ", 5) == 0)
            {
                if (prev != 0 &&  !OutUUCP(F, receiver, start, prev))
                {
                    ret = 0;
                    break;
                }
                start = prev;
                receiver[0] = '\0'; from[0] = '\0'; subject[0] = '\0';
            }
            else if (strnicmp(str, "To: ", 4) == 0)
                strcpy(receiver, str+4);
            else if (strnicmp(str, "Subject: ", 9) == 0)
                strncpy(subject, str+9, sizeof(subject)-1);
            else if (strnicmp(str, "Reply-To: ", 10) == 0)
                strncpy(from, str+10, sizeof(from)-1);
            else if (from[0] == '\0' && strnicmp(str, "From: ", 6) == 0)
                strncpy(from, str+6, sizeof(from)-1);
        }

        prev = ftell(F);
    }
    if (ret && prev != 0 && !OutUUCP(F, receiver, start, ftell(F))) ret = 0;
    fclose(F);

    return ret;
}

int ScanUUCP(char *path)
{
    DIR *dirp;
    struct dirent *direntp;
    char *strp, tmp[256];

#ifdef VBOARD
    if (!OpenMsgArea(pNetmailArea->wAreaNum))
#else
    if (!OpenMsgArea(pNetmailArea->pszPath,1))
#endif
    {
        printf("Could not open message area `%s`\n",pMsgAreaRecord->szAreaTag);
        WriteLog('!',"Could not open message area `%s`",pMsgAreaRecord->szAreaTag);
        return 0;
    }

    dirp = opendir(path);
    if (dirp == NULL)
    {
        WriteLog('!', "Can't open path `%s`", path);
        printf("Can't open path `%s`\n", path);
        return 0;
    }

    for (;;)
    {
        direntp = readdir(dirp);
        if (direntp == NULL) break;

        strp = strrchr(direntp->d_name, '.');
        if (strp == NULL) continue;
        if (toupper(*(++strp)) == 'D' && *(++strp) == '\0')
        {
            if (ReadUUCP(path, direntp->d_name))
            {
                sprintf(tmp, "%s"SSLASH"%s", path, direntp->d_name);
                remove(tmp);
            }
        }
    }
    closedir(dirp);
    CloseMsgArea();

    return 1;
}

int InitEmail(char *fname)
{
    FILE *F;
    char str[256], *strp, *strp2;
    LIST_REC *rec, *tmp;
    unsigned long line;

    F = fopen(fname, "rb");
    if (F == NULL)
    {
        WriteLog('*', "Can't open export list file `%s`", fname);
        return 0;
    }

    first = NULL; rec = NULL; line = 0;
    while (_fgets(str, sizeof(str), F) != NULL)
    {
        line++;
        strp = strchr(str, ' ');
        if (strp == NULL)
        {
            WriteLog('!', "Error in export list file %s line %lu\n", fname, line);
            printf("Error in export list file %s line %lu\n", fname, line);
            return 0;
        }

        *strp++ = '\0';
        tmp = (LIST_REC *) malloc(sizeof(LIST_REC));
        if (tmp == NULL)
        {
            printf("Not enough memory to read export list!\n");
            return 0;
        }
        str2addr(str, &tmp->zone, &tmp->net, &tmp->node, &tmp->point);
        tmp->next = NULL;

        strp2 = strchr(strp, ' ');
        if (strp2 == NULL)
        {
            WriteLog('!', "Error in export list file %s line %lu\n", fname, line);
            printf("Error in export list file %s line %lu\n", fname, line);
            free(tmp);
            return 0;
        }
        *strp2++ = '\0';
        strcpy(tmp->alias, strp);
        strcpy(tmp->name, strp2);
        if (first == NULL) first = tmp;
        if (rec != NULL) rec->next = tmp;
        rec = tmp;
    }

    fclose(F);
    return 1;
}
