#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <ctype.h>
#include "files.h"

#ifndef __linux__
#  include <dos.h>  /* sleep() */
#endif

#ifdef __linux__
FILE_BUF *OpenFile(char *szFileName, int iMode, int )
#else
FILE_BUF *OpenFile(char *szFileName, int iMode, int iShareMode)
#endif
{
    FILE_BUF *pFileBuf;
    int hFile;

    hFile = sopen(szFileName, iMode, iShareMode);
    if (hFile == -1) return NULL;

    pFileBuf = new FILE_BUF;
    if (pFileBuf == NULL)
    {
        printf("Not enough memory to allocate %d bytes!\n", sizeof(FILE_BUF));
        close(hFile);
        return NULL;
    }

    pFileBuf->hFile = hFile;

    pFileBuf->lBufStart = 1;
    pFileBuf->lBufEnd = 0;

    pFileBuf->lSeekPos = 0;
    pFileBuf->dwFileSize = lseek(hFile,0,SEEK_END);
    lseek(hFile,0,SEEK_SET);

    return pFileBuf;
}

#ifdef __linux__
FILE_BUF *CreateFile(char *szFileName, int iMode)
#else
FILE_BUF *CreateFile(char *szFileName, int iMode)
#endif
{
    FILE_BUF *pFileBuf;
    int hFile;

    hFile = creat(szFileName, iMode);
    if (hFile == -1) return NULL;

    pFileBuf = new FILE_BUF;
    if (pFileBuf == NULL) return NULL;

    pFileBuf->hFile = hFile;

    pFileBuf->lBufStart = 1;
    pFileBuf->lBufEnd = 0;

    pFileBuf->lSeekPos = 0;
    pFileBuf->dwFileSize = lseek(hFile,0,SEEK_END);
    lseek(hFile,0,SEEK_SET);

    return pFileBuf;
}

void CloseFile(FILE_BUF *pFile)
{
    close(pFile->hFile);
    delete pFile;
}

int ReadFile(FILE_BUF *pFile, void *sBuffer, int iBufSize)
{
    int iReaded,iPos;

    //printf("Someone wants to read %d bytes starting at %ld\n",iBufSize,pFile->lSeekPos);
    if (iBufSize > (int) sizeof(pFile->sBuffer))
    {
        printf("\nReadFile() : Buffer too small! This should NEVER happen!!\n");
        exit(1);
    }

    if (pFile->lBufStart > pFile->lSeekPos || pFile->lBufEnd < pFile->lSeekPos+iBufSize-1)
    {
        lseek(pFile->hFile,pFile->lSeekPos,SEEK_SET);
        iReaded = read(pFile->hFile,pFile->sBuffer,sizeof(pFile->sBuffer));
        if (iReaded < 1) return -1;

        pFile->lBufStart = pFile->lSeekPos;
        pFile->lBufEnd = pFile->lSeekPos+iReaded-1;

        iPos = 0;
        if (iReaded > iBufSize) iReaded = iBufSize;
        pFile->lSeekPos += iBufSize;
    }
    else
    {
        iPos = (int) (pFile->lSeekPos - pFile->lBufStart);
        pFile->lSeekPos += iBufSize;
        iReaded = iBufSize;
    }

    //printf("Gave 'em position %d\n",iPos);
    memcpy(sBuffer, (char *) pFile->sBuffer+iPos, iReaded);

    return iReaded;
}

int WriteFile(FILE_BUF *file, void *buffer, int bufsize)
{
    int readed;

    file->lBufStart = 1;
    file->lBufEnd = 0;

    lseek(file->hFile, file->lSeekPos, SEEK_SET);
    readed = write(file->hFile, buffer, bufsize);
    file->dwFileSize = lseek(file->hFile, 0, SEEK_END);
    if (readed > 0) file->lSeekPos += readed;

    return readed;
}

long SeekFile(FILE_BUF *pFile, long lFilePos, int iType)
{
    switch (iType)
    {
        case SEEK_SET:
            pFile->lSeekPos = lFilePos;
            break;
        case SEEK_CUR:
            pFile->lSeekPos += lFilePos;
            break;
        case SEEK_END:
            pFile->lSeekPos = pFile->dwFileSize-lFilePos;
            break;
    }
    //printf("Someone wants to seek to position %ld\n",pFile->lSeekPos);

    return pFile->lSeekPos;
}

int CopyFile(char *szSourceFile, char *szDestFile)
{
    FILE *Fsrc,*Fdst;
    char *buf;
    size_t readed;

    if ((Fsrc = fopen(szSourceFile,"rb")) == NULL) return 0;
    if ((Fdst = fopen(szDestFile,"w+b")) == NULL) { fclose(Fsrc); return 0; }
    buf = new char[4096];

    do
    {
        readed = fread(buf,1,4096,Fsrc);
        fwrite(buf,1,readed,Fdst);
    }
    while (readed == 4096);

    delete[] buf;
    fclose(Fsrc); fclose(Fdst);

    return 1;
}

int LockFile(int hFile)
{
    int tries;

    tries = 0;
    while (flock(hFile,0,1) != 0)
    {
        tries++;
        if (tries == 5)
        {
            printf("\nPlease wait, trying to lock file...\n");
        }
        if (tries == 60*5)
        {
            printf("Okay! Now I've had it! Couldn't lock it in 5 minutes..\n");
            return 0;
        }
        sleep(1);
    }

    return 1;
}

/* Get string from file, remove all CRs and LFs */
char *_fgets(char *buf, size_t n, FILE *fp)
{
    char *strp;

    if (fgets(buf,n,fp) == NULL) return NULL;
    strp = buf;
    while (*strp != '\0')
    {
        if (*strp == '\r' || *strp == '\n')
        {
            *strp = '\0';
            break;
        }
        strp++;
    }
    return buf;
}

#ifndef __linux__
void ChangeDrive(char drv)
{
    unsigned drives;

    _dos_setdrive(toupper(drv)-'A'+1, &drives);
}
#endif
