#ifndef __FILES_H
#define __FILES_H

/* Includes */
#include <fcntl.h>
#include <sys/stat.h>

#ifdef __linux__
#  include <unistd.h>
#  include <dirent.h>
#else
#  include <io.h>
#  include <share.h>
#  include <direct.h>
#endif

/* Defines */
#ifdef __linux__
#  define O_BINARY 0
#  define SH_DENYNO 0
#  define SH_DENYRW 0
#  define READWRITE 0660
#  define ISDIR(a) S_ISDIR(a.st_mode)
#  define flock(a,b,c) lockf(a,F_LOCK,b)
#  define funlock(a,b,c) lockf(a,F_ULOCK,b)
#  define sopen(a,b,c) open(a,b)
#  define ftrunc(a,b) ftruncate(a,b)
inline int MakeDir(char *path)
{
    int ret;
    ret = mkdir(path, 0770);
    if (ret == 0) chmod(path, 0770);
    return ret;
}
#  define setmode(a,b)
#else
#  define READWRITE S_IREAD | S_IWRITE
#  define ISDIR(a) (((a.st_attr) & _A_SUBDIR) == _A_SUBDIR)
#  define flock(a,b,c) lock(a,b,c)
#  define funlock(a,b,c) unlock(a,b,c)
#  define MakeDir(a) mkdir(a)
#  define ftrunc(a,b) chsize(a,b)
#endif

typedef struct
{
    int hFile;

    long lBufStart,lBufEnd;
    char sBuffer[16384];

    long lSeekPos;
    unsigned long dwFileSize;
}
FILE_BUF;

FILE_BUF *OpenFile(char *szFileName, int iMode, int iShareMode);
FILE_BUF *CreateFile(char *szFileName, int iMode);
void CloseFile(FILE_BUF *pFile);
int ReadFile(FILE_BUF *pFile, void *sBuffer, int iBufSize);
int WriteFile(FILE_BUF *pFile, void *sBuffer, int iBufSize);
long SeekFile(FILE_BUF *pFile, long lFilePos, int iType);

int CopyFile(char *szSourceFile, char *szDestFile);

int LockFile(int hFile);
#define UnLockFile(a) funlock(a,0,1);

char *_fgets(char *buf, size_t n, FILE *fp);

void ChangeDrive(char drv);

#endif
