#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

/* ------------- */
#ifndef VBOARD
#include "jam.h"
extern long crc32tab[256];
#else
#include <time.h>
#include "vboard.h"
#endif
/* ------------- */

#include "os.h"
#include "files.h"
#include "config.h"
#include "mareas.h"
#include "logfile.h"

#ifdef __linux__
#  define stricmp strcasecmp
#endif

unsigned long dwMessages;

time_t tMsgDate;
char szMsgFrom[36];
char szMsgTo[36];
char szMsgSubj[256];
#ifndef VBOARD
char szMsgDest[100];
char szMsgID[80];
char szReplyID[80];
int kludges;
char msgkludge[100][80];
#endif
char fMsgSent, fMsgPrivate;

unsigned long wMsgTextLength;
MSG_TEXT *pFirstMsgText = NULL;

unsigned short wOrigZone, wOrigNet, wOrigNode, wOrigPoint;
unsigned short wDestZone, wDestNet, wDestNode, wDestPoint;

#ifdef VBOARD
static unsigned short wMsgOrigZone;
static unsigned short wMsgOrigNet;
static unsigned short wMsgOrigNode;

static int hHead = -1, hText = -1;
static int iOpenArea = -1;
static MESSAGE *Message = NULL;

/* Functions needed from BBSAHuff.c ... */
extern "C"
{
    void SaveCompressed(unsigned char *Input, unsigned short InputSize, unsigned int Handle);
    void LoadCompressed(unsigned int Handle, unsigned char *Destination, unsigned short *ExpandSize);
};

int OpenMsgArea(int iAreaNumber)
{
    char szPath[256];

    if (Message == NULL)
    {
        Message = new MESSAGE;
        if (Message == NULL)
        {
            WriteLog('!', "OpenMsgArea() : malloc( %d ) : Not enough memory!", sizeof(MESSAGE));
            return 0;
        }
    }

    if (iOpenArea == iAreaNumber) return 1;
    iOpenArea = -1;

    if (hHead != -1) CloseMsgArea();

    /* Open header file */
    sprintf(szPath,"%smsghead.%03x",szMsgAreaPath,iAreaNumber);
    hHead = sopen(szPath, O_RDWR | O_BINARY, SH_DENYNO);
    if (hHead == -1)
    {
        printf("Error opening file '%s'\n",szPath);
        WriteLog('!',"Error opening file '%s'",szPath);
        return 0;
    }

    /* Open text file */
    sprintf(szPath,"%smsgtext.%03x",szMsgAreaPath,iAreaNumber);
    hText = sopen(szPath, O_RDWR | O_BINARY, SH_DENYNO);
    if (hText == -1)
    {
        printf("Error opening file '%s'\n",szPath);
        WriteLog('!',"Error opening file '%s'",szPath);
        close(hHead); hHead = -1;
        return 0;
    }
    iOpenArea = iAreaNumber;

    /* Get number of messages */
    dwMessages = lseek(hHead,0,SEEK_END) / sizeof(MSGHEAD);

    return 1;
}

void DeInitText(void)
{
    MSG_TEXT *pText;

    while (pFirstMsgText != NULL)
    {
        pText = pFirstMsgText->pNext;
        delete pFirstMsgText;
        pFirstMsgText = pText;
    }
}

void CloseMsgArea(void)
{
    if (hHead != -1) { close(hHead); hHead = -1; }
    if (hText != -1) { close(hText); hText = -1; }
}

void CloseMessages(void)
{
    if (Message != NULL)
    {
        delete[] Message;
        Message = NULL;
    }
}

int ReadMessage(unsigned long dwMsgNumber)
{
    MSGHEAD MsgHead;
    SUBMSGHEAD SubMsgHead;
    MSG_TEXT *pText,*pTempText;
    int iToRead,iReaded;
    unsigned long wTotalReaded;

    /* Read header */
    lseek(hHead, (dwMsgNumber-1) * sizeof(MSGHEAD), SEEK_SET);
    if (read(hHead, &MsgHead, sizeof(MsgHead)) != sizeof(MsgHead)) return 0;

    /* Don't allow reading killed messaged */
    if (MsgHead.OffsetAttr & MASK_KILLED) return 0;
    if (MsgHead.OffsetAttr & MASK_SENT) return 0;

    lseek(hText, MsgHead.OffsetAttr & MASK_OFFSET, SEEK_SET);
    wMsgTextLength = 0;

    if (MsgHead.OffsetAttr & MASK_PACKED)
    {
#ifdef COMPRESSED
        LoadCompressed(hText, (unsigned char *)Message, (unsigned short *) &wMsgTextLength);
        wMsgTextLength -= sizeof(SUBMSGHEAD);
        memcpy(&SubMsgHead, &Message->SubMsgHead, sizeof(SubMsgHead));
#else
        return 0;
#endif
    }
    else
    {
        /* Read header from text file */
        if (read(hText,&wMsgTextLength,sizeof(unsigned short)) != sizeof(unsigned short)) return 0;
        if (read(hText,&SubMsgHead,sizeof(SubMsgHead)) != sizeof(SubMsgHead)) return 0;
    }

    /* Copy to outworld */
    strcpy(szMsgFrom, (char *) SubMsgHead.MsgFrom);
    strcpy(szMsgTo, (char *) SubMsgHead.MsgTo);
    strcpy(szMsgSubj, (char *) SubMsgHead.Subject);
    tMsgDate = SubMsgHead.MsgDate;
    wMsgOrigZone = SubMsgHead.From.Zone;
    wMsgOrigNet = SubMsgHead.From.Net;
    wMsgOrigNode = SubMsgHead.From.Node;
    fMsgPrivate = (MsgHead.OffsetAttr & MASK_PRIVATE) != 0;
    fMsgSent = (MsgHead.OffsetAttr & MASK_SENT) != 0;

    DeInitText();

    wTotalReaded = 0; pText = NULL;
    while (wTotalReaded < wMsgTextLength)
    {
        /* Get memory for new message text block */
        pTempText = new MSG_TEXT;
        if (pTempText == NULL)
        {
            WriteLog('!', "ReadMessage() : malloc( %d ) : Not enough memory!\n", sizeof(MSG_TEXT));
            printf("ReadMessage() : malloc( %d ) : Not enough memory!\n", sizeof(MSG_TEXT));
            exit(1);
        }
        pTempText->pNext = NULL;
        if (pFirstMsgText == NULL) pFirstMsgText = pTempText;
        if (pText != NULL) pText->pNext = pTempText;
        pText = pTempText;

        /* Read message text */
        iToRead = (int) (wMsgTextLength-wTotalReaded);
        if (iToRead > (int) sizeof(pText->sText)) iToRead = sizeof(pText->sText);
        if (MsgHead.OffsetAttr & MASK_PACKED)
        {
            /* Packed message - move from memory */
            memcpy(pText->sText, Message->MsgTxt+wTotalReaded, iToRead);
            iReaded = iToRead;
        }
        else
        {
            /* Unpacked message - Read from file */
            iReaded = read(hText, pText->sText, iToRead);
        }
        wTotalReaded += iReaded;
        if (iReaded != iToRead)
        {
            /* Could not read all of file! */
            printf("Warning: Message text too short! Should be %lu bytes, got only %lu!\n",wMsgTextLength,wTotalReaded);
            WriteLog('!',"Warning: Message text too short! Should be %lu bytes, got only %lu!",wMsgTextLength,wTotalReaded);
            wMsgTextLength = wTotalReaded;
            break;
        }
    }

    return 1;
}

unsigned short UserNumber(char *szUser)
{
    char szPath[256];
    unsigned short wUser;
    int hFile;
    USER User;

    /* Open userbase */
    sprintf(szPath,"%susers.dat",szMsgAreaPath);
    hFile = sopen(szPath, O_RDONLY | O_BINARY, SH_DENYNO);
    if (hFile == -1)
    {
        WriteLog('*',"Userbase %s not found",szPath);
        return 0;
    }

    /* Search user */
    wUser = 0;
    while (read(hFile, &User, sizeof(USER)) == sizeof(USER))
    {
        wUser++;
        if (stricmp((char *) User.Name, szUser) == 0 && (User.SFlags & SFLAG_KILLED) == 0)
        {
            close(hFile);
            return wUser;
        }
    }

    close(hFile);
    return 0;
}

int WriteMessage(int iTextOver, int skiplf)
{
    MSG_TEXT *pMsgText;
    MSGHEAD MsgHead;
    SUBMSGHEAD SubMsgHead;
    char sBuf[256],*strp,last,fWrite;
    unsigned iPos;
    int iLinePos,iLastSpace;
    long lWritten, oldpos;

    /* Lock message base */
    if (!LockFile(hHead))
    {
        printf("Can't lock message base!\n");
        WriteLog('!',"Can't lock message base!");
        return 0;
    }

    if (!LockFile(hText))
    {
        UnLockFile(hHead);
        printf("Can't lock message base!\n");
        WriteLog('!',"Can't lock message base!");
        return 0;
    }

    memset(&SubMsgHead, 0, sizeof(SubMsgHead));

    /* Set sender */
    SubMsgHead.MsgFrom[sizeof(SubMsgHead.MsgFrom)-1] = '\0';
    strncpy((char *) SubMsgHead.MsgFrom, szMsgFrom, sizeof(SubMsgHead.MsgFrom)-1);

    /* Set receiver */
    SubMsgHead.MsgTo[sizeof(SubMsgHead.MsgTo)-1] = '\0';
    strncpy((char *) SubMsgHead.MsgTo, szMsgTo, sizeof(SubMsgHead.MsgTo)-1);

    /* Set subject */
    SubMsgHead.Subject[sizeof(SubMsgHead.Subject)-1] = '\0';
    strncpy((char *) SubMsgHead.Subject, szMsgSubj, sizeof(SubMsgHead.Subject)-1);

    /* Set date */
    SubMsgHead.MsgDate = tMsgDate;

    /* Set origin address */
    SubMsgHead.From.Zone = wMsgOrigZone;
    SubMsgHead.From.Net = wMsgOrigNet;
    SubMsgHead.From.Node = wMsgOrigNode;

    /* Set MsgHead */
    memset(&MsgHead, 0, sizeof(MsgHead));
    MsgHead.MsgFromN = UserNumber(szMsgFrom);
    MsgHead.MsgToN = UserNumber(szMsgTo);
    MsgHead.OffsetAttr = lseek(hText, 0, SEEK_END);
    if (fMsgPrivate) MsgHead.OffsetAttr |= MASK_PRIVATE;
    if (fMsgSent) MsgHead.OffsetAttr |= MASK_SENT;
    if (fSavePacked)
    {
        /* Save message packed */
        MsgHead.OffsetAttr |= MASK_PACKED;
        memset(Message, 0, sizeof(MESSAGE));
        memcpy(&Message->SubMsgHead, &SubMsgHead, sizeof(SubMsgHead));
    }
    else
    {
        /* Save message unpacked */
        /* Write message text length */
        if (write(hText, &wMsgTextLength, sizeof(unsigned short)) != sizeof(unsigned short))
        {
            /* Error writing subheader */
            printf("Could not write message length! Out of disk space?\n");
            WriteLog('!',"Could not write message length! Out of disk space?");
            goto __err_writing;
        }

        /* Write SubMsgHead */
        if (write(hText, &SubMsgHead, sizeof(SubMsgHead)) != sizeof(SubMsgHead))
        {
            /* Error writing subheader */
            printf("Could not write message (sub)header! Out of disk space?\n");
            WriteLog('!',"Could not write message (sub)header! Out of disk space?");

        __err_writing:
            UnLockFile(hHead);
            UnLockFile(hText);
            exit(1);
        }
    }

    /* Write message text */
    lWritten = 0; iPos = 0; iLinePos = 0; last = '\0'; iLastSpace = -1; fWrite = 0;
    pMsgText = pFirstMsgText;
    while (pMsgText != NULL)
    {
        strp = pMsgText->sText+iTextOver; iTextOver = 0;
        while (*strp != '\0' && (unsigned) (strp-pMsgText->sText) < sizeof(pMsgText->sText))
        {
            if ((skiplf && *strp == 10) || (!skiplf && *strp == 13))
            {
                strp++;
                continue;
            }

            if ((skiplf && *strp == 13) || (!skiplf && *strp == 10))
            {
                sBuf[iPos++] = '\0';
                fWrite = 1; iLinePos = 0;
                last = '\0';
            }
            else
            {
                if (*strp == ' ') iLastSpace = iPos;
                sBuf[iPos++] = *strp;
                last = *strp; iLinePos++;
            }

            if (iLinePos == 80)
            {
                /* Gotta split line */

                if (iLastSpace == -1)
                {
                    /* No spaces in line, raw split */
                    sBuf[iPos++] = '\0';
                    last = '\0'; iLinePos = 0;
                }
                else
                {
                    /* Split from last space */
                    sBuf[iLastSpace] = '\0'; iLinePos = 79-iLastSpace;
                }
                fWrite = 1;
            }

            if (fWrite)
            {
                iLastSpace = -1;
                fWrite = 0;

                oldpos = lWritten;
                if (lWritten+iPos > 64000)
                {
                    iPos = (unsigned) (64000-lWritten);
                    lWritten = 64000;
                    WriteLog('*', "Message longer than 64000 bytes, truncating.");
                }
                else
                    lWritten += iPos;

                if (fSavePacked)
                {
                    /* Packed - Save to memory */
                    memcpy(Message->MsgTxt+oldpos, sBuf, iPos);
                }
                else
                {
                    /* Unpacked - Write to disk */
                    if (write(hText,sBuf,iPos) != iPos)
                    {
                        printf("Could not write message text! Out of disk space?\n");
                        WriteLog('!',"Could not write message text! Out of disk space?");
                        goto __err_writing;
                    }
                }
                iPos = 0;
                if (lWritten == 64000) break;
            }
            strp++;
        }

        if (lWritten == 64000) break;
        pMsgText = pMsgText->pNext;
    }

    /* Last nuls */
    if (last != '\0') sBuf[iPos++] = '\0';

    if (iPos > 0)
    {
        oldpos = lWritten;
        if (lWritten+iPos > 64000)
        {
            iPos = (unsigned) (64000-lWritten);
            lWritten = 64000;
            WriteLog('*', "Message longer than 64000 bytes, truncating.");
        }
        else
            lWritten += iPos;

        if (fSavePacked)
        {
            /* Packed - Save to memory */
            memcpy(Message->MsgTxt+oldpos, sBuf, iPos);
        }
        else
        {
            /* Unpacked - Write to disk */
            if (write(hText,sBuf,iPos) != iPos)
            {
                printf("Could not write message text (rest)! Out of disk space?\n");
                WriteLog('!',"Could not write message text (rest)! Out of disk space?");
                goto __err_writing;
            }
        }
    }

    wMsgTextLength = lWritten;

#ifdef COMPRESSED
    if (fSavePacked)
    {
        /* Finally save packed message to disk */
        SaveCompressed((unsigned char *)Message, (unsigned short) (sizeof(SUBMSGHEAD)+wMsgTextLength), hText);
    }
    else
    {
#endif
        /* Write message text length */
        lseek(hText,MsgHead.OffsetAttr & MASK_OFFSET,SEEK_SET);
        if (write(hText, &wMsgTextLength, sizeof(unsigned short)) != sizeof(unsigned short))
        {
            /* Error writing subheader */
            printf("Could not write message length (second time)! Out of disk space?\n");
            WriteLog('!',"Could not write message length (second time)! Out of disk space?");
            goto __err_writing;
        }
#ifdef COMPRESSED
    }
#endif

    /* Write MsgHead */
    lseek(hHead, 0, SEEK_END);
    if (write(hHead, &MsgHead, sizeof(MsgHead)) != sizeof(MsgHead))
    {
        /* Error writing header */
        printf("Could not write message header! Out of disk space?\n");
        WriteLog('!',"Could not write message header! Out of disk space?");
        goto __err_writing;
    }

    /* Unlock message base */
    UnLockFile(hHead);
    UnLockFile(hText);

    dwMessages = lseek(hHead,0,SEEK_END) / sizeof(MSGHEAD);

    return 1;
}

int WriteMessageHeader(unsigned long dwMsgNumber)
{
    MSGHEAD MsgHead;
    SUBMSGHEAD SubMsgHead;
    unsigned short len;

    /* Read header */
    lseek(hHead, (dwMsgNumber-1) * sizeof(MSGHEAD), SEEK_SET);
    if (read(hHead, &MsgHead, sizeof(MsgHead)) != sizeof(MsgHead)) return 0;

    memset(&SubMsgHead, 0, sizeof(SubMsgHead));

    /* Set sender */
    SubMsgHead.MsgFrom[sizeof(SubMsgHead.MsgFrom)-1] = '\0';
    strncpy((char *) SubMsgHead.MsgFrom, szMsgFrom, sizeof(SubMsgHead.MsgFrom)-1);

    /* Set receiver */
    SubMsgHead.MsgTo[sizeof(SubMsgHead.MsgTo)-1] = '\0';
    strncpy((char *) SubMsgHead.MsgTo, szMsgTo, sizeof(SubMsgHead.MsgTo)-1);

    /* Set subject */
    SubMsgHead.Subject[sizeof(SubMsgHead.Subject)-1] = '\0';
    strncpy((char *) SubMsgHead.Subject, szMsgSubj, sizeof(SubMsgHead.Subject)-1);

    /* Set date */
    SubMsgHead.MsgDate = tMsgDate;

    /* Set origin address */
    SubMsgHead.From.Zone = wMsgOrigZone;
    SubMsgHead.From.Net = wMsgOrigNet;
    SubMsgHead.From.Node = wMsgOrigNode;

    if (MsgHead.OffsetAttr & MASK_PACKED)
    {
        lseek(hText, MsgHead.OffsetAttr & MASK_OFFSET, SEEK_SET);
        LoadCompressed(hText, (unsigned char *)Message, (unsigned short *) &len);
        memcpy(&Message->SubMsgHead, &SubMsgHead, sizeof(SubMsgHead));

        MsgHead.OffsetAttr &= MASK_ALL;
        MsgHead.OffsetAttr |= lseek(hText, 0, SEEK_END);
        SaveCompressed((unsigned char *)Message, (unsigned short) (sizeof(SUBMSGHEAD)+len), hText);

        /* Write header */
        lseek(hHead, (dwMsgNumber-1) * sizeof(MSGHEAD), SEEK_SET);
        if (write(hHead, &MsgHead, sizeof(MsgHead)) != sizeof(MsgHead)) return 0;
    }
    else
    {
        lseek(hText, MsgHead.OffsetAttr & MASK_OFFSET, SEEK_SET);
        write(hText, &SubMsgHead, sizeof(SubMsgHead));
    }
    return 1;
}

int SetMessageSent(unsigned long dwMsgNumber)
{
    MSGHEAD MsgHead;

    /* Read header */
    lseek(hHead, (dwMsgNumber-1) * sizeof(MSGHEAD), SEEK_SET);
    if (read(hHead, &MsgHead, sizeof(MsgHead)) != sizeof(MsgHead)) return 0;

    MsgHead.OffsetAttr |= MASK_SENT;

    /* Write header */
    lseek(hHead, (dwMsgNumber-1) * sizeof(MSGHEAD), SEEK_SET);
    if (write(hHead, &MsgHead, sizeof(MsgHead)) != sizeof(MsgHead)) return 0;

    return 1;
}
#else

static int Fjhr = -1, Fjdx = -1, Fjdt = -1;
static char sOpenArea[256] = "";
static int Fmastat = -1;

int OpenMsgArea(char *fname, int create)
{
    char tmp[256];
    JAMHDRINFO jamhdr;

    if (fname == NULL)
    {
        WriteLog('!',"OpenMsgArea) : fname == NULL ?? This should NEVER happen!");
        return 0;
    }
    if (strcmp(sOpenArea,fname) == 0) return 1;
    sOpenArea[0] = '\0'; dwMessages = 0;

    if (Fjhr != -1) CloseMsgArea();

    /* Open message base files */
    sprintf(tmp,"%s"EXT_HDRFILE, fname); Fjhr = sopen(tmp, O_RDWR | O_BINARY, SH_DENYNO);
    sprintf(tmp,"%s"EXT_IDXFILE, fname); Fjdx = sopen(tmp, O_RDWR | O_BINARY, SH_DENYNO);
    sprintf(tmp,"%s"EXT_TXTFILE, fname); Fjdt = sopen(tmp, O_RDWR | O_BINARY, SH_DENYNO);

    if (create && (Fjhr == -1 && Fjdx == -1 && Fjdt == -1))
    {
        /* Create message base files */
        sprintf(tmp,"%s"EXT_HDRFILE, fname); Fjhr = creat(tmp, O_RDWR | S_IREAD | S_IWRITE);
        sprintf(tmp,"%s"EXT_IDXFILE, fname); Fjdx = creat(tmp, O_RDWR | S_IREAD | S_IWRITE);
        sprintf(tmp,"%s"EXT_TXTFILE, fname); Fjdt = creat(tmp, O_RDWR | S_IREAD | S_IWRITE);

        if (Fjhr != -1)
        {
            setmode(Fjhr,O_BINARY);
            setmode(Fjdx,O_BINARY);
            setmode(Fjdt,O_BINARY);

            LockFile(Fjhr);

            memset(&jamhdr,0,sizeof(jamhdr));
            strcpy(jamhdr.Signature,HEADERSIGNATURE);
            jamhdr.DateCreated = time(NULL);
            jamhdr.PasswordCRC = 0xffffffff;
            if (write(Fjhr,&jamhdr,sizeof(jamhdr)) != sizeof(jamhdr))
            {
                close(Fjhr);
                Fjhr = -1;
            }

            UnLockFile(Fjhr);
        }
    }

    if (Fjhr == -1 || Fjdx == -1 || Fjdt == -1)
    {
        close(Fjhr);
        close(Fjdx);
        close(Fjdt);
        return 0;
    }

    /* Get number of messages */
    dwMessages = (unsigned) (lseek(Fjdx,0,SEEK_END) / sizeof(JAMIDXREC));

    strcpy(sOpenArea, fname);

    return 1;
}

void CloseMsgArea(void)
{
    /* Close message base files */
    if (Fjhr != -1) { close(Fjhr); Fjhr = -1; }
    if (Fjdx != -1) { close(Fjdx); Fjdx = -1; }
    if (Fjdt != -1) { close(Fjdt); Fjdt = -1; }
}

void CloseMessages(void)
{
    if (Fmastat != -1) { close(Fmastat); Fmastat = -1; }
}

int ReadMessage(unsigned long dwMsgNumber)
{
    JAMIDXREC jamidx;
    JAMHDR msgrec;
    JAMSUBFIELD *subfield;
    unsigned len,sublen;
    unsigned long wTotalReaded;
    char *data;
    MSG_TEXT *pText,*pTempText;
    int iToRead,iReaded;

    if (pFirstMsgText != NULL) DeInitText();

    if (dwMsgNumber == 0) return 0;

    /* Read .JHR file position from .HDX file */
    lseek(Fjdx,(dwMsgNumber-1)*sizeof(jamidx),SEEK_SET);
    if (read(Fjdx,&jamidx,sizeof(jamidx)) == 0) return 0;
    if (jamidx.HdrOffset == 0xffffffff) return 0;

    /* Read message record from .JHR file */
    lseek(Fjhr,jamidx.HdrOffset,SEEK_SET);
    if (read(Fjhr,&msgrec,sizeof(msgrec)) == 0) return 0;

    if (strcmp(msgrec.Signature,HEADERSIGNATURE) != 0) return 0;
    if (msgrec.Revision != CURRENTREVLEV) return 0;

    if (msgrec.Attribute & MSG_DELETED) return 0;
    if ((msgrec.Attribute & MSG_LOCAL) == 0) return 0;
    if (msgrec.Attribute & MSG_SENT) return 0;

    fMsgPrivate = (msgrec.Attribute & MSG_PRIVATE) != 0;
    fMsgSent = (msgrec.Attribute & MSG_SENT) != 0;

    tMsgDate = (time_t) msgrec.DateWritten;
    wMsgTextLength = msgrec.TxtLen;

    szMsgFrom[0] = '\0'; szMsgTo[0] = '\0'; szMsgSubj[0] = '\0';
    szMsgDest[0] = '\0'; szMsgID[0] = '\0'; szReplyID[0] = '\0';

    kludges = 0;
    /* Read subfields */
    if (msgrec.SubfieldLen > 0)
    {
        data = new char[msgrec.SubfieldLen];
        if (data == NULL)
        {
            WriteLog('!', "ReadMessage() : subfield : malloc( %lu ) : Not enough memory!", msgrec.SubfieldLen);
            printf("ReadMessage() : subfield : malloc( %lu ) : Not enough memory!\n", msgrec.SubfieldLen);
            exit(1);
        }

        /* Read subfields into memory */
        read(Fjhr,data,(unsigned) msgrec.SubfieldLen);

        sublen = 0;
        while (sublen < msgrec.SubfieldLen)
        {
            subfield = (JAMSUBFIELD *) (data+sublen);
            len = (unsigned) subfield->DatLen;

            switch (subfield->LoID)
            {
                case JAMSFLD_SENDERNAME:
                    if (len > sizeof(szMsgFrom)-1) len = sizeof(szMsgFrom)-1;
                    szMsgFrom[len] = '\0';
                    memcpy(szMsgFrom,subfield->Buffer,len);
                    break;
                case JAMSFLD_RECVRNAME:
                    if (len > sizeof(szMsgTo)-1) len = sizeof(szMsgTo)-1;
                    szMsgTo[len] = '\0';
                    memcpy(szMsgTo,subfield->Buffer,len);
                    break;
                case JAMSFLD_SUBJECT:
                    if (len > sizeof(szMsgSubj)-1) len = sizeof(szMsgSubj)-1;
                    szMsgSubj[len] = '\0';
                    memcpy(szMsgSubj,subfield->Buffer,len);
                    break;
                case JAMSFLD_DADDRESS:
                    if (len > sizeof(szMsgDest)-1) len = sizeof(szMsgDest)-1;
                    szMsgDest[len] = '\0';
                    memcpy(szMsgDest,subfield->Buffer,len);
                    break;
                case JAMSFLD_MSGID:
                    if (len > sizeof(szMsgID)-1) len = sizeof(szMsgID)-1;
                    szMsgID[len] = '\0';
                    memcpy(szMsgID,subfield->Buffer,len);
                    break;
                case JAMSFLD_REPLYID:
                    if (len > sizeof(szReplyID)-1) len = sizeof(szReplyID)-1;
                    szReplyID[len] = '\0';
                    memcpy(szReplyID,subfield->Buffer,len);
                    break;
                case JAMSFLD_FTSKLUDGE:
                    if (kludges == 100) break;
                    if (len > sizeof(msgkludge[0])-1) len = sizeof(msgkludge[0])-1;
                    msgkludge[kludges][len] = '\0';
                    memcpy(msgkludge[kludges],subfield->Buffer,len);
                    kludges++;
                    break;
            }

            sublen += sizeof(JAMBINSUBFIELD)+subfield->DatLen;
        }

        delete[] data;
    }
    lseek(Fjdt,msgrec.TxtOffset,SEEK_SET);

    wTotalReaded = 0; pText = NULL;
    while (wTotalReaded < wMsgTextLength)
    {
        /* Get memory for new message text block */
        pTempText = new MSG_TEXT;
        if (pTempText == NULL)
        {
            WriteLog('!', "ReadMessage() : msgtext : malloc( %d ) : Not enough memory!", sizeof(MSG_TEXT));
            printf("ReadMessage() : msgtext : malloc( %d ) : Not enough memory!\n", sizeof(MSG_TEXT));
            exit(1);
        }
        pTempText->pNext = NULL;
        if (pFirstMsgText == NULL) pFirstMsgText = pTempText;
        if (pText != NULL) pText->pNext = pTempText;
        pText = pTempText;

        /* Read message text */
        iToRead = (int) (wMsgTextLength-wTotalReaded);
        if (iToRead > (int) sizeof(pText->sText)) iToRead = sizeof(pText->sText);
        iReaded = read(Fjdt, pText->sText, iToRead);
        wTotalReaded += iReaded;
        if (iReaded != iToRead)
        {
            /* Could not read all of file! */
            printf("Warning: Message text too short! Should be %lu bytes, got only %lu!\n",wMsgTextLength,wTotalReaded);
            WriteLog('!',"Warning: Message text too short! Should be %lu bytes, got only %lu!",wMsgTextLength,wTotalReaded);
            wMsgTextLength = wTotalReaded;
            break;
        }
    }

    return 1;
}

void add_subfield(char *subdata, char *fielddata, unsigned subtype, JAMHDR *msgrec)
{
    JAMBINSUBFIELD subrec;

    if (fielddata[0] == '\0') return;

    subrec.LoID = (unsigned short) subtype;
    subrec.HiID = 0;
    subrec.DatLen = strlen(fielddata);

    memcpy(subdata,&subrec,sizeof(subrec));
    memcpy(subdata+sizeof(subrec),fielddata,(unsigned) subrec.DatLen);
    msgrec->SubfieldLen += subrec.DatLen+sizeof(subrec);
}

int lo_func(int chr)
{
    switch (chr)
    {
        case '':
            return '';
        case '':
            return '';
        case '':
            return '';
        case '':
            return '';
        case '':
            return '';
        default:
            return tolower(chr);
    }
}

unsigned long lo_crc32(char *str)
{
    unsigned long crc;

    crc = -1UL;
    while (*str != '\0')
    {
        crc = (crc >> 8) ^ crc32tab [(int) ((crc^lo_func(*str)) & 0xffUL)];
        str++;
    }
    return crc;
}

unsigned long crc32(char *str)
{
    unsigned long crc;

    crc = -1UL;
    while (*str != '\0')
    {
        crc = (crc >> 8) ^ crc32tab [(int) ((crc^*str) & 0xffUL)];
        str++;
    }
    return crc;
}

#if 0
typedef struct
{
    unsigned long crc;
    unsigned long messages;
}
MASTAT_REC;

MASTAT_REC malist[1000];
int mastart = 1, maend = 0;
unsigned long mastat_counter;

void update_area_messages(char *path, unsigned long messages)
{
    char tmp[256];
    MASTAT_REC ma;
    int n, readed, pos;
    unsigned long counter;

    if (Fmastat == -1)
    {
        sprintf(tmp, "%s"SSLASH"msgareas.idx", YakData);
        Fmastat = sopen(tmp, O_RDWR | O_BINARY, SH_DENYNO);
        if (Fmastat == -1)
        {
            mastat_counter = 0;
            Fmastat = creat(tmp, READWRITE);
            if (Fmastat == -1)
            {
                printf("Can't create file '%s'\n", tmp);
                WriteLog('!', "Can't create file '%s'", tmp);
                return;
            }
            write(Fmastat, &mastat_counter, sizeof(mastat_counter));
            lseek(Fmastat, 0, SEEK_SET);
        }
        read(Fmastat, &mastat_counter, sizeof(mastat_counter));
    }

    ma.crc = crc32(path);
    ma.messages = messages;

    if (!LockFile(Fmastat))
        WriteLog('!', "update_area_messages(): Can't lock file %s"SSLASH"msgareas.idx", YakData);

    /* Update counter */
    lseek(Fmastat, 0, SEEK_SET);
    read(Fmastat, &counter, sizeof(counter));
    if (counter != mastat_counter)
    {
        /* Counter changed, clear buffer */
        mastart = 1; maend = 0;
        mastat_counter = counter;
    }
    mastat_counter++;
    lseek(Fmastat, 0, SEEK_SET);
    write(Fmastat, &mastat_counter, sizeof(mastat_counter));

    pos = 0;
    for (;; pos++)
    {
        if (pos < mastart || pos > maend)
        {
            lseek(Fmastat, pos*sizeof(MASTAT_REC), SEEK_SET);
            readed = read(Fmastat, malist, sizeof(malist));
            if (readed < (int) sizeof(MASTAT_REC)) break;
            mastart = pos;
            maend = pos+readed/sizeof(MASTAT_REC);
        }
        for (n = 0; n < maend-mastart; n++)
        {
            if (malist[n].crc == ma.crc)
            {
                lseek(Fmastat, pos*sizeof(MASTAT_REC), SEEK_CUR);
                if (write(Fmastat, &ma, sizeof(MASTAT_REC)) != sizeof(MASTAT_REC))
                    WriteLog('!', "update_area_messages(): Can't write to file %s"SSLASH"msgareas.idx", YakData);
                UnLockFile(Fmastat);
                return;
            }
        }
    }
    lseek(Fmastat, 0, SEEK_END);
    write(Fmastat, &ma, sizeof(MASTAT_REC));
    UnLockFile(Fmastat);
}
#endif

int WriteMessage(int iTextOver, int /*skiplf*/)
{
    JAMHDRINFO jamhdr;
    JAMIDXREC jamidx;
    JAMHDR msgrec;
    MSG_TEXT *pMsgText;
    unsigned long fjhr_pos,msgnum;
    char *subbuf,*strp,*text,addr[30];
    int len;

    subbuf = new char[4096];
    if (subbuf == NULL)
    {
        WriteLog('!', "WriteMessage() : malloc( 4096 ) : Not enough memory!");
        printf("WriteMessage() : malloc( 4096 ) : Not enough memory!\n");
        return 0;
    }

    /* Lock message base */
    if (LockFile(Fjhr) == 0)
    {
        printf("Can't lock message base\n");
        WriteLog('!',"Can't lock message base");
        delete[] subbuf;
        return 0;
    }

    /* Read message base header */
    lseek(Fjhr,0,SEEK_SET);
    if (read(Fjhr,&jamhdr,sizeof(jamhdr)) != sizeof(jamhdr))
    {
        printf("Can't read message base header\n");
        WriteLog('!',"Can't read message base header");
        UnLockFile(Fjhr);
        delete[] subbuf;
        return 0;
    }

    /* Read message number of previous message */
    if (lseek(Fjdx,0,SEEK_END) == 0)
    {
        /* No messages in message base */
        jamidx.HdrOffset = lseek(Fjhr,0,SEEK_END);
    }
    else
    {
        do
        {
            lseek(Fjdx,-sizeof(jamidx),SEEK_CUR);
            if (read(Fjdx,&jamidx,sizeof(jamidx)) != sizeof(jamidx))
            {
                jamidx.HdrOffset = lseek(Fjhr,0,SEEK_END);
                break;
            }
            if (lseek(Fjdx,-sizeof(jamidx),SEEK_CUR) == 0 && jamidx.HdrOffset == 0xffffffff)
            {
                /* No messages in message base */
                jamidx.HdrOffset = lseek(Fjhr,0,SEEK_END);
                break;
            }
        }
        while (jamidx.HdrOffset == 0xffffffff);
    }

    lseek(Fjhr,jamidx.HdrOffset,SEEK_SET);
    if (read(Fjhr,&msgrec,sizeof(msgrec)) == 0) msgrec.MsgNum = 0;

    msgnum = msgrec.MsgNum+1;

    /* Create message header */
    memset(&msgrec,0,sizeof(msgrec));
    strcpy(msgrec.Signature,HEADERSIGNATURE);
    msgrec.Revision = CURRENTREVLEV;
    msgrec.DateWritten = tMsgDate;
    msgrec.MsgNum = msgnum;
    msgrec.PasswordCRC = 0xffffffff;
    msgrec.Attribute = MSG_LOCAL | MSG_SENT;
    if (fMsgPrivate) msgrec.Attribute |= MSG_PRIVATE;
    if (fMsgSent) msgrec.Attribute |= MSG_SENT;

    msgrec.TxtOffset = lseek(Fjdt,0,SEEK_END);

    /* Write message text */
    pMsgText = pFirstMsgText; wMsgTextLength = 0;
    while (pMsgText != NULL)
    {
        text = strp = pMsgText->sText+iTextOver;
        while (*strp != '\0' && strp < text+sizeof(pMsgText->sText)-iTextOver) strp++;
        len = (int) (strp-text); iTextOver = 0;
        if (write(Fjdt,text,len) != len)
        {
            printf("Can't write message text\n");
            WriteLog('!',"Can't write message text");
            UnLockFile(Fjhr);
            return 0;
        }
        wMsgTextLength += len;

        pMsgText = pMsgText->pNext;
    }

    msgrec.TxtLen = wMsgTextLength;

    /* Create subfields */
    add_subfield(subbuf+msgrec.SubfieldLen,szMsgFrom,JAMSFLD_SENDERNAME,&msgrec);
    sprintf(addr, "%u:%u/%u.%u", wOrigZone, wOrigNet, wOrigNode, wOrigPoint);
    add_subfield(subbuf+msgrec.SubfieldLen,addr,JAMSFLD_OADDRESS,&msgrec);
    add_subfield(subbuf+msgrec.SubfieldLen,szMsgTo,JAMSFLD_RECVRNAME,&msgrec);
    sprintf(addr, "%u:%u/%u.%u", wDestZone, wDestNet, wDestNode, wDestPoint);
    add_subfield(subbuf+msgrec.SubfieldLen,addr,JAMSFLD_DADDRESS,&msgrec);
    add_subfield(subbuf+msgrec.SubfieldLen,szMsgSubj,JAMSFLD_SUBJECT,&msgrec);
    add_subfield(subbuf+msgrec.SubfieldLen,szMsgID,JAMSFLD_MSGID,&msgrec);
    add_subfield(subbuf+msgrec.SubfieldLen,szReplyID,JAMSFLD_REPLYID,&msgrec);

    /* Write message header */
    fjhr_pos = lseek(Fjhr,0,SEEK_END);

    if (write(Fjhr,&msgrec,sizeof(msgrec)) != sizeof(msgrec))
    {
        printf("Can't write message header\n");
        WriteLog('!',"Can't write message header");
        UnLockFile(Fjhr);
        delete[] subbuf;
        return 0;
    }

    /* Write subfields */
    if (write(Fjhr,subbuf,(unsigned) msgrec.SubfieldLen) != (int) msgrec.SubfieldLen)
    {
        printf("Can't write subfields\n");
        WriteLog('!',"Can't write subfields");
        UnLockFile(Fjhr);
        delete[] subbuf;
        return 0;
    }
    delete[] subbuf;

    /* Update .JDX file */
    lseek(Fjdx,0,SEEK_END);
    jamidx.UserCRC = lo_crc32(szMsgTo);
    jamidx.HdrOffset = fjhr_pos;
    if (write(Fjdx,&jamidx,sizeof(jamidx)) != sizeof(jamidx))
    {
        printf("Can't write message index\n");
        WriteLog('!',"Can't write message index");
        UnLockFile(Fjhr);
        return 0;
    }

    /* Update ModCounter */
    jamhdr.ModCounter++;
    lseek(Fjhr,0,SEEK_SET);
    write(Fjhr,&jamhdr,sizeof(jamhdr));

    /* Update message number counters */
    dwMessages = (unsigned) (lseek(Fjdx,0,SEEK_END) / sizeof(JAMIDXREC));
    //update_area_messages(sOpenArea, dwMessages);

    /* Unlock message base */
    UnLockFile(Fjhr);

    return 1;
}

int SetMessageSent(unsigned long dwMsgNumber)
{
    JAMHDRINFO jamhdr;
    JAMIDXREC jamidx;
    JAMHDR msgrec;

    /* Lock message base */
    if (LockFile(Fjhr) == 0)
    {
        printf("Can't lock message base\n");
        WriteLog('!',"Can't lock message base");
        return 0;
    }

    /* Read message base header */
    lseek(Fjhr,0,SEEK_SET);
    if (read(Fjhr,&jamhdr,sizeof(jamhdr)) != sizeof(jamhdr))
    {
        printf("Can't read message header\n");
        WriteLog('!',"Can't read message header");
        UnLockFile(Fjhr);
        return 0;
    }

    /* Read index */
    lseek(Fjdx,(dwMsgNumber-1)*sizeof(jamidx),SEEK_SET);
    if (read(Fjdx,&jamidx,sizeof(jamidx)) != sizeof(jamidx))
    {
        printf("Can't read message index\n");
        WriteLog('!',"Can't read message index");
        UnLockFile(Fjhr);
        return 0;
    }

    /* Read message header */
    lseek(Fjhr,jamidx.HdrOffset,SEEK_SET);
    if (read(Fjhr,&msgrec,sizeof(msgrec)) != sizeof(msgrec)) msgrec.MsgNum = 0;

    /* Update flags */
    msgrec.Attribute |= MSG_SENT;

    /* Write message header */
    lseek(Fjhr,jamidx.HdrOffset,SEEK_SET);
    if (write(Fjhr,&msgrec,sizeof(msgrec)) != sizeof(msgrec))
    {
        printf("Can't write message header\n");
        WriteLog('!',"Can't write message header");
        UnLockFile(Fjhr);
        return 0;
    }

    /* Update ModCounter */
    jamhdr.ModCounter++;
    lseek(Fjhr,0,SEEK_SET);
    write(Fjhr,&jamhdr,sizeof(jamhdr));

    dwMessages = (unsigned) (lseek(Fjdx,0,SEEK_END) / sizeof(JAMIDXREC));

    /* Unlock message base */
    UnLockFile(Fjhr);

    return 1;
}

void DeInitText(void)
{
    MSG_TEXT *pText;

    while (pFirstMsgText != NULL)
    {
        pText = pFirstMsgText->pNext;
        delete pFirstMsgText;
        pFirstMsgText = pText;
    }
}
#endif
