/* To-do:
  - .msg netmaileja osaisi lukea
  - badechot voisi tossailla
  - auto area create
  - CHARSET & REPLY kludge
  - Kiva interfeissi
*/

int dupepos;

#ifdef __linux__
#  define SEMAPHORE "/tmp/xtossing.sem"
#  define CONFIG_FILE "/etc/xtoss.conf"
#  define DEFAULT_DUPE_FILE "/var/lib/xtoss_messages.dupes"
#  define OUTPUT_PATH "/tmp/xtoss"
#else
#  define SEMAPHORE "xtossing.sem"
#  define CONFIG_FILE "xtoss.cfg"
#  define DEFAULT_DUPE_FILE "messages.dup"
#  define OUTPUT_PATH "."
#endif

#define CRCLENGTH 7
unsigned char CRCSTRING[] = "\x36\x56\x51\x41\x66\x82\xb2xxxx";

/* ANSI-C */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

/* Posix */
#include <sys/stat.h>
#include <stdlib.h>

/* OS/2 */
#ifdef __OS2__
#  define INCL_DOS
#  include <os2.h>
#elif defined (__linux__)
#  include <sys/wait.h>
#else
#  include <i86.h>
#endif

#ifdef __DOS__
#  include "swap.h"

int shell(char *str)
{
    char tmp[256];

    strcpy(tmp, str);
    str = strchr(tmp, ' ');
    if (str != NULL) *str++ = '\0'; else str = "";
    return swap(tmp, str);
}
#else
#  define shell system
#endif

#include "os.h"
#include "crc32.h"
#include "files.h"
#include "config.h"
#include "mareas.h"
#include "logfile.h"
#include "areafix.h"
#include "version.h"
#include "email.h"

#ifdef __linux__
#  define stricmp strcasecmp
#  define strnicmp strncasecmp
#endif

static int hDupe;
static unsigned uOutFiles, uMsgsImported, uMsgsExported, uDupes, uBads;

char *szMonth[12] =
{
    "Jan","Feb","Mar","Apr","May","Jun",
    "Jul","Aug","Sep","Oct","Nov","Dec"
};

static char *sWeekDay[7] =
{
    "su","mo","tu","we","th","fr","sa"
};

typedef struct
{
    unsigned short wOrigNode;
    unsigned short wDestNode;
    unsigned short wYear,wMonth,wDay;
    unsigned short wHour,wMin,wSec;
    unsigned short wBaud;
    unsigned short wPacketType;
    unsigned short wOrigNet;
    unsigned short wDestNet;
    unsigned char byProdCode;
    unsigned char bySearialNo;
    char sPassword[8];
    unsigned short wQOrigZone;
    unsigned short wQDestZone;

    /* FSC-0039 */
    char sFiller[2];
    unsigned short wCapValid;  /* Byte swapped of capword */
    unsigned char byPrdCodH;
    unsigned char byPvMinor;
    unsigned short wCapWord;
    unsigned short wOrigZone;
    unsigned short wDestZone;
    unsigned short wOrigPoint;
    unsigned short wDestPoint;
    unsigned long dwProdData;
}
PKT_HEADER;

typedef struct
{
    unsigned short wProdCode;
    unsigned short wOrigNode;
    unsigned short wDestNode;
    unsigned short wOrigNet;
    unsigned short wDestNet;
    unsigned short wAttrib;
    unsigned short wCost;
    /*char datetime[];*/
}
MSG_HEADER;

typedef struct
{
    char from[36];
    char to[36];
    char subj[72];
    char date[20];
    unsigned short times_read;
    unsigned short dest_node;
    unsigned short orig_node;
    unsigned short cost;
    unsigned short orig_net;
    unsigned short dest_net;
    unsigned short dest_zone;
    unsigned short orig_zone;
    unsigned short dest_point;
    unsigned short orig_point;
    unsigned short replyto;
    unsigned short attrib;
    unsigned short nextreply;
}
STORED_MSG_HEADER;

typedef struct
{
    int hFile;
    unsigned uNodePos;
    char szFileName[13];
}
OUTPUT_REC;

typedef struct SEEN_REC
{
    struct SEEN_REC *pNext;
    struct SEEN_REC *pPrev;

    unsigned short wNet, wNode;
}
SEEN_REC;

typedef struct PATH_REC
{
    struct PATH_REC *pNext;

    unsigned short wNet, wNode;
}
PATH_REC;

typedef struct /* Dupe file record */
{
    unsigned short wOrigZone, wOrigNet, wOrigNode, wOrigPoint;
    unsigned long HdrCRC;
    unsigned long TxtCRC;
}
DUPE_REC;

static MSGAREA_REC *pLastMsgArea;
static OUTPUT_REC output[256];
static SEEN_REC *pFirstSeenBy, *pSeenBy;
static PATH_REC *pFirstPath, *pPath;
static unsigned short wLastSeenByNet, wLastPathNet;
static unsigned uSeenBys, uPaths;
static DUPE_REC *duperecs;
static int dupes;

unsigned long ulLastTime;
unsigned long CurrentMessage;

/* Hex comparation for VBoard's packer configuration */
int HexCompare(char *pszText, char *pszSearch)
{
    char sHexTable[] = "0123456789ABCDEF";
    char szHexString[100],*pStr;
    if (*pszSearch == '\0') return 0;

    /* Copy string -> hex string */
    pStr = szHexString;
    while (*pszText != '\0')
    {
        *pStr++ = sHexTable[(unsigned char) *pszText >> 4];
        *pStr++ = sHexTable[*pszText & 0x0f];
        pszText++;
    }
    *pStr = '\0';

    /* Compare strings */
    pStr = szHexString;
    while (*pszSearch != '\0')
    {
        if (*pszSearch != '?')
        {
            if (*pszSearch != *pStr) return 0;
        }
        else
        {
            if (*pStr == '\0') return 0;
        }
        pszSearch++; pStr++;
    }

    return 1;
}

/* Is message in dupe file? */
int IsDupe(DUPE_REC *pDupe)
{
    int n;

    for (n = 0; n < dupes; n++)
        if (memcmp(pDupe, &duperecs[n], sizeof(DUPE_REC)) == 0) return 1;

    return 0;
}

/* Add message to dupe file */
void AddDupe(DUPE_REC *pDupe)
{
    if (dupes == max_dupes) dupepos = 0;
    lseek(hDupe, 4+dupepos*sizeof(DUPE_REC), SEEK_SET);
    if (write(hDupe, pDupe, sizeof(DUPE_REC)) != sizeof(DUPE_REC))
    {
        /* Error writing.. */
        printf("Warning! Couldn't write to dupe file!!\n");
        WriteLog('!',"Warning! Couldn't write to dupe file!!");
    }
    memcpy(&duperecs[dupepos], pDupe, sizeof(DUPE_REC));
    dupepos++;
}

/* Unpack given file to current directory */
int UnPackFile(char *pszFileName)
{
    int hPacket,fPros,iPacker;
    char szID[32],szString[256],*pStr,*pOut;

    printf("Unpacking file `%s`\n",pszFileName);
    WriteLog('-',"Unpacking file `%s`",pszFileName);
    hPacket = sopen(pszFileName, O_RDONLY | O_BINARY, SH_DENYRW);
    if (hPacket == -1) return 0;

    if (read(hPacket,szID,32) == -1)
    {
        close(hPacket);
        return 0;
    }
    close(hPacket);

    pPackerRecord = pFirstPacker; iPacker = 0;
    while (pPackerRecord != NULL)
    {
        iPacker++;
        if (HexCompare(szID,pPackerRecord->szIdentify))
        {
            /* Found packer! */
            break;
        }
        pPackerRecord = pPackerRecord->pNext;
    }
    if (pPackerRecord == NULL)
    {
        printf("UnPackFile() : Unknown packer format.\n");
        WriteLog('!',"UnPackFile() : Unknown packer format.");
        return 0;
    }

    /* Convert unpacker string */
    pStr = pPackerRecord->pszUnPackString;
    pOut = szString; fPros = 0;
    for (; *pStr != '\0'; pStr++)
    {
        if (fPros)
        {
            if (toupper(*pStr) == 'F')
            {
                /* List of files to unpack */
            }
            else if (toupper(*pStr) == 'P')
            {
                /* Packet name */
                pOut += sprintf(pOut,"%s",pszFileName);
            }
            else if (*pStr == '%')
            {
                /* Just a '%' character */
                *pOut++ = '%';
            }
            else
            {
                printf("Unknown parameter in packer #%d : `%c`\n",iPacker,*pStr);
                WriteLog('!',"Unknown parameter in packer #%d : `%c`",iPacker,*pStr);
                return 0;
            }
            fPros = 0;
            continue;
        }
        if (*pStr == '%') fPros = 1;
        else *pOut++ = *pStr;
    }
    *pOut = '\0';

    /* Run unpacker */
    return shell(szString) == 0;
}

/* Unpack all packed message packets from inbound directory */
int UnPackMail(char *pszPath)
{
    DIR *pDir;
    struct dirent *pDirent;
    struct stat StatBuf;
    char szString[256],szOldPath[256],*pExt;
    int iDay;

    /* Open directory */
    pDir = opendir(pszPath);
    if (pDir == NULL)
    {
        printf("Could not open dir `%s`\n",pszPath);
        WriteLog('!',"Could not open dir `%s`",pszPath);
        return 0;
    }

    /* Change to temp path */
    getcwd(szOldPath,sizeof(szOldPath));
    strcpy(szString,pszTemp); szString[strlen(szString)-1] = '\0';
#ifndef __linux__
    if (szString[1] == ':') ChangeDrive(szString[0]);
#endif
    chdir(szString);

    /* Unpack packets */
    for (;;)
    {
        /* Read entry */
        pDirent = readdir(pDir);
        if (pDirent == NULL) break;

        /* Get extension */
        pExt = strrchr(pDirent->d_name,'.');
        if (pExt == NULL) continue;

        pExt++;
        if (stricmp(pExt,"PKT") == 0)
        {
            /* Unpacked .pkt file */
            sprintf(szString,"%s%s",pszPath,pDirent->d_name);
            if (rename(szString,pDirent->d_name) != 0)
            {
                if (CopyFile(szString,pDirent->d_name)) remove(szString);
            }
            continue;
        }

        for (iDay = 0; iDay < 7; iDay++)
        {
            if (toupper(sWeekDay[iDay][0]) == toupper(*pExt) &&
                toupper(sWeekDay[iDay][1]) == toupper(pExt[1]) &&
                pExt[2] >= '0' && pExt[2] <= '9' && pExt[3] == '\0')
            {
                /* Found packed message packet */
                sprintf(szString,"%s%s",pszPath,pDirent->d_name);
                if (stat(szString,&StatBuf) != 0) continue;
                if (ISDIR(StatBuf)) continue;

                if (!UnPackFile(szString))
                {
                    printf("Could not unpack file `%s`\n",szString);
                    WriteLog('!',"Could not unpack file `%s`",szString);
                    break;
                }
                remove(szString);
                break;
            }
        }
    }

#ifndef __linux__
    if (szOldPath[1] == ':') ChangeDrive(szOldPath[0]);
#endif
    chdir(szOldPath);
    closedir(pDir);

    return 1;
}

static MSG_TEXT *pFirstText;

/* Write message to message base */
int TossMessage(char *szSender, char *szReceiver, char *szSubject, char *szDateTime, DUPE_REC *pDupe)
{
    MSG_TEXT *temp;
    static int iMsgs;
    struct tm Tim;
    char szTemp[81],szAreaTag[61],*pStr;
    int i,iTextPos,dupe;

    dupe = IsDupe(pDupe);
    if (dupe && pDupeMsgArea == NULL)
    {
        /* No dupe message area, kill it.. */
        uDupes++;
        printf("D"); fflush(stdout);
        return 2;
    }

    /* Get message date */
    memset(&Tim,0,sizeof(Tim));
    szDateTime[13] = ' '; szDateTime[16] = ' ';
    sscanf(szDateTime,"%d %s %d %d %d %d",&Tim.tm_mday,szTemp,&Tim.tm_year,
           &Tim.tm_hour,&Tim.tm_min,&Tim.tm_sec);

    /* Get month */
    for (i = 0; i < 12; i++)
    {
        if (stricmp(szMonth[i], szTemp) == 0)
        {
            Tim.tm_mon = i;
            break;
        }
    }

    /* And here it is! */
    tMsgDate = mktime(&Tim);

    iTextPos = 0;
    /* Get message area */
    szAreaTag[0] = '\0';
    if (strncmp(pFirstText->sText,"AREA:",5) != 0)
    {
        /* Netmail.. */
        if (!dupe)
        {
            /* Check if this is form something@in.somewhere */
            pStr = szReceiver; i = 0;
            while (*pStr != '\0')
            {
                if (*pStr == '@')
                {
                    if (i != 0) break;
                    i = 1;
                }
                if (*pStr == '.' && i == 1) i = 2;
                pStr++;
            }
            if (i == 2 || stricmp(szReceiver, "UUCP") == 0)
            {
                /* OK, was an email address, send it as email */
                pFirstMsgText = pFirstText;
                SendEmail(szSender, szReceiver, szSubject, 0);
            }
            else if (stricmp(szReceiver,"AREAFIX") == 0 || stricmp(szReceiver,"XTOSS") == 0 || stricmp(szReceiver,"CONFMGR") == 0)
            {
                /* AreaFix message */
                pFirstMsgText = pFirstText;
                AreaFixMessage(szSubject);
            }
            pMsgAreaRecord = pNetmailArea;
        }
    }
    else
    {
        /* Echomail.. */
        szAreaTag[sizeof(szAreaTag)-1] = '\0';
        memcpy(szAreaTag, pFirstText->sText+5, sizeof(szAreaTag)-1);

        pStr = strchr(szAreaTag, 13);
        if (pStr != NULL) *pStr = '\0';

        /* Search area tag */
        pMsgAreaRecord = pFirstMsgArea;
        while (pMsgAreaRecord != NULL && !dupe)
        {
#ifndef VBOARD
            if (stricmp(pMsgAreaRecord->szAreaTag, szAreaTag) == 0 && pMsgAreaRecord->pszPath != NULL)
#else
            if (stricmp(pMsgAreaRecord->szAreaTag, szAreaTag) == 0)
#endif
            {
                /* Got it! */
                iTextPos = strlen(szAreaTag)+6; /* AREA:<text><CR> */
                break;
            }
            pMsgAreaRecord = pMsgAreaRecord->pNext;
        }

        if (pMsgAreaRecord == NULL && !dupe)
        {
            /* Msg area not found -> bad */
            pMsgAreaRecord = pBadMsgArea;
            uBads++;
            /* Add FROM and TO fields.. */
            temp = new MSG_TEXT;
            if (temp == NULL)
            {
                WriteLog('!', "TossMessage() : msgtext : malloc( %d ) : Not enough memory!", sizeof(MSG_TEXT));
                printf("TossMessage() : msgtext : malloc( %d ) : Not enough memory!\n", sizeof(MSG_TEXT));
                exit(1);
            }
            temp->pNext = pFirstText;
            sprintf(temp->sText,"FROM: %u:%u/%u.%u\rTO: %u:%u/%u.%u\r",
                    wOrigZone, wOrigNet, wOrigNode, wOrigPoint,
                    wDestZone, wDestNet, wDestNode, wDestPoint);
            pFirstText = temp;
        }
    }

    /* Save message to dupe area */
    if (dupe) pMsgAreaRecord = pDupeMsgArea;

    if (pLastMsgArea != pMsgAreaRecord)
    {
        /* Print area name */
        printf("\n%s: 1",pMsgAreaRecord->szAreaTag); iMsgs = 1;
        pLastMsgArea = pMsgAreaRecord;
    }
    else
    {
        /* Cursor back */
        sprintf(szTemp,"%d",iMsgs); pStr = szTemp;
        while (*pStr != '\0') *pStr++ = 8;
        printf(szTemp);
        iMsgs++;
        printf("%d",iMsgs);
        fflush(stdout);
    }

    if (pMsgAreaRecord == NULL)
    {
        /* Crash-protection */
        printf("pMsgAreaRecord == NULL ?? This should NEVER happen!!!\n");
        WriteLog('!',"pMsgAreaRecord == NULL ?? This should NEVER happen!!!\n");
        exit(1);
    }

    pFirstMsgText = pFirstText;

#ifdef VBOARD
    if (!OpenMsgArea(pMsgAreaRecord->wAreaNum))
#else
    if (!OpenMsgArea(pMsgAreaRecord->pszPath,1))
#endif
    {
        printf("Could not open message area `%s`\n",pMsgAreaRecord->szAreaTag);
        WriteLog('!',"Could not open message area `%s`",pMsgAreaRecord->szAreaTag);
        return 0;
    }

    /* Set WriteMessage() information */
    strcpy(szMsgFrom, szSender);
    strcpy(szMsgTo, szReceiver);
#ifdef VBOARD
    if (pMsgAreaRecord == pNetmailArea)
    {
        sprintf(szMsgSubj, "%u:%u/%u.%u %s", wOrigZone, wOrigNet, wOrigNode, wOrigPoint, szSubject);
    }
    else
#endif
        strcpy(szMsgSubj, szSubject);

    fMsgSent = 1;
    return WriteMessage(iTextPos, 1);
}

/* Open/create .out file */
int OpenOutputPacket(unsigned uNodePos)
{
    PKT_HEADER PktHdr;
    int hFile;

    time_t _tim;
    struct tm *tim;

    struct stat statbuf;
    char fname[256];
    unsigned num;

    /* Check if file is already made */
    for (num=0; num<uOutFiles; num++)
    {
        if (output[num].uNodePos == uNodePos) return num;
    }

    num = 1;
    pNodeRecord = pFirstNode;
    while (uNodePos > num)
    {
        pNodeRecord = pNodeRecord->pNext;
        num++;
    }

    _tim = time(NULL);
    tim = localtime(&_tim);

    if (stat(OUTPUT_PATH, &statbuf) != 0) MakeDir(OUTPUT_PATH);

    /* Change name until file don't exist */
    hFile = -1;
    do
    {
        if (hFile > 0) close(hFile);
        sprintf(fname, OUTPUT_PATH SSLASH"%08lx.out", _tim);
        if (stat(fname, &statbuf) == 0)
        {
            hFile = sopen(fname, O_RDONLY | O_BINARY, SH_DENYRW);
            if (hFile == -1) hFile = 0;
        }
        else
            hFile = -1;
        _tim++;
    }
    while (hFile != -1);

    /* Create file */
    hFile = creat(fname, READWRITE);
    if (hFile == -1)
    {
        printf("Can't create file `%s`\n",fname);
        WriteLog('!',"Can't create file `%s`",fname);
        return -1;
    }
    setmode(hFile, O_BINARY);

    /* Create packet header */
    memset(&PktHdr,0,sizeof(PktHdr));
    PktHdr.wOrigNode = pNodeRecord->wYourNode;
    PktHdr.wDestNode = pNodeRecord->wNode;
    PktHdr.wYear = (unsigned short) tim->tm_year; PktHdr.wMonth = (unsigned short) (tim->tm_mon-1); PktHdr.wDay = (unsigned short) tim->tm_mday;
    PktHdr.wHour = (unsigned short) tim->tm_hour; PktHdr.wMin = (unsigned short) tim->tm_min; PktHdr.wSec = (unsigned short) tim->tm_sec;
    PktHdr.wPacketType = 2;
    PktHdr.wOrigNet = pNodeRecord->wYourNet;
    PktHdr.wDestNet = pNodeRecord->wNet;
    PktHdr.byProdCode = 0;
    memcpy(PktHdr.sPassword, pNodeRecord->szPassword, strlen(pNodeRecord->szPassword));
    PktHdr.wQOrigZone = pNodeRecord->wYourZone;
    PktHdr.wQDestZone = pNodeRecord->wZone;
    PktHdr.wOrigZone = pNodeRecord->wYourZone;
    PktHdr.wDestZone = pNodeRecord->wZone;
    PktHdr.wOrigPoint = pNodeRecord->wYourPoint;
    PktHdr.wDestPoint = pNodeRecord->wPoint;
    PktHdr.wCapWord = 1; /* FSC-0039 compatible */
    PktHdr.wCapValid = 256; /* FSC-0039 compatible */
    if (write(hFile,&PktHdr,sizeof(PktHdr)) != sizeof(PktHdr))
    {
        printf("OpenOutputPacket() - Couldn't write to file `%s`\n",fname);
        WriteLog('!',"OpenOutputPacket() - Couldn't write to file `%s`",fname);
        close(hFile);
        remove(fname);
        return -1;
    }

    output[uOutFiles].hFile = hFile;
    output[uOutFiles].uNodePos = uNodePos;
    strcpy(output[uOutFiles].szFileName, fname);
    uOutFiles++;

    return uOutFiles-1;
}

/* Get netmail route address */
int GetRouteVia(ADDR_REC *pAddr, ADDR_REC *pViaAddr)
{
    char szAddr[30];
    char *pRoute;
    int closest,closest_num,iNodePos;
    int num,waiting;

    if (pAddr->wPoint == 0)
        sprintf(szAddr, "%u:%u/%u", pAddr->wZone, pAddr->wNet, pAddr->wNode);
    else
        sprintf(szAddr, "%u:%u/%u.%u", pAddr->wZone, pAddr->wNet, pAddr->wNode, pAddr->wPoint);

    closest = 0; closest_num = 0;
    pNodeRecord = pFirstNode; iNodePos = 1;
    while (pNodeRecord != NULL)
    {
        num = 0; waiting = 0; pRoute = pNodeRecord->szRoute;
        for (;;)
        {
            switch (*pRoute)
            {
                case '\0':
                case ' ':
                    if (szAddr[num] == '\0')
                    {
                        closest_num = iNodePos;
                        closest = num+1;
                        break;
                    }
                    num = 0; waiting = 0;
                    break;
                case '*':
                    if (waiting) break;
                    if (closest < num)
                    {
                        closest_num = iNodePos;
                        closest = num;
                    }
                    waiting = 1;
                    break;
                case '.':
                    if (waiting) break;
                    if (szAddr[num] == '\0') break;
                    if (szAddr[num] != '.')
                    {
                        num = 0;
                        waiting = 1;
                    }
                    num++;
                    break;
                default:
                    if (waiting) break;
                    if (*pRoute != szAddr[num])
                    {
                        num = 0;
                        waiting = 1;
                    }
                    num++;
                    break;
            }
            if (*pRoute == '\0') break;
            pRoute++;
        }
        pNodeRecord = pNodeRecord->pNext; iNodePos++;
    }

    if (closest == 0)
    {
        printf("\nDon't know how to route to %s\n",szAddr);
        WriteLog('*',"Don't know how to route to %s",szAddr);
        return 0;
    }

    pNodeRecord = pFirstNode; iNodePos = 1;
    while (iNodePos < closest_num)
    {
        if (pNodeRecord == NULL)
        {
            printf("\nGetRouteVia() : pNodeRecord == NULL ?? This should NEVER happen!");
            WriteLog('!',"GetRouteVia() : pNodeRecord == NULL ?? This should NEVER happen!");
            exit(1);
        }
        pNodeRecord = pNodeRecord->pNext; iNodePos++;
    }

    pViaAddr->wZone = pNodeRecord->wZone;
    pViaAddr->wNet = pNodeRecord->wNet;
    pViaAddr->wNode = pNodeRecord->wNode;
    pViaAddr->wPoint = pNodeRecord->wPoint;
    printf("\nExporting netmail to %s via %u:%u/%u.%u\n", szAddr,
           pViaAddr->wZone, pViaAddr->wNet, pViaAddr->wNode, pViaAddr->wPoint);
    WriteLog('+',"Exporting netmail to %s via %u:%u/%u.%u", szAddr,
             pViaAddr->wZone, pViaAddr->wNet, pViaAddr->wNode, pViaAddr->wPoint);

    if (pViaAddr->wNode == pAddr->wNode &&
        pViaAddr->wNet == pAddr->wNet &&
        pViaAddr->wPoint == pAddr->wPoint &&
        pViaAddr->wZone == pAddr->wZone)
    {
        /* Direct netmail in there.. */
        return 2;
    }

    return 1;
}

/* Add new SEEN-BY to list */
void SortAddSeenBy(unsigned short wNet, unsigned short wNode)
{
    SEEN_REC *pTempSeenBy,*pLastSeenBy;

    /* Create SEEN-BY record */
    pTempSeenBy = new SEEN_REC;
    if (pTempSeenBy == NULL)
    {
        printf("SortAddSeenBy() : malloc( %d ) : Not enough memory!\n",sizeof(SEEN_REC));
        WriteLog('!',"SortAddSeenBy() : malloc( %d ) : Not enough memory!",sizeof(SEEN_REC));
        exit(1);
    }
    pTempSeenBy->pNext = NULL;
    pTempSeenBy->pPrev = NULL;
    pTempSeenBy->wNet = wNet;
    pTempSeenBy->wNode = wNode;
    uSeenBys++;

    /* Seek to right position */
    pSeenBy = pFirstSeenBy; pLastSeenBy = NULL;
    while (pSeenBy != NULL)
    {
        if (wNet < pSeenBy->wNet || (wNet == pSeenBy->wNet && wNode < pSeenBy->wNode))
        {
            /* Insert new record here */
            break;
        }
        if (wNet == pSeenBy->wNet && wNode == pSeenBy->wNode)
        {
            /* Already in there! */
            delete pTempSeenBy;
            uSeenBys--;
            return;
        }
        pLastSeenBy = pSeenBy;
        pSeenBy = pSeenBy->pNext;
    }

    /* Update linked list */
    pTempSeenBy->pNext = pSeenBy;
    if (pSeenBy == NULL)
    {
        pTempSeenBy->pPrev = pLastSeenBy;
        if (pLastSeenBy == NULL)
        {
            /* This is the first record */
            pFirstSeenBy = pTempSeenBy;
        }
        else
        {
            pLastSeenBy->pNext = pTempSeenBy;
        }
    }
    else
    {
        if (pSeenBy == pFirstSeenBy) pFirstSeenBy = pTempSeenBy;
        pTempSeenBy->pPrev = pSeenBy->pPrev;
        if (pSeenBy->pPrev != NULL) pSeenBy->pPrev->pNext = pTempSeenBy;
        pSeenBy->pPrev = pTempSeenBy;
    }
}

/* Remove memory used by seen-by lines */
void DestroySeenByList(void)
{
    while (pFirstSeenBy != NULL)
    {
        pSeenBy = pFirstSeenBy->pNext;
        delete pFirstSeenBy;
        pFirstSeenBy = pSeenBy;
    }
    uSeenBys = 0; wLastSeenByNet = 0;
}

/* List of SEEN-BYs to one string */
char *SeenByToString(void)
{
    char *pString,*pStr;

    /* Get memory for string */
    pString = new char[uSeenBys*12];
    if (pString == NULL)
    {
        printf("SeenByToString() : malloc( %d ) : Not enough memory!\n",uPaths*12);
        WriteLog('!',"SeenByToString() : malloc( %d ) : Not enough memory!",uPaths*12);
        exit(1);
    }

    /* Put addresses to string */
    pSeenBy = pFirstSeenBy; pStr = pString; wLastSeenByNet = 0;
    while (pSeenBy != NULL)
    {
        if (pSeenBy->wNet != wLastSeenByNet || pSeenBy == pFirstSeenBy)
        {
            /* Add net */
            pStr += sprintf(pStr, "%u/",pSeenBy->wNet);
            wLastSeenByNet = pSeenBy->wNet;
        }
        /* Add node */
        pStr += sprintf(pStr, "%u",pSeenBy->wNode);

        pSeenBy = pSeenBy->pNext;
        if (pSeenBy != NULL) *pStr++ = ' ';
    }
    *pStr = '\0';

    /* Remove SEEN-BYs from memory */
    DestroySeenByList();

    return pString;
}

/* Add list of SEEN-BYs */
void AddSeenBy(char *szList)
{
    unsigned short wNet, wNode;

    while (*szList != '\0')
    {
        /* Skip spaces */
        while (*szList == 13 || *szList == ' ') szList++;
        if (*szList == '\0') break;

        /* Get string */
        wNet = wLastSeenByNet; wNode = 0;
        while (*szList != 13 && *szList != ' ' && *szList != '\0')
        {
            if (*szList >= '0' && *szList <= '9') wNode = (unsigned short) (wNode*10 + (*szList-'0'));
            if (*szList == '/')
            {
                wNet = wNode;
                wNode = 0;
            }
            szList++;
        }
        wLastSeenByNet = wNet;
        SortAddSeenBy(wNet, wNode);
    }
}

/* Add list of paths */
void AddPath(char *szList)
{
    unsigned short wNet, wNode;
    PATH_REC *pTempPath;

    while (*szList != '\0')
    {
        /* Skip spaces */
        while (*szList == 13 || *szList == ' ') szList++;
        if (*szList == '\0') break;

        /* Get string */
        wNet = wLastPathNet; wNode = 0;
        while (*szList != 13 && *szList != ' ' && *szList != '\0')
        {
            if (*szList >= '0' && *szList <= '9') wNode = (unsigned short) (wNode*10 + (*szList-'0'));
            if (*szList == '/')
            {
                wNet = wNode;
                wNode = 0;
            }
            szList++;
        }
        wLastPathNet = wNet;

        /* Add to list */
        pTempPath = new PATH_REC;
        if (pTempPath == NULL)
        {
            printf("AddPath() : malloc( %d ) : Not enough memory\n", sizeof(PATH_REC));
            WriteLog('!',"AddPath() : malloc( %d ) : Not enough memory", sizeof(PATH_REC));
            exit(1);
        }
        pTempPath->pNext = NULL;
        pTempPath->wNet = wNet;
        pTempPath->wNode = wNode;
        if (pPath != NULL) pPath->pNext = pTempPath;
        if (pFirstPath == NULL) pFirstPath = pTempPath;
        pPath = pTempPath;
        uPaths++;
    }
}

/* Remove memory used by paths */
void DestroyPathList(void)
{
    while (pFirstPath != NULL)
    {
        pPath = pFirstPath->pNext;
        delete pFirstPath;
        pFirstPath = pPath;
    }
    uPaths = 0; wLastPathNet = 0;
}

/* List of SEEN-BYs to one string */
char *PathToString(void)
{
    char *pString,*pStr;

    /* Get memory for string */
    pString = new char[uPaths*12];
    if (pString == NULL)
    {
        printf("PathToString() : malloc( %d ) : Not enough memory!\n",uPaths*12);
        WriteLog('!',"PathToString() : malloc( %d ) : Not enough memory!",uPaths*12);
        exit(1);
    }

    /* Put addresses to string */
    pPath = pFirstPath; pStr = pString; wLastPathNet = 0;
    while (pPath != NULL)
    {
        if (pPath->wNet != wLastPathNet || pPath == pFirstPath)
        {
            /* Add net */
            pStr += sprintf(pStr, "%u/",pPath->wNet);
            wLastPathNet = pPath->wNet;
        }
        /* Add node */
        pStr += sprintf(pStr, "%u",pPath->wNode);

        pPath = pPath->pNext;
        if (pPath != NULL) *pStr++ = ' ';
    }
    *pStr = '\0';

    /* Remove PATHs from memory */
    DestroyPathList();

    return pString;
}

/* Print string to file and max. 79 characters per line */
void SplitPrintf(int hFile, char *fmt, char *data, int fmt_len)
{
    char *pbuf,*strp,buf[80];

    pbuf = data;
    for (;;)
    {
        if ((int) strlen(pbuf) > 79-fmt_len)
        {
            buf[79-fmt_len] = '\0';
            strncpy(buf, pbuf, 79-fmt_len);
            strp = strrchr(buf, ' ');
            if (strp == NULL)
            {
                printf("Something's REALLY wrong with kludge `%s`\n",data);
                WriteLog('!',"Something's REALLY wrong with kludge `%s`",data);
                break;
            }
            strp = pbuf+(strp-buf);
            *strp++ = '\0';
        }
        else
        {
            strp = NULL;
        }

        sprintf(buf,fmt,pbuf);
        write(hFile,buf,strlen(buf));
        if (strp == NULL) break;
        pbuf = strp;
    }
}

/* Export message */
int ExportMessage(int fTransit)
{
    MSG_HEADER MsgHdr;
    MSG_TEXT *pMsgText;
    ADDR_REC *pAddr,*pTempAddr;
    ADDR_REC DestAddr,ViaAddr;
    int hOut, iBufPos, match;
    unsigned long ulTime, oldtextlen;
    unsigned uNodePos,uBytes;
    char *pSubject,*pStr,sBuf[82],fNoWrite,fOrigin;
    time_t _tim;
    struct tm *pTim;
#ifdef VBOARD
    char sAddress[30];
    long lReaded;
#else
    char lastchr;
    int num;
#endif

    match = 0;
    if (pMsgAreaRecord == pNetmailArea)
    {
        /* Get netmail destination address from subject */
        DestAddr.pNext = NULL;

        if (!fTransit)
        {
#ifdef VBOARD
            /* Local message, get destination from subject */
            pSubject = strchr(szMsgSubj,' ');
            if (pSubject != NULL) *pSubject++ = '\0'; else pSubject = "";
            if (!str2addr(szMsgSubj,&DestAddr.wZone,&DestAddr.wNet,&DestAddr.wNode,&DestAddr.wPoint))
            {
                /* Error in netmail */
                printf("\nError in netmail destination address: `%s`\n",szMsgSubj);
                WriteLog('*',"Error in netmail destination address: `%s`",szMsgSubj);
                return 2;
            }
#else
            pSubject = szMsgSubj;
            if (!str2addr(szMsgDest,&DestAddr.wZone,&DestAddr.wNet,&DestAddr.wNode,&DestAddr.wPoint))
            {
                /* Error in netmail */
                printf("\nError in netmail destination address: `%s`\n",szMsgDest);
                WriteLog('*',"Error in netmail destination address: `%s`",szMsgDest);
                return 2;
            }
#endif
        }
        else
        {
            /* Transit message, get destination from saved kludges */
            DestAddr.wZone = wDestZone;
            DestAddr.wNet = wDestNet;
            DestAddr.wNode = wDestNode;
            DestAddr.wPoint = wDestPoint;
            pSubject = szMsgSubj;

            if (DestAddr.wZone == 0)
            {
                /* No INTL-kludge -> already in destination */
                return 1;
            }
        }

        /* Check if this was one of our own addresses.. */
        pAddr = pFirstAddr;
        while (pAddr != NULL)
        {
            if (pAddr->wZone == DestAddr.wZone && pAddr->wNet == DestAddr.wNet &&
                pAddr->wNode == DestAddr.wNode && pAddr->wPoint == DestAddr.wPoint)
            {
                /* Was ours, netmail in destination.. */
                return 1;
            }
            pAddr = pAddr->pNext;
        }

        /* And route it via ... */
        pAddr = &ViaAddr;
        match = GetRouteVia(&DestAddr, pAddr);
        if (!match)
        {
            /* Don't know how to route here.. */
            return 2;
        }
        pAddr->pNext = NULL;
    }
    else
    {
        pAddr = pMsgAreaRecord->pExport;
        pSubject = szMsgSubj;
    }

    while (pAddr != NULL)
    {
        /* Don't send message back where it came from.. */
        if (pAddr->wZone == wOrigZone && pAddr->wNet == wOrigNet &&
            pAddr->wNode == wOrigNode && pAddr->wPoint == wOrigPoint)
        {
            pAddr = pAddr->pNext;
            continue;
        }

        /* Search correct NODE_REC */
        uNodePos = 1; pNodeRecord = pFirstNode;
        while (pNodeRecord != NULL)
        {
            if (pNodeRecord->wZone == pAddr->wZone &&
                pNodeRecord->wNet == pAddr->wNet &&
                pNodeRecord->wNode == pAddr->wNode &&
                pNodeRecord->wPoint == pAddr->wPoint)
            {
                /* Got it! */
                break;
            }
            pNodeRecord = pNodeRecord->pNext;
            uNodePos++;
        }

        if (pNodeRecord == NULL)
        {
            printf("\nExportMessage() : pNodeRecord == NULL! This should NEVER happen!\n");
            WriteLog('!',"ExportMessage() : pNodeRecord == NULL! This should NEVER happen!");
            exit(1);
        }

        /* Create message header */
        memset(&MsgHdr,0,sizeof(MsgHdr));
        MsgHdr.wProdCode = 2;
        if (pMsgAreaRecord == pNetmailArea)
        {
            MsgHdr.wOrigNet = pNodeRecord->wYourNet;
            MsgHdr.wOrigNode = pNodeRecord->wYourNode;
        }
        else
        {
            MsgHdr.wOrigNet = pMsgAreaRecord->OwnAddr.wNet;
            MsgHdr.wOrigNode = pMsgAreaRecord->OwnAddr.wNode;
        }
        MsgHdr.wDestNet = pAddr->wNet;
        MsgHdr.wDestNode = pAddr->wNode;
        MsgHdr.wAttrib = fMsgPrivate ? 1 : 0;

        /* Open/create .out file */
        hOut = OpenOutputPacket(uNodePos);
        if (hOut == -1) return 0;

        /* Write msg header */
        if (write(output[hOut].hFile,&MsgHdr,sizeof(MsgHdr)) != sizeof(MsgHdr))
        {
            printf("\nExportMessage() - Couldn't write to .pkt file!\n");
            WriteLog('!',"ExportMessage() - Couldn't write to .pkt file!");
            return 0;
        }

        pTim = localtime((time_t *) &tMsgDate);

        /* Write message date */
        sprintf(sBuf,"%02d %s %02d  %02d:%02d:%02d",
                pTim->tm_mday,szMonth[pTim->tm_mon],pTim->tm_year%100,
                pTim->tm_hour,pTim->tm_min,pTim->tm_sec);
        write(output[hOut].hFile,sBuf,strlen(sBuf)+1);

        /* Just another internal error checking.. */
        if (pSubject == NULL)
        {
            printf("\nExportMessage() : pSubject == NULL! This should NEVER happen!\n");
            WriteLog('!',"ExportMessage() : pSubject == NULL! This should NEVER happen!");
            exit(1);
        }

        /* Write from-, to- and subject fields */
        write(output[hOut].hFile, szMsgTo, strlen(szMsgTo)+1);
        write(output[hOut].hFile, szMsgFrom, strlen(szMsgFrom)+1);
        write(output[hOut].hFile, pSubject, strlen(pSubject)+1);

        if (!fTransit)
        {
            /* Local message - add kludges etc. */

            if (pMsgAreaRecord != pNetmailArea)
            {
                /* Echomail area, write area tag */
                sprintf(sBuf,"AREA:%s\r", pMsgAreaRecord->szAreaTag);
                write(output[hOut].hFile, sBuf, strlen(sBuf));

#ifdef VBOARD
                /* Write MSGID */
                ulTime = time(NULL);
                while (ulTime <= ulLastTime) ulTime++;
                ulLastTime = ulTime;
                sprintf(sBuf,"\x01MSGID: %u:%u/%u.%u %08lx\r",
                        pMsgAreaRecord->OwnAddr.wZone, pMsgAreaRecord->OwnAddr.wNet,
                        pMsgAreaRecord->OwnAddr.wNode, pMsgAreaRecord->OwnAddr.wPoint, ulTime);
                write(output[hOut].hFile, sBuf, strlen(sBuf));
#endif
            }
            else
            {
                /* Netmail kludges */

                /* INTL */
                if (match != 2)
                {
                    sprintf(sBuf,"\x01INTL %u:%u/%u %u:%u/%u\r",
                            DestAddr.wZone, DestAddr.wNet, DestAddr.wNode,
                            pNodeRecord->wYourZone, pNodeRecord->wYourNet,
                            pNodeRecord->wYourNode
                           );
                    write(output[hOut].hFile, sBuf, strlen(sBuf));

                    if (DestAddr.wPoint != 0)
                    {
                        /* TOPT */
                        sprintf(sBuf,"\x01TOPT %u\r", DestAddr.wPoint);
                        write(output[hOut].hFile, sBuf, strlen(sBuf));
                    }
                    if (pNodeRecord->wYourPoint != 0)
                    {
                        /* FMPT */
                        sprintf(sBuf,"\x01FMPT %u\r",pNodeRecord->wYourPoint);
                        write(output[hOut].hFile, sBuf, strlen(sBuf));
                    }
                }

#ifdef VBOARD
                /* Write MSGID */
                ulTime = time(NULL);
                while (ulTime <= ulLastTime) ulTime++;
                ulLastTime = ulTime;
                sprintf(sBuf,"\x01MSGID: %u:%u/%u.%u %08lx\r",
                        pNodeRecord->wYourZone, pNodeRecord->wYourNet,
                        pNodeRecord->wYourNode, pNodeRecord->wYourPoint, ulTime);
                write(output[hOut].hFile, sBuf, strlen(sBuf));
#endif
            }

#ifndef VBOARD
            if (szMsgID[0] == '\0')
            {
                /* Create MSGID if BBS didn't make it. */
                ulTime = time(NULL);
                while (ulTime <= ulLastTime) ulTime++;
                ulLastTime = ulTime;
                sprintf(sBuf,"\x01MSGID: %u:%u/%u.%u %08lx\r",
                        pNodeRecord->wYourZone, pNodeRecord->wYourNet,
                        pNodeRecord->wYourNode, pNodeRecord->wYourPoint, ulTime);
                write(output[hOut].hFile, sBuf, strlen(sBuf));
            }
            else
            {
                /* Write MSGID */
                sprintf(sBuf,"\x01MSGID: %s\r", szMsgID);
                write(output[hOut].hFile, sBuf, strlen(sBuf));
            }

            if (szReplyID[0] != '\0')
            {
                /* Write REPLYID */
                sprintf(sBuf,"\x01REPLY: %s\r", szReplyID);
                write(output[hOut].hFile, sBuf, strlen(sBuf));
            }

            /* Write other kludges */
            for (num=0; num<kludges; num++)
            {
                sprintf(sBuf,"\x01%s\r", msgkludge[num]);
                write(output[hOut].hFile, sBuf, strlen(sBuf));
            }
#endif
            /*if (reply_kludge[0] != '\0')
            {*/
                /* Write REPLY */
            /*    sprintf(buf,"\x01REPLY: %s\r",reply_kludge);
                write(output[hOut].hFile,buf,strlen(buf));
            }*/

            /* Write message text */
            pMsgText = pFirstMsgText;
#ifdef VBOARD
            lReaded = 0;
#else
            lastchr = '\r';
#endif
            oldtextlen = wMsgTextLength;
            while (pMsgText != NULL)
            {
#ifdef VBOARD
                /* Have to change all \0 chars to CRs */
                pStr = pMsgText->sText;
                while ((unsigned long) lReaded < wMsgTextLength && pStr < pMsgText->sText+sizeof(pMsgText->sText))
                {
                    if (*pStr == '\0') *pStr = 13;
                    pStr++; lReaded++;
                }
                uBytes = (int) (pStr-pMsgText->sText);
#else
                if (wMsgTextLength > sizeof(pMsgText->sText))
                    uBytes = sizeof(pMsgText->sText);
                else
                    uBytes = (unsigned) wMsgTextLength;
                lastchr = pMsgText->sText[uBytes];
#endif

                if (write(output[hOut].hFile, pMsgText->sText, uBytes) != (int) uBytes)
                {
                    printf("\nCouldn't write %d bytes of message text, out of disk space?\n",uBytes);
                    WriteLog('!',"Couldn't write %d bytes of message text, out of disk space?",uBytes);
                    return 0;
                }

                wMsgTextLength -= uBytes;
                pMsgText = pMsgText->pNext;
            }
            wMsgTextLength = oldtextlen;

#ifdef VBOARD
            if (pMsgAreaRecord != pNetmailArea)
            {
                /* Add Tearline */
                if (szTearLine[0] == '\0')
                    strcpy(sBuf, "---\r");
                else
                    sprintf(sBuf, "--- %s\r", szTearLine);
                write(output[hOut].hFile, sBuf, strlen(sBuf));

                /* Add origin */
                if (pMsgAreaRecord->szOrigin[0] == '\0')
                {
                    /* Use default origin line */
                    if (pMsgAreaRecord->OwnAddr.wPoint != 0)
                    {
                        /* With point */
                        sprintf(sAddress, " (%u:%u/%u.%u)\r",
                                pMsgAreaRecord->OwnAddr.wZone, pMsgAreaRecord->OwnAddr.wNet,
                                pMsgAreaRecord->OwnAddr.wNode, pMsgAreaRecord->OwnAddr.wPoint);
                    }
                    else
                    {
                        /* Without point */
                        sprintf(sAddress, " (%u:%u/%u)\r",
                                pMsgAreaRecord->OwnAddr.wZone, pMsgAreaRecord->OwnAddr.wNet,
                                pMsgAreaRecord->OwnAddr.wNode);
                    }
                    sprintf(sBuf, " * Origin: %s", szOrigin);
                    sBuf[80-strlen(sAddress)] = '\0';
                    strcat(sBuf, sAddress);
                }
                else
                {
                    /* Area specific origin line */
                    sprintf(sBuf, " * Origin: %s\r", pMsgAreaRecord->szOrigin);
                }
                write(output[hOut].hFile, sBuf, strlen(sBuf));
            }
#else
            if (lastchr != '\r') write(output[hOut].hFile, "\r", 1);
#endif
        }
        else
        {
            /* Transit message */

            /* Write message text */
            pMsgText = pFirstMsgText; iBufPos = 0; fOrigin = 0;
            while (pMsgText != NULL)
            {
                pStr = pMsgText->sText;
                while (*pStr != '\0' && pStr < pMsgText->sText+sizeof(pMsgText->sText))
                {
                    sBuf[iBufPos++] = *pStr;

                    if (*pStr == 13 || *++pStr == '\0' || iBufPos == sizeof(sBuf)-1)
                    {
                        fNoWrite = 0;
                        if (*pStr == 13 || *pStr == '\0')
                        {
                            if (!fOrigin && strncmp(sBuf," * Origin:",10) == 0) fOrigin = 1;
                            else if (fOrigin && strncmp(sBuf,"SEEN-BY:",8) == 0)
                            {
                                /* Found SEEN-BY line! */
                                fNoWrite = 1;
                                sBuf[iBufPos] = '\0';
                                AddSeenBy(sBuf+8);
                            }
                            else if (strncmp(sBuf,"\x01PATH:",6) == 0)
                            {
                                /* Found PATH line! */
                                fNoWrite = 1;
                                sBuf[iBufPos] = '\0';
                                AddPath(sBuf+6);
                            }
                        }
                        if (!fNoWrite)
                        {
                            if (write(output[hOut].hFile, sBuf, iBufPos) != iBufPos)
                            {
                                printf("\nCouldn't write %d bytes of message text, out of disk space?\n",iBufPos);
                                WriteLog('!',"Couldn't write %d bytes of message text, out of disk space?",iBufPos);
                                return 0;
                            }
                        }
                        if (*pStr == 13) pStr++;
                        iBufPos = 0;
                    }
                }

                pMsgText = pMsgText->pNext;
            }

            if (iBufPos > 0)
            {
                printf("\nExportMessage() : iBufPos > 0 ?? This should NEVER happen!!\n");
                WriteLog('!',"ExportMessage() : iBufPos > 0 ?? This should NEVER happen!!");
                exit(1);
            }
        }

        if (pMsgAreaRecord == pNetmailArea)
        {
            if (match != 2)
            {
                /* Add VIA kludge to netmails */
                _tim = time(NULL);
                pTim = localtime(&_tim);
                sprintf(sBuf,"\x01Via %u:%u/%u.%u @%02d%02d%02d%02d%02d%02d XToss "VERSION"\r",
                        /* Address */
                        pNodeRecord->wYourZone,pNodeRecord->wYourNet,
                        pNodeRecord->wYourNode,pNodeRecord->wYourPoint,
                        /* Date */
                        pTim->tm_year, pTim->tm_mon+1, pTim->tm_mday,
                        /* Time */
                        pTim->tm_hour, pTim->tm_min, pTim->tm_sec);
                if (write(output[hOut].hFile, sBuf, strlen(sBuf)) != (int) strlen(sBuf))
                {
                    printf("\nCouldn't write VIA kludge, out of disk space?\n");
                    WriteLog('!',"Couldn't write VIA kludge, out of disk space?");
                    return 0;
                }
            }
        }
        else
        {
            /* Echomail, add addresses to PATH and SEEN-BY kludges */

            /* Add export addresses to seen-by */
            pTempAddr = pMsgAreaRecord->pExport;
            while (pTempAddr != NULL)
            {
                SortAddSeenBy(pTempAddr->wNet, pTempAddr->wNode);
                pTempAddr = pTempAddr->pNext;
            }

            /* Add your address to seen-by */
            SortAddSeenBy(pMsgAreaRecord->OwnAddr.wNet, pMsgAreaRecord->OwnAddr.wNode);
            //WriteLog('?',"debug - AddSeenBy(%u,%u)", pMsgAreaRecord->OwnAddr.wNet, pMsgAreaRecord->OwnAddr.wNode);

            pStr = SeenByToString();
            //WriteLog('?',"debug - pstr = %s", pStr);
            SplitPrintf(output[hOut].hFile, "SEEN-BY: %s\r", pStr, 9);
            delete[] pStr;

            if (pMsgAreaRecord->OwnAddr.wPoint == 0)
            {
                /* Add to path if not a point */
                sprintf(sBuf, "%u/%u", pMsgAreaRecord->OwnAddr.wNet, pMsgAreaRecord->OwnAddr.wNode);
                //WriteLog('?',"debug - AddPath(%s)",sBuf);
                AddPath(sBuf);
                pStr = PathToString();
                //WriteLog('?',"debug - pstr = %s", pStr);
                SplitPrintf(output[hOut].hFile, "\x01PATH: %s\r", pStr, 7);
                delete[] pStr;
            }
        }

        /* Zero to end of message */
        write(output[hOut].hFile, "", 1);

        uMsgsExported++;
        pAddr = pAddr->pNext;
    }

    return 1;
}

/* Toss messages found in .pkt file */
int ReadPktFile(char *pszFileName, int fBadFile)
{
    PKT_HEADER PktHeader;
    MSG_HEADER MsgHeader;
    MSG_TEXT *pText,*pTempText;
    FILE_BUF *pFile;
    DUPE_REC dupe;
    int iLen,fEOT,iRet,iBufPos;
    char szDateTime[21];
    char szReceiver[36],szSender[36],szSubject[72],sBuf[81],*pStr,*pStr2;

    /* Open .pkt file */
    pFile = OpenFile(pszFileName, O_RDONLY | O_BINARY, SH_DENYNO);
    if (pFile == NULL)
    {
        printf("Can't open file `%s`\n", pszFileName);
        WriteLog('!', "Can't open file `%s`", pszFileName);
        return 0;
    }

    printf("\nTossing file `%s`: ",pszFileName);
    WriteLog('-',"Tossing file `%s`",pszFileName);
    pLastMsgArea = NULL;

    /* Read packet header */
    if (ReadFile(pFile, &PktHeader, sizeof(PKT_HEADER)) != sizeof(PKT_HEADER))
    {
        printf("Could not read packet header!\n");
        WriteLog('*',"Could not read packet header!");
        CloseFile(pFile);
        return 0;
    }

    /* Check validity */
    if (PktHeader.wPacketType != 2)
    {
        printf("Invalid PacketType: %d\n", PktHeader.wPacketType);
        WriteLog('*',"Invalid PacketType: %d", PktHeader.wPacketType);
        CloseFile(pFile);
        return 0;
    }

    printf("\nFrom %u:%u/%u.%u to %u:%u/%u.%u\n",
             PktHeader.wOrigZone, PktHeader.wOrigNet,
             PktHeader.wOrigNode, PktHeader.wOrigPoint,
             PktHeader.wDestZone, PktHeader.wDestNet,
             PktHeader.wDestNode, PktHeader.wDestPoint);
    WriteLog('+',"From %u:%u/%u.%u to %u:%u/%u.%u",
             PktHeader.wOrigZone, PktHeader.wOrigNet,
             PktHeader.wOrigNode, PktHeader.wOrigPoint,
             PktHeader.wDestZone, PktHeader.wDestNet,
             PktHeader.wDestNode, PktHeader.wDestPoint);

    pNodeRecord = pFirstNode;
    while (pNodeRecord != NULL && !fBadFile)
    {
        if (PktHeader.wOrigZone == pNodeRecord->wZone &&
            PktHeader.wOrigNet == pNodeRecord->wNet &&
            PktHeader.wOrigNode == pNodeRecord->wNode &&
            PktHeader.wOrigPoint == pNodeRecord->wPoint)
        {
            /* Found node, verify password. */
            memcpy(sBuf,PktHeader.sPassword,8); sBuf[8] = '\0';
            if (stricmp(sBuf, pNodeRecord->szPassword) != 0)
            {
                /* Invalid password! */
                printf("Wrong password `%s` <-> `%s` from %u:%u/%u.%u\n",
                       sBuf, pNodeRecord->szPassword,
                       PktHeader.wOrigZone, PktHeader.wOrigNet,
                       PktHeader.wOrigNode, PktHeader.wOrigPoint);
                WriteLog('*',"Wrong password `%s` <-> `%s` from %u:%u/%u.%u",
                       sBuf, pNodeRecord->szPassword,
                       PktHeader.wOrigZone, PktHeader.wOrigNet,
                       PktHeader.wOrigNode, PktHeader.wOrigPoint);
                return 0;
            }
            printf("ok\n");
            break;
        }
        pNodeRecord = pNodeRecord->pNext;
    }

    if (pNodeRecord == NULL && !fBadFile)
    {
        /* Node not found from configuration. */
        printf("Warning! Unknown node: %u:%u/%u.%u\n",
               PktHeader.wOrigZone, PktHeader.wOrigNet,
               PktHeader.wOrigNode, PktHeader.wOrigPoint);
        WriteLog('*',"Warning! Unknown node: %u:%u/%u.%u",
               PktHeader.wOrigZone, PktHeader.wOrigNet,
               PktHeader.wOrigNode, PktHeader.wOrigPoint);
    }

    /* Read .pkt file */
    iRet = 1;
    while (ReadFile(pFile, &MsgHeader, sizeof(MSG_HEADER)) == sizeof(MSG_HEADER))
    {
        /* Check message validity */
        if (MsgHeader.wProdCode != 2)
        {
            printf("Invalid ProdCode in file `%s`: %d\n", pszFileName, MsgHeader.wProdCode);
            WriteLog('*',"Invalid ProdCode in file `%s`: %d", pszFileName, MsgHeader.wProdCode);
            CloseFile(pFile);
            return 0;
        }

        /* Get message date */
        szDateTime[20] = '\0';
        if (ReadFile(pFile, szDateTime,20) != 20)
        {
        __err_eof:
            printf("Error in file `%s`: Unexpected end of file\n", pszFileName);
            WriteLog('*',"Error in file `%s`: Unexpected end of file", pszFileName);
            CloseFile(pFile);
            break;
        }

        /* Seek back if date didn't take all 20 bytes. */
        iLen = strlen(szDateTime);
        if (iLen < 19) SeekFile(pFile, iLen-19, SEEK_CUR);

        /* Get receiver */
        if (ReadFile(pFile, szReceiver, 36) == 0) goto __err_eof;
        iLen = strlen(szReceiver);
        if (iLen < 35) SeekFile(pFile, iLen-35, SEEK_CUR);

        /* Get sender */
        if (ReadFile(pFile,szSender, 36) == 0) goto __err_eof;
        iLen = strlen(szSender);
        if (iLen < 35) SeekFile(pFile, iLen-35, SEEK_CUR);

        /* Get subject */
        if (ReadFile(pFile,szSubject, 72) == 0) goto __err_eof;
        iLen = strlen(szSubject);
        if (iLen < 71) SeekFile(pFile, iLen-71, SEEK_CUR);

        /* Calculate header CRC */
        dupe.HdrCRC = 0xffffffff;
        pStr = szDateTime; while (*pStr != '\0') CalcCRC(dupe.HdrCRC, *pStr++);
        pStr = szReceiver; while (*pStr != '\0') CalcCRC(dupe.HdrCRC, *pStr++);
        pStr = szSender; while (*pStr != '\0') CalcCRC(dupe.HdrCRC, *pStr++);
        pStr = szSubject; while (*pStr != '\0') CalcCRC(dupe.HdrCRC, *pStr++);

        /* Get message text */
        pText = NULL; pFirstText = NULL;
        fEOT = 0; iBufPos = 0;
#ifndef VBOARD
        szMsgID[0] = '\0'; szReplyID[0] = '\0';
#endif
        wOrigZone = PktHeader.wOrigZone; wOrigNet = MsgHeader.wOrigNet; wOrigNode = MsgHeader.wOrigNode; wOrigPoint = PktHeader.wOrigPoint;
        wDestZone = PktHeader.wDestZone; wDestNet = MsgHeader.wDestNet; wDestNode = MsgHeader.wDestNode; wDestPoint = PktHeader.wDestPoint;
        dupe.TxtCRC = 0xffffffff;
        while (!fEOT)
        {
            /* New record */
            pTempText = new MSG_TEXT;
            if (pTempText == NULL)
            {
                printf("ReadPktFile() : malloc( %d ) : Not enough memory for message text!\n", sizeof(MSG_TEXT));
                WriteLog('!',"ReadPktFile() : malloc( %d ) : Not enough memory for message text!", sizeof(MSG_TEXT));
                exit(1);
            }
            pTempText->pNext = NULL;
            if (pText != NULL) pText->pNext = pTempText;
            if (pFirstText == NULL) pFirstText = pTempText;
            pText = pTempText;

            /* Read text */
            pText->sText[sizeof(pText->sText)-1] = '\0';
            if (ReadFile(pFile, pText->sText, sizeof(pText->sText)-1) == -1) goto __err_eof;

            /* Search for kludges */
            pStr = pText->sText;
            while (*pStr != '\0')
            {
                CalcCRC(dupe.TxtCRC, *pStr);
                if (iBufPos != -1) sBuf[iBufPos++] = *pStr;
                if (iBufPos == sizeof(sBuf)) iBufPos = -1;
                if (*pStr == 13)
                {
                    if (iBufPos > 0)
                    {
                        sBuf[iBufPos-1] = '\0';
                        if (strncmp(sBuf, "\x01INTL ",6) == 0)
                        {
                            /* Origin/destination address got */
                            pStr2 = strchr(sBuf+6, ' ');
                            if (pStr2 != NULL)
                            {
                                *pStr2++ = '\0';
                                str2addr(sBuf+6, &wDestZone, &wDestNet, &wDestNode, &wDestPoint);
                                str2addr(pStr2, &wOrigZone, &wOrigNet, &wOrigNode, &wOrigPoint);
                            }
                        }
                        else if (strncmp(sBuf, "\x01FMPT ",6) == 0)
                        {
                            /* Origin point address got */
                            if (sscanf(sBuf+6, "%hu", &wOrigPoint) != 1) wOrigPoint = 0;
                        }
                        else if (strncmp(sBuf, "\x01TOPT ",6) == 0)
                        {
                            /* Destination point address got */
                            if (sscanf(sBuf+6, "%hu", &wDestPoint) != 1) wDestPoint = 0;
                        }
#ifndef VBOARD
                        else if (strncmp(sBuf, "\x01MSGID: ",8) == 0)
                        {
                            /* Destination point address got */
                            strcpy(szMsgID, sBuf+8);
                        }
                        else if (strncmp(sBuf, "\x01REPLY: ",8) == 0)
                        {
                            /* Destination point address got */
                            strcpy(szReplyID, sBuf+8);
                        }
#endif
                    }
                    iBufPos = 0;
                }
                pStr++;
            }

            /* Seek back if message text was smaller than this */
            iLen = (int) (pStr-pText->sText);
            if (iLen < (int) sizeof(pText->sText)-2)
            {
                SeekFile(pFile,(long) iLen-sizeof(pText->sText)+2,SEEK_CUR);
                fEOT = 1;
            }
        }

        /* Toss message to message base */
        fMsgPrivate = (char) (MsgHeader.wAttrib & 1);
        dupe.wOrigZone = wOrigZone; dupe.wOrigNet = wOrigNet;
        dupe.wOrigNode = wOrigNode; dupe.wOrigPoint = wOrigPoint;
        iRet = TossMessage(szSender, szReceiver, szSubject, szDateTime, &dupe);
        if (iRet == 1)
        {
            if (ExportMessage(1) != 1)
            {
                iRet = 0;
            }
            else
            {
                AddDupe(&dupe);
                uMsgsImported++;
            }
        }

        /* Release memory used by message text */
        while (pFirstText != NULL)
        {
            pText = pFirstText->pNext;
            delete pFirstText;
            pFirstText = pText;
        }
        pFirstMsgText = NULL;

        if (iRet == 0) break;
    }

    CloseFile(pFile);
    if (iRet) remove(pszFileName);

    return iRet != 0;
}

/* Search all .pkt files and toss them */
int TossMessages(void)
{
    char szFileName[256];
    DIR *pDir;
    struct dirent *pDirent;
    struct stat StatBuf;
    unsigned long dpos;

    /* Create dupe file */
    hDupe = sopen(pszDupeFile, O_RDWR | O_BINARY, SH_DENYNO);
    if (hDupe == -1)
    {
        hDupe = creat(pszDupeFile, READWRITE);
        if (hDupe == -1)
        {
            printf("Could not create file `%s`\n", pszDupeFile);
            WriteLog('!',"Could not create file `%s`", pszDupeFile);
            return 0;
        }
        setmode(hDupe, O_BINARY);
        dpos = 0; write(hDupe, &dpos, 4);
    }

    /* Read dupe position */
    read(hDupe, &dpos, 4);
    dupepos = (int) dpos;

    dupes = (int) (lseek(hDupe, 0, SEEK_END)-4)/sizeof(DUPE_REC);
    duperecs = new DUPE_REC[max_dupes];
    if (duperecs == NULL)
    {
        printf("Not enough memory for dupe records (%d bytes needed)\n", sizeof(DUPE_REC)*max_dupes);
        WriteLog('!', "Not enough memory for dupe records (%d bytes needed)", sizeof(DUPE_REC)*max_dupes);
        return 0;
    }
    dupes = read(hDupe, duperecs, dupes*sizeof(DUPE_REC))/sizeof(DUPE_REC);

    /* Open directory */
    pDir = opendir(pszTemp);
    if (pDir == NULL)
    {
        printf("Could not open dir `%s`\n", pszTemp);
        WriteLog('!',"Could not open dir `%s`", pszTemp);
        return 0;
    }

    /* Read directory */
    for (;;)
    {
        pDirent = readdir(pDir);
        if (pDirent == NULL) break;

        sprintf(szFileName,"%s%s",pszTemp,pDirent->d_name);
        if (stat(szFileName,&StatBuf) != 0) continue;
        if (ISDIR(StatBuf)) continue;

        ReadPktFile(szFileName, 0);
    }

    /* All finished */
    close(hDupe);
    closedir(pDir);

    return 1;
}

int scan_area(void)
{
    int iRet;
    char *pStr;
    unsigned long dwMsg;

#ifdef VBOARD
    if (!OpenMsgArea(pMsgAreaRecord->wAreaNum))
#else
    if (!OpenMsgArea(pMsgAreaRecord->pszPath,0))
#endif
    {
        /* Could not open message area */
        printf("Could not open message area `%s`\n",pMsgAreaRecord->szAreaTag);
        WriteLog('!',"Could not open message area `%s`",pMsgAreaRecord->szAreaTag);
    }
    else
    {
        /* Read messages from there */
        printf("%s\n",pMsgAreaRecord->szAreaTag);
        for (dwMsg = 1; dwMsg <= dwMessages; dwMsg++)
        {
            CurrentMessage = dwMsg;
            if (!ReadMessage(dwMsg)) continue;

            iRet = 0;
            if (pMsgAreaRecord == pNetmailArea)
            {
                /* Check if this is form something@in.somewhere */
                pStr = szMsgTo;
                while (*pStr != '\0')
                {
                    if (*pStr == '@')
                    {
                        if (iRet != 0) break;
                        iRet = 1;
                    }
                    if (*pStr == '.' && iRet == 1) iRet = 2;
                    pStr++;
                }
                if (iRet == 2 || stricmp(szMsgTo, "UUCP") == 0)
                {
                    /* OK, was an email address, send it as email */
                    if (SendEmail(szMsgFrom, szMsgTo, szMsgSubj, 1))
                        iRet = 3;
                }
            }

            if (iRet != 3)
            {
                iRet = ExportMessage(0);
                if (!iRet) return 0;
            }

            /* Set message sent */
            if (iRet != 2) SetMessageSent(dwMsg);
        }
        printf("                                                            \r");
    }

    return 1;
}

/* Export messages from message base */
int ScanMessages(void)
{
    FILE *F;
    struct stat statbuf;
    char area[256];

    if (pszBaseFile != NULL && stat(pszBaseFile, &statbuf) == 0)
    {
        /* Scan only areas listed in base file */
        sprintf(area, "%sbasefile.tmp", pszTemp);
        if (!CopyFile(pszBaseFile, area))
        {
            printf("Can't rename file %s to %s\n", pszBaseFile, area);
            WriteLog('!',"Can't rename file %s to %s", pszBaseFile, area);
        }
        else
        {
            remove(pszBaseFile);

            F = fopen(area, "rb");
            if (F == NULL)
            {
                printf("Can't open file %s\n", area);
                WriteLog('!',"Can't open file %s", area);
                rename(area, pszBaseFile);
            }
            else
            {
                printf("Scanning only areas listed on file %s\n", pszBaseFile);
                while (_fgets(area, sizeof(area), F) != NULL)
                {
                    pMsgAreaRecord = pFirstMsgArea;
                    while (pMsgAreaRecord != NULL)
                    {
                        if (stricmp(area, pMsgAreaRecord->szAreaTag) == 0)
                        {
                            if (!scan_area()) return 0;
                            break;
                        }
                        pMsgAreaRecord = pMsgAreaRecord->pNext;
                    }
                }
                fclose(F);

                sprintf(area, "%sbasefile.tmp", pszTemp);
                remove(area);
            }
        }

        return 1;
    }

    pMsgAreaRecord = pFirstMsgArea;
    while (pMsgAreaRecord != NULL)
    {
        if (!scan_area()) return 0;
        pMsgAreaRecord = pMsgAreaRecord->pNext;
    }

    return 1;
}

char *get_domain(NODE_REC *node)
{
    int match;
    ADDR_REC *best;

    if (domain_outbounds)
    {
        pAddrRecord = pFirstAddr; match = 0;
        while (pAddrRecord != NULL)
        {
            if (pAddrRecord->wZone == node->wZone)
            {
                if (match < 1) { match = 1; best = pAddrRecord; }
                if (pAddrRecord->wNet == node->wNet)
                {
                    if (match < 2) { match = 2; best = pAddrRecord; }
                    if (pAddrRecord->wNode == node->wNode)
                    {
                        if (match < 3) { match = 3; best = pAddrRecord; }
                        if (pAddrRecord->wPoint == node->wPoint)
                        {
                            match = 4; best = pAddrRecord;
                            break;
                        }
                    }
                }
            }
            pAddrRecord = pAddrRecord->pNext;
        }
    }

    if (!domain_outbounds || pAddrRecord == NULL || pAddrRecord->szDomain[0] == '\0')
        return NULL;
    else
        return pAddrRecord->szDomain;
}

int create_msg_file(char *path, char *subj)
{
    int f;
    STORED_MSG_HEADER msg;
    time_t _tim;
    struct tm *tim;
    char str[80];

    f = creat(path, READWRITE);
    if (f == -1) return 0;
    setmode(f, O_BINARY);

    memset(&msg, 0, sizeof(msg));
    strcpy(msg.from, "ARCmail");
    strcpy(msg.to, "Sysop");
    strcpy(msg.subj, subj);

    _tim = time(NULL);
    tim = localtime(&_tim);

    /* Write message date */
    sprintf(msg.date,"%02d %s %02d  %02d:%02d:%02d",
            tim->tm_mday,szMonth[tim->tm_mon],tim->tm_year%100,
            tim->tm_hour,tim->tm_min,tim->tm_sec);
    msg.orig_zone = pNodeRecord->wYourZone;
    msg.orig_net = pNodeRecord->wYourNet;
    msg.orig_node = pNodeRecord->wYourNode;
    msg.orig_point = pNodeRecord->wYourPoint;
    msg.dest_zone = pNodeRecord->wZone;
    msg.dest_net = pNodeRecord->wNet;
    msg.dest_node = pNodeRecord->wNode;
    msg.dest_point = pNodeRecord->wPoint;
    msg.attrib = 0x0001 | 0x0010 | 0x0080 | 0x1000; /* private | file attached | killsent | local */

    /* Write .msg header */
    write(f, &msg, sizeof(msg));

    /* Write INTL kludge */
    sprintf(str, "\x01""INTL %d:%d/%d", pNodeRecord->wZone, pNodeRecord->wNet, pNodeRecord->wNode);
    if (pNodeRecord->wPoint != 0) sprintf(str+strlen(str), ".%d", pNodeRecord->wPoint);
    sprintf(str+strlen(str), " %d:%d/%d", pNodeRecord->wYourZone, pNodeRecord->wYourNet, pNodeRecord->wYourNode);
    if (pNodeRecord->wYourPoint != 0) sprintf(str+strlen(str), ".%d", pNodeRecord->wYourPoint);
    write(f, str, strlen(str));

    /* Write FLAGS kludge */
    write(f, "\r\x01""FLAGS KFS DIR\r", 17);

    close(f);
    return 1;
}

/* Get/add packet from .?LO file */
char *GetFloName(char *pszFileName)
{
    FILE *Flo;
    char szOutb[256],szPath[256],tmp[256],*ext;
    int len, i;

    time_t _tim;
    struct tm *pTim;
    struct stat statbuf;

    _tim = time(NULL);
    pTim = localtime(&_tim);

    /* Outbound directory name */
    if (pNodeRecord->wZone == pFirstAddr->wZone)
        strcpy(szOutb, pszOutbound);
    else
    {
        ext = strrchr(pszOutbound, SLASH);
        if (ext == NULL)
            szOutb[0] = '\0';
        else
        {
            *ext = '\0';
            strcpy(szOutb, pszOutbound);
            *ext = SLASH;
        }
        ext = get_domain(pNodeRecord);
        if (no_outboundzones && ext != NULL)
            sprintf(szOutb+strlen(szOutb), SSLASH"%s", ext);
        else
            sprintf(szOutb+strlen(szOutb), SSLASH"outbound.%03x", pNodeRecord->wZone);
    }

    if (pNodeRecord->wPoint != 0)
    {
        /* Point, put to .PNT directory. */
        if (low_flo)
            sprintf(szOutb+strlen(szOutb),SSLASH"%04x%04x.pnt", pNodeRecord->wNet, pNodeRecord->wNode);
        else
            sprintf(szOutb+strlen(szOutb),SSLASH"%04x%04x.PNT", pNodeRecord->wNet, pNodeRecord->wNode);
    }

    /* Create directory if doesn't exist */
    if (stat(szOutb, &statbuf) != 0)
    {
        if (MakeDir(szOutb) != 0)
        {
            printf("Could not create directory `%s`\n",szOutb);
            WriteLog('!',"Could not create directory `%s`\n",szOutb);
        }
    }
    strcat(szOutb,SSLASH);

    /* .FLO file name */
    if (pNodeRecord->wPoint == 0)
        sprintf(szPath,"%s%04x%04x.",szOutb,pNodeRecord->wNet,pNodeRecord->wNode);
    else
        sprintf(szPath,"%s%08x.",szOutb,pNodeRecord->wPoint);

    for (i=0; i<10; i++)
    {
        sprintf(tmp,"%s%s%d",szPath,sWeekDay[pTim->tm_wday],i);
        if (stat(tmp, &statbuf) != 0 || statbuf.st_size != 0) break;
    }

    ext = strrchr(tmp, SLASH);
    if (ext == NULL) ext = tmp; else ext++;
    strcpy(pszFileName, ext);

    if (stat(tmp, &statbuf) != 0 || statbuf.st_size == 0)
    {
        if (statbuf.st_size == 0) remove(tmp);

        if (bink_outb)
        {
            /* .flo outbounds */
            strcat(szPath, low_flo ? "flo" : "FLO");
            Flo = fopen(szPath,"a+b");
            if (Flo == NULL)
            {
                printf("GetFloName() : Could not create file `%s`\n",szPath);
                WriteLog('!',"GetFloName() : Could not create file `%s`",szPath);
                return NULL;
            }
            chmod(szPath, 0770);
#ifndef __linux__
            if (always_slash)
            {
                /* \ -> / conversion */
                ext = tmp;
                while (*ext != '\0')
                {
                    if (*ext == '\\') *ext = '/';
                    *ext++;
                }

                if (tmp[1] == ':')
                    fprintf(Flo,"#%s\n",tmp+2);
                else
                    fprintf(Flo,"#%s\n",tmp);
            }
            else
#endif
                fprintf(Flo,"#%s\n",tmp);
            fclose(Flo);
        }
        else
        {
            /* .msg outbounds */
            for (i = 1;;i++)
            {
                sprintf(szPath, "%s%d.msg", pszMsgPath, i);
                if (stat(szPath, &statbuf) != 0) break;
            }
            create_msg_file(szPath, tmp);
        }
    }

    /* Remove yesterday empty message packets.. */
    i = pTim->tm_wday-1;
    if (i < 0) i = 6;
    tmp[strlen(tmp)-3] = '\0';
    sprintf(tmp+strlen(tmp), "%s0", sWeekDay[i]);
    len = strlen(tmp)-1;
    for (i=0; i<10; i++)
    {
        tmp[len] = (char) (i+'0');
        if (stat(tmp, &statbuf) == 0 && statbuf.st_size == 0)
            remove(tmp);
    }

    return pszFileName;
}

void FinishOutputPackets(void)
{
    unsigned num,num2;
    char outb[256],tmp[256],tmp2[256],fname[13],outname[256],*strp,fPros,*pOut,*pStr;
    int hFile;
    PKT_HEADER PktHdr;

    DIR *dirp;
    struct dirent *direntp;

    /* Close output files */
    for (num = 0; num < uOutFiles; num++)
    {
        close(output[num].hFile);
    }

    /* Open directory */
    dirp = opendir(OUTPUT_PATH);
    if (dirp == NULL)
    {
        printf("FinishOutputPackets() : Couldn't open directory `"OUTPUT_PATH"`!!\n");
        WriteLog('!',"FinishOutputPackets() : Couldn't open directory `"OUTPUT_PATH"`!!");
        return;
    }

    /* Scan *.out files */
    for (;;)
    {
        /* Read record */
        direntp = readdir(dirp);
        if (direntp == NULL) break;

        /* Check if it is .out file.. */
        if (strlen(direntp->d_name) > 12) continue; /* Name too long */
        strp = strrchr(direntp->d_name,'.');
        if (strp == NULL) continue; /* No extension */
        if (stricmp(strp,".out") != 0) continue; /* Not .out file */

        /* Open file */
        sprintf(tmp2, OUTPUT_PATH SSLASH"%s", direntp->d_name);
        hFile = sopen(tmp2, O_RDWR | O_BINARY, SH_DENYRW);
        if (hFile == -1)
        {
            printf("Couldn't open output file `%s`\n",tmp2);
            WriteLog('!',"Couldn't open output file `%s`",tmp2);
            continue;
        }

        /* Read header */
        if (read(hFile,&PktHdr,sizeof(PktHdr)) != sizeof(PktHdr))
        {
            /* Empty file, remove it. */
            close(hFile);
            remove(tmp);
            continue;
        }

        lseek(hFile,-3,SEEK_END);
        if (read(hFile,tmp,3) != 3 || tmp[0] != '\0' || tmp[1] != '\0' || tmp[2] != '\0')
        {
            /* Add 2 or 3 nuls to end of file */
            num = tmp[2] == '\0' ? 2 : 3;
            tmp[0] = '\0'; tmp[1] = '\0'; tmp[2] = '\0';
            write(hFile, tmp, num);
        }

        close(hFile);

        /* Find node record */
        pNodeRecord = pFirstNode;
        while (pNodeRecord != NULL)
        {
            if (pNodeRecord->wZone == PktHdr.wDestZone && pNodeRecord->wNet == PktHdr.wDestNet &&
                pNodeRecord->wNode == PktHdr.wDestNode && pNodeRecord->wPoint == PktHdr.wDestPoint)
            {
                /* Node record found. */
                break;
            }
            pNodeRecord = pNodeRecord->pNext;
        }

        if (pNodeRecord == NULL)
        {
            /* Destination address unknown.. */
            printf("%u:%u/%u.%u address not found from export addresses\n",PktHdr.wDestZone,PktHdr.wDestNet,PktHdr.wDestNode,PktHdr.wDestPoint);
            WriteLog('*',"%u:%u/%u.%u address not found from export addresses",PktHdr.wDestZone,PktHdr.wDestNet,PktHdr.wDestNode,PktHdr.wDestPoint);
            continue;
        }

        /* Pack exported messages */
        if (GetFloName(fname) == NULL) continue;

        /* Outbound directory name */
        if (pNodeRecord->wZone == pFirstAddr->wZone)
            sprintf(outb,"%s"SSLASH, pszOutbound);
        else
        {
            strp = strrchr(pszOutbound, SLASH);
            if (strp == NULL)
                outb[0] = '\0';
            else
            {
                *strp = '\0';
                strcpy(outb, pszOutbound);
                *strp = SLASH;
            }

            strp = get_domain(pNodeRecord);
            if (no_outboundzones && strp != NULL)
                sprintf(outb+strlen(outb), SSLASH"%s"SSLASH, strp);
            else
                sprintf(outb+strlen(outb), SSLASH"outbound.%03x"SSLASH, pNodeRecord->wZone);
        }

        if (pNodeRecord->wPoint != 0)
        {
            /* Point, put to .PNT directory. */
            if (low_flo)
                sprintf(outb+strlen(outb),"%04x%04x.pnt"SSLASH, pNodeRecord->wNet, pNodeRecord->wNode);
            else
                sprintf(outb+strlen(outb),"%04x%04x.PNT"SSLASH, pNodeRecord->wNet, pNodeRecord->wNode);
        }

        /* Find packer */
        num2 = 1;
        pPackerRecord = pFirstPacker;
        while (pNodeRecord->iPacker > (int) num2)
        {
            pPackerRecord = pPackerRecord->pNext;
            num2++;
        }

        /* New name from .out to .pkt */
        strcpy(outname, tmp2);
        strp = strrchr(outname,'.');
        if (strp == NULL)
        {
            printf("FinishOutputPackets() : strp == NULL ???? This should NEVER happen!\n");
            WriteLog('!',"FinishOutputPackets() : strp == NULL ???? This should NEVER happen!");
            exit(1);
        }
        strcpy(strp+1,"pkt");

        /* Get packer execution string */
        pStr = pPackerRecord->pszPackString;
        pOut = tmp; fPros = 0;
        for (; *pStr != '\0'; pStr++)
        {
            if (fPros)
            {
                if (toupper(*pStr) == 'F')
                {
                    /* List of files to pack */
                    pOut += sprintf(pOut,"%s",outname);
                }
                else if (toupper(*pStr) == 'P')
                {
                    /* Packet name */
                    pOut += sprintf(pOut,"%s%s",outb,fname);
                }
                else if (*pStr == '%')
                {
                    /* Just a '%' character */
                    *pOut++ = '%';
                }
                else
                {
                    printf("Unknown parameter in packer #%d : `%c`\n",pNodeRecord->iPacker,*pStr);
                    WriteLog('!',"Unknown parameter in packer #%d : `%c`",pNodeRecord->iPacker,*pStr);
                    return;
                }
                fPros = 0;
                continue;
            }
            if (*pStr == '%') fPros = 1;
            else *pOut++ = *pStr;
        }
        *pOut = '\0';

        /* Rename .out to .pkt */
        if (rename(tmp2, outname) != 0)
        {
            printf("FinishOutputPackets() : Error renaming file `%s` to `%s`\n",tmp2,outname);
            WriteLog('!',"FinishOutputPackets() : Error renaming file `%s` to `%s`",tmp2,outname);
            continue;
        }

        /* Pack files */
        if (shell(tmp) != 0)
        {
            printf("Error executing packer! Command line: %s\n", tmp);
            WriteLog('!',"Error executing packer! Command line: %s", tmp);

            /* Rename back to .out */
            if (rename(outname, tmp2) != 0)
            {
                printf("FinishOutputPackets() : Error renaming file `%s` to `%s`\n",outname,tmp2);
                WriteLog('!',"FinishOutputPackets() : Error renaming file `%s` to `%s`",outname,tmp2);
                continue;
            }

            continue;
        }
        sprintf(tmp, "%s%s", outb, fname);
        chmod(tmp, 0660);

        remove(outname);
    }

    uOutFiles = 0;
    printf("\n\nImported: %u   Exported: %u\n", uMsgsImported, uMsgsExported);
    printf("Dupes: %u   Bad: %u\n", uDupes, uBads);

    WriteLog('+',"Imported: %u   Exported: %u  Dupes: %u   Bad: %u", uMsgsImported, uMsgsExported, uDupes, uBads);
}

void help(void)
{
    printf("Usage: xtoss [scan] [toss] [uucp]\n\n");
    printf(" scan - Scan for outbound mail\n");
    printf(" toss - Toss inbound mail\n\n");
    printf(" uucp - Toss uucp mail\n");
}

int CheckCRC(char *fname)
{
    unsigned long CRC1,CRC2;
    unsigned char sCRC[4];
    int hFile,iReaded,iReadPos,pos;
    unsigned char buf[256],*strp;

    hFile = open(fname, O_RDONLY | O_BINARY);
    if (hFile == -1)
    {
        printf("Can't open file `%s`\n",fname);
        return 0;
    }

    pos = 0; CRC1 = 0xffffffff;
    for (;;)
    {
        iReaded = read(hFile, buf, sizeof(buf));
        if (iReaded == 0) break;

        strp = buf; iReadPos = 0;
        while (iReadPos < iReaded)
        {
            if (pos < 0)
            {
                if (pos == -1) { sCRC[0] = *strp; pos--; }
                else if (pos == -2) { sCRC[1] = *strp; pos--; }
                else if (pos == -3) { sCRC[2] = *strp; pos--; }
                else if (pos == -4) { sCRC[3] = *strp; pos = 255; memcpy(&CRC2,sCRC,4); }
            }
            else
            {
                CalcCRC(CRC1, *strp);
                if (pos != 255)
                {
                    if (*strp == CRCSTRING[pos])
                    {
                        pos++;
                        if (pos == CRCLENGTH) pos = -1; /* Found CRC position! */
                    }
                    else
                    {
                        pos = 0;
                    }
                }
            }
            strp++; iReadPos++;
        }
    }

    close(hFile);
    if (CRC1 != CRC2)
    {
        printf("Illegal CRC in file!\n");
    }
    return CRC1 == CRC2;
}

int main(int argc, char *argv[])
{
    int arg,year;
    time_t _tim;
    struct tm *pTim;
    int hOpen;
    struct stat statbuf;

#ifdef __linux__
    pid_t pid;

    /* Set correct user id */
    if (stat(CONFIG_FILE, &statbuf) == 0)
        seteuid(statbuf.st_uid);
#endif

    printf("\nXToss "VERSION" - Copyright (c) 1997 Timo Sirainen\n");

#ifndef __linux__
    if (!CheckCRC(argv[0])) return 2;
#endif

    /* Read configuration and initialize variables */
    if (!ReadConfig(CONFIG_FILE)) return 1;
#ifndef YAK
    if (registered)
        printf("Registered to: %s\n\n", regname);
    else
        printf("Unregistered DEMO version\n\n");
#endif
    if (pszExportList != NULL && !InitEmail(pszExportList)) return 1;
    if (pszDupeFile == NULL) pszDupeFile = DEFAULT_DUPE_FILE;

    if (argc == 1)
    {
        help();
        return 0;
    }

    if (pszLogFile != NULL && !OpenLog(pszLogFile,0))
    {
        printf("Could not open log file!\n");
        return 1;
    }
#ifdef VBOARD
    if (pszVBLogFile != NULL && !OpenLog(pszVBLogFile,1))
    {
        WriteLog('*',"Warning: Could not open VBoard's log file!");
        printf("Warning: Could not open VBoard's log file!\n");
    }
#endif

#ifdef __linux__
    MakeDir("/tmp/xtoss");
    /* Check if semaphore exists.. */
    hOpen = sopen(SEMAPHORE, O_RDONLY | O_BINARY, SH_DENYNO);
    if (hOpen != -1)
    {
        if (read(hOpen, &pid, sizeof(pid)) == sizeof(pid))
        {
            /* It does, wait until its owner has died.. */
            printf("Another XToss is already active, waiting it to exit..\n");
            waitpid(pid, 0, 0);
        }
        close(hOpen);
    }
#else
    /* Check if semaphore exists.. */
    if (stat(SEMAPHORE,&statbuf) == 0)
    {
        /* It does, check if it can be deleted.. */
        if (remove(SEMAPHORE) != 0)
        {
            /* Wait until the other XToss finishes.. */
            printf("Another XToss is already active, waiting it to exit..\n");
            while (stat(SEMAPHORE,&statbuf) == 0 && remove(SEMAPHORE) != 0)
            {
#ifdef __OS2__
                DosSleep(1);
#else
                delay(1000);
#endif
            }
        }
    }
#endif

    /* Create semaphore.. */
    hOpen = creat(SEMAPHORE, READWRITE);
    if (hOpen == -1)
    {
        WriteLog('*',"Warning: Could not create XToss's semaphore file!");
        printf("Warning: Could not create XToss's semaphore file!\n");
    }

    uOutFiles = 0; uMsgsImported = 0; uMsgsExported = 0; uDupes = 0; uBads = 0;
    pFirstSeenBy = NULL; pSeenBy = NULL; wLastSeenByNet = 0; uSeenBys = 0;
    pFirstPath = NULL; pPath = NULL; wLastPathNet = 0; uPaths = 0;
    ulLastTime = 0;

    _tim = time(NULL);
    pTim = localtime(&_tim);
    year = pTim->tm_year;
    if (year < 70) year += 1900; else year += 2000;
    WriteLog('-',"--- XToss "VERSION" --- %02d.%02d.%02d ---",pTim->tm_mday,pTim->tm_mon,pTim->tm_year);

    /* Check command line parameters */
    for (arg = 1; arg < argc; arg++)
    {
        if (stricmp(argv[arg], "SCAN") == 0)
        {
            if (!ScanMessages()) return 1;
            continue;
        }
        if (stricmp(argv[arg], "UUCP") == 0)
        {
            if (!ScanUUCP(pszSpool)) return 1;
            continue;
        }
        if (stricmp(argv[arg], "TOSS") == 0)
        {
            if (pszInbound != NULL && !UnPackMail(pszInbound)) return 1;
            if (pszSafeIn != NULL && !UnPackMail(pszSafeIn)) return 1;
            if (!TossMessages()) return 1;
            continue;
        }
        help();
        return 0;
    }

    CloseMsgArea();
    CloseMessages();
    DeInitText();

    /* Rename and pack .out files */
    FinishOutputPackets();

    CloseLog();
    close(hOpen);

    remove(SEMAPHORE);

    return 0;
}
