/*
 *  How VBoard message files work?
 *
 *  MsgHead.xxx:
 *  - Stores only most important information, must be kept small for speed
 *    and storage reasons.
 *  - Stores user numbers, not names.
 *  - Stores message attributes and offset to MsgText.xxx.
 *    - Atributes are OR:ed into offset, so upper 5 bits are used.
 *    - This limits MsgText.xxx size to 128 MB. (Should be enougth)
 *  - Stores reply to number.
 *  - Because one MsgHead item is only 10 bytes long you can speed up
 *    sequential operations by reading lots of them with one read operation.
 *
 *
 *  MsgText.xxx:
 *  - Stores SubMsgHeader and message text.
 *  - SubMsgHeader stores all non critical information about message.
 *  - Same MsgText.xxx file can contain both compressed and uncompressed
 *    messages, so before loading you must check MASK_PACKED attribute.
 *  - For unpacked messages there is 2 bytes at the begining of each
 *    message telling how many bytes to read data. You must read message
 *    text this amount of data and SubMsgHeader. SubMsgHeader is always
 *    located just after size bytes.
 *  - For compressed messages LoadCompressed() function handles all
 *    necessary things, but it reports SubMsgHeader size added to message
 *    size. You must substract this before passing message forward.
 *
 *
 *  Message text:
 *  - Message text is strings terminated with '\0'. There is no index
 *    telling where each line begins, but functions for locating this are
 *    provided. Usually sequential access is suffient and most efficient.
 *  - Last line of message text must also be terminated with '\0'. Some
 *    early versions of VBoard incorrectly did not add this terminating
 *    null to the end of message, so you should make sure that all messages
 *    are terminated correctly.
 *
 *  (C) Vesa Suonp 1993-1997. All rights reserved.
 *
 */

#ifndef BBS_MSGIO_ICLUDED
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
//#include <share.h>
//#include <io.h>
#include <time.h>

//#include "C:\SOURCE\BBS\BBSDEF.h"       /* Include BBS defines */
//#include "C:\SOURCE\COMMON\VSTRING.h"   /* Include own string routines */
//#include "C:\SOURCE\COMMON\VDIR.h"      /* Include own directory routines */

#ifdef __OS2__
#define INCL_NOPMAPI
#define INCL_DOSFILEMGR
#include <os2.h>
#endif

#include "bbsdef.h"
//#include "C:\SOURCE\BBS\BBS.H"
//#include "C:\SOURCE\BBS\BBSFUNC.H"
#endif /* BBSMsgIO */

#define FASTLOADHEADERS 100

/* Functions needed from BBSAHuff.c ... */
extern void    SaveCompressed(unsigned char *Input, unsigned short InputSize, unsigned int Handle);
extern void    LoadCompressed(unsigned int Handle, unsigned char *Destination, unsigned short *ExpandSize);
extern void    LoadCompressedN(unsigned int Handle, unsigned char *Destination, unsigned short ExpandToSize);

/* Local functions ... */
unsigned int   LoadMessage(unsigned int ConfNum, unsigned int MsgNum);
unsigned int   SaveMessage(unsigned int ConfNum, unsigned int MsgNum);
unsigned int   LoadMsgHeader(unsigned int ConfNum, unsigned int MsgNum, MSGHEAD *MsgH, SUBMSGHEAD *SubMsgH);
unsigned int   SaveMsgHeader(unsigned int ConfNum, unsigned int MsgNum, MSGHEAD *MsgH);
void           UpdateSubMsgHeader(unsigned int ConfNum, unsigned int MsgNumber);
unsigned int   FastLoadMsgHead(unsigned int ConfNum, unsigned int StartNum, MSGHEAD *Buffer, unsigned int *HeadHandle, unsigned int *TextHandle, unsigned char *EofSignal);
void           FastLoadMsg(unsigned long OffsetAttr, unsigned int Handle);
void           FastLoadSubHeader(unsigned long OffsetAttr, unsigned int Handle, SUBMSGHEAD *ScanSub);
unsigned int   FindFirstReply(unsigned int ConfNum, unsigned int MsgNum, unsigned char *ReplyCount);
unsigned int   FindNextReply(unsigned int ConfNum, unsigned int FromMsg, unsigned int OriginalMsg);
void           AddReply(unsigned int ConfNum, unsigned int Original, unsigned int ReplyMsg);
void           SaveLastRead(unsigned int ConfNum, CONFACCESS *Access);
unsigned int   GetFirstMsg(unsigned int ConfNumber);
unsigned int   GetLastMsg(unsigned int ConfNumber);
void           LoadConf(unsigned int ConfNum, CONF *ConfSets, CONFACCESS *Access);
unsigned int   FindConf(unsigned char *Pattern, unsigned int StartConf, CONF *FConf, CONFACCESS *FUAcc);

/* Using message text ... */
unsigned char *GetNLine(unsigned int LineNum);
unsigned char *GetNLine2(unsigned int LineNum);
unsigned char *GetNLine3(unsigned int LineNum);

unsigned int   CountMsgLines(void);
unsigned char *GetNextMsgLine(void);
unsigned char *GetNextMsgLine2(void);
unsigned char *GetNextMsgLineAndLen(unsigned int *LineLen);



/**************************************************************************
 Function: InitNextMessageLine()                                          *
  Purpose: Macro to initialize GetNextLine function.                      *
  Returns: Message start.                                                 */

#define InitNextMessageLine()   NextLine=Message->MsgTxt
#define MsgEd_UpdateLine()      MsgEd_ChangeLine(&Ed, Ed.y)


/**************************************************************************
 Function: CountMsgLines()                                                *
  Purpose: Count how many lines is in the message.                        *
  Returns: Number of lines in message. 0-64000                            */

unsigned int CountMsgLines(void)
{
  unsigned int Lines=FALSE;

  InitNextMessageLine();
  while (GetNextMsgLine()) ++Lines;
  return Lines;
}


/**************************************************************************
 Function: GetNextMessageLine()                                           *
  Purpose: To get next message line.                                      *
  Returns: Pointer to next message line or NULL if no lines available.    */

unsigned char *GetNextMsgLine(void)
{
  unsigned char *Save;

  Save=NextLine;
  NextLine+=strlen(Save)+1;
  return Save<(Message->MsgTxt+Message->BytesInMsg) ? Save : NULL;
}


/**************************************************************************
 Function: GetNextMessageLine2()                                          *
  Purpose: To get next message line.                                      *
  Returns: Pointer to next message line or "\0" if no lines available.    */

unsigned char *GetNextMsgLine2(void)
{
  unsigned char *Save;

  Save=NextLine;
  NextLine+=strlen(Save)+1;
  return Save<(Message->MsgTxt+Message->BytesInMsg) ? Save : "";
}


/**************************************************************************
 Function: GetNextMessageLineAndLen()                                     *
  Purpose: To get next message line.                                      *
  Returns: Line lenght and Pointer to next line lengt                     */

unsigned char *GetNextMsgLineAndLen(unsigned int *LineLen)
{
  unsigned char *Save;

  Save=NextLine;
  *LineLen=strlen(Save);
  NextLine+=(*LineLen+1);
  return Save<(Message->MsgTxt+Message->BytesInMsg) ? Save : NULL;
}



/**************************************************************************
 Function:                                                                *
  Purpose:                                                                *
  Returns:                                                                */

unsigned char *GetNLine(unsigned int LineNum)
{
  unsigned int   i;
  unsigned char *p;

  for (p=Message->MsgTxt,i=1;i<LineNum;++i) p+=strlen(p)+1;
  return p;
}


/**************************************************************************
 Function:                                                                *
  Purpose:                                                                *
  Returns:                                                                */

unsigned char *GetNLine2(unsigned int LineNum)
{
  unsigned int   i;
  unsigned char *p;

  for (p=Message->MsgTxt,i=1;i<LineNum;++i) p+=strlen(p)+1;
  return p<(Message->MsgTxt+Message->BytesInMsg) ? p : "";
}


/**************************************************************************
 Function:                                                                *
  Purpose:                                                                *
  Returns:                                                                */

unsigned char *GetNLine3(unsigned int LineNum)
{
  unsigned int   i;
  unsigned char *p;

  for (p=Message->MsgTxt,i=1;i<LineNum;++i) p+=strlen(p)+1;
  return p<(Message->MsgTxt+Message->BytesInMsg) ? p : NULL;
}





/**************************************************************************
 Function: LoadMessage()                                                  *
  Purpose: Loads random message.                                          *
  Returns: 0 = All ok, message loaded.                                    *
           1 = Message private, can't access.                             *
           2 = Message killed.                                            *
           3 = Message removed from base.                                 *
           8 = Failed to load header.                                     *
           9 = Failed to load sub header or text.                         */

unsigned int LoadMessage(unsigned int ConfNum, unsigned int MsgNum)
{
  unsigned char  FileName[FILENAMELEN];
  unsigned int   Handle;

  /* Read MSGHeader ... */
  sprintf(FileName, "MSGHEAD.%03x", ConfNum);
  Handle=FileOpen(FileName, O_RDWR | O_BINARY, SH_DENYNO);
  lseek(Handle, LONGOFFSET(MsgNum-1, MSGHEAD), SEEK_SET);
  if (read(Handle, &MsgHead, sizeof(MSGHEAD))!=sizeof(MSGHEAD)) {
    close(Handle);
    return MSG_NOTEXISTS;
  }

  /* Test that message is available ... */
  if (MsgHead.OffsetAttr & MASK_KILLED) { close(Handle); return MSG_KILLED; }
  if (!(MsgHead.OffsetAttr & MASK_OFFSET)) { close(Handle); return MSG_REMOVED; }
  if ((MsgHead.OffsetAttr & MASK_PRIVATE) && !(User.SysOp & SYSOP_PRIVATEMAIL) && !(UserStatus & US_TEMPSYSOP)) {
    if (MsgHead.MsgFromN!=UserNumber && !IsItMe(MsgHead.MsgToN)) { close(Handle); return MSG_PRIVATE; }
  }

  /* If message is to current user, mark it readed ... */
  if (MsgHead.MsgToN==UserNumber && !(MsgHead.OffsetAttr & MASK_READ)) {
    MsgHead.OffsetAttr|=MASK_READ;
    lseek(Handle, LONGOFFSET(MsgNum-1, MSGHEAD), SEEK_SET);
    write(Handle, &MsgHead, sizeof(MSGHEAD));
  } close(Handle);

  /* Read message data ... */
  sprintf(FileName, "MSGTEXT.%03x", ConfNum);
  Handle=FileOpen(FileName, O_RDONLY | O_BINARY, SH_DENYNO);

  /* Seek to begining of message ... */
  lseek(Handle, (MsgHead.OffsetAttr & MASK_OFFSET), SEEK_SET);

  /* If message is saved with packed format ... */
  if (MsgHead.OffsetAttr & MASK_PACKED) {
    LoadCompressed(Handle, (unsigned char *)Message, &Message->BytesInMsg);
    Message->BytesInMsg-=sizeof(SUBMSGHEAD);
  } else {
    read(Handle, &Message->BytesInMsg, sizeof(unsigned short));
    read(Handle, Message, sizeof(SUBMSGHEAD)+Message->BytesInMsg);
  } close(Handle);

  /* Fix unterminated messages ... */
  *(Message->MsgTxt+Message->BytesInMsg)='\0';

  /* Mark message in memory ... */
  Message->MsgInMem=MsgNum;

  /* All done ... */
  return FALSE;
}


/**************************************************************************
 Function: SaveMessage()                                                  *
  Purpose: Saves message.                                                 *
   Params: If MsgNum is FALSE then append to last message.                *
  Returns: Saved message number, or 0 if failed.                          */

unsigned int SaveMessage(unsigned int ConfNum, unsigned int MsgNum)
{
  unsigned char  FileName[FILENAMELEN];
  unsigned int   Handle;

  unsigned short i;
  unsigned char *p1=Message->MsgTxt; /* Target */
  unsigned char *p2=Message->MsgTxt; /* Source */
  unsigned char *pt=Message->MsgTxt;

  /* Remove spaces from end of lines and remove extra empty lines ... */
  for (i=0;i<Message->BytesInMsg;++i) {
    switch (*p2) {
      case '\0': /* Next line */
        /* Kill trailing spaces ... */
        while (*(p1-1)==' ' && p1>Message->MsgTxt) --p1;
        /* Kill empty lines ... */
        if (NConfig.NCOpt1 & NCOPT1_KILLEMPTYLINES) {
          if (p1==Message->MsgTxt || (*(p1-1)=='\0' && *(p1-2)=='\0')) {
            ++p2;
          } else *p1++=*p2++;
        } else *p1++=*p2++;
        break;

      case ' ':
        *p1++=*p2++; /* Space is just like all other chars, execpt it */
        break;       /* Is not accepted as a last char in message! */

      default:   /* Just a char */
        *p1++=*p2++; pt=p1;
        break;
    }
  }

  /* Remove empty lines from end of the message (if KillLines not set) */
  while (pt>Message->MsgTxt && *pt=='\0' && *(pt-1)=='\0') --pt;

  ++pt; *pt='\0';                         /* Add message terminating null */
  Message->BytesInMsg=pt-Message->MsgTxt; /* Readjust message size. */
  if (Message->BytesInMsg<2) return 0;    /* Do not save empty message */

  /* Save message to MsgText.xxx ... */
  sprintf(FileName, "MSGTEXT.%03x", ConfNum);
  Handle=FileOpen(FileName, O_WRONLY | O_BINARY | O_APPEND, SH_DENYWR);

  MsgHead.OffsetAttr&=MASK_ALL;
  MsgHead.OffsetAttr|=filelength(Handle);

  /* If we are using compressed message base use compression ... */
  if (GConfig.GBitToggles & GBIT_MESSAGECOMPRESSION) {
    MsgHead.OffsetAttr|=MASK_PACKED;
    SaveCompressed((unsigned char *)Message, sizeof(SUBMSGHEAD)+Message->BytesInMsg, Handle);
  } else {
    MsgHead.OffsetAttr&=(~MASK_PACKED);
    write(Handle, &Message->BytesInMsg, sizeof(unsigned short));
    write(Handle, Message, sizeof(SUBMSGHEAD)+Message->BytesInMsg);
  } close(Handle);

  /* Save message header to MsgHead.xxx */
  sprintf(FileName, "MSGHEAD.%03x", ConfNum);
  if (ConfNum==Conf_Post || ConfNum==Conf_NetMail) MsgHead.OffsetAttr|=MASK_PRIVATE;

  /* If using predefined message number ... */
  if (MsgNum) {
    Handle=FileOpen(FileName, O_RDWR | O_BINARY, SH_DENYNO);
    lseek(Handle, LONGOFFSET(MsgNum-1, MSGHEAD), SEEK_SET);
    write(Handle, &MsgHead, sizeof(MSGHEAD));
  } else {
    Handle=FileOpen(FileName, O_WRONLY | O_BINARY | O_APPEND, SH_DENYWR);
    write(Handle, &MsgHead, sizeof(MSGHEAD));
    MsgNum=(filelength(Handle)/sizeof(MSGHEAD));
  } close(Handle);

  /* All done, return message number ... */
  return MsgNum;
}



/**************************************************************************
 Function: LoadMsgHeader.                                                 *
  Purpose: ReWrites message header(s).                                    *
  Returns: 0 if successful, any other fails.                              */

unsigned int LoadMsgHeader(unsigned int ConfNum, unsigned int MsgNum, MSGHEAD *MsgH, SUBMSGHEAD *SubMsgH)
{
  unsigned char  FileName[FILENAMELEN];
  unsigned int   Handle;
  unsigned int   BytesRead;

  sprintf(FileName, "MSGHEAD.%03x", ConfNum);
  Handle=FileOpen(FileName, O_RDONLY | O_BINARY, SH_DENYNO);

  lseek(Handle, LONGOFFSET(MsgNum-1, MSGHEAD), SEEK_SET);
  BytesRead=read(Handle, MsgH, sizeof(MSGHEAD));
  close(Handle);

  if (BytesRead!=sizeof(MSGHEAD)) return MSG_NOTEXISTS;
  if (MsgH->OffsetAttr & MASK_KILLED) return MSG_KILLED;
  if (!(MsgH->OffsetAttr & MASK_OFFSET)) return MSG_REMOVED;


  /*********************** SUB HEADER *******************************/

  if (SubMsgH!=NULL) {
    sprintf(FileName, "MSGTEXT.%03x", ConfNum);
    Handle=FileOpen(FileName, O_RDONLY | O_BINARY, SH_DENYNO);

    if (MsgH->OffsetAttr & MASK_PACKED) {
      lseek(Handle, (MsgH->OffsetAttr & MASK_OFFSET), SEEK_SET);
      LoadCompressedN(Handle, (unsigned char *)SubMsgH, sizeof(SUBMSGHEAD));
    } else {
      lseek(Handle, (MsgH->OffsetAttr & MASK_OFFSET)+sizeof(unsigned short), SEEK_SET);
      read(Handle, SubMsgH, sizeof(SUBMSGHEAD));
    } close(Handle);
  } return FALSE;
}


/**************************************************************************
 Function: SaveMsgHeader()                                                *
  Purpose: ReWrites message header(s).                                    *
  Returns: 0 if successful, any other fails.                              */

unsigned int SaveMsgHeader(unsigned int ConfNum, unsigned int MsgNum, MSGHEAD *MsgH)
{
  unsigned char  FileName[FILENAMELEN];
  unsigned int   Handle;

  sprintf(FileName, "MSGHEAD.%03x", ConfNum);
  Handle=FileOpen(FileName, O_WRONLY | O_BINARY, SH_DENYNO);
  lseek(Handle, LONGOFFSET(MsgNum-1, MSGHEAD), SEEK_SET);
  write(Handle, MsgH, sizeof(MSGHEAD));
  close(Handle);

  /*********************** SUB HEADER ******************************
  if (SubMsgH!=NULL) {
    sprintf(FileName, "MSGTEXT.%03x", ConfNum);
    Handle=FileOpen(FileName, O_WRONLY | O_BINARY, SH_DENYNO);
    lseek(Handle, (MsgH->OffsetAttr & MASK_OFFSET)+2, SEEK_SET);
    write(Handle, SubMsgH, sizeof(SUBMSGHEAD));
    close(Handle);
  } */
  return FALSE;
}


void UpdateSubMsgHeader(unsigned int ConfNum, unsigned int MsgNumber)
{
  unsigned char  FileName[FILENAMELEN];
  unsigned int   Handle;

  /* If this was packed message then repack and save ... */
  if (MsgHead.OffsetAttr & MASK_PACKED) {
    /* Whole message must be resaved ... */
    sprintf(FileName, "MSGTEXT.%03x", ConfNum);
    Handle=FileOpen(FileName, O_WRONLY | O_BINARY | O_APPEND, SH_DENYWR);
    MsgHead.OffsetAttr&=MASK_ALL;
    MsgHead.OffsetAttr|=MASK_PACKED;
    MsgHead.OffsetAttr|=filelength(Handle);
    SaveCompressed((unsigned char *)Message, sizeof(SUBMSGHEAD)+Message->BytesInMsg, Handle);
    close(Handle);

    /* Resave message header to MsgHead.xxx */
    sprintf(FileName, "MSGHEAD.%03x", ConfNum);
    Handle=FileOpen(FileName, O_RDWR | O_BINARY, SH_DENYNO);
    lseek(Handle, LONGOFFSET(MsgNumber-1, MSGHEAD), SEEK_SET);
    write(Handle, &MsgHead, sizeof(MSGHEAD));
    close(Handle);
  } else {
    /* If message was not packed this is easy job ... */
    sprintf(FileName, "MSGTEXT.%03x", ConfNum);
    Handle=FileOpen(FileName, O_WRONLY | O_BINARY, SH_DENYNO);
    lseek(Handle, (MsgHead.OffsetAttr & MASK_OFFSET)+2, SEEK_SET);
    write(Handle, &Message->SubMsgHead, sizeof(SUBMSGHEAD));
    close(Handle);
  } return;
}



/**************************************************************************
 Function: FastLoadMessageHead()                                          *
  Purpose: Load number of msg heads. MAX_FASTLOAD                         *
    Notes: Opens both handles if not already open, and closes when no     *
           more messages are available.                                   *
  Returns: Number of message headers read.                                */

unsigned int FastLoadMsgHead(unsigned int ConfNum, unsigned int StartNum, MSGHEAD *Buffer, unsigned int *HeadHandle, unsigned int *TextHandle, unsigned char *EofSignal)
{
  unsigned int BytesRead;

  if (!*HeadHandle) {
    unsigned char  FileName[FILENAMELEN];
    sprintf(FileName, "MSGHEAD.%03x", ConfNum);
    *HeadHandle=FileOpen(FileName, O_BINARY | (*EofSignal==OPEN_WITH_RDWR ? O_RDWR : O_RDONLY), SH_DENYNO);
    if (*EofSignal==OPEN_WITH_RDWR) *EofSignal=FALSE;
  }

  lseek(*HeadHandle, LONGOFFSET(StartNum-1, MSGHEAD), SEEK_SET);
  BytesRead=read(*HeadHandle, Buffer, sizeof(MSGHEAD)*FASTLOADHEADERS);

  if (BytesRead<(sizeof(MSGHEAD)*FASTLOADHEADERS)) *EofSignal=TRUE;

  if (TextHandle!=NULL && !*TextHandle) {
    unsigned char  FileName[FILENAMELEN];
    sprintf(FileName, "MSGTEXT.%03x", ConfNum);
    *TextHandle=FileOpen(FileName, O_BINARY | O_RDONLY, SH_DENYNO);
  }
  return BytesRead/sizeof(MSGHEAD);
}


/**************************************************************************
 Function: FastLoadMessage.                                               *
  Purpose: Loads message really fast.                                     *
  Returns: 0 if successful, any other fails.                              */

void FastLoadMsg(unsigned long OffsetAttr, unsigned int Handle)
{
  lseek(Handle, (OffsetAttr & MASK_OFFSET), SEEK_SET);

  if (OffsetAttr & MASK_PACKED) {
    LoadCompressed(Handle, (unsigned char *)Message, &Message->BytesInMsg);
    Message->BytesInMsg-=sizeof(SUBMSGHEAD);
  } else {
    read(Handle, &Message->BytesInMsg, sizeof(unsigned short));
    read(Handle, Message, sizeof(SUBMSGHEAD)+Message->BytesInMsg);
  }

  /* Fix untruncated messages */
  *(Message->MsgTxt+Message->BytesInMsg)='\0';
  return;
}


/**************************************************************************
 Function: FastLoadSubHeader.                                             *
  Purpose: Loads message really fast.                                     *
  Returns: Nothing.                                                       */

void FastLoadSubHeader(unsigned long OffsetAttr, unsigned int Handle, SUBMSGHEAD *ScanSub)
{
  if (OffsetAttr & MASK_PACKED) {
    lseek(Handle, (OffsetAttr & MASK_OFFSET), SEEK_SET);
    LoadCompressedN(Handle, (unsigned char *)ScanSub, sizeof(SUBMSGHEAD));
  } else {
    lseek(Handle, (OffsetAttr & MASK_OFFSET)+2, SEEK_SET);
    read(Handle,  ScanSub, sizeof(SUBMSGHEAD));
  } return;
}



#define FISRTREPLY_LOADSIZE   128
#define FIRSTREPLY_CACHESIZE  512
unsigned int FindFirstReply(unsigned int ConfNum, unsigned int MsgNum, unsigned char *ReplyCount)
{
  static unsigned short CacheConf=CONF_NOT_PRESENT;
  static unsigned short CacheStart;
  static unsigned short CacheEnd;
  static unsigned short CacheTable[FIRSTREPLY_CACHESIZE];
  static unsigned char  CacheTableReplyCount[FIRSTREPLY_CACHESIZE];

  /* If wrong conf or no cache hit ... */
  if (CacheConf!=ConfNum || MsgNum<CacheStart || MsgNum>CacheEnd) {
    MSGHEAD      FlMsgHdr[FISRTREPLY_LOADSIZE];
    unsigned int Loaded;
    unsigned int Handle;
    unsigned int ThisMsg;

    /* Set variables ... */
    CacheConf=ConfNum;

    /* If used to invalidate cache ... */
    if (CacheConf==CONF_NOT_PRESENT) {
      CacheConf=CONF_NOT_PRESENT-1;
      return 0;
    }

    CacheStart=MsgNum;
    CacheEnd=MsgNum+FIRSTREPLY_CACHESIZE;
    ThisMsg=MsgNum;

    sprintf((unsigned char *)CacheTable, "MSGHEAD.%03x", ConfNum);
    Handle=FileOpen((unsigned char *)CacheTable, O_RDONLY | O_BINARY, SH_DENYNO);
    lseek(Handle, LONGOFFSET(ThisMsg-1, MSGHEAD), SEEK_SET);

    /* Clear table ... */
    memset(CacheTable, 0, sizeof(CacheTable));
    memset(CacheTableReplyCount, 0, sizeof(CacheTableReplyCount));

    do {
      unsigned int j;

      /* Load more headers to analyze ... */
      Loaded=read(Handle, FlMsgHdr, MIN(CacheEnd-ThisMsg, FISRTREPLY_LOADSIZE)*sizeof(MSGHEAD))/sizeof(MSGHEAD);

      for(j=0;j<Loaded;++j,++ThisMsg) {
        /* If this message is a reply to another .. */
        if (FlMsgHdr[j].ReplyTo &&
            FlMsgHdr[j].ReplyTo>=CacheStart &&
            FlMsgHdr[j].ReplyTo<CacheEnd) {

          /* Add replycount ... */
          ++CacheTableReplyCount[FlMsgHdr[j].ReplyTo-CacheStart];

          /* If no previous reply marked mark this one as a first reply ... */
          if (!CacheTable[FlMsgHdr[j].ReplyTo-CacheStart]) CacheTable[FlMsgHdr[j].ReplyTo-CacheStart]=ThisMsg;
        }
      }
    } while (Loaded==FISRTREPLY_LOADSIZE && ThisMsg<CacheEnd);
    close(Handle);
    if (ThisMsg<CacheEnd) CacheEnd=ThisMsg;
  }

  /* Return correct value .... */
  if (ReplyCount!=NULL) *ReplyCount=CacheTableReplyCount[MsgNum-CacheStart];
  return CacheTable[MsgNum-CacheStart];
}


/**************************************************************************
 Function: FindNextReply()                                                *
  Purpose: Seeks next reply number.                                       *
  Returns: False if none found, or found message number.                  */

unsigned int FindNextReply(unsigned int ConfNum, unsigned int FromMsg, unsigned int OriginalMsg)
{
  unsigned short i;
  unsigned char  FileName[FILENAMELEN];
  unsigned int   FileHandle;
  MSGHEAD        VMsgHead;

  sprintf(FileName, "MSGHEAD.%03x", ConfNum);
  FileHandle=FileOpen(FileName, O_RDONLY | O_BINARY, SH_DENYNO);

  lseek(FileHandle, LONGOFFSET(FromMsg, MSGHEAD), SEEK_SET);
  for (i=FromMsg+1;read(FileHandle, &VMsgHead, sizeof(MSGHEAD))==sizeof(MSGHEAD);++i) {
    if (!CanISeeMsg(&VMsgHead)) continue;
    if (VMsgHead.ReplyTo==OriginalMsg) {
      close(FileHandle);
      return i;
    }
  } close(FileHandle);
  return FALSE;
}


void SaveLastRead(unsigned int ConfNum, CONFACCESS *Access)
{
  unsigned int  ConfHandle;
  unsigned char FileName[FILENAMELEN];

  sprintf(FileName, "MSGUSER.%03x", ConfNum);
  ConfHandle=FileOpen(FileName, O_WRONLY | O_BINARY, SH_DENYNO);

  lseek(ConfHandle, LONGOFFSET(UserNumber-1, CONFACCESS), SEEK_SET);
  write(ConfHandle, Access, sizeof(CONFACCESS));
  close(ConfHandle);
  return;
}


/**************************************************************************
 Function: FindConf()                                                     *
  Purpose: Finds first conference that matches Pattern (Wildcards!).      *
  Returns: Conference number or CONF_NOT_PRESENT if not found.            */

unsigned int FindConf(unsigned char *Pattern, unsigned int StartConf, CONF *FConf, CONFACCESS *FUAcc)
{
  unsigned char TempAreaName[41];
  unsigned int  ConfInfoHandle;
  CONF          FiConf;
  CONFACCESS    FiMsgUser;

  StrToUpper(Pattern);
  strcat(Pattern, "*");

  ConfInfoHandle=FileOpen("CONFINFO.DAT", O_BINARY | O_RDONLY, SH_DENYNO);
  if (StartConf) lseek(ConfInfoHandle, LONGOFFSET(StartConf, CONF), SEEK_SET);

  for(;read(ConfInfoHandle, &FiConf, sizeof(FiConf))==sizeof(FiConf);++StartConf) {
    strcpy(TempAreaName, FiConf.Name);
    StrToUpper(TempAreaName);
    if (WildStrCmp(Pattern, TempAreaName)) continue;

    LoadConf(StartConf, NULL, &FiMsgUser);
    if (!(User.SysOp & SYSOP_CONFERENCES) &&
        !(UserStatus & US_TEMPSYSOP) &&
        !(FiMsgUser.Access & (CA_READ | CA_GRP_READ)) && (FiConf.Flags1 & CONFFLAG1_SECRET)) continue;

      close(ConfInfoHandle);
      if (FConf!=NULL) memcpy(FConf, &FiConf, sizeof(FiConf));
      if (FUAcc!=NULL) memcpy(FUAcc, &FiMsgUser, sizeof(FiMsgUser));
      return StartConf;
  } close(ConfInfoHandle);
  return CONF_NOT_PRESENT;
}



/**************************************************************************
 Function: GetFirstMsg()                                                  *
  Purpose: Gets first message in the conference.                          *
  Returns: Always 1.                                                      */

unsigned int GetFirstMsg(unsigned int ConfNum)
{
  return ConfNum ? 1 : 1;
}


/**************************************************************************
 Function: GetLastMsg()                                                   *
  Purpose: Gets last message on conference using directory entry.         *
  Returns: Return last message number.                                    */

unsigned int GetLastMsg(unsigned int ConfNum)
{
  unsigned char FileName[FILENAMELEN];
  #if defined (__MSDOS__) || defined(__WIN32__)
  struct FFBLK Dir;
  #endif
  #ifdef __OS2__
  FILEFINDBUF FileInfo;
  #endif

  sprintf(FileName, "%sMSGHEAD.%03x", NConfig.MainDir, ConfNum);

  #if defined (__MSDOS__) || defined(__WIN32__)
  FindFirst(FileName, &Dir, FALSE);
  FindClose(&Dir);
  return (Dir.ff_fsize/sizeof(MSGHEAD));
  #endif
  #ifdef __OS2__
  DosQueryPathInfo(FileName, FIL_STANDARD, &FileInfo, sizeof(FileInfo));
  return (FileInfo.cbFile/sizeof(MsgHead));
  #endif
}


void LoadConf(unsigned int ConfNum, CONF *ConfSets, CONFACCESS *Access)
{
  unsigned int  ConfHandle;
  unsigned char ConfName[FILENAMELEN];

  if (ConfSets!=NULL) {
    sprintf(ConfName, "MSGTEXT.%03x", ConfNum);
    ConfHandle=FileOpen(ConfName, O_RDONLY | O_BINARY, SH_DENYNO);
    read(ConfHandle, ConfSets, sizeof(CONF));
    close(ConfHandle);
  }

  if (Access!=NULL) {
    sprintf(ConfName, "MSGUSER.%03x", ConfNum);
    ConfHandle=FileOpen(ConfName, O_RDONLY | O_BINARY, SH_DENYNO);
    lseek(ConfHandle, LONGOFFSET(UserNumber-1, CONFACCESS), SEEK_SET);
    read(ConfHandle, Access, sizeof(CONFACCESS));
    close(ConfHandle);
  }
  return;
}

void SetNewUserConfAccess(void)
{
  unsigned int   Handle;
  unsigned int   i;

  Handle=FileOpen("CONFINFO.DAT", O_BINARY | O_RDONLY, SH_DENYNO);

  for(i=FALSE;read(Handle, &Conf, sizeof(CONF))==sizeof(CONF);++i) {
    unsigned short LastMsg;
    CONFACCESS NewMsgUser;

    NewMsgUser.Access=Conf.NewUserAccess;
    LastMsg=GetLastMsg(i);
    if (LastMsg>Conf.Scan) {
      NewMsgUser.LastRead=LastMsg-Conf.Scan;
    } else NewMsgUser.LastRead=0;
    SaveLastRead(i, &NewMsgUser);
  } close(Handle);
  return;
}



