// aliaksessa sotkee color koodit

#include <stdio.h>
#include <string.h>

#include "config.h"
#include "bbs_func.h"
#include "files.h"
#include "keyb.h"
#include "modem.h"
#include "nodes.h"
#include "output.h"
#include "userbase.h"
#include "timeslic.h"
#include "memory.h"
#include "logfile.h"

#define MNODE_LOGIN 1
#define MNODE_LOGOUT 2

#define local_mline(a) output("\r@X07@CLREOL@%s@CR@@X1F@CLREOL@", a);

static int fgrp, lpos;
static unsigned long oldgrpsize;
static char grpstr[256], *nodemsg = NULL;
static int nodemsglen = 0;

void write_grpline(char *str)
{
    unsigned long oldpos;
    unsigned long len;

    FileLock(fgrp);

    oldpos = FileSeek(fgrp, 0, SEEK_CUR);
    FileSeek(fgrp, 0, SEEK_END);
    len = strlen(str) + ((unsigned long) nodenum << 16);
    FileWrite(fgrp, &len, 4);
    FileWrite(fgrp, str, len & 0xffff);
    FileSeek(fgrp, oldpos, SEEK_SET);

    FileUnlock(fgrp);
}

int init_group(void)
{
    char *strp;

    /* Open multinode file */
    sprintf(grpstr, "%smnode.dat", common_path);
    fgrp = FileOpen(grpstr, O_RDWR | O_BINARY, SH_DENYNO);
    if (fgrp == -1)
    {
        /* Create file */
        fgrp = FileCreate(grpstr, CREATE_MODE);
        if (fgrp == -1)
        {
            write_log("Can't create file %s", grpstr);
            return 0;
        }
    }
    else
        FileSeek(fgrp, 0, SEEK_END); /* Seek to end of multinode chat */

    oldgrpsize = FileSeek(fgrp, 0, SEEK_CUR);

    if (usrsub.grpalias[0] == 0)
    {
        /* No groupchat alias set yet.. */
        if (usrsub.alias[0] != 0)
        {
            /* Use normal alias */
            strcpy(usrsub.grpalias, usrsub.alias);
        }
        else
        {
            /* No alias set! Use first name */
            strcpy(usrsub.grpalias, usrsub.name);
            strp = strchr(usrsub.grpalias, ' ');
            if (strp != NULL) *strp = '\0';
        }
    }

    nodemsglen = 512;
    nodemsg = (char *) _malloc(512);

    if (usrsub.grpcolor == 0) usrsub.grpcolor = 7;

    sprintf(grpstr, "\x01%s\x01%s", usrsub.name, usrsub.grpalias);
    write_grpline(grpstr);

    output("@X07@CLR@@X17@DOWN:%d@@CLREOL@", user.ScreenLen);
    local_mline("Type /? for help.");

    return 1;
}

void close_group(void)
{
    int node;

    /* Send exit text */
    sprintf(grpstr, "\x02%s\x01%s", usrsub.name, usrsub.grpalias);
    write_grpline(grpstr);
    output("@X07");

    FileLock(fgrp);

    /* Check if anyone is in groupchat */
    for (node=1; node<=nodes; node++)
    {
        if (read_nodefile(node,&noderec) == 0) break;
        if (noderec.doing == DOING_GROUPCHAT || noderec.doing == DOING_DOWNLOAD_GROUP)
        {
            node = 0;
            break;
        }
    }

    if (node != 0)
    {
        /* Nope, there wasn't, truncate mnode.dat */
        FileTrunc(fgrp, 0);
    }
    FileUnlock(fgrp);
    FileClose(fgrp);

    _free(nodemsg); nodemsg = NULL; nodemsglen = 0;
}

char *check_grpfile(int own)
{
    struct stat statbuf;
    unsigned long len;
    int line;
    char tmp[256], *strp;

    if (fstat(fgrp, &statbuf) == 0 && statbuf.st_size != oldgrpsize)
    {
        for (;;)
        {
            if (FileRead(fgrp, &len, 4) == 4)
            {
                line = len >> 16;
                len &= 0xffff;
                if (len > nodemsglen)
                {
                    if (nodemsg != NULL) _free(nodemsg);
                    nodemsg = _malloc((int) len);
                    nodemsglen = len;
                }
            }
            else
            {
                /* No more messages.. */
                FileSeek(fgrp, oldgrpsize, SEEK_SET);
                break;
            }

            if (FileRead(fgrp, nodemsg, (int) len) == (int) len)
            {
                oldgrpsize = FileSeek(fgrp, 0, SEEK_CUR);
                if (!own || line != nodenum)
                {
                    nodemsg[len] = '\0';
                    switch (nodemsg[0])
                    {
                        case MNODE_LOGIN:
                            /* Groupchat login */
                            strcpy(tmp, nodemsg);
                            strp = strchr(tmp+1, 1);
                            if (strp == NULL) strp = ""; else *strp++ = '\0';
                            sprintf(nodemsg, "@X0F%s (%s) liittyi ryhmkeskusteluun.", tmp+1, strp);
                            return nodemsg;

                        case MNODE_LOGOUT:
                            /* Groupchat logout */
                            strcpy(tmp, nodemsg);
                            strp = strchr(tmp+1, 1);
                            if (strp == NULL) strp = ""; else *strp++ = '\0';
                            sprintf(nodemsg, "@X0F%s (%s) lhti ryhmkeskustelusta.", tmp+1, strp);
                            return nodemsg;

                        default:
                            /* Normal nodetext */
                            return nodemsg;
                    }
                }
            }
            else
            {
                /* No more messages.. */
                FileSeek(fgrp, oldgrpsize, SEEK_SET);
                break;
            }
        }
    }

    return NULL;
}

void draw_prompt(int startpos, int aliaspos)
{
    char ch;

    grpstr[lpos] = '\0';
    if (startpos == aliaspos)
    {
        /* Display alias with different color.. */
        ch = grpstr[aliaspos];
        grpstr[aliaspos] = '\0';
        output("@X1F");
        outtext(grpstr+4);
        grpstr[aliaspos] = ch;
    }

    output("@X17");
    outtext(grpstr+startpos);
    output("@CLREOL@");
}

int convert_line(char *str, char *startup)
{
    char tmp[256], *strp;
    int quit, num;

    quit = 0; strp = str;
    while (*str >= '0' && *str <= '9') str++;
    if (*str == '>')
    {
        /* Private node message. */
        *str = ' ';
        send_nodemsg(strp);
    }
    else if (*str == '/')
    {
        str++;
        if (stricmp(str, "HELP") == 0 || stricmp(str, "?") == 0)
        {
            /* Type help. */
            output("/HELP            - Help!!@CR@");
            output("/WHO             - Who's online?@CR@");
            output("/LIST            - List users in groupchat?@CR@");
            output("/COMMENT=comment - Change your comment@CR@");
            output("/COLOR=color#    - Change your color@CR@");
            output("/QUIT            - Quit groupchat@CR@");
            output("/ALIAS=new_alias - Change your alias@CR@");
            output("@CR@To send private node messages, type 'nodenumber>type here your text'@CR@");
        }
        else if (stricmp(str, "QUIT") == 0 || stricmp(str, "Q") == 0)
        {
            /* Quit */
            quit = 1;
        }
        else if (stricmp(str, "WHO") == 0 || stricmp(str, "W") == 0)
        {
            /* Who's online */
            display_who();
            output("@CR@");
        }
        else if (stricmp(str, "LIST") == 0 || stricmp(str, "L") == 0)
        {
            /* List users in groupchat */
            output("@X0F #  Name (alias) <comment>@CR@");
            output("@X08------------------------------------------------------------------------@CR@");

            for (num = 1; num<=nodes; num++)
            {
                if (read_nodefile(num, &noderec) == 0) break;
                if (noderec.doing == DOING_GROUPCHAT || noderec.doing == DOING_DOWNLOAD_GROUP)
                {
                    strp = strchr(noderec.data, 1);
                    if (strp == NULL) continue;
                    *strp++ = '\0';
                    strcpy(tmp1_str, strp);

                    if (noderec.doing == DOING_DOWNLOAD_GROUP)
                    {
                        strp = strchr(noderec.data, ';');
                        if (strp == NULL) continue;
                        strp++;
                    }
                    else
                        strp = noderec.data;
                    strcpy(tmp2_str, strp);
                    output("@X07%2d @X0F@WHO_NAME@ @X08(@X0A@TMP1@@X08) <@X07@TMP2@@X08>@CR@", num);
                }
            }
            output("@CR@");
        }
        else if (strnicmp(str, "ALIAS=", 6) == 0 || strnicmp(str, "A=", 2) == 0)
        {
            /* Change alias */
            strcpy(tmp, usrsub.grpalias);
            strcpy(usrsub.grpalias, str[1] == '=' ? str+2 : str+6);
            if (stricmp(usrsub.grpalias, tmp) != 0)
            {
                sprintf(grpstr, "%s (%s) is now known as %s", usrsub.name, tmp, usrsub.grpalias);
                write_grpline(grpstr);

                read_nodefile(nodenum, &noderec);
                if (noderec.doing == DOING_DOWNLOAD_GROUP)
                {
                    strp = strchr(noderec.data, ';');
                    if (strp != NULL) *strp = '\0';
                    sprintf(noderec.data+strlen(noderec.data), ";%s\x01%s", usrsub.grpcomment, usrsub.grpalias);
                }
                else
                    sprintf(noderec.data, "%s\x01%s", usrsub.grpcomment, usrsub.grpalias);

                update_nodefile(0);
            }
        }
        else if (strnicmp(str, "COMMENT=", 8) == 0 || strnicmp(str, "CO=", 3) == 0)
        {
            /* Change comment */
            strcpy(usrsub.grpcomment, str[2] == '=' ? str+5 : str+8);

            read_nodefile(nodenum, &noderec);
            if (noderec.doing == DOING_DOWNLOAD_GROUP)
            {
                strp = strchr(noderec.data, ';');
                if (strp != NULL) *strp = '\0';
                sprintf(noderec.data+strlen(noderec.data), ";%s\x01%s", usrsub.grpcomment, usrsub.grpalias);
            }
            else
                sprintf(noderec.data, "%s\x01%s", usrsub.grpcomment, usrsub.grpalias);

            update_nodefile(0);
        }
        else if (strnicmp(str, "COLOR=", 6) == 0 || strnicmp(str, "C=", 2) == 0)
        {
            /* Change color */
            sscanf(str[1] == '=' ? str+2 : str+6, "%d", &num);
            if ((num & 0x0f) != 0) usrsub.grpcolor = (unsigned char) (num & 0x0f);
        }
    }
    else
    {
        /* Normal groupchat text */
        write_grpline(startup);
    }
    return quit;
}

void group_chat(void)
{
    char nodefile[256], *strp;
    int aliaspos, startpos, quit, hit;
    unsigned char ch;
    struct stat statbuf;

    if (!init_group()) return;

    noderec.doing = DOING_GROUPCHAT;
    sprintf(noderec.data, "%s\x01%s", usrsub.grpcomment, usrsub.grpalias);
    update_nodefile(0);

    aliaspos = startpos = lpos = sprintf(grpstr, "@X0%X%s: ", usrsub.grpcolor, usrsub.grpalias);

    sprintf(nodefile, "%snodemsg.%d", common_path, nodenum);

    quit = 0;
    while (!quit)
    {
        if (!carr_det()) break;

        /* Scan for groupchat messages */
        hit = 0;
        do
        {
            strp = check_grpfile(0);
            if (strp != NULL)
            {
                if (!hit)
                {
                    output("\r@X07@CLREOL@");
                    hit = 1;
                }
                ch = strp[4]; strp[4] = '\0';
                output(strp);
                strp[4] = ch;
                outtext(strp+4);
                output("@CR@");
            }
        }
        while (strp != NULL);
        if (hit) draw_prompt(startpos, aliaspos);

        /* Scan for node messages */
        if (stat(nodefile, &statbuf) == 0 && statbuf.st_size > 0)
        {
            /* Found node file! */
            if (display_nodefile(nodefile, "\r@X07@CLREOL@"))
            {
                output("@CR@");
                draw_prompt(startpos, aliaspos);
            }
        }

        hit = get_event();
        if (hit)
        {
            /* Key pressed */
            ch = (unsigned char) (hit == 1 ? getch() : mdm_getch());
            switch (ch)
            {
                case 8:
                    if (lpos == startpos) break;
                    outtext("\b \b"); lpos--;
                    break;
                case 13:
                    if (lpos == aliaspos) break;
                    grpstr[lpos] = '\0';
                    lpos = aliaspos;
                    output("\r@X07@CLREOL@");
                    quit = convert_line(grpstr+lpos, grpstr);
                    output("\r@X1F");
                    aliaspos = startpos = lpos = sprintf(grpstr, "@X0%X%s: ", usrsub.grpcolor, usrsub.grpalias);
                    outtext(grpstr+4);
                    output("@X17@CLREOL@");
                    break;

                default:
                    if (ch < 32 || lpos >= 255) break;
                    if (lpos-startpos+(startpos == aliaspos ? aliaspos : 0) == user.ScreenWidth-1)
                    {
                        output("\r@CLREOL@"); startpos = lpos;
                    }
                    outchr(ch); grpstr[lpos++] = ch;
                    break;
            }
        }
        else
            give_timeslice();
    }
    close_group();

    noderec.doing = DOING_NOTHING;
    update_nodefile(0);
}
