/* YAK - Copyright (c) 1997 Timo Sirainen - read license.txt */

/* os.c - OS-dependent functions */

#ifdef __OS2__
#  define INCL_WIN
#  define INCL_KBD
#include <os2.h>
#endif

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "files.h"
#include "keyb.h"
#include "modem.h"
#include "config.h"

#ifdef UNIX

char *strupr(char *str)
{
    while (*str != '\0')
    {
        *str = toupper(*str);
        str++;
    }
    return str;
}

char *strlwr(char *str)
{
    while (*str != '\0')
    {
        *str = tolower(*str);
        str++;
    }
    return str;
}

#ifdef WANT_STRCHR
char *r_strchr(char *s, int c)
{
    while (*s)
    {
        if (*s == c) return s;
        s++;
    }
    return NULL;
}

char *r_strrchr(char *s, int c)
{
    char *start;

    start = s;
    while (*s) s++;
    while (s > start)
    {
        if (*s == c) return s;
        s--;
    }
    return NULL;
}
#endif

#ifdef WANT_STRSTR
char *r_strstr(const char *haystack, const char *needle)
{
    const char *last,*pos;

    last = haystack; pos = needle;
    while (*haystack)
    {
        if (*haystack == *pos)
        {
            if (pos == needle) last = haystack;
            pos++;
            if (*pos == '\0') return (char *) last;
        }
        else if (pos != needle)
        {
            haystack = last;
            pos = needle;
        }
        haystack++;
    }
    return NULL;
}
#endif

#endif

char *stristr(char *haystack, char *needle)
{
    char *last,*pos;

    last = haystack; pos = needle;
    while (*haystack)
    {
        if (toupper(*haystack) == toupper(*pos))
        {
            if (pos == needle) last = haystack;
            pos++;
            if (*pos == '\0') return (char *) last;
        }
        else if (pos != needle)
        {
            haystack = last;
            pos = needle;
        }
        haystack++;
    }
    return NULL;
}

#ifndef UNIX
int get_event(void)
{
    if (kbhit()) return 1;
    if (mdm_kbhit()) return 2;
    return 0;
}
#endif

void pathchg(char *path)
{
    char tmp[1024], *str;

#ifdef __NO_DRIVES__
    char *after;

    if (path[1] == ':')
    {
        after = (path[2] == SLASH || path[2] == OTHER_SLASH) ? path+3 : path+2;
        str = dirlist[toupper(path[0])];
        if (str == NULL)
            strcpy(tmp, after);
        else
            sprintf(tmp, "%s"SSLASH"%s", str, after);
    }
    else
#endif
        strcpy(tmp, path);

    /* Change wrong slashes to right slashes */
    str = tmp;
    while (*str != '\0')
    {
        if (*str == OTHER_SLASH) *str = SLASH;
        str++;
    }
    strcpy(path, tmp);
}

#ifdef __OS2__
APIRET16 APIENTRY16 Win16SetTitle (HAB hab, PSZ title);

void set_title(char *title)
{
    HSWITCH hsw;
    SWCNTRL sw;
    PID pid;
    TID tid;

    hsw = WinQuerySwitchHandle(NULLHANDLE, getpid());

    if (WinQuerySwitchEntry(hsw, &sw) == 0)
    {
        strncpy (sw.szSwtitle, title, MAXNAMEL - 1);
        sw.szSwtitle[MAXNAMEL-1] = 0;

        WinQueryWindowProcess(sw.hwnd, &pid, &tid);

        WinChangeSwitchEntry(hsw, &sw);
    }
    Win16SetTitle(0, title);
}

char *get_title(char *title)
{
    HSWITCH hsw;
    SWCNTRL sw;

    hsw = WinQuerySwitchHandle(NULLHANDLE, getpid());

    if (WinQuerySwitchEntry(hsw, &sw) == 0)
        strcpy(title, sw.szSwtitle);

    return title;
}
#endif

#ifdef __NT__
void set_title(char *title)
{
    SetConsoleTitle(title);
}

char *get_title(char *title)
{
    GetConsoleTitle(title,80);
    return title;
}
#endif

#ifdef __OS2__
void _initkbd(void)
{
    /* Set keyboard to binary mode */
    KBDINFO kbstInfo;

    kbstInfo.cb = sizeof(kbstInfo);
    KbdGetStatus(&kbstInfo,0);
    kbstInfo.fsMask &= ~KEYBOARD_ASCII_MODE;
    kbstInfo.fsMask |= KEYBOARD_BINARY_MODE;
    KbdSetStatus(&kbstInfo,0);
}
#endif
