/* Driver for MSLP (Multiplexed Serial Link Protocol) */

#define VER 3

#ifdef _AMIGA

/* Amiga headers for SAS/C 6.00 */

#include <exec/types.h>
#include <intuition/intuition.h>
#include <proto/exec.h>
#include <proto/dos.h>
#include <clib/timer_protos.h>
#include <clib/intuition_protos.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>

#include "mslp.h"
#include "crc.h"
#include "serial.h"
#include "amiga.h"

#endif

#ifdef LINUX

/* Linux headers. */

#include <string.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <termio.h>
#include <stdio.h>
#include <stdlib.h>

#include "../modem.h"
#include "../keyb.h"
#include "mslp.h"
#include "crc.h"
#include "smodem.h"

#endif

/* OS/2 headers. */

#ifdef __OS2__
#define INCL_DOS
#include <os2.h>
#endif

#include <string.h>
#include <unistd.h>
#include <conio.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include "../modem.h"
#include "../keyb.h"
#include "mslp.h"
#include "crc.h"
#include "smodem.h"

#define CAN 24

#define HALT 'z'
#define CAPSULE '{'
#define CANCEL '|'
#define FRAME '}'
#define XCAPSULE (CAPSULE^0x20)
#define CAPSTART '!'
#define CAPSIZE 85

/*
	paketin loppu CAPSULE,FRAME
	paketin sisEllE:
		CAPSULE,FRAME -> CAPSULE,XCAPSULE,FRAME
		CAPSULE,CANCEL -> CAPSULE,XCAPSULE,CANCEL
		CAPSULE,XCAPSULE -> CAPSULE,XCAPSULE,XCAPSULE
*/

#define MAXCH 32
#define MINBLOCK 8
#define MAXBLOCK 1024
#define BUFFSIZE (MAXCH + MAXCH + 1 + MAXBLOCK + 2 + 2 + 8)
#define RSBUFFER 4096
#define WORKSIZE 256
#define MAXRETRY 15
#define NCHBIT 0xc0
#define NAKBIT 0xc0
#define ACKBIT 0x80
#define ENQBIT 0x40
#define PHSBIT 0x20
#define CHBIT 0x1f

struct channel
{
	int channel;            /* kanavan numero 0-31 */
	int fast;               /* korkea prioriteetti / vastaanotto tEysi */
	int enq;				/* paketin kuittaus on pyydetty */
	int phase;              /* toimintavaihe 0/PHSBIT */
	int ack;                /* paketti on saatu ok ACKBIT/NAKBIT */
	int count;              /* merkkilaskuri */
	int free;               /* tilaa vapaana */
	int used;               /* tilaa kEytetty */
	int retries;			/* uusintakertojen mEErE */
	char _far *start;       /* puskurin alku */
	char _far *end;         /* puskurin loppu */
	char _far *read;        /* lukukohta */
	char _far *write;       /* kirjoituskohta */
	struct list *list;      /* aktiivinen lista */
	struct channel *next;   /* ketjun seuraava */
	struct channel *prev;   /* ketjun edellinen */
};

struct list
{
	struct channel *first;
	struct channel *last;
};

//#ifdef ELITE
//unsigned long localserial = 0xffffffff;
//#else
unsigned long localserial = 0L;
//#endif

unsigned long remoteserial = 0L;

int RsStatus;

int mux_quit;

unsigned char mslp_capsule[256];
unsigned char mslp_mode[6];

static struct channel * muxtx[MAXCH];
static struct channel txchannel[MAXCH];
static unsigned char txpacket[BUFFSIZE];

static struct channel * muxrx[MAXCH];
static struct channel rxchannel[MAXCH];
static unsigned char rxpacket[BUFFSIZE];

time_t timer;
static time_t send_timer;
static time_t receive_timer;
static time_t ack_timer;

static int usecapsule;

char mslp_link_error[] = "MSLP data link error.";
char mslp_link_failure[] = "MSLP data link failure.";
char outmem[] = "Out of memory.";
static unsigned char frame[2] = {CAPSULE, FRAME};

int fifoin = 8;
int fifoout = 16;
int telix_error = 0;
unsigned cancelkey = 3;
#ifndef SMODEM
static int break_pressed = 0;
unsigned Key;
#endif

/* setupissa alustettavat muuttujat */

static struct list received;
static struct list sendfast;
static struct list resendfast;
static struct list send;
static struct list resend;
static struct list sent;
static struct list lost;

int mux_max_window;
int mux_window_size;
int mux_packet_size;
int mux_max_size;
int mux_min_size;

long mux_packets_sent;
long mux_bytes_sent;
long mux_packets_lost;

long mux_packets_received;
long mux_bytes_received;
long mux_packets_missed;

static int adaptive_count;
static char *errormsg;
static int mux_errorwait;

int display_messages = 1;
int return_code = 0;
int mux_errorlevel;

#ifdef PRO
char *initstring;
#endif

static unsigned char *send_ptr;
static unsigned char *send_end;

static unsigned char *rec_ptr;
static int rec_ok;
static int rec_syn;
static int rec_halt;


#ifdef _AMIGA

void qcputs(char *str)
{
	if (display_messages) PutStr(str);
}

time_t get_time(void)
{
	struct timeval kalatime;
	time_t kalakala;

	GetSysTime(&kalatime);
	kalakala=kalatime.tv_secs+252460800;

	timer=kalakala;
	return kalakala;
}

/* Amiga version omat muuttujat ja rutiinit */

int ReadKey(void)
{
	return 0;
}

void bell(void)
{
	DisplayBeep(0);
}

#endif

#ifdef LINUX

/* LINUX version rutiinit */

/* Definet siirretty mslp.h:hon jotta ne toimii muuallakin. */

/* piipaa */

void give_up(void)
{
	/*fd_set rfds;
	struct timeval tv;

	tv.tv_sec = 0;
	tv.tv_usec = 10000;

	FD_ZERO(&rfds);
	FD_SET(0, &rfds);

	if (select(1, &rfds, NULL, NULL, &tv) < 0)
	{
		fprintf(stderr, "select(): %s.\r\n", strerror(errno));
	}*/
}

#endif

void qcputs(char *str)
{
    if (display_messages) cputs(str);
}

int quiet = 0;

#ifdef __DOS__
void give_up(void)
{
}
#endif

#ifdef __OS2__
void give_up(void)
{
    DosSleep(1);
}

void bell(void)
{
    if (!quiet) DosBeep(1000,500);
}

int ReadKey(void)
{
    if (kbhit_nowait())
    {
        Key = getch();
        return 1;
    }
    return 0;
}
#endif

#ifdef GENERIC

/* nEppEimien luku */

#ifndef __OS2__
int ReadKey(void)
{
	if (kbhit())
	{
		Key = getch();
		return 1;
	}
	return 0;
}
#endif

#endif

/* viiveluuppi */

void makedelay(int sec)
{
	time_t start;
	start = get_time();
	while (elapsed(start) <= sec)
	{
		give_up();
		get_time();
	}
}

/* poisto listasta */

static void _fastcall list_remove(struct channel *ch)
{
	if (!ch->list) return;
	if (!ch->next)
		ch->list->last = ch->prev;
	else
		ch->next->prev = ch->prev;
	if (!ch->prev)
		ch->list->first = ch->next;
	else
		ch->prev->next = ch->next;
	ch->list = NULL;
}

/* lisEys listan loppuun */

static void _fastcall list_append(struct list *list, struct channel *ch)
{
	if (ch->list) list_remove(ch);
	if (!list->last)
	{
		list->last = list->first = ch;
		ch->prev = NULL;
	}
	else
	{
		list->last->next = ch;
		ch->prev = list->last;
		list->last = ch;
	}
	ch->next = NULL;
	ch->list = list;
}

/* listan ensimmEisen otto */

static struct channel * _fastcall list_first(struct list *list)
{
	struct channel *ch = list->first;
	if (!ch) return NULL;
	list->first->list = NULL;
	list->first = list->first->next;
	if (list->first)
		list->first->prev = NULL;
	else
		list->last = NULL;
	return ch;
}

/* crc:n laskenta */

static unsigned short _fastcall calculate_crc(unsigned char *ptr, unsigned char *end)
{
	unsigned short crc = CRCINIT;
	while (ptr < end) crc = CRC(crc, *ptr++);
	return crc;
}

/* kasvatetaan pakettia */

void ack_packet(void)
{
	if (++adaptive_count >= 128)
	{
		mux_packet_size *= 2;
		adaptive_count = 0;
	}
	else
		mux_packet_size += 1;
	if (mux_packet_size > mux_max_size) mux_packet_size = mux_max_size;
}

/* pienennetEEn pakettia */

void nak_packet(void)
{
	if (adaptive_count)
		mux_packet_size -= mux_packet_size / 16;
	else
		mux_packet_size /= 2;
	if (mux_packet_size < mux_min_size) mux_packet_size = mux_min_size;
	adaptive_count = 0;
}

/* lEhetyspaketin muodostaminen */

static unsigned char * send_packet(void)
{
	unsigned char *p;
	int i, c;
	struct channel *ch;
	unsigned short crc;

	if (rec_halt) return NULL;

	p = txpacket;

	while (received.first)
	{
		ch = list_first(&received);
		*p++ = (unsigned char) (ch->channel | ch->ack | ch->phase);
	}

	while (lost.first)
	{
		ch = list_first(&lost);
		*p++ = (unsigned char) (ch->channel | ENQBIT | ch->phase);
		list_append(&sent, ch);
		ch->enq = 1;
                if (ch->retries++ > MAXRETRY)
                {
                    mux_shutdown(mslp_link_failure, 2, 1);
                    return NULL;
                }
	}

	if (mux_window_size < mux_max_window)
	{
		for(;;)
		{
			ch = list_first(&resendfast);
			if (!ch)
			{
				ch = list_first(&sendfast);
				if (!ch)
				{
					ch = list_first(&resend);
					if (!ch)
					{
						ch = list_first(&send);
						if (!ch) break;
					}
				}
            }
			if (!ch->used) continue;

			*p++ = (unsigned char) (ch->channel | ch->phase);

			i = ch->used;
			if (i > mux_packet_size) i = mux_packet_size;
			c = ch->end - ch->read;
			if (i > c)
			{
				_fmemcpy(p, ch->read, c);
				_fmemcpy(p + c, ch->start, i - c);
			}
			else
			{
				_fmemcpy(p, ch->read, i);
			}
			mux_window_size += i;
			ch->count = i;
			ch->enq = 0;
			p += i;
			if (i == ch->used && ch->fast == MUX_OPT_TEXT) *p++ = 0;
			if (!sent.first) ack_timer = timer;
			list_append(&sent, ch);
			goto send;
		}
	}

	if (sent.last && elapsed(ack_timer) > 4)
	{
		*p++ = (unsigned char) (sent.last->channel | ENQBIT | sent.last->phase);
		sent.last->enq = 1;
		ack_timer = timer;
                if (sent.last->retries++ > MAXRETRY)
                {
                    mux_shutdown(mslp_link_failure, 2, 1);
                    return NULL;
                }
	}

	if (p == txpacket && elapsed(send_timer) < 5) return NULL;

send:
	crc = calculate_crc(txpacket, p);
	*p++ = CRC1(crc);
	*p++ = CRC2(crc);

	return p;
}

/* datan lEhetys paketista */

static void send_data(void)
{
	static unsigned char buffer[WORKSIZE];
	static unsigned char *tptr;
	static tsize = 0;
	static int bits;
	static int dle;
	unsigned long l;
	unsigned char *p;
	unsigned char c;
	unsigned u;
	int i;
				
	for (;;)
	{
		if (tsize)
		{
                        i = mdm_dataout_nowait(tptr, tsize);
			tsize -= i;
			tptr += i;
			if (tsize) return;
		}
		if (send_ptr < send_end)
		{   
			//if ((int) TxUsedSize > mux_buffer) return;
			p = buffer;
			do
			{
				if (usecapsule & OUTMAX)
				{
					l = *send_ptr++;
					for (u = 1; u < 4 && send_ptr < send_end; u++) l = (l << 8) | *send_ptr++;
					i = u;
					while (u)
					{   
						p[u--] = (unsigned char) (l % CAPSIZE + CAPSTART);
						l /= CAPSIZE;
					}   
					*p = (unsigned char) (l + CAPSTART);
					p += i + 1;
					continue;
				}
				else if (usecapsule & OUT7BIT)
				{
					if (!bits)
					{   
						u = 0;
						for (bits = 0; bits < 7; bits++)
						{   
							u = ((u & 0xff00) | (send_ptr[bits])) << 1;
						}
						c = (unsigned char)((u >> 8) ^ 0x20);
					}
					else
					{
						bits--;
						c = (unsigned char) (*send_ptr++ & 0x7f);
					}
				}
				else
				{
					c = *send_ptr++;
				}
				if (usecapsule & OUTCAP)
				{
					if (mslp_capsule[c])
					{
						*p++ = CAPSULE;
						c ^= 0x20;
					}
				}
                else 
                {   
	                if (dle)
					{
						if (c == HALT || c == FRAME || c == CANCEL || c == XCAPSULE)
						{
							*p++ = XCAPSULE;
						}
					}
					dle = (c == CAPSULE);
				}
				*p++ = c;
			}
			while (p < (buffer + sizeof buffer - 4) && send_ptr < send_end);
			tptr = buffer;
			tsize = p - buffer;
			//TxData(buffer, p - buffer);
			send_timer = timer;
		}
		else
		{
			if (send_ptr)
			{
				tptr = frame;
				tsize = sizeof frame;
				//TxData(frame, sizeof frame);
                                i = mdm_dataout_nowait(tptr, tsize);
				tsize -= i;
				tptr += i;
			}
                        send_end = send_packet();
                        if (mux_quit) return;
			if (!send_end)
			{
				send_ptr = send_end; /* NULL */
				if (tsize)
				{
					i = mdm_dataout_nowait(tptr, tsize);
					tsize -= i;
					tptr += i;
				}
				return;
			}
			send_ptr = txpacket;
			bits = 0;
			dle = 0;

			if (!tsize && elapsed(send_timer) > 1)
			{
				tptr = frame;
				tsize = sizeof frame;
			}
		}
	}
}

/* vastaanotetun paketin kEsittely */

static void _fastcall receive_packet(unsigned char *end)
{
	int i, c;
	unsigned char *p;
	int ph;
	struct channel *ch, *cc;

	receive_timer = timer;

        if (calculate_crc(rxpacket, end) != CRCCHECK) return;

	end -= 2;
	for (p = rxpacket; p < end; p++)
	{
		ph = *p & PHSBIT;
		c = *p & NCHBIT;

		if (c == ACKBIT)
		{
			ack_timer = timer;
			ch = muxtx[*p & CHBIT];
			if (!ch) goto channel_error;
			if (ph == ch->phase && ch->count)
			{
				mux_bytes_sent += ch->count;
				ch->used -= ch->count;
				ch->free += ch->count;
				mux_window_size -= ch->count;
				i = ch->end - ch->read;
				if (ch->count >= i)
					ch->read = ch->start + ch->count - i;
				else
					ch->read += ch->count;
				ch->phase ^= PHSBIT;
				ch->count = 0;
				if (ch->list == &sent) for(;;)
				{
					cc = list_first(&sent);
					if (cc == ch) break;
					list_append(&lost, cc);
				}
				else
				{
                                    mux_shutdown(mslp_link_error, 2, 1);
                                    return;
				}
				/* else goto channel_error; */

				if (ch->used)
				{
					if (ch->fast)
						list_append(&sendfast, ch);
					else
						list_append(&send, ch);
				}
				mux_packets_sent++;
				ch->retries = 0;
				ack_packet();
			}
		}
		else if (c == NAKBIT)
		{
			ack_timer = timer;
			ch = muxtx[*p & CHBIT];
			if (!ch) goto channel_error;
			if (ph == ch->phase && ch->count)
			{
				mux_window_size -= ch->count;
				ch->count = 0;
				ch->phase ^= PHSBIT;
				if (ch->list == &sent) for(;;)
				{
					cc = list_first(&sent);
					if (cc == ch) break;
					list_append(&lost, cc);
				}
                                else
                                {
                                    mux_shutdown(mslp_link_error, 2, 1);
                                    return;
                                }
				/* else goto channel_error; */

				if (ch->enq)
				{
					if (ch->fast)
						list_append(&resendfast, ch);
					else
						list_append(&resend, ch);
				}
				else
				{
					list_append(&send, ch);
				}
				mux_packets_lost++;
				nak_packet();
			}
		}
		else if (c == ENQBIT)
		{
			ch = muxrx[*p & CHBIT];
			if (!ch) goto channel_error;
			list_append(&received, ch);
			if (ph != ch->phase)
			{
				ch->ack = NAKBIT;
				ch->phase = ph;
				mux_packets_missed++;
				nak_packet();
			}
		}
		else
		{
			ch = muxrx[*p++ & CHBIT];
			if (!ch) goto channel_error;
			list_append(&received, ch);
			ch->phase = ph;
			c = end - p;
			if (c > ch->free)
			{
				ch->ack = NAKBIT;
				mux_packets_missed++;
				nak_packet();
			}
			else
			{
				i = ch->end - ch->write;
				if (c >= i)
				{
					_fmemcpy(ch->write, p, i);
					_fmemcpy(ch->start, p + i, c - i);
					ch->write = ch->start + c - i;
				}
				else
				{
					_fmemcpy(ch->write, p, c);
					ch->write += c;
				}
				ch->free -= c;
				ch->used += c;
				ch->ack = ACKBIT;
				mux_bytes_received += c;
				mux_packets_received++;
				ack_packet();
			}
			return;

		channel_error:
                    mux_shutdown(mslp_link_error, 2, 1);
                    break;
		}
	}
}

/* paketin vastaanotto */

static void receive_data(void)
{
	static unsigned char buffer[WORKSIZE];
	static int bits;
	static unsigned long lwork;
	static unsigned char bitc;
	unsigned char c, *p, *q;
	int i;

	while ((i = mdm_datain(buffer, sizeof buffer)) > 0)
	{
		p = buffer;
		q = buffer + i;
		if (rec_syn) goto syn1;
		do
		{
			if (rec_ptr >= rxpacket + sizeof rxpacket - 3)
			{
				rec_ok = 0;
				rec_ptr = rxpacket;
			}
			c = *p++;
			if (rec_halt)
			{
				if (c == CAN)
				{
                                    if (rec_halt++ >= 4)
                                    {
                                        mux_shutdown("MSLP data link interrupted.", 1, 1);
                                        return;
                                    }
                                    p++;
                                    continue;
				}
				rec_halt = 1;
			}
			if (c == CAPSULE)
			{
				if (p >= q)
				{
					rec_syn = 1;
					break;
			syn1:
					rec_syn = 0;
					c = CAPSULE;
				}
				switch (*p)
				{
				case FRAME:
					if (rec_ok)
					{
						if (usecapsule & INPMAX)
						{
							if (bits)
							{
								i = --bits;
								while (bits)
								{
									rec_ptr[--bits] = (unsigned char)(lwork & 0xff);
									lwork >>= 8;
								}
								rec_ptr += i;
							}
						}
                                                receive_packet(rec_ptr);
                                                if (mux_quit) return;
					}
					if (rec_halt)
					{
						receive_timer = get_time();
						rec_halt = 0;
					}
					bits = 0;
					lwork = 0L;
					rec_ok = 1;
					rec_ptr = rxpacket;
					p++;
					continue;

				case CANCEL:
                                    if (rec_ok && rec_ptr == rxpacket)
                                    {
                                        mux_shutdown("MSLP data link terminated.", 1, 0);
                                        return;
                                    }
                                    rec_ok = 0;
                                    rec_ptr = rxpacket;
                                    p++;
                                    continue;

				case HALT:
				    rec_halt = 1;
					rec_ok = 0;
					rec_ptr = rxpacket;
					p++;
					continue;

				case XCAPSULE:
					p++;
					break;

				default:
					if (usecapsule & INPCAP)
					{
						c = (unsigned char)(*p++ ^ 0x20);
					}
				}
			}
			else
			{
				if ((usecapsule & INPCAP) && (mslp_capsule[c] & INPCAP)) continue;
			}
			if (usecapsule & INPMAX)
			{
				if (c < CAPSTART || c >= CAPSTART + CAPSIZE) continue;
				lwork = lwork * CAPSIZE + (c - CAPSTART);

				if (++bits >= 5)
				{
					bits--;
					while (bits)
					{
						rec_ptr[--bits] = (unsigned char)(lwork & 0xff);
						lwork >>= 8;
					}
					rec_ptr += 4;
				}
				continue;
			}
			else if (usecapsule & INP7BIT)
			{
				if (!bits)
				{
					lwork = c ^ 0x20;
					bits = 7;
					continue;
				}
				c = (unsigned char)(c | ((bitc <<= 1) & 0x80));
				bits--;
			}
			*rec_ptr++ = c;
		}
		while (p < q);
	}
}

/* linjan varmistus */

void test_condition(void)
{
	static time_t carrtime;
	static time_t checktime;

        if (carr_det())
        {
            carrtime = timer;
        }
        else if (elapsed(carrtime) > 1)
        {
            mux_shutdown("Carrier lost.", 0, 1);
            mux_quit = 1;
            return;
	}

	if (elapsed(checktime) > 30) receive_timer = timer;
	checktime = timer;

	if (!rec_halt && elapsed(receive_timer) > 60)
	{
            mux_shutdown("MSLP data link time-out.", 1, 1);
            mux_quit = 1;
            return;
	}

#ifndef SMODEM
	if (break_pressed)
	{
            mux_shutdown("Ctrl+C pressed.", 1, 0);
            mux_quit = 1;
	}
#endif
}

/* virhekEsittely */

void mux_shutdown(char *msg, int level, int wait)
{
	//TxFlush();
	errormsg = msg;
	mux_errorlevel = level;
        mux_errorwait = wait;
        mux_quit = 1;
}

/* rs-portin sulku */

void shutdown(void)
{
	time_t start, rectime;
	unsigned char pc = 0;
	unsigned char c;
	static unsigned char string[] = {CAPSULE, FRAME, CAPSULE, CANCEL, CAPSULE, FRAME};

	cputs("\r\n");
	cputs(errormsg);
	cputs("\r\n");

	if (mux_errorlevel > 0)
	{
		mdm_dataout(string, sizeof string);
		start = rectime = get_time();
		if (mux_errorlevel > 1)
		{
			cputs("Waiting for MSLP termination... ");
			for (;;)
			{
				get_time();

				if (elapsed(start) >= 60 || elapsed(rectime) >= 15)
				{
					cputs("time-out\r\n");
					break;
				}

				if (kbhit() && getch() == (int) cancelkey)
				{
					cputs("cancelled\r\n");
					break;
				}

				if (!mdm_datain(&c, 1))
				{
					give_up();
					continue;
				}

				rectime = timer;
				if (pc == CAPSULE && c == CANCEL)
				{
					cputs("ok.\r\n");
					break;
				}
				pc = c;
			}
		}
	}

	if (mux_errorwait)
	{
		start = get_time();
		while (elapsed(start) < 5)
		{
			if (kbhit() && getch() == (int) cancelkey) break;
			give_up();
			get_time();
		}
	}
}


/* konversio niblestE hexaksi */
int ntox(int n)
{
	return (n > 9) ? (n + 'A' - 10) : (n + '0');
}

/* konversio hexasta nibleksi */
int xton(int x)
{
	return (x > '9') ? (x - 'A' + 10) : (x - '0');
}

/* rs-portin avaus */

int mux_setup(char *app, char ver)
{
	unsigned char *r = rxpacket, *q, *s;
	char t[2], pc;
	unsigned crc;
	time_t retry_timer;
	int setup = 0;
	int init = 0;
	int i;
	char c = 0;

	usecapsule = 0;
        if (localserial) usecapsule |= SERKEY;

	/* globaalien muuttujien alustus */

	received.first = NULL;
	received.last = NULL;
	sendfast = received;
	resendfast = received;
	send = received;
	resend = received;
	sent = received;
	lost = received;

        mux_quit = 0;
	mux_max_window = 16 * 1024;
	mux_window_size = 0;
	//mux_buffer = 2048;
	mux_packet_size = 1024;
	mux_max_size = 1024;
	mux_min_size = 16;

	mux_packets_sent = 0;
	mux_bytes_sent = 0;
	mux_packets_lost = 0;

	mux_packets_received = 0;
	mux_bytes_received = 0;
	mux_packets_missed = 0;

	adaptive_count = 96;
	errormsg = "MSLP data link closed.";
	mux_errorwait = 0;
	mux_errorlevel = 2;

	send_ptr = NULL;
	send_end = NULL;

	rec_ptr = rxpacket;
	rec_ok = 0;
	rec_syn = 0;
	rec_halt = 0;

	receive_timer = get_time();
        test_condition();
        if (mux_quit) return 0;
	mslp_capsule[CAPSULE] = 0;
	memcpy(txpacket + 1, app, 8);
	s = txpacket;
        *s++ = VER;	/* mslp versio */
        *s++ = ver;	/* app versio */
        *s++ = (char) usecapsule;
   	if (usecapsule & OUTCAP)
	{
		for (i = 0; i < 256; i++)
		{
			if (mslp_capsule[i] & OUTCAP) *s++ = (char) i;
		}
		*s++ = CAPSULE;
	}
   	if (usecapsule & INPCAP)
	{
		for (i = 0; i < 256; i++)
		{
			if (mslp_capsule[i] & INPCAP) *s++ = (char) i;
		}
		*s++ = CAPSULE;
	}
//	if (usecapsule & SERKEY)
//	{
//		memcpy(s, &localserial, 4);
//		s += 4;
//	}
	crc = calculate_crc(txpacket, s);
	*s++ = CRC1(crc);
	*s++ = CRC2(crc);

        i = -1;
	for(;;)
	{
		if (!init)
		{
			mdm_dataout(app, strlen(app));
			mdm_chrout('*');
			q = txpacket;
			while (q < s)
			{
				t[0] = (char) ntox((*q >> 4));
				t[1] = (char) ntox((*q & 0x0f));
				mdm_dataout(t, 2);
				q++;
			}
			mdm_dataout("*\r", 2);
		}

		retry_timer = get_time();
		while(elapsed(retry_timer) < 5)
		{
			get_time();
			if (kbhit())
			{
				if (getch() == (int) cancelkey)
				{
                                    mux_shutdown("Initialization cancelled.", 1, 0);
                                    return 0;
				}
			}
			test_condition();

			pc = c;
			if (!mdm_datain(&c, 1))
			{
				give_up();
				continue;
			}
			if (display_messages) putch(c);

			if ((c == CAN && pc == CAN && !setup) || (pc == CAPSULE && c == CANCEL) )
			{
                            mux_shutdown("Initialization interrupted.", 1, 0);
                            return 0;
			}

			if (pc == CAPSULE && c == FRAME)
			{
				if (setup)
				{
					mdm_dataout(frame, sizeof frame);
					goto start;
				}
				init = 1;
				i = -1;
				break;
			}

			if (c != '*')
			{
				if (i >= 0)
				{
					if (i++ & 1)
					{
						*r++ = (unsigned char) ((xton(pc) << 4) | xton(c));
						if (r >= rxpacket + sizeof rxpacket) i = -1;
					}
				}
			}
			else
			{
				crc = calculate_crc(rxpacket, r);
    			        if (crc != CRCCHECK) goto retry;
				if (!setup)
				{
					if (rxpacket[0] != VER)
					{
                                            mux_shutdown("Incompatible program version.", 1, 1);
                                            return 0;
					}
					if (rxpacket[2] & OUT7BIT) usecapsule |= INP7BIT;
					if (rxpacket[2] & INP7BIT) usecapsule |= OUT7BIT;
					if (rxpacket[2] & OUTMAX) usecapsule |= INPMAX;
					if (rxpacket[2] & INPMAX) usecapsule |= OUTMAX;

					q = rxpacket + 3;
					if (rxpacket[2] & OUTCAP)
					{
						usecapsule |= INPCAP;
						while (*q != CAPSULE)
						{
							mslp_capsule[*q] |= INPCAP;
							q++;
						}
						q++;
	                }

					if (rxpacket[2] & INPCAP)
					{
						usecapsule |= OUTCAP;
						while (*q != CAPSULE)
						{
							mslp_capsule[*q] |= OUTCAP;
							q++;
						}
						q++;
					}
//					if (rxpacket[2] & SERKEY) memcpy(&remoteserial, q, 4);
				}
				mdm_dataout(frame, sizeof frame);
				if (init) goto start;
				setup = 1;

			retry:
				i = 0;
				r = rxpacket;
				break;
			}
		}
        }
start:

	mslp_capsule[CAPSULE] |= usecapsule & (OUTCAP | INPCAP);

	if (usecapsule & INPMAX)
		strcpy((char *) mslp_mode, "7X");
	else
	{
		if (usecapsule & INP7BIT) strcpy((char *) mslp_mode, "7"); else strcpy((char *) mslp_mode, "8");
		if (usecapsule & INPCAP) strcat((char *) mslp_mode, "C");
	}
	strcat((char *) mslp_mode, "/");
	if (usecapsule & OUTMAX)
		strcat((char *) mslp_mode, "7X");
	else
	{
		if (usecapsule & OUT7BIT) strcat((char *) mslp_mode, "7"); else strcat((char *) mslp_mode, "8");
		if (usecapsule & OUTCAP) strcat((char *) mslp_mode, "C");
	}
	qcputs("\r\nMSLP data link established.");
	qcputs(" Data link i/o mode ");
	qcputs((char *) mslp_mode);
	qcputs(" bits.\r\n");
	return rxpacket[1]; /* palautetaan remote app versio */
}

/* pakettien "tausta-ajo" */

void mux_run(void)
{
	get_time();
	receive_data();
	test_condition();
	send_data();
}

/* lEhetyskanavan avaus */

void mux_open_tx_channel(int channel, int size, int opt)
{
	struct channel *ch = &txchannel[channel];
	char _far *buffer;
	buffer = (char _far *) _fmalloc(size);
        if (!buffer)
        {
            mux_shutdown(outmem, 2, 1);
            return;
        }
	muxtx[channel] = ch;
	ch->retries = 0;
	ch->channel = channel;
	ch->start = buffer;
	ch->read = buffer;
	ch->write = buffer;
	ch->end = buffer + size;
	ch->free = size;
	ch->used = 0;
	ch->enq = 0;
	ch->phase = 0;
	ch->ack = 0;
	ch->count = 0;
	ch->retries = 0;
	ch->fast = opt;
}

/* lEhetyskanavan vapaa */

int _fastcall mux_tx_free(int channel)
{
	return muxtx[channel]->free;
}

/* lEhetyskanavan kEytetty */

int _fastcall mux_tx_used(int channel)
{
	return muxtx[channel]->used;
}

/* tekstilEhetyksen valmiustila */

int _fastcall mux_tx_ready(int channel)
{
	if (!muxtx[channel]->free) return 0;
	if (muxtx[channel]->used == muxtx[channel]->count) return 0;
	return 1;
}

/* lEhetys kanavaan */

int _fastcall mux_tx_data(int channel, char *ptr, int size)
{
	struct channel *ch = muxtx[channel];
	int u;

	if (size > ch->free) size = ch->free;
	if (size)
	{
		u = ch->end - ch->write;
		if (size >= u)
		{
			_fmemcpy(ch->write, ptr, u);
			_fmemcpy(ch->start, ptr + u, size - u);
			ch->write = ch->start + size - u;
		}
		else
		{
			_fmemcpy(ch->write, ptr, size);
			ch->write += size;
		}
		ch->free -= size;
		ch->used += size;
		if (!ch->list)
		{
			if (ch->fast)
				list_append(&sendfast, ch);
			else
				list_append(&send, ch);
		}
	}
	return size;
}

/* lEhetyskanavan nollaus */

void _fastcall mux_tx_flush(int channel)
{
	struct channel *ch = muxtx[channel];
	ch->free += ch->used;
	if (ch->count)
	{
		mux_window_size -= ch->count;
		ch->count = 0;
		ch->phase ^= PHSBIT;
		list_remove(ch);
	}
	ch->used = 0;
	ch->read = ch->write = ch->start;
}

/* vastaanottokanavan avaus */

void mux_open_rx_channel(int channel, int size)
{
	struct channel *ch = &rxchannel[channel];
	char _far *buffer = (char _far *) _fmalloc(size);
        if (!buffer)
        {
            mux_shutdown(outmem, 2, 1);
            return;
        }
	muxrx[channel] = ch;
	ch->channel = channel;
	ch->phase = PHSBIT;
	ch->ack = NAKBIT;
	ch->start = buffer;
	ch->read = buffer;
	ch->write = buffer;
	ch->end = buffer + size;
	ch->free = size;
	ch->used = 0;
}

/* vastaanottokanavan kEytetty */

int _fastcall mux_rx_used(int channel)
{
	return muxrx[channel]->used;
}

/* vastaanottokanavan vapaa */

int _fastcall mux_rx_free(int channel)
{
	return muxrx[channel]->free;
}

/* vastaanotto kanavasta */

int _fastcall mux_rx_data(int channel, char *ptr, int size)
{
	struct channel *ch = muxrx[channel];
	int u;

	if (size > ch->used) size = ch->used;
	if (size)
	{
		u = ch->end - ch->read;
		if (size >= u)
		{
			_fmemcpy(ptr, ch->read, u);
			_fmemcpy(ptr + u, ch->start, size - u);
			ch->read = ch->start + size - u;
		}
		else
		{
			_fmemcpy(ptr, ch->read, size);
			ch->read += size;
		}
		ch->free += size;
		ch->used -= size;
	}
	return size;
}

/* vastaanottokanavan nollaus */

void _fastcall mux_rx_flush(int channel)
{
	struct channel *ch = muxrx[channel];
	ch->read = ch->write = ch->start;
	ch->free += ch->used;
	ch->used = 0;
}

void close_channels(void)
{
	int i;

        _free(txchannel[CH_SYSTEM].start);
        _free(rxchannel[CH_SYSTEM].start);

	for (i = CH_FIRST_OUTPUT; i <= CH_LAST_OUTPUT; i++)  /* tekstin siirtokanavat */
	{
            _free(txchannel[i].start);
            _free(rxchannel[i].start);
	}

        _free(txchannel[CH_DOWNLOAD].start);
        _free(rxchannel[CH_DOWNLOAD].start);

        _free(txchannel[CH_UPLOAD].start);
        _free(rxchannel[CH_UPLOAD].start);

        _free(txchannel[CH_CONTROL].start);
        _free(rxchannel[CH_CONTROL].start);

	for (i = CH_FIRST_FILE; i <= CH_LAST_FILE; i++) /* tiedostojen siirtokanavat */
	{
            _free(txchannel[i].start);
            _free(rxchannel[i].start);
	}
}
