#pragma check_stack(off)

#define TRUE 1
#define FALSE 0

/* optimointia joillekin muuttujille */
#define _register static

/************************ serial.lib funktiot **********************/

void _cdecl _interrupt _far Serial(void); /* keskeytysrutiini */

#define TxSize() TxUsedSize
#define TxFree() TxFreeSize

void _fastcall TxInsert(unsigned Size);
void TxFlush(void);
int TxEmpty(void);

int _fastcall TxChar(char c);
unsigned _fastcall TxData(const char *Buffer,unsigned Size);
unsigned _fastcall TxFarData(const char _far *Buffer, unsigned Size);
unsigned _fastcall TxFile(int Handle, unsigned Size);
int _fastcall TxStr(const char *c);
int _fastcall TxFarStr(const char _far *c);
unsigned _fastcall TxMaxBlock(unsigned Size);

#define RxSize() RxUsedSize
#define RxFree() RxFreeSize

void _fastcall RxDelete(unsigned Size);
void RxFlush(void);

char _fastcall RxSeek(unsigned Pos);

int RxChar(void);
unsigned _fastcall RxData(char *Buffer,unsigned Size);
unsigned _fastcall RxFarData(char far *Buffer, unsigned Size);
unsigned _fastcall RxFile(int Handle, unsigned Size);
unsigned _fastcall RxMaxBlock(unsigned Size);

unsigned RsSetUp(int addr, int irq, unsigned baud,
	char parity, int bits, int stops, char flow, int infifo, int outfifo,
	char *RxBuffer, unsigned RBSize, char *TxBuffer, unsigned TBSize);

/* Kello-oskillaattorin taajuus 1.8432 Mhz / 16 = 115200 Hz */
#define SERIAL_CLOCK 115200L
#define BAUD(bps) (unsigned)(SERIAL_CLOCK/bps)

/*
	Asetukset:
		addr = portin osoite
		irq = 0..7
		baud = BAUD(9600)
		parity = 'N', 'E', 'O', 'M' tai 'S'
		bits = 7 tai 8
		stops = 1 tai 2
		flow = 'W', 'H', 'S' tai 'B'
		infifo = 8
		outfifo = 16
		RxBuffer = rdata[]
		RBSize = sizeof(rdata)
		TxBuffer = tdata[]
		TBSize = sizeof(tdata)
*/

void _fastcall RsShutDown(int resetstate);
void _fastcall RsSetDtr(int dtr);
void _fastcall RsSetRts(int rts);
void _fastcall RsSetXon(int xon);
int _fastcall RsRing(int clear);

/* Suorat osoittimet buffereihin */
extern char *RxBufStart;                /* vastaanottopuskurialue */
extern char *RxBufEnd;
extern char *RxRead;                    /* lukuosoitin puskuriin */

extern char *TxBufStart;                /* lhetyspuskurialue */
extern char *TxBufEnd;
extern char *TxWrite;                   /* kirjoitusosoitin puskuriin */

extern unsigned RxFreeSize;             /* puskurin koko */
extern unsigned RxUsedSize;             /* puskurin koko */
extern unsigned TxFreeSize;             /* puskurin koko */
extern unsigned TxUsedSize;             /* puskurin koko */

/* RsStatus bitit */

extern int RsStatus;            /* statustiedot */
extern char Use7bit;            /* 7 bitti kytss */

/* line status bitit */
#define DATAREADY 0x0001

/* line error bitit */
#define OVERRUN 0x0002
#define PARITY 0x0004
#define FRAMING 0x0008
#define BREAK 0x0010

/* line status bitit */
#define THEMPTY 0x0020

/* kttelymoodit */
#define USERTSCTS 0x0040
#define USEXONXOFF 0x0080

/* bufferin virhe */
#define OVERFLOW 0x0100

/* lhettimen busy status */
#define FORCEDBUSY 0x0200

/* modem status */
#define TERI 0x0400

/* lhettimen busy status */
#define XOFFBUSY 0x0800

/* prosessoitu modem status */
#define CTSBUSY 0x1000

/* modem status */
#define DSR 0x2000

/* vastaanottimen busy tila */
#define RXBUSY 0x4000

/* modem status */
#define DCD 0x8000

#define RsCarrier() (RsStatus & DCD)
#define RsReady() (RsStatus & DSR)
#define RsBusy() (RsStatus & (CTSBUSY | XOFFBUSY))

int RxError(void);
void TxBreak(void);
