void debug(char * text);

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#ifdef UNIX
#include <termios.h>
#include <utime.h>
#else
#define cputs puts
#endif
#include "mslp.h"
#include "smodem.h"
#include "terminal.h"
#include "../output.h"

#define READWRITE (0666)

#include "../logfile.h"
#include "../files.h"
#include "../keyb.h"
#include "../os.h"

long speed;

static char dl_buff[HEADER_LEN + 2 * PATH_LEN];
static char ul_buff[HEADER_LEN + PATH_LEN];

struct dir_path _far *ulpath_first;
struct dir_path _far *ulpath_last;

struct dir_path _far *dlpath_first;
struct dir_path _far *dlpath_last;

static int dlhandle;
static int ulhandle;
char dlpath[PATH_LEN];
char defpath[PATH_LEN];

static time_t dl_start = 0;
static time_t ul_start = 0;

static int control_res = 0;

unsigned long localcheck = 0;
unsigned long remotecheck = 0;

int setup_packet = MSG_HUGE;
static int yell_mode = 0;
int quietstart = -1;
int quietstop = 0;
static int transfer = 0;
static int forced_dl = 0;
time_t start_time;
time_t idle_start_time;
int includeaborted = 0;
int removefiles = 0;
int deletefiles = 0;
int ulreq = 1;
int overwrite = 0;
int renumber = 0;
int restrict = 1;
int wildcard = 1;
int dlreq = 0;
int resume = 1;
int messages = 1;
int exit_delay_time = 0;
int idle_delay_time = 0;
int autoexit = 0;

static int fatal_exit = 0;
int txcps = 0;
int rxcps = 0;
int cur_rxcps = 0;
int cur_txcps = 0;
int update_packet = 0;
struct file_entry _far *dl = NULL;
static struct file_entry _far *ul = NULL;
static int send_abort_ul = 0;
static int send_abort_dl = 0;
static int dlch = CH_FIRST_FILE;
static int dlcount = BUCKETSIZE;
static int ulch = CH_FIRST_FILE;
static int ulcount = BUCKETSIZE;

long cur_download_bytes = 0L;
long cur_download_left = 0L;
long cur_upload_bytes = 0L;
long cur_upload_left = 0L;

static long upload_files = 0L;
static long upload_bytes = 0L;
static long uploaded_bytes = 0L;
static long upfiles_left = 0L;
static long upbytes_left = 0L;

static long download_files = 0L;
static long download_bytes = 0L;
static long downloaded_bytes = 0L;
long downfiles_left = 0L;
static long downbytes_left = 0L;

int logfile = -1;
char Buffer[BUCKETSIZE] = "";

char terminal_prompt[40] = "";
int promptlen;
char input[1000];

#ifdef SCHAT			
int remote_charset = 0;
#endif

static char packets[] = "TSCMLH";

static char *ul_ptr = ul_buff;
static char *dl_ptr = dl_buff;

struct file_list send_upload_list = {NULL, NULL};
struct file_list sent_upload_list = {NULL, NULL};
struct file_list upload_list = {NULL, NULL};
struct file_list uploading_list = {NULL, NULL};

struct file_list download_rename_list = {NULL, NULL};
struct file_list download_list = {NULL, NULL};
struct file_list downloaded_list = {NULL, NULL};

struct file_list send_request_list = {NULL, NULL};
struct file_list sent_request_list = {NULL, NULL};
struct file_list request_path_list = {NULL, NULL};

static struct file_list error_list = {NULL, NULL};

unsigned upload_id = 0;

char *logpath = NULL;
char termpath[PATH_LEN] = "SMODEM.$$$";

static void system_run(void);
static void upload_run(void);
static void download_run(void);
static void control_run(void);
static void status_run(void);
static void close_dl(void);
static void close_ul(void);
void insert_log(char mode, struct file_entry _far *_new, int cps);

#ifdef SCHAT	
int busyrun;
#endif

int smodem_run(void)
{
    int i;

        i = kbhit() ? getch() : 0;
	if (i)
        {
                Key = i;
		yell_mode = 0;
		return i;
	}

        mux_run();
        if (mux_quit) return 0;
	system_run();
	control_run();
	download_run();
	upload_run();
        status_run();
        if (mux_quit) return 0;
	mux_run();
        if (mux_quit) return 0;

#ifndef SCHAT
	terminal_run();
#endif
	UpdateStat(1);

        i = kbhit() ? getch() : 0;
	if (i)
	{
                Key = i;
		yell_mode = 0;
		return i;
	}
#ifdef SCHAT	
	if (!busyrun) 
#endif
		give_up();
	return 0;
}

void exit_smodem(void)
{
	close_dl();
	close_ul();

	if (logfile != -1) 
	{
		close(logfile);
	}

	if (fatal_exit)
	{
		sprintf(Buffer, "Fatal internal error %d.\r\n", fatal_exit);
		cputs(Buffer);
	}
#ifdef SMODEM
#ifndef SCHAT
	setcwd(defpath);
#endif
#endif
}

static void fatal_error(int i)
{
        fatal_exit = i;
        write_log("smodem : fatal_error(%d)", i);
	exit(return_code);
}

char * _fastcall strspace(char *p)
{
	p = strchr(p, ' ');
	if (p)
	{
		*p = 0;
		do p++; while (*p == ' ');		
	}
	return p;
}

static void _fastcall file_list_append(
	struct file_list *list, struct file_entry _far *entry)
{
	if (!entry) fatal_error(3);
	if (!list->last)
	{
		list->last = list->first = entry;
		entry->prev = NULL;
	}
	else
	{
		list->last->next = entry;
		entry->prev = list->last;
		list->last = entry;
	}
	entry->next = NULL;
}

static struct file_entry _far * _fastcall file_list_first(struct file_list *list)
{
        struct file_entry _far *entry = list->first;
	if (!entry) return NULL;

        list->first = list->first->next;
        if (list->first)
                list->first->prev = NULL;
        else
		list->last = NULL;
        entry->prev = entry->next = NULL;
	return entry;
}

#ifdef __linux__
void _splitpath(char *path, char *drive, char *dir, char *fname, char *ext)
{
    char *strp, *ext2;

    *drive = '\0';
    strp = strrchr(path, '/');
    if (strp == NULL)
    {
        *dir = '\0';
        strp = path;
    }
    else
    {
        *strp = '\0';
        strcpy(dir, path);
        *strp++ = '/';
    }
    ext2 = strrchr(strp, '.');
    if (ext2 == NULL)
    {
        *ext = '\0';
        strcpy(fname, strp);
    }
    else
    {
        *ext2 = '\0';
        strcpy(fname, strp);
        strcpy(ext, ext2+1);
        *ext2 = '.';
    }
}

void _makepath(char *path, char *drive, char *dir, char *fname, char *ext)
{
    drive = drive;

    if (*dir)
    {
        if (*ext)
            sprintf(path, "%s/%s.%s", dir, fname, ext);
        else
            sprintf(path, "%s/%s", dir, fname);
    }
    else
    {
        if (*ext)
            sprintf(path, "%s.%s", fname, ext);
        else
            strcpy(path, fname);
    }
}

void dosify(char *path)
{
    while (*path != '\0')
    {
        if (*path == '/') *path = '\\';
        path++;
    }
}

void unixify(char *path)
{
    while (*path != '\0')
    {
        if (*path == '\\') *path = '/';
        path++;
    }
}

void stampdate(char *fname, unsigned short date, unsigned short time)
{
    struct utimbuf buf;
    struct tm tim;

    memset(&tim, 0, sizeof(tim));
    tim.tm_sec = (time & 0x1f)*2;
    tim.tm_min = (time >> 5) & (0x1f+32);
    tim.tm_hour = time >> 11;
    tim.tm_mday = (date & 0x1f);
    tim.tm_mon = ((date >> 5) & 0x0f)-1;
    tim.tm_year = (date >> 9)+80;
    buf.actime = buf.modtime = mktime(&tim);
    utime(fname, &buf);
}
#endif

unsigned short makedostime(time_t _tim)
{
    struct tm *tim;

    tim = localtime(&_tim);
    return (tim->tm_sec/2) + (tim->tm_min << 5) + (tim->tm_hour << 11);
}

unsigned short makedosdate(time_t _tim)
{
    struct tm *tim;

    tim = localtime(&_tim);
    return (tim->tm_mday) + ((tim->tm_mon+1) << 5) + ((tim->tm_year-80) << 9);
}

#if 0
int upload(char *file, char *lpath, char *rpath, char status)
{
#if defined (MSDOS) || defined (__OS2__)
	struct find_t buffer;
#else
	struct stat st;
#endif 
	int len;
	struct file_entry _far *_new;
        char path[2 * PATH_LEN];
        char drive[3];
	char dir[PATH_LEN];
	char fname[PATH_LEN];
	char ext[PATH_LEN];
	int count = 0;
	struct dir_path _far *list = NULL;
	
	return_code |= 2;
	if (!wildcard && (strchr(file, '?') || strchr(file, '*'))) return 0;
	
	_splitpath(file, drive, dir, fname, ext);
	if ((!status && restrict) || (!*drive && !*dir))
	{
		*drive = 0;
		*dir = 0;
		if (lpath && *lpath)
			strcpy(dir, lpath);
		else
			list = ulpath_first;
	}
	do
	{
		if (list) 
		{
			_fstrcpy(dir, list->path);
			list = list->next;
		}
		if (!*drive && !*dir) 
#ifdef SMODEM
		strcpy(dir, defpath);
#else
		*dir = 0;
#endif
		_makepath(path, drive, dir, fname, ext);
#if defined (MSDOS) || defined (__OS2__)
		if (_dos_findfirst(path, _A_NORMAL | _A_RDONLY, &buffer)) continue;
		do
		{
			_makepath(path, drive, dir, buffer.name, "");
	
			if (rpath && *rpath)
			{
				strcat(path, " ");
				strcat(path, rpath);
			}
#else
			if (stat(path, &st) < 0) continue;
#endif
			len = strlen(path) + 1;
			_new = (struct file_entry _far *) _fmalloc(sizeof *_new + len);
			if (!_new) break;
#if defined (MSDOS) || defined (__OS2__)
			_fmemcpy(&_new->time, &buffer.wr_time, 8);
#endif
#ifdef LINUX
			_new->time = makedostime(st.st_mtime);
			_new->date = makedosdate(st.st_mtime);
			_new->size = st.st_size;
#endif
#ifdef AMIGA
			strcpy(_new->name, path);
			_new->size = get_file_size(path);
#endif
			_fmemcpy(_new->name, path, len);
			_new->status = status;
			//_new->id = ++upload_id;
			_new->start = 0L;
			file_list_append(&send_upload_list, _new);
			count++;
#if defined (MSDOS) || defined (__OS2__)
		}
		while (!_dos_findnext(&buffer));
#endif
	}
	while (list);
	return count;
}
#else
int upload(char *file, char *lpath, char *rpath, char status)
{
	struct stat st;
	int len;
	struct file_entry _far *_new;
        char path[2*PATH_LEN];

        lpath = rpath;
        return_code |= 2;
        strcpy(path, file);

        if (stat(path, &st) < 0) return 0;

        len = strlen(path) + 1;
        _new = (struct file_entry _far *) _fmalloc(sizeof *_new + len);
        if (!_new) return 0;
        _new->time = makedostime(st.st_mtime);
        _new->date = makedosdate(st.st_mtime);
        _new->size = st.st_size;
        _fmemcpy(_new->name, path, len);
        _new->status = status;
        //_new->id = ++upload_id;
        _new->start = 0L;
        file_list_append(&send_upload_list, _new);

	return 1;
}
#endif

int download(char *file, char *rpath, char *lpath)
{
	struct file_entry _far *_new;
	int len;
        char path[2 * PATH_LEN];

        lpath = rpath;
	strcpy(path, file);
	len = strlen(path) + 1;
	_new = (struct file_entry _far *) _fmalloc(sizeof *_new + len);
	if (!_new) return 0;
	_fmemcpy(_new->name, path, len);
	_new->status = 1;
	file_list_append(&send_request_list, _new);
	return 1;
}

int downloadname(char *file)
{
	struct file_entry _far *_new;
	int len;

	forced_dl = 1;
	return_code |= 1;

	len = strlen(file) + 1;
	_new = (struct file_entry _far *) _fmalloc(sizeof *_new + len);
	if (!_new) return 0;
	_fmemcpy(_new->name, file, len);
	file_list_append(&download_rename_list, _new);
	return 1;
}

int xtoi(char *hex)
{
	int c, d, val = 0;
        int i = 0;

	while (isspace(*hex)) hex++;
	for(c = 4; c ; c--)
	{
		if (*hex >= '0' && *hex <= '9')
			d = *hex - '0';
		else if (*hex >= 'a' && *hex <= 'f')
			d = *hex - 'a' + 10;
		else if (*hex >= 'A' && *hex <= 'F')
			d = *hex - 'A' + 10;
		else
		{
			if (!i) return -1;
			break;
		}
		val = 16 * val + d;
		hex++;
		i = 1;
	}
	return val;
}

int set_packets(int msg)
{
#ifdef SMODEM
	SetMenuFlags(&MaxSizeMenu, setup_packet, NORMAL);
#endif	
	setup_packet = msg;
	switch (setup_packet)
	{
	case MSG_TINY:
		mux_max_size = 256;
		mux_max_window = 1;
		break;
	case MSG_SMALL:
		mux_max_size = 256;
		mux_max_window = 512;
		break;
	case MSG_COMPACT:
		mux_max_size = 256;
		mux_max_window = 1024;
		break;
	case MSG_MEDIUM:
		mux_max_size = 256;
		mux_max_window = 4096;
		break;
	case MSG_LARGE:
		mux_max_size = 1024;
		mux_max_window = 4096;
		break;
	case MSG_HUGE:
		mux_max_size = 1024;
		mux_max_window = 16384;
		break;
	default:
		return -1;
	}
#ifdef SMODEM
	SetMenuFlags(&MaxSizeMenu, setup_packet, CHECKED);
#endif
	mux_packet_size = mux_max_size;
	return 0;
}

static void dl_cps_calc(long size)
{
	static time_t dl_last_time;
	static long dl_last_bytes;
	long l, e;

	if (!downfiles_left)
	{
		dl_start = 0;
		downloaded_bytes = 0;
		dl_last_bytes = 0;
		return;
	}
	
	l = downloaded_bytes + size;
	if (!l) return;
		
	if (!dl_start)
	{
		dl_start = timer;
		dl_last_time = timer;
	}
	
	e = elapsed(dl_start);
	if (e < 1) e = 1;
	rxcps = (int)(l / e);

	if (e < 10) 
	{
		cur_rxcps = rxcps;
		return;
	}
	
	e = elapsed(dl_last_time);
	if (e < 60) return;
	cur_rxcps = (int)((l - dl_last_bytes) / e);
	dl_last_bytes = l;
	dl_last_time = timer;
}

static void ul_cps_calc(long size)
{
	static time_t ul_last_time;
	static long ul_last_bytes;
	long l, e;

	if (!upfiles_left)
	{
		ul_start = 0;
		uploaded_bytes = 0;
		ul_last_bytes = 0;
		return;
	}
	
	l = uploaded_bytes + size;
	if (!l) return;
		
	if (!ul_start)
	{
		ul_start = timer;
		ul_last_time = timer;
	}
	
	e = elapsed(ul_start);
	if (e < 1) e = 1;
	txcps = (int)(l / e);
	
	if (e < 10) 
	{
		cur_txcps = txcps;
		return;
	}
	
	e = elapsed(ul_last_time);
	if (e < 60) return;
	cur_txcps = (int)((l - ul_last_bytes) / e);
	ul_last_bytes = l;
	ul_last_time = timer;
}

void UpdateStat(int mode)
{
	static time_t last_update = 0;

	long cur_rx_size;
	long cur_tx_size;
        int i;
#ifdef SMODEM
        int rp, tp;
#endif

	if (mode && last_update == timer) return;
	last_update = timer;

#ifdef SMODEM
	Win = &MenuStat;
	SetColor(mcolor[0]);
	sprintf(Buffer, "%2.2d:%2.2d:%2.2d", dostime.hour, dostime.minute, dostime.second);
	SetWindowStr(0, 51, Buffer, 8);
#endif

	cur_rx_size = 0L;
	cur_tx_size = 0L;
	for (i = CH_FIRST_FILE; i <= CH_LAST_FILE; i++)
	{
		cur_rx_size += mux_rx_used(i);
		cur_tx_size -= mux_tx_used(i);
	}
    
    dl_cps_calc(cur_rx_size);
    ul_cps_calc(cur_tx_size);

#if defined(PRO) || defined(SCHAT)
 	if (quietstart >= 0)
 	{   
 		if ((!quiet) == (quietstop > quietstart))
 		{   
			if (dostime.hour >= quietstop && dostime.hour < quietstart) change_quiet();
 		}
 		else
 		{
 			if (dostime.hour >= quietstart || dostime.hour < quietstop) change_quiet();
 		}
 	}
#endif

	/*if (yell_mode)
	{
		yell_mode--;
		bell();
	}*/

#ifdef _AMIGA
	amiga_update_downcps(cur_rxcps);
	amiga_update_upcps(cur_txcps);
#endif

#ifdef SMODEM
	SetColor(mcolor[0]);
	rp = display_time(cur_rxcps, download_bytes, downbytes_left - cur_rx_size, 16);
	tp = display_time(cur_txcps, upload_bytes, upbytes_left - cur_tx_size, 41);

	/*if (!ViewStat) return;

	Win = &Statistics;*/
	
	display_stat(2, 16, cur_download_bytes);
	display_stat(2, 32, cur_upload_bytes);

	display_stat(3, 16, cur_download_left);
	display_stat(3, 32, cur_upload_left);

	display_stat(4, 16, download_files);
	display_stat(4, 32, upload_files);

	display_stat(5, 16, download_bytes);
	display_stat(5, 32, upload_bytes);

	display_stat(6, 16, downfiles_left);
	display_stat(6, 32, upfiles_left);

	display_stat(7, 16, downbytes_left - cur_rx_size);
	display_stat(7, 32, upbytes_left - cur_tx_size);

	display_stat(8, 16, (long) rxcps);
	display_stat(8, 32, (long) txcps);

	display_stat(9, 16, (long) cur_rxcps);
        display_stat(9, 32, (long) cur_txcps);

	display_stat(1, 68, (long) mux_max_window);
	display_stat(2, 68, (long) mux_window_size);
	display_stat(3, 68, (long) mux_packet_size);
	display_stat(4, 68, mux_packets_received);
	display_stat(5, 68, mux_bytes_received);
	display_stat(6, 68, mux_packets_missed);
	display_stat(7, 68, mux_packets_sent);
	display_stat(8, 68, mux_bytes_sent);
	display_stat(9, 68, mux_packets_lost);
	
	display_bar(rp, 26);
	display_bar(tp, 42);
#endif
}

void set_prompt(char *str)
{
	int i;
	strncpy(terminal_prompt, str, sizeof terminal_prompt - 1);
	i = strlen(terminal_prompt);
#ifdef _AMIGA
	amiga_changeprompt(terminal_prompt);
#endif
#ifdef SMODEM
	InputDialogField[0].Col = i;
	InputDialogField[0].Width = 80 - i;
	if (Dlg == &InputDialog)
		UpdateDialogField(InputDialogField);
#endif
#ifdef LINUX	
	/*promptlen = strlen(terminal_prompt);
	lsd_printprompt();*/
#endif
        promptlen = i > 0 ? (i+1) : 0;
}

static void system_run(void)
{
	static char buff[80];
	static char system_mode = '*';
	static char *systemptr = buff;
	char c;
	
	while (mux_rx_used(CH_SYSTEM))
	{
		mux_rx_data(CH_SYSTEM, &c, 1);
		if (!system_mode)
		{   
			switch (c)
			{
			case 'B':
#ifdef SCHAT
				chatstatus.graphics = 'N';
#endif
				break;
				
			case 'Y':
				yell_mode = 30;
				break;

			case 'Q':
#ifdef SMODEM
				SetMenuFlags(&SystemMenu, MSG_YELL, DISABLED);
#endif
				break;
											
			case 'q':
#ifdef SMODEM
				SetMenuFlags(&SystemMenu, MSG_YELL, NORMAL);
#endif
				break;
				
			case 'D':
#ifndef SCHAT
#ifdef SMODEM
				SetMenuFlags(&SystemMenu, MSG_DOWNLOAD, NORMAL);
#endif
#endif
				break;

			case 'U':
#ifndef SCHAT
#ifdef SMODEM
				SetMenuFlags(&SystemMenu, MSG_UPLOAD, NORMAL);
#endif
#endif
				break;

			case 'R':
#ifndef SCHAT
#ifdef SMODEM
				SetDialogFlags(&UploadDialog, MSG_REMOTE, NORMAL);
				SetDialogFlags(&DownloadDialog, MSG_REMOTE, NORMAL);
#endif
#endif
				break;

			case 'L':    
			case 'C':
			case 'S':
			case 'M':
			case 'P':
				system_mode = c;
				systemptr = buff;
			}
		}
		else switch (system_mode)
		{
			case 'L':    
				terminal_lines = c;
				system_mode = 0;          
				break;

			case 'C':
				terminal_columns = c;
				system_mode = 0;
				break;

			case 'S':
#ifdef SCHAT			
				/* character set I=IBM L=LATIN1 */
				remote_charset = c;
#endif
				system_mode = 0;
				break;

			case 'M':
				if (strchr(packets, setup_packet) > strchr(packets, c)) update_packet = 1;
				set_packets(c);
				system_mode = 0;
				break;

			case 'P':
				if (systemptr >= buff + sizeof terminal_prompt) fatal_error(5);
				*systemptr++ = c;
				if (!c)
				{
					set_prompt(buff);
					system_mode = 0;
				}
				break;

			case '*':
				*systemptr++ = c;
				if (systemptr >= buff + 4) 
				{
					remotecheck = *(unsigned long *) buff;
					system_mode = 0;
				}
				break;

			default:
				system_mode = 0;
		}
	}
	if (update_packet && mux_tx_free(CH_SYSTEM) >= 2)
	{
		update_packet = 0;
		Buffer[0] = 'M';
		Buffer[1] = (char)setup_packet;
		mux_tx_data(CH_SYSTEM, Buffer, 2);
	}
}

void insert_download(struct file_entry _far *_new, long res, char status)
{
	if (!downfiles_left)
	{
		download_files = 0;
		download_bytes = 0;
	}
	_new->status = status;
	_new->start = res;
	file_list_append(&download_list, _new);
	downfiles_left++;
	download_files++;
	download_bytes += _new->size;
	downbytes_left += _new->size - res;
	mux_tx_data(CH_CONTROL, &status, 1);
	transfer = 1;
	start_time = timer;
}

static void upload_run(void)
{
	char c, *p;
	struct file_entry _far *_new;
        int i, len;
        unsigned short ti,da;
	struct file_entry _far *fsp;
	char drive[3];
	char dir[PATH_LEN];
	char fname[PATH_LEN];
	char ext[6];
	char apu[PATH_LEN];
	char extn[6];
	long res;
        int renum;
	struct dir_path _far *list;
    
	while (mux_rx_used(CH_DOWNLOAD) && mux_tx_free(CH_CONTROL))
	{
		mux_rx_data(CH_DOWNLOAD, &c, 1);
		if (ul_ptr >= ul_buff + sizeof ul_buff) fatal_error(1);
		*ul_ptr++ = c;
		if (ul_ptr > ul_buff + HEADER_LEN && !c)
		{
			ul_ptr = ul_buff;
#ifdef LINUX
			unixify(ul_buff + HEADER_LEN);
#endif
			_new = file_list_first(&download_rename_list);
			if (_new) 
				_fstrcpy(ul_buff + HEADER_LEN, _new->name);
			else 
				if (forced_dl) goto toomany;
			
			_splitpath(ul_buff + HEADER_LEN, drive, dir, fname, ext);
			if ((!_new && restrict) || (!*drive && !*dir))
			{
				*drive = 0;
				if (*dlpath) strcpy(dir, dlpath); else 
#ifdef SMODEM				
				strcpy(dir, defpath);
#else
				*dir = 0;
#endif
			}
                        renum = 0;
                        if (_new != NULL) _free(_new);
		
		retrydl:
			_makepath(ul_buff + HEADER_LEN, drive, dir, fname, ext);
			len = strlen(ul_buff + HEADER_LEN) + 1;
			_new = (struct file_entry _far *) _fmalloc(sizeof *_new + len);
			if (!_new)
			{
			toomany:
				mux_tx_data(CH_CONTROL, "t", 1);
				break;
			}
			_fmemcpy(&_new->time, ul_buff, HEADER_LEN + len);
#ifdef _AMIGA /* motorola hi- low- words. */
			_new->size = (long)swaplongs((ULONG)_new->size);
#endif
			if (!ulreq)
			{
				c = 'q';
				goto uperror;
			}
    
    		        //if (!_fstricmp(dl->name, ul_buff + HEADER_LEN)) goto duperr;
    		        fsp = download_list.first;
			while (fsp)
			{
				if (!_fstricmp(fsp->name, ul_buff + HEADER_LEN))
				{
				//duperr:
					c = 'i';
					goto uperror;
				}				
				fsp = fsp->next;
			}

			list = dlpath_first;
			while(list)
			{
				_fstrcpy(apu, list->path);
				_makepath(Buffer, "", apu, fname, ext);
				if (!access(Buffer, 0)) goto existerror;
				list = list->next;
			}

			i = sopen(ul_buff + HEADER_LEN, O_BINARY | O_RDWR, SH_DENYRW, 0);
			if (i != -1)
			{
				if (isatty(i))
				{
					close(i);
					goto existerror;
				}
				if (resume)
				{
#if defined (MSDOS) || defined (__OS2__)
					_dos_getftime(i, &da, &ti);
					if (da == (_new->date & 0x01ff) && ti == _new->time)
#endif
                                        {
						res = lseek(i, 0, SEEK_END);
						if (res < _new->size)
						{
							close(i);
							insert_download(_new, res, 'r');
#ifdef _AMIGA  /* motorola 680x0 hi/low words */
							{
								int apu = res;
								apu = (int)swaplongs((ULONG)apu);
								mux_tx_data(CH_CONTROL, (char *)(&apu), 4);
							}
#else
							mux_tx_data(CH_CONTROL, (char *)&res, 4);
#endif
							break;
						}
						if (res == _new->size)
						{
							close(i);
							c = 'x';
							goto uperror;
						}
					}
				}
				close(i);

				if (!overwrite || unlink(ul_buff + HEADER_LEN) == -1)
				{   
				existerror:
					if (renumber && renum < 1000)
					{
						_ffree(_new);
						if (!*ext) strcpy(ext, ".");
                                                while (strlen(ext) < 4) strcat(ext, "0");
                                                sprintf(extn, "%d", renum++);
						//itoa(renum++, extn, 10);
						strcpy(ext + 4 - strlen(extn), extn);
						goto retrydl;
					}
					c = 'e';
				uperror:
					_new->status = c;
					if (includeaborted) insert_log('E', _new, txcps);
					file_list_append(&error_list, _new);
					mux_tx_data(CH_CONTROL, &c, 1);
					break;
				}
			}
			insert_download(_new, 0L, 'u');
			break;
		}
	}

	if (send_abort_ul) return;

	while (mux_tx_free(CH_DOWNLOAD) >= 256)
	{
		_new = file_list_first(&send_upload_list);
		if (!_new) break;
#ifdef _AMIGA /* Motorola 680x0 -> Intel 80x86 */
		_new->size = (long)swaplongs((ULONG)_new->size);
#endif
		_fmemcpy(Buffer, &_new->time, HEADER_LEN + _fstrlen(_new->name) + 1);
		p = strspace(Buffer + HEADER_LEN);
		_splitpath(Buffer + HEADER_LEN, drive, dir, fname, ext);
		if (p) strcpy(dir, p); else *dir = 0;
		_makepath(Buffer + HEADER_LEN, "", dir, fname, ext);
#ifdef LINUX
                dosify(Buffer + HEADER_LEN);
#endif
		len = HEADER_LEN + strlen(Buffer + HEADER_LEN) + 1;
		mux_tx_data(CH_DOWNLOAD, Buffer, len);
#ifdef _AMIGA /* Intel 80x86 -> Motorola 680x0 */
		_new->size = (long)swaplongs((ULONG)_new->size);
#endif  
                file_list_append(&sent_upload_list, _new);
	}

	if (ul)
	{   
		if (cur_upload_left > ulcount) i = ulcount; else i = (int) cur_upload_left;
		if (mux_tx_free(ulch) >= i)
		{
			if (read(ulhandle, Buffer, i) != i)
			{
				abort_upload('1');
				return;
			}
			mux_tx_data(ulch, Buffer, i);
			ulcount -= i;
			upbytes_left -= i;
			cur_upload_left -= i;
			uploaded_bytes += i;
			if (!ulcount)
			{
				ulcount = BUCKETSIZE;
				if (++ulch > CH_LAST_FILE) ulch = CH_FIRST_FILE;
			}
			if (!cur_upload_left)
			{
				close(ulhandle);
				ulhandle = 0;
				cur_upload_bytes = 0L;
				cur_upload_left = 0L;
				ul = NULL;
				goto ulnext;
			}
		}
	}
	else
	{
	ulnext:
		ul = file_list_first(&upload_list);
		if (ul)
		{
			_fstrcpy(Buffer, ul->name);
			strspace(Buffer);
			ulhandle = sopen(Buffer, O_BINARY | O_RDONLY, SH_DENYWR, 0);
			if (ulhandle == -1)
			{
				ulhandle = 0;
				abort_upload('2');
				return;
			}
			if (lseek(ulhandle, ul->start, SEEK_SET) == -1)
			{
				abort_upload('3');
				return;
			}
			cur_upload_bytes = ul->size;
			cur_upload_left = ul->size - ul->start;
                        file_list_append(&uploading_list, ul);
#ifdef _AMIGA
			open_upload_window(); 
			update_upload_window();
#endif
//#ifdef SMODEM
			update_upload();
//#endif
		}
	}
}

static void download_run(void)
{
	char c;
	struct file_entry _far *_new;
	int i, len;

	while (mux_rx_used(CH_UPLOAD) && mux_tx_free(CH_CONTROL))
	{
		mux_rx_data(CH_UPLOAD, &c, 1);
		if (dl_ptr >= dl_buff + sizeof dl_buff) fatal_error(2);
		*dl_ptr++ = c;
		if (!c)
		{
			dl_ptr = dl_buff;
			if (!dlreq)
			{
				c = 'p';
				goto dlerror;
			}
#ifdef LINUX
			unixify(dl_buff);
#endif
			if (!upload(dl_buff, NULL, NULL, 0))
			{
				c = 'm';
			dlerror:
				len = strlen(dl_buff) + 1;
				_new = (struct file_entry _far *) _fmalloc(sizeof *_new + len);
				if (_new)
				{
					_new->status = c;
					_fmemcpy(_new->name, dl_buff, len);
					file_list_append(&error_list, _new);
				}
			}
			else
				c = 'd';
			mux_tx_data(CH_CONTROL, &c, 1);
			break;
		}
	}

	if (send_abort_dl) return;

	while (mux_tx_free(CH_UPLOAD) >= 256)
	{
		_new = file_list_first(&send_request_list);
		if (!_new) break;
		_fstrcpy(Buffer, _new->name);
#ifdef LINUX		
		dosify(Buffer);
#endif
		mux_tx_data(CH_UPLOAD, Buffer, strlen(Buffer) + 1);
                file_list_append(&sent_request_list, _new);
	}

	if (dl)
	{
		if (cur_download_left > dlcount) i = dlcount; else i = (int) cur_download_left;
		if (mux_rx_used(dlch) >= i)
		{
			mux_rx_data(dlch, Buffer, i);
			if (write(dlhandle, Buffer, i) != i)
			{
				abort_download('5');
				return;
			}
			dlcount -= i;
			downbytes_left -= i;
			cur_download_left -= i;
			downloaded_bytes += i;
			
			if (!dlcount)
			{
				dlcount = BUCKETSIZE;
				if (++dlch > CH_LAST_FILE) dlch = CH_FIRST_FILE;
			}
			if (!cur_download_left)
			{
				close_dl();
				file_list_append(&downloaded_list, dl);
				dl = NULL;
				if (--downfiles_left) goto dlnext;
#ifdef _AMIGA
				update_download_window();
#endif
//#ifdef SMODEM
				update_download();
//#endif
			}
		}
	}
	else
	{
	dlnext:
		dl = file_list_first(&download_list);
		if (dl)
		{
			_fstrcpy(Buffer, dl->name);
			strspace(Buffer);
			if (dl->status == 'r')
				dlhandle = sopen(Buffer, O_BINARY | O_RDWR, SH_DENYRW, 0);
                        else
                                dlhandle = sopen(Buffer, O_BINARY | O_RDWR | O_CREAT | O_EXCL, SH_DENYRW, READWRITE);
                        if (dlhandle == -1)
			{
				dlhandle = 0;
				abort_download('2');
				return;
			}
			if (lseek(dlhandle, 0L, SEEK_END) != dl->start)
			{
				abort_download('3');
				return;
			}
			cur_download_bytes = dl->size;
			cur_download_left = dl->size - dl->start;
#if defined (MSDOS) || defined (__OS2__)
			_dos_setftime(dlhandle, dl->date, dl->time);
#endif
#ifdef LINUX
			stampdate(dl->name, dl->date, dl->time);
#endif
#ifdef _AMIGA
			open_download_window();
			update_download_window();
#endif
//#ifdef SMODEM
			update_download();
//#endif
		}
	}
}

static void insert_upload(long start)
{
	struct file_entry _far *_new;

	_new = file_list_first(&sent_upload_list);
	if (!_new) return;
	if (!upfiles_left)
	{
		upload_files = 0;
		upload_bytes = 0;
	}
	_new->status = 0;
	_new->start = start;
	upfiles_left++;
	upload_files++;
	upload_bytes += _new->size;
	upbytes_left += _new->size - start;
        file_list_append(&upload_list, _new);
	transfer = 1;
	start_time = timer;
}

void insert_log(char mode, struct file_entry _far *_new, int cps)
{
	int i;
	char entry[PATH_LEN + 80];
	    
	if (logfile != -1) 
	{
		if (!cps) cps = 9999;
		_fstrcpy(Buffer, _new->name);
		strspace(Buffer);
		i = sprintf(entry,
			"%c %6ld %5ld bps %4d cps   0 errors     0 %4d %s %lu",
			mode, _new->size, speed, cps,
			mux_packet_size, Buffer, remoteserial);
#ifdef SCHAT
		if (addusername) 
		{
			strcat(entry, " ");
			strcat(entry, UserName);
		}   
#endif
		strcat(entry, "\n");
		write(logfile, entry, strlen(entry));
	}
}

static void close_dl(void)
{
	if (dlhandle) 
	{
#if defined (MSDOS) || defined (__OS2__)
		if (cur_download_left) dl->date &= 0x01ff;
		dl->size -= cur_download_left;
		_dos_setftime(dlhandle, dl->date, dl->time);
#endif
		close(dlhandle);
#ifdef LINUX
		stampdate(dl->name, dl->date, dl->time);
#endif
		dlhandle = 0;
		if (cur_download_left)
		{   
			if (removefiles)
			{
				_fstrcpy(Buffer, dl->name);
				unlink(Buffer);
			}
			else if (includeaborted) insert_log('E', dl, rxcps);
		}
		else
		{
			insert_log('S', dl, rxcps);
		}
	}
	cur_download_bytes = 0L;
	cur_download_left = 0L;
}

static void close_ul(void)
{
	struct file_entry _far *_new;
	if (ulhandle)
	{
		uploading_list.first->size -= cur_upload_left;
		close(ulhandle);
		ulhandle = 0;
	}
	
	cur_upload_bytes = 0L;
	cur_upload_left = 0L;
    
        _new = uploading_list.first;
        while(_new)
        {
		insert_log('e', _new, txcps);
    	        _new = _new->next;
	}
	if (upload_list.first) 
		insert_log('e', upload_list.first, txcps);
	else if (sent_upload_list.first) 
	{
		sent_upload_list.first->size = 0;
		insert_log('e', sent_upload_list.first, txcps);
	}
	else if (send_upload_list.first) 
	{   
		send_upload_list.first->size = 0;
		insert_log('e', send_upload_list.first, txcps);
	}
}

void abort_download(char error)
{
        void *ptr;

	if (dl)
	{
		close_dl();
		dl->status = error;
		file_list_append(&error_list, dl);
		dl = NULL;
	}
	send_abort_dl = 1;

	{
		while ((ptr = file_list_first(&send_request_list)) != NULL) _free(ptr);
		while ((ptr = file_list_first(&sent_request_list)) != NULL) _free(ptr);
		mux_tx_flush(CH_UPLOAD);
	}
}

void abort_upload(char error)
{
	struct file_entry _far *_new;
	int i;

	close_ul();

	if (ul)
	{
		ul = NULL;
		_new = file_list_first(&uploading_list);
		if (_new)
		{
			_new->status = error;
                        file_list_append(&error_list, _new);
		}
	}

	send_abort_ul = 1;

	{
		while ((_new = file_list_first(&send_upload_list)) != NULL) _free(_new);
		while ((_new = file_list_first(&sent_upload_list)) != NULL) _free(_new);
		while ((_new = file_list_first(&upload_list)) != NULL) _free(_new);
		while ((_new = file_list_first(&uploading_list)) != NULL) _free(_new);
    
		mux_tx_flush(CH_DOWNLOAD);
		upfiles_left = 0;
		upbytes_left = 0;
	}
	for (i = CH_FIRST_FILE; i <= CH_LAST_FILE; i++) mux_tx_flush(i);
	ulch = CH_FIRST_FILE;
	ulcount = BUCKETSIZE;
#ifdef _AMIGA
	update_upload_window();
#endif
//#ifdef SMODEM
	update_upload();
//#endif
}

static void control_run(void)
{
	static long start;
	static char *ptr;
	struct file_entry _far *_new;
	char c;
	int i;

	while (mux_rx_data(CH_CONTROL, &c, 1))
	{
		if (control_res)
		{
			*ptr++ = c;
			if (!--control_res)
			{
				insert_upload(start);
			}
		}
		else switch (c)
		{
		case 'a':   /* abort upload */
			mux_rx_flush(CH_UPLOAD);
			dl_ptr = dl_buff;
			abort_upload('0');
			break;

		case 'b':   /* abort download */
			if (dl)
			{
				close_dl();
				dl->status ='4';
                                file_list_append(&error_list, dl);
				dl = NULL;
			}
			while ((_new = file_list_first(&download_list)) != NULL) _free(_new);
			while ((_new = file_list_first(&downloaded_list)) != NULL) _free(_new);
			for (i = CH_FIRST_FILE; i <= CH_LAST_FILE; i++) mux_rx_flush(i);
			dlch = CH_FIRST_FILE;
			dlcount = BUCKETSIZE;
			mux_rx_flush(CH_DOWNLOAD);
			ul_ptr = ul_buff;
			downfiles_left = 0;
			downbytes_left = 0;
#ifdef _AMIGA
			update_download_window();
#endif
//#ifdef SMODEM
			update_download();
//#endif
			send_abort_dl = 3;
			break;

		case 'o':   /* upload aborted */
			send_abort_ul = 0;
			break;

		case 'c':   /* file uploaded */
			if (send_abort_ul) break;
			_new = file_list_first(&uploading_list);
			if (!_new) break;
			upfiles_left--;
			insert_log('s', _new, txcps);
			if (deletefiles)
			{
				_fstrcpy(Buffer, _new->name);
				strspace(Buffer);
				unlink(Buffer);
			}
			if (!upfiles_left)
			{
				_new->status = 'U';
				file_list_append(&error_list, _new);
				return_code &= ~2;
			}
			else
			{
				_ffree(_new);
			}

#ifdef _AMIGA
			update_upload_window();
#endif
//#ifdef SMODEM
			update_upload();
//#endif
			break;

		case 'r':   /* ul resume upload */
			if (send_abort_ul) break;
			control_res = 4;
			ptr = (char *) &start;
			break;

		case 'i':   /* duplicate ignored */
		case 'q':   /* ul disabled */
		case 'e':   /* ul file exist */
		case 't':   /* ul too many files */
			if (send_abort_ul) break;
			_new = file_list_first(&sent_upload_list);
			if (!_new) break;
			if (c != 'e' || includeaborted) insert_log('e', _new, txcps);
			_new->status = c;
			file_list_append(&error_list, _new);
			break;

		case 'u':   /* ul ok */
			if (send_abort_ul) break;
			insert_upload(0L);
			break;                                   

		case 'x':   /* file already transferred */
			if (send_abort_ul) break;
			_new = file_list_first(&sent_upload_list);
			if (!_new) break;
			if (deletefiles)
			{
				_fstrcpy(Buffer, _new->name);                       
				strspace(Buffer);
				unlink(Buffer);
			}
			_new->status = c;
			file_list_append(&error_list, _new);
			break;

		case 'p':   /* dl disabled */
		case 'm':   /* dl missing */
			if (send_abort_dl) break;
			_new = file_list_first(&sent_request_list);
			if (!_new) break;
			_new->status = c;
			file_list_append(&error_list, _new);
			break;

		case 'd':   /* dl ok */
			if (send_abort_dl) break;
			_ffree(file_list_first(&sent_request_list));
			break;
		}
	}
}

static void status_run(void)
{
	struct file_entry _far *_new;

	while (mux_tx_free(CH_CONTROL))
	{
		if (send_abort_ul == 1)
		{
			send_abort_ul = 2;
			mux_tx_data(CH_CONTROL, "b", 1);
			continue;
		}
		if (send_abort_dl == 1)
		{
			send_abort_dl = 2;
			mux_tx_data(CH_CONTROL, "a", 1);
			continue;
		}
		if (send_abort_dl == 3)
		{
			send_abort_dl = 0;
			mux_tx_data(CH_CONTROL, "o", 1);
			continue;
		}
		_new = file_list_first(&downloaded_list);
		if (_new)
		{
			mux_tx_data(CH_CONTROL, "c", 1);
			if (!downfiles_left)
			{
				_new->status = 'D';
                                file_list_append(&error_list, _new);
				return_code &= ~1;
			}
			else
			{
				_ffree(_new);
			}
			continue;
		}
		break;
	}

	if (upfiles_left || downfiles_left) 
	{
		start_time = timer;
		return;
	}

	if (autoexit)
	{
		if (!transfer && elapsed(start_time) < 120) return;
		if ((autoexit & 2) && elapsed(idle_start_time) < idle_delay_time) return;
		if ((autoexit & 1) && elapsed(start_time) < exit_delay_time) return;
                if (mux_tx_used(CH_CONTROL)) return;
                mux_quit = 1;
	}
}

void message_run(void)
{
	struct file_entry _far *_new;
#ifndef SCHAT
	char apu[20];
#endif
	
	_new = file_list_first(&error_list);
	if (_new)
	{
		_fstrcpy(Buffer, _new->name);
		strspace(Buffer);
#ifndef SCHAT
		switch (_new->status)
		{
		case '0':
			ErrMsg("Upload aborted", Buffer);
			break;
		case '1':
			ErrMsg("File read error", Buffer);
			break;
		case '2':
			ErrMsg("File open error", Buffer);
			break;
		case '3':
			ErrMsg("File size error", Buffer);
			break;
		case '4':
			ErrMsg("Download aborted", Buffer);
			break;
		case '5':
			ErrMsg("File write error", Buffer);
			break;
		case 'p':
			ErrMsg("Download request denied", Buffer);
			break;
		case 'q':
			ErrMsg("Upload request denied", Buffer);
			break;
		case 'm':
			ErrMsg("Requested file not found", Buffer);
			break;
		case 'e':
			ErrMsg("File already exist", Buffer);
			break;
		case 'i':
			ErrMsg("Duplicate file ignored", Buffer);
			break;
		case 'x':
			ErrMsg("File already transferred", Buffer);
			break;
		case 't':
			ErrMsg("Too many files to transfer", Buffer);
			break;
		case 'D':
			sprintf(apu, "%ld file(s) received", download_files);
			ErrMsg("Download completed", apu);
			break;
		case 'U':
			sprintf(apu, "%ld file(s) sent", upload_files);
			ErrMsg("Upload completed", apu);
			break;
		}
#endif
		_ffree(_new); 
	}
}

void open_channels(void)
{
	int i;
	
        mux_open_tx_channel(CH_SYSTEM, 256, MUX_OPT_FAST);  /* chat ohjauskanava */
	mux_open_rx_channel(CH_SYSTEM, 256);

	for (i = CH_FIRST_OUTPUT; i <= CH_LAST_OUTPUT; i++)  /* tekstin siirtokanavat */
	{
		mux_open_tx_channel(i, 256, MUX_OPT_TEXT);
		mux_open_rx_channel(i, 514);
	}

	mux_open_tx_channel(CH_DOWNLOAD, 512, MUX_OPT_FAST); /* download pyyntkanava */
	mux_open_rx_channel(CH_DOWNLOAD, 1024);

	mux_open_tx_channel(CH_UPLOAD, 512, MUX_OPT_FAST); /* upload pyyntkanava */
	mux_open_rx_channel(CH_UPLOAD, 1024);

	mux_open_tx_channel(CH_CONTROL, 256, MUX_OPT_FAST); /* tiedostojen siirron kuittauskanava */
	mux_open_rx_channel(CH_CONTROL, 512);

	for (i = CH_FIRST_FILE; i <= CH_LAST_FILE; i++) /* tiedostojen siirtokanavat */
	{
		mux_open_tx_channel(i, 2048, MUX_OPT_SLOW);
		mux_open_rx_channel(i, 2048);
	}

}

void send_setup(void)
{
   	static char buff[] = "S C L ";

	mux_tx_data(CH_SYSTEM, (char *) &localcheck, 4);
   	
	if (dlreq)
	{
		mux_tx_data(CH_SYSTEM, "D", 1);
	}
	if (ulreq)
	{
		mux_tx_data(CH_SYSTEM, "U", 1);
	}   
	if (!restrict) 
	{
		mux_tx_data(CH_SYSTEM, "R", 1);
	}
#ifdef SMODEM	
	if (MonoVideo == 1)
	{
		mux_tx_data(CH_SYSTEM, "B", 1);
	}
#endif

#ifdef SMODEM
	buff[1] = 'I';
	buff[3] = 80;
#ifdef SCHAT
	buff[5] = 0;
#else
	buff[5] = (char) Terminal.Height;
#endif
#endif
#ifdef LINUX
	buff[1] = 'L';
	buff[3] = scrwidth;
	buff[5] = scrlength;
#endif
#ifdef _AMIGA
	buff[1] = 'L';
	buff[3] = ?;
	buff[5] = ?;
#endif
#ifdef __OS2__
        buff[1] = 'I';
        buff[3] = 80;
        buff[5] = 25;
#endif
#ifdef GENERIC
        buff[1] = 'S';
        buff[3] = 80;
        buff[5] = 25;
#endif
	mux_tx_data(0, buff, 6);
}

void deinit_list(struct file_entry *list)
{
    struct file_entry *tmp;

    while (list != NULL)
    {
        tmp = list->next;
        _free(list);
        list = tmp;
    }
}

void close_lists(void)
{
    deinit_list(send_upload_list.first);
    deinit_list(sent_upload_list.first);

    deinit_list(upload_list.first);
    deinit_list(uploading_list.first);

    deinit_list(download_rename_list.first);
    deinit_list(download_list.first);
    deinit_list(downloaded_list.first);

    deinit_list(send_request_list.first);
    deinit_list(sent_request_list.first);
    deinit_list(request_path_list.first);

    deinit_list(error_list.first);
}

void execute_smodem(char *dszlog, char *dllist, unsigned long bpsrate, int passive)
{
    int file;
    char str[MAX_PATH_LEN];
    static char ver[30];
    strcpy(ver, "\x18SMODEM\x19 "VERSION" ");

    dl_start = 0; ul_start = 0; control_res = 0; localcheck = 0;
    remotecheck = 0; setup_packet = MSG_HUGE; yell_mode = 0; quietstart = -1;
    quietstop = 0; transfer = 0; forced_dl = 0; includeaborted = 0;
    removefiles = 0; deletefiles = 0; ulreq = 1; overwrite = 0; renumber = 0;
    restrict = 1; wildcard = 1; dlreq = 0; resume = 1; messages = 1;
    exit_delay_time = 0; idle_delay_time = 0; autoexit = 2;

    fatal_exit = 0; txcps = 0; rxcps = 0; cur_rxcps = 0; cur_txcps = 0;
    update_packet = 0; dl = NULL; ul = NULL; send_abort_ul = 0;
    send_abort_dl = 0; dlch = CH_FIRST_FILE; dlcount = BUCKETSIZE;
    ulch = CH_FIRST_FILE; ulcount = BUCKETSIZE;

    cur_download_bytes = 0L; cur_download_left = 0L; cur_upload_bytes = 0L;
    cur_upload_left = 0L;

    upload_files = 0L; upload_bytes = 0L; uploaded_bytes = 0L; upfiles_left = 0L;
    upbytes_left = 0L;

    download_files = 0L; download_bytes = 0L; downloaded_bytes = 0L;
    downfiles_left = 0L; downbytes_left = 0L;

    logfile = -1; Buffer[0] = '\0'; terminal_prompt[0] = '\0';
    strcpy(packets, "TSCMLH");

    ul_ptr = ul_buff; dl_ptr = dl_buff;

    send_upload_list.first = send_upload_list.last = NULL;
    sent_upload_list.first = sent_upload_list.last = NULL;
    upload_list.first = upload_list.last = NULL;
    uploading_list.first = uploading_list.last = NULL;

    download_rename_list.first = download_rename_list.last = NULL;
    download_list.first = download_list.last = NULL;
    downloaded_list.first = downloaded_list.last = NULL;

    send_request_list.first = send_request_list.last = NULL;
    sent_request_list.first = sent_request_list.last = NULL;
    request_path_list.first = request_path_list.last = NULL;

    error_list.first = error_list.last = NULL;

    upload_id = 0; return_code = 0; strcpy(termpath, "SMODEM.$$$");

    idle_delay_time = 0; idle_start_time = start_time - idle_delay_time;

    logpath = dszlog;
    speed = bpsrate;

    file = dllist == NULL ? -1 : open(dllist, O_RDONLY | O_BINARY);
    if (file != -1)
    {
        while (_fgets(str, sizeof(str), file) != NULL)
        {
            upload(str, NULL, NULL, 1);
        }
        close(file);
    }

    /* start of of msdos.c */
    start_time = get_time();

    if (logpath && *logpath)
    {
        logfile = sopen(logpath, O_APPEND | O_TEXT | O_CREAT | O_RDWR, SH_DENYNO, S_IREAD | S_IWRITE);
        if (logfile == -1)
        {
            write_log("Can not open file transfer log file: %s", logpath);
            cputs("Can not open file transfer log file: ");
            cputs(logpath);
            cputs("\r\n");
            return;
        }
    }
    if (passive < 1)
    {
        *strchr(ver, '\x19') = '\x18';
    }
    if (passive > 1)
    {
        *strchr(ver, '\x18') = ' ';
        *strchr(ver, '\x19') = ' ';
    }

    mux_setup(ver, VERNUM);

    open_channels();
    /* end of msdos.c */

    if (!mux_quit) terminal_init();

    close_channels();
    close_lists();

    if (logfile != -1) close(logfile);
    if (ulhandle > 0) close(ulhandle);
    if (dlhandle > 0) close(dlhandle);

    write_log("SModem return code %d", return_code);
}
