#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#define WriteWindowChar(c) outchr(c)

#include "mslp.h"
#include "smodem.h"
#include "terminal.h"
#include "../output.h"
#include "../grpchat.h"
#include "../userbase.h"

static char inbuffer[1024], tmpbuf[1024];

int terminal_lines = -1;
int terminal_columns = 80;
static int term_output = CH_FIRST_OUTPUT;
static int term_input = CH_LAST_OUTPUT;
static char *inptr = inbuffer;
static int term_xpos, term_ypos;

void scroll_win(void)
{
    if (term_ypos < scrlength-1) return;

    loc_output("@SCROLL_UP:13@@GOTO:1,%d@@INSLINE@@CLREOL@", scrlength-1);
    term_ypos--;
}

void split_text(char *p, int size)
{
        int j, i;

        if (!size) goto empty;

        scroll_win(); loc_output("@GOTO:%d,%d@", term_xpos+1, term_ypos+1);
        for (i = size - 1; i ; i--) 
        {
                if (p[i] < ' ')
                {
                        while (size) 
                        {
                                size--;
                                WriteWindowChar(*p++);
                        }
                        goto empty;
                }
        }

        while(size)
        {
                scroll_win(); loc_output("@GOTO:%d,%d@", term_xpos+1, term_ypos+1);
                i = scrwidth;
                if (i > size) i = size;
                if (size > scrwidth)
                {
                        j = i;
                        while (p[i] != ' ')
                        {
                                if (i < scrwidth / 2)
                                {
                                        i = j;
                                        break;
                                }
                                i--;
                        }
                }
                size -= i;
                while (i--)
                {             
                        term_xpos++;
                        WriteWindowChar(*p++);
                }
                if (size)
                {
                        if (*p == ' ')
                        {
                                p++;
                                size--;
                        }
                }

        empty:
#ifdef SMODEM
                StoreLine();
#endif
                term_xpos = 0; term_ypos++;
                loc_output("@GOTO:%d,%d@", term_xpos+1, term_ypos+1);
        }
}

void terminal_run(void)
{
        char c;
        int lpos;

#ifdef PRO      
        if (print) return;
#endif
        if (terminal_lines < 0) return;

#ifdef SMODEM
        Win = &Terminal;
#endif
        while (mux_rx_data(term_output, &c, 1))
        {
                if (!c)
                {
                        term_output++;
                        if (term_output > CH_LAST_OUTPUT) term_output = CH_FIRST_OUTPUT;
                }
                else
                {
                        if (terminal_lines > 0)
                        {
                                if (c == '\r' || c == '\n')
                                {   
#ifdef PRO
                                        if (commandpath && commandmode && *inbuffer == '/')
                                        {
                                                *inptr = 0;
                                                runcommand(commandpath, inbuffer + 1);
                                                print = 1;                        
                                                Win = &Terminal;
                                        }                                       
#endif
                                        scroll_win();
                                        loc_output("@X07@GOTO:%d,%d@", term_xpos+1, term_ypos+1);
                                        lpos = sprintf(tmpbuf, "@X0%X%s: ", usrsub.grpcolor, usrsub.grpalias);
                                        memcpy(tmpbuf+lpos, inbuffer, (int) (inptr - inbuffer));
                                        tmpbuf[(int) (inptr - inbuffer)+lpos] = '\0';
                                        convert_line(tmpbuf+lpos, tmpbuf);
                                        split_text(inbuffer, inptr - inbuffer);
                                        inptr = inbuffer;
                                }
                                else
                                {   
                                        if (inptr < inbuffer + sizeof inbuffer - 1)     *inptr++ = c;       
                                }
                        }
                        else
                        {
#ifdef SMODEM
                                if (c == '\n') StoreLine();
#endif
                                WriteWindowChar(c);
                        }
                }
        }
}

#include "../logfile.h"

void init_interface(void);
void deinit_interface(void);

void terminal_init(void)
{
        char tmp[80];
#ifdef PRO
	FILE *msgfp;
#endif

        terminal_lines = -1;
        terminal_columns = 80;
        term_output = CH_FIRST_OUTPUT;
        term_input = CH_LAST_OUTPUT;
        inptr = inbuffer;
        term_ypos = scrlength-1;

        send_setup();

        /* Set prompt */
        sprintf(tmp, "P%s: ", usrsub.grpalias);
        mux_tx_data(CH_SYSTEM, tmp, strlen(tmp)+1);

        init_interface();
        init_group();
	while (keyboard_run() && !mux_quit) ;
        close_group();
        deinit_interface();
        shutdown();
}

void terminal_entry(char *str)
{
	char *s = str;
	int k, i, j, n;

	idle_start_time = timer;

	/*if (
#ifdef PRO	
	print || 
#endif
		terminal_lines < 0)
	{
		bell();
		return;
	}*/

#ifdef PRO
	if (commandpath && commandmode && *str == '/')
	{
		runcommand(commandpath, str + 1);
		print = 1;
	}					
#endif

	n = i = strlen(str);
	strcpy(str + i, "\r");
	j = i++;
        k = term_input;
	while (j > 0)
	{	
		if (!mux_tx_ready(k))
		{
			k++;
			if (k > CH_LAST_OUTPUT) k = CH_FIRST_OUTPUT;
			if (mux_tx_used(k))
			{
				str[n] = 0;
				//bell();
				return;
			}
		}
		j -= mux_tx_free(k);
	}

	while (i)
	{	
		if (!mux_tx_ready(term_input))
		{
			term_input++;
			if (term_input > CH_LAST_OUTPUT) term_input = CH_FIRST_OUTPUT;
		}
		j = mux_tx_data(term_input, s, i);
		s += j;
		i -= j;
	}
	
#ifndef GENERIC
	if (terminal_lines > 0)
	{
#ifdef SMODEM
		Win = &Terminal;
		Terminal.Color = Terminal.WinColor[1];
#endif
                loc_output("@X0F");
		split_text(str, n);
#ifdef SMODEM
		Terminal.Color = Terminal.WinColor[0];
#endif
	}
#endif
	*str = 0;
	return;
}
