/* YAK - Copyright (c) 1997 Timo Sirainen - read license.txt */

/* tetris.c - Tetris one of my first C programs.. Made in 1995, modified a bit
              to work with YAK */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "os.h"
#include "output.h"
#include "keyb.h"
#include "modem.h"

#define blocks  7

#define xblocks 10              // 10x19 screen
#define yblocks 19

#define maxx    80             // X-size of screen
#define maxy    25             // Y-size of screen

#define xsize   2              // x-pixels in one block
#define ysize   1              // y-pixels in one block

#define startx  ((maxx / 2) - ((xblocks/2)*xsize))
#define endx    ((maxx / 2) + ((xblocks/2)*xsize)-1)

#define starty  ((maxy / 2) - ((yblocks/2)*ysize))
#define endy    ((maxy / 2) + ((yblocks/2)*ysize))

#define putpix(x, y, col) output("@GOTO:%d,%d@@X%X7 ", x, y, col);
/*#define putpix(x,y,c) (scr[(x)+(y)*80]=((c)<<8)+219)*/

#ifdef UNIX
int tet_get_event(void)
{
    static fd_set fdset;
    struct timeval wait;

    wait.tv_sec = 0;
    wait.tv_usec = 10000;

    FD_ZERO(&fdset);
    if (!null) FD_SET(0, &fdset);
    FD_SET(hCom, &fdset);

    if (select(hCom+1, &fdset, NULL, NULL, &wait) >= 1)
    {
        if (!null && FD_ISSET(0, &fdset)) return 1;
        if (FD_ISSET(hCom, &fdset)) return 2;
    }

    return 0;
}
#else
int tet_get_event(void)
{
    usleep(10000);
    return get_event();
}
#endif

char palikka[4*blocks][4][4] = {
                          /* xxxx */
                          { { 0,1,0,0 },
                            { 0,1,0,0 },
                            { 0,1,0,0 },
                            { 0,1,0,0 },
                          },
                          { { 0,0,0,0 },
                            { 0,0,0,0 },
                            { 1,1,1,1 },
                            { 0,0,0,0 },
                          },
                          { { 0,1,0,0 },
                            { 0,1,0,0 },
                            { 0,1,0,0 },
                            { 0,1,0,0 },
                          },
                          { { 0,0,0,0 },
                            { 0,0,0,0 },
                            { 1,1,1,1 },
                            { 0,0,0,0 },
                          },
                          /*
                             x
                             x
                             xx
                          */
                          { { 0,0,0,0 },
                            { 1,0,0,0 },
                            { 1,0,0,0 },
                            { 1,1,0,0 },
                          },
                          { { 0,0,0,0 },
                            { 0,0,0,0 },
                            { 0,0,1,0 },
                            { 1,1,1,0 },
                          },
                          { { 0,0,0,0 },
                            { 1,1,0,0 },
                            { 0,1,0,0 },
                            { 0,1,0,0 },
                          },
                          { { 0,0,0,0 },
                            { 0,0,0,0 },
                            { 1,1,1,0 },
                            { 1,0,0,0 },
                          },
                          /*
                              x
                              x
                             xx
                          */
                          { { 0,0,0,0 },
                            { 0,1,0,0 },
                            { 0,1,0,0 },
                            { 1,1,0,0 },
                          },
                          { { 0,0,0,0 },
                            { 0,0,0,0 },
                            { 1,1,1,0 },
                            { 0,0,1,0 },
                          },
                          { { 0,0,0,0 },
                            { 1,1,0,0 },
                            { 1,0,0,0 },
                            { 1,0,0,0 },
                          },
                          { { 0,0,0,0 },
                            { 0,0,0,0 },
                            { 1,0,0,0 },
                            { 1,1,1,0 },
                          },
                          /*  x
                            xxx */
                          { { 0,0,0,0 },
                            { 0,1,0,0 },
                            { 1,1,1,0 },
                            { 0,0,0,0 },
                          },
                          { { 0,0,0,0 },
                            { 0,1,0,0 },
                            { 1,1,0,0 },
                            { 0,1,0,0 },
                          },
                          { { 0,0,0,0 },
                            { 0,0,0,0 },
                            { 1,1,1,0 },
                            { 0,1,0,0 },
                          },
                          { { 0,0,0,0 },
                            { 0,1,0,0 },
                            { 0,1,1,0 },
                            { 0,1,0,0 },
                          },
                          /* xx
                             xx */
                          { { 0,0,0,0 },
                            { 0,0,0,0 },
                            { 0,1,1,0 },
                            { 0,1,1,0 },
                          },
                          { { 0,0,0,0 },
                            { 0,0,0,0 },
                            { 0,1,1,0 },
                            { 0,1,1,0 },
                          },
                          { { 0,0,0,0 },
                            { 0,0,0,0 },
                            { 0,1,1,0 },
                            { 0,1,1,0 },
                          },
                          { { 0,0,0,0 },
                            { 0,0,0,0 },
                            { 0,1,1,0 },
                            { 0,1,1,0 },
                          },
                          /*  xx
                             xx */
                          { { 0,0,0,0 },
                            { 0,0,0,0 },
                            { 0,1,1,0 },
                            { 1,1,0,0 },
                          },
                          { { 0,0,0,0 },
                            { 1,0,0,0 },
                            { 1,1,0,0 },
                            { 0,1,0,0 },
                          },
                          { { 0,0,0,0 },
                            { 0,0,0,0 },
                            { 0,1,1,0 },
                            { 1,1,0,0 },
                          },
                          { { 0,0,0,0 },
                            { 1,0,0,0 },
                            { 1,1,0,0 },
                            { 0,1,0,0 },
                          },
                          /* xx
                              xx */
                          { { 0,0,0,0 },
                            { 0,0,0,0 },
                            { 1,1,0,0 },
                            { 0,1,1,0 },
                          },
                          { { 0,0,0,0 },
                            { 0,1,0,0 },
                            { 1,1,0,0 },
                            { 1,0,0,0 },
                          },
                          { { 0,0,0,0 },
                            { 0,0,0,0 },
                            { 1,1,0,0 },
                            { 0,1,1,0 },
                          },
                          { { 0,0,0,0 },
                            { 0,1,0,0 },
                            { 1,1,0,0 },
                            { 1,0,0,0 },
                          }
                        };

char ruutu[10][19];
int bst,ben,yst,yen,block;
long lines;

void draw_background(char backnum)
{
int x,y;

  for(x=startx-1;x<=endx+1;x++) putpix(x,starty-1,8); // Laatikko
  for(x=startx-1;x<=endx+1;x++) putpix(x,endy+1,8);
  for(y=starty-1;y<=endy+1;y++) putpix(startx-1,y,8);
  for(y=starty-1;y<=endy+1;y++) putpix(endx+1,y,8);

  output("@GOTO:1,1@@X08.------------------.@CR@");
  output("| Points:          |@CR@");
  output("| Lines:           |@CR@");
  output("| Level:           |@CR@");
  output("| Next             |@CR@");
  output("|                  |@CR@");
  output("|                  |@CR@");
  output("|                  |@CR@");
  output("|                  |@CR@");
  output("`------------------'@CR@@X00 \b");
}

void draw_linesinc(long lines)
{
  output("@GOTO:12,3@@X08%lu@GOTO:1,11@", lines);
}

void drawoneblock(char x,char y,char col)
{
int x1,y1;

  for(y1=starty+(y*ysize);y1<=starty+(y*ysize)+ysize-1;y1++)
    for(x1=startx+(x*xsize);x1<=startx+(x*xsize)+xsize-1;x1++)
      putpix(x1,y1,col);
}

void drawoneblockxy(char x,char y,char col)
{
int x1,y1;

  for(y1=y*ysize;y1<=(y*ysize)+ysize-1;y1++)
    for(x1=x*xsize;x1<=(x*xsize)+xsize-1;x1++)
      putpix(x1,y1,col);
}

void draw_nextblock(int block)
{
int x1,y1,c;

  c=(block>>2)+1;
  for (y1=0;y1<=3;y1++)
    for (x1=0;x1<=3;x1++)
      if (palikka[block][y1][x1]>0)
        drawoneblockxy(4+x1,5+y1,c); else
        drawoneblockxy(4+x1,5+y1,0);
}

void drawblock(int block, int x, int y)
{
int x1,y1,sy,c;

  if (y>=0) sy=0; else sy=abs(y);

  c=(block>>2)+1;
  for (y1=sy;y1<=3;y1++)
    for (x1=0;x1<=3;x1++)
      if (palikka[block][y1][x1]>0) {
        ruutu[x+x1][y+y1]=c;
        drawoneblock(x+x1,y+y1,c);
      }
}

void clearblock(int block, int x, int y)
{
int x1,y1,sy;

  if (y>=0) sy=0; else sy=abs(y);

  for (y1=sy;y1<=3;y1++)
    for (x1=0;x1<=3;x1++)
      if (palikka[block][y1][x1]>0) {
        ruutu[x+x1][y+y1]=0;
        drawoneblock(x+x1,y+y1,0);
      }
}

char getstart(int block)
{
int x,y;

  for (x=0;x<=3;x++)
    for (y=3;y>0;y--)
      if (palikka[block][y][x]>0) return(x);
  return(3);
}

char getend(int block)
{
int x,y;

  for (x=3;x>0;x--)
    for (y=3;y>0;y--)
      if (palikka[block][y][x]>0) return(x);
  return(0);
}

char getystart(int block)
{
int x,y;

  for (y=0;y<=3;y++)
    for (x=0;x<=3;x++)
      if (palikka[block][y][x]>0) return(y);
  return(3);
}

char getyend(int block)
{
int x,y;

  for (y=3;y>0;y--)
    for (x=0;x<=3;x++)
      if (palikka[block][y][x]>0) return(y);
  return(0);
}

char canput(int x, int y)
{
int x1,y1,sy;

if (y>=0) sy=0; else sy = y<0 ? -y : y;

  for (y1=sy;y1<=3;y1++)
    for (x1=0;x1<=3;x1++)
      if ((palikka[block][y1][x1]>0) && (ruutu[x+x1][y+y1]>0)) return(0);
  return(1);
}

void init_block()
{
  bst=getstart(block);
  ben=getend(block);

  yst=getystart(block);
  yen=getyend(block);
}

void dotetris()
{

signed char xpos,ypos,x,y,z,ytmp;
unsigned char stop,oneline[4],quit,c,nextblock;

int wait, hit, waittime, level;

  srand(time(NULL));

  lines=0;
  memset(ruutu,0,sizeof(ruutu));
  quit=0;

  level=1;
  ypos=-3;
  xpos=4;
  block=(rand() % blocks) << 2;
  nextblock=(rand() % blocks) << 2;
  draw_nextblock(nextblock);

  init_block();

  waittime = 40;
  do {
    drawblock(block,xpos,ypos);
    wait=waittime;
    while ((wait>0) && (!quit)) {
      wait--;

      hit = tet_get_event();
      if (hit)
        {
          c = (unsigned char) (hit == 2 ? mdm_getch() : getch());
          switch (c) {
            case 0:   c = (unsigned char) (hit == 2 ? mdm_getch() : getch());
                      switch (c) {
                        case 'M': if (xpos<9-ben) {
                                    z=0;
                                    y=ypos+yst;
                                    if (y<0) y=0;
                                    for (;y<=ypos+yen;y++)
                                      for (x=3;x>=0;x--)
                                        if (palikka[block][y-ypos][x]>0)
                                          {
                                            if (ruutu[xpos+x+1][y]>0) z=1;
                                            break;
                                          }
                                    if (z==1) break;
                                    clearblock(block,xpos,ypos);
                                    xpos++;
                                    drawblock(block,xpos,ypos);
                                    break;
                                  } else break;
                        case 'K': if (xpos>0-bst) {
                                    z=0;
                                    y=ypos+yst;
                                    if (y<0) y=0;
                                    for (;y<=ypos+yen;y++)
                                      for (x=0;x<=3;x++)
                                        if (palikka[block][y-ypos][x]>0)
                                          {
                                            if (ruutu[xpos+x-1][y]>0) z=1;
                                            break;
                                          }
                                    if (z==1) break;
                                    clearblock(block,xpos,ypos);
                                    xpos--;
                                    drawblock(block,xpos,ypos);
                                    break;
                                  } else break;
                        case 'H': {
                                    y=block;

                                    clearblock(block,xpos,ypos);

                                    x=(block & 3)+1;
                                    block&=~3;
                                    if (x==4) x=0;
                                    block|=x;
again:
                                    init_block();
                                    if ((xpos+bst<0) || (xpos+ben>xblocks-1)) {
                                      block=y;
                                      goto again;
                                    }

                                    if (ypos+yen>yblocks-1) {
                                      block=y;
                                      goto again;
                                    }

                                    if (!canput(xpos,ypos)) {
                                      block=y;
                                      goto again;
                                    }

                                    drawblock(block,xpos,ypos);
                                    break;
                                  }
                        case 'P': {
                                    wait=0;
                                    break;
                                  }
                      }
                      break;
            case 18:   for (y=yblocks-1;y>0;y--)               // Redraw screen
                         for (x=0;x<=xblocks-1;x++)            // !! NOT OPTIMIZED !!
                           drawoneblock(x,y,ruutu[x][y]);
                       break;
            case 27:   quit=1;
                       break;
          }
          output("@GOTO:1,11@");
        }
    }

    stop=0;
    if (ypos>=0) ytmp=0; else ytmp=abs(ypos);
    for (x=0;x<=3;x++)
      for (y=3;y>=ytmp;y--)
        if (palikka[block][y][x]>0) {
          if (ruutu[xpos+x][ypos+y+1]>0) stop=1;
          break;
        }
    if ((ypos<yblocks-yen-1) && (!stop)) {
      clearblock(block,xpos,ypos);
      ypos++;
    } else {
      if (ypos+yst<=0) quit=1;
      x=0;
      for (y=ypos+yst;y<=ypos+yen;y++) {
        oneline[x]=y; x++;
        for (xpos=0;xpos<=xblocks-1;xpos++)
          if (ruutu[xpos][y]==0) {
            x--;
            break;
          }
        }

      if (x>0) {
        for (z=0;z<x;z++) {
          lines++;
          draw_linesinc(lines);
          ypos=oneline[z];
          for (y=ypos;y>0;y--)               // Move lines down
            for (xpos=0;xpos<=xblocks-1;xpos++)
              ruutu[xpos][y]=ruutu[xpos][y-1];

          for (xpos=0;xpos<=xblocks-1;xpos++)     // Clear uppest line
            ruutu[xpos][0]=0;

          for (xpos=0;xpos<=xblocks-1;xpos++)
            drawoneblock(xpos,ypos,0);
        }

        if (lines / 10 > level-1) {
          level++;
          if (waittime > 2) waittime-=3;
          output("@GOTO:12,4@@X08%lu@GOTO:1,11@", level);
        }
        output("@GOTO:1,11@");
        usleep(200000);

        for (y=yblocks-1;y>0;y--)               // Redraw screen
          for (xpos=0;xpos<=xblocks-1;xpos++)   // !! NOT OPTIMIZED !!
            drawoneblock(xpos,y,ruutu[xpos][y]);
      }

      ypos=-3;
      xpos=4;
      block=nextblock;
      init_block();
      nextblock=(rand() % blocks) << 2;
      draw_nextblock(nextblock);
      output("@GOTO:1,11@");
    }
  } while (!quit);
}

void play_tetris(void)
{
  output("@X07@CLR@");
  draw_background(1);

  dotetris();

  output("@X07@CLR@");
  output("Sait %lu rivi pois@CR@", lines);
}
