#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/time.h>

#include "os.h"
#include "output.h"
#include "files.h"
#include "keyb.h"
#include "logfile.h"
#include "modem.h"
#include "timeslic.h"
#include "bbs_func.h"
#include "config.h"
#include "nodes.h"

int get_event(void)
{
    static fd_set fdset;
    struct timeval wait;

    wait.tv_sec = 0;
    wait.tv_usec = 500000;

    if (!carr_det()) return 0;

    FD_ZERO(&fdset);
    if (!null) FD_SET(Fin, &fdset);
    FD_SET(hCom, &fdset);

    if (select(((hCom > Fin || null) ? hCom : Fin)+1, &fdset, NULL, NULL, &wait) >= 1)
    {
        if (!null && FD_ISSET(Fin, &fdset)) return 1;
        if (FD_ISSET(hCom, &fdset)) return 2;
    }

    return 0;
}

void execute_protocol(char *path, int usestd)
{
    char *arg[50],*strp;
    int pid,num;

    KbdDeInit();

    pid = fork();
    if (pid == -1)
    {
        output("\r\nCouldn't fork()! Out of memory??");
        write_log("execute_protocol() : Couldn't fork()! Out of memory??");
        KbdInit();
        return;
    }

    if (pid == 0)
    {
        /* Run protocol */
        arg[0] = path;
        num = 1;
        for (;;)
        {
            strp = strchr(arg[num-1],' ');
            if (strp == NULL) break;
            *strp = '\0';
            arg[num++] = strp+1;
        }
        arg[num] = NULL;
        if (usestd)
        {
            dup2(hCom, STDIN_FILENO);
            dup2(hCom, STDOUT_FILENO);
            dup2(hCom, STDERR_FILENO);
        }
        execvp(arg[0],&arg[0]);
    }

    while (waitpid(pid,NULL,WNOHANG) == 0)
    {
        if (!carr_det())
        {
            /* No carrier! */
            sleep(5);
            if (waitpid(pid,NULL,WNOHANG) == 0)
            {
                /* It can't kill itself, so we have to do the dirty job.. */
                kill(pid,SIGHUP);
            }
            break;
        }
        sleep(1);
        inactivity = 0; told_inactivity = 0; last_tim = last_min = time(NULL);
    }

    KbdInit();
}

static int fout = STDOUT_FILENO, ferr = STDERR_FILENO;

void output_modem(void)
{
    tty_old();
    fout = dup(STDOUT_FILENO);
    ferr = dup(STDERR_FILENO);
    dup2(hCom, STDOUT_FILENO);
    dup2(hCom, STDERR_FILENO);
}

void output_screen(void)
{
    dup2(fout, STDOUT_FILENO); close(fout);
    dup2(ferr, STDERR_FILENO); close(ferr);
    tty_raw();
}

void idiot_mode(char *name)
{
    char *arg[6];
    char buf[80];
    FILE *F;
    struct stat statbuf;
    int tries;

    write_log("Started idiot_mode() with application %s",name);
    output("\r\nStarting %s .. this will take a couple of seconds..",name);

    /* Check if dosemu is still in background */
    if (dosemu_pid != -1 && waitpid(dosemu_pid,NULL,WNOHANG) != 0)
    {
        /* Nope, it wasn't. */
        dosemu_pid = -1;
    }

    if (dosemu_pid == -1)
    {
        /* Start DOS emulator */
        dosemu_pid = fork();
        if (dosemu_pid == -1)
        {
            output("\r\nCouldn't fork()! Out of memory??");
            write_log("idiot_mode() : Couldn't fork()! Out of memory??");
            return;
        }

        if (dosemu_pid == 0)
        {
            /* Start DOS emulator */
            arg[0] = "dos";
            arg[1] = "-F";
            sprintf(buf, "/etc/dosemu%d.conf", nodenum);
            arg[2] = buf;
            arg[3] = "-o";
            arg[4] = "/tmp/debug";
            arg[5] = NULL;
            execvp(arg[0],&arg[0]);
        }
    }
    else
    {
        kill(dosemu_pid, SIGCONT);
    }

    sprintf(buf,"%sack.%d",common_path,nodenum);
    remove(buf);

    /* Create starting semafore to dosemu */
    sprintf(buf,"%swaiting.%d",common_path,nodenum);
    F = fopen(buf, "w+b"); fwrite(name,1,1,F); fclose(F);
    chmod(buf, 0660);

    /* Wait for ack-file from dosemu.. */
    tries = 0;
    sprintf(buf,"%sack.%d",common_path,nodenum);
    while (stat(buf,&statbuf) != 0 && tries < 10)
    {
        outchr('.');
        tries++;
        sleep(1);
    }
    output("ok!\r\n");

    sprintf(buf,"%sack.%d",common_path,nodenum); remove(buf);

    if (tries == 30)
    {
        /* Didn't got ack-file in 30 seconds, something must've happened to
         dosemu.. */
        output("What?? DOS-emulator must have crashed, why doesn't it work???\r\n");
    }
    else
    {
        /* Update node file */
        noderec.doing = DOING_DOOR;
        strcpy(noderec.data,name);
        update_nodefile(0);

        /* Wait until door quits */
        for (;;)
        {
            sleep(1);

            /* Check if dosemu is still in there.. */
            if (waitpid(dosemu_pid,NULL,WNOHANG) != 0)
            {
                output("\r\n\n@X07Great, DOS-emulator crashed..\r\n");
                dosemu_pid = -1;
                break;
            }

            /* Check if door was quit. */
            sprintf(buf,"%swaiting.%d",common_path,nodenum);
            if (stat(buf,&statbuf) != 0) break;
        }
        inactivity = 0; told_inactivity = 0; last_tim = last_min = time(NULL);
    }

    sprintf(buf,"%swaiting.%d",common_path,nodenum); remove(buf);

    /* Update node file */
    noderec.doing = DOING_NOTHING;
    update_nodefile(0);

    if (dosemu_pid != -1) kill(dosemu_pid, SIGSTOP);
}
